/***************************************************************************
                          attr.cpp  -  description
                             -------------------
    begin                : Tue Mar 18 2003
    copyright            : (C) 2003 by John Redmond
    email                : john_redmond@optusnet.com.au
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "attr.h"
#include "text.h"
#include "element.h"
#include "document.h"
#include "namednodemap.h"
#include "writer.h"

using std::cout;
using std::endl;

namespace limpid {

Attr::Attr(): ParentNode() {
  try {
    nodeContent_ = new AttributeContent("", *this);
  }
  catch (std::bad_alloc) {
    delete nodeContent_;
    throw;
  }
}

Attr::Attr(const DOMString& name): ParentNode() {
  try {
    nodeContent_ = new AttributeContent(name, *this);
  }
  catch (std::bad_alloc) {
    delete nodeContent_;
    throw;
  }
}

Attr::Attr(const DOMString& namespaceURI, const DOMString& name): ParentNode() {
  try {
    nodeContent_ = new AttributeContent(name, *this);
    setNamespaceURI(namespaceURI);
  }
  catch (std::bad_alloc) {
    delete nodeContent_;
    throw;
  }
}

Attr::Attr(const Attr& refNode): ParentNode(refNode) {
}

Attr& Attr::operator=(const Attr& refNode) {
  replaceContent(refNode);
  return *this;
}

DOMString Attr::getRawValue() const {
  if (!isValid()) return "";

  DOMString value;
  AttributeContent *contentPtr = static_cast<AttributeContent *>(nodeContent_);
  if (!contentPtr->childNodes_) return value;

  int childCount = contentPtr->childNodes_->getLength();
  for (int i = 0; i < childCount; i++) {
    Node& node = contentPtr->childNodes_->item(i);//i);
    short type = node.getNodeType();
    if (type != Node::TEXT_NODE && type != Node::ENTITY_REFERENCE_NODE) continue;
    value += node.getDisplayString();
  }

  return value;
}

DOMString Attr::getValue() const {
  DOMString text = getRawValue().resolveEntities();
  return getOwnerDocument().resolveEntities(text);
}

void Attr::setValue(const DOMString& value) {
  if (!isValid()) return;

  AttributeContent *contentPtr = static_cast<AttributeContent *>(nodeContent_);
  if (contentPtr->childNodes_) delete contentPtr->childNodes_;
  contentPtr->childNodes_ = new NodeList(*this);

  Text text = (const_cast<DOMString &>(value)).normalizeWhitespace();
  contentPtr->childNodes_->addNode(text);
}

Element& Attr::getOwnerElement() const {
  if (nodeContent_->parentContent_)
    return static_cast<Element &>(*nodeContent_->parentContent_->ownerNode_);
  else return Element::failElement;
}

DOMString Attr::getNamespaceURI() const {
  AttributeContent *contentPtr = static_cast<AttributeContent *>(nodeContent_);
  if (contentPtr->namespaceURI_) return contentPtr->namespaceURI_;
  else if (!contentPtr->prefix_) return DOMString::failString;
  else if (contentPtr->parentContent_ && contentPtr->parentContent_->ownerNode_) {
    Element *ownerElement =
      static_cast<Element *>(contentPtr->parentContent_->ownerNode_);
    DOMString prefix = "xmlns:";
    prefix += contentPtr->prefix_;
    return ownerElement->findNamespaceURI(prefix);
  }
  else return "";
}

Node& Attr::cloneNode(bool) const {
  Attr *clone = new Attr(getNodeName());
  clone->nodeContent_->documentContent_ = nodeContent_->documentContent_;

  AttributeContent *contentPtr = static_cast<AttributeContent *>(nodeContent_);
  if (contentPtr->childNodes_) { // always copy child nodes in this case
    size_t length = contentPtr->childNodes_->getLength();
    for (size_t i = 0; i < length; i++) {
      Node& child = contentPtr->childNodes_->item(i).cloneNode(true);
      clone->appendChild(child);
    }
  }

  getOwnerDocument().cacheNode(*clone);
  return *clone;
}

DOMString Attr::getDisplayString() const {
  DOMString s = getNodeName();
  s += "=\"";
  s += getValue().escapeDelimiters();
  s += "\"";

  return s;
}

void Attr::setNamespaceURI(const DOMString& namespaceURI) {
  AttributeContent *contentPtr = static_cast<AttributeContent *>(nodeContent_);
  contentPtr->namespaceURI_ = namespaceURI.copyChars();
}

Attr Attr::failAttr("#fail");

}
