/***************************************************************************
                          attributevalidator.cpp  -  description
                             -------------------
    begin                : Wed Jan 7 2004
    copyright            : (C) 2004 by John Redmond
    email                : john_redmond@optusnet.com.au
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "attr.h"
#include "namednodemap.h"
#include "attributevalidator.h"
#include "liststream.h"
#include "nodestream.h"
#include "attributestream.h"
#include "writer.h"

namespace limpid {

AttributeValidator::AttributeValidator():
      writer(new Writer), elementMap(new ElementMap) {
}

void AttributeValidator::setup(Document& document) {
  try {
    SimpleMap simpleTypes;
    collectSimpleTypes(document, simpleTypes);
    collectContent(document, simpleTypes);
  }
  catch (std::bad_alloc) {
    throw;
  }
}

void AttributeValidator::collectSimpleTypes(Document& document, SimpleMap& simpleTypes) {
  Element& documentElement = document.getDocumentElement();
  ListStream stream(documentElement);

  while (true) {
    Node& node = stream.nextNode();
    if (!node.isValid()) break;
    else if (node.getLocalName() == "simpleType" &&
        node.getNamespaceURI() == "http://www.w3c.org/2001/XMLSchema") {
      Element &simpleElement = static_cast<Element &>(node);
      DOMString key = simpleElement.getAttribute("name");
      if (key.empty()) continue;
      simpleTypes.insert(make_pair(key, &node));
    }
  }
}

void AttributeValidator::collectContent(Document& document, SimpleMap& simpleTypes) {
  Element& documentElement = document.getDocumentElement();
  ListStream stream(documentElement);

  while (true) {
    Node& node = stream.nextNode();
    if (!node.isValid()) break;

    DOMString localName = node.getLocalName();
    if ((localName == "complexType" || localName == "element")  &&
        node.getNamespaceURI() == "http://www.w3c.org/2001/XMLSchema") {
      AttributeMap attributeMap;
      addAttributeSpecs(node, simpleTypes, attributeMap);
      if (!attributeMap.empty()) {
        Element &complexElement = static_cast<Element &>(node);
        DOMString key = complexElement.getAttribute("name");
        elementMap->insert(make_pair(key, attributeMap));
      }
    }
  }
}

void AttributeValidator::addAttributeSpecs(Node& parentNode,
      SimpleMap& simpleTypes, AttributeMap& attributeMap) {
  Element &element = static_cast<Element &>(parentNode);
  NodeStream stream(element);

  while (true) {
    Node& node = stream.nextNode();
    if (!node.isValid()) break;

    DOMString localName = node.getLocalName();
    if (node.getLocalName() != "attribute" ||
        node.getNamespaceURI() != "http://www.w3c.org/2001/XMLSchema")
      continue;

    Element &attrElement = static_cast<Element &>(node);
    DOMString key = attrElement.getAttribute("name");
    if (!key.empty()) {
      addSimpleSpec(key, attrElement, attributeMap);
      continue;
    }

    key = attrElement.getAttribute("type");
    if (key.empty()) continue;

    SimpleMap::iterator p = simpleTypes.find(key);
    if (p != simpleTypes.end()) {
      Element &simpleElement = static_cast<Element &>(*p->second);
      addReferenceSpec(key, attrElement, simpleElement, attributeMap);
    }
  }
}

void AttributeValidator::addSimpleSpec(DOMString& key,
      Element& attrElement, AttributeMap& attributeMap) {
  DOMString defaultVal = attrElement.getAttribute("value");
  AttributeSpec specs(defaultVal);

  DOMString use = attrElement.getAttribute("use");
  specs.required = (use != "optional");
  specs.fixed = (use == "fixed");
  DOMString type = attrElement.getAttribute("type");
  specs.type = getType(type);

  attributeMap.insert(make_pair(key, specs));
}

void AttributeValidator::addReferenceSpec(DOMString& key,
      Element& attrElement, Element& referredElement, AttributeMap& attributeMap) {
  DOMString defaultVal = attrElement.getAttribute("default");
  AttributeSpec spec(defaultVal);
  attributeMap.insert(make_pair(key, spec));

  DOMString use = attrElement.getAttribute("use");
  spec.required = (use != "optional");
  spec.fixed = (use == "fixed");


  Node& restrictionNode = attrElement;
  if (attrElement.getFirstChild().isValid())
    restrictionNode = attrElement.getFirstChild();
  else
    restrictionNode = referredElement;
  if (!restrictionNode.isValid()) return; // cannot proceed

  Element& restrictionElement = static_cast<Element &>(restrictionNode);
  DOMString base = restrictionElement.getAttribute("base");
  spec.type = getType(base);

  ListStream enumStream(restrictionNode);
  while (true) {
    Node& node = enumStream.nextNode();
    if (!node.isValid()) break;

    if (node.getNodeName() == "xs:enumeration" &&
        node.getNamespaceURI() == "http://www.w3c.org/2001/XMLSchema") {
      Element &enumElement = static_cast<Element &>(node);
      DOMString value = enumElement.getAttribute("value");
      if (!value.empty())
        spec.addEnumeration(value);
    }
  }
}

AttributeSpec::Type AttributeValidator::getType(DOMString& typeString) {
  if (typeString == "xs:string") return AttributeSpec::STRING;
  if (typeString == "xs:ENTITY") return AttributeSpec::ENTITY;
  if (typeString == "xs:ENTITIES") return AttributeSpec::ENTITIES;
  if (typeString == "xs:NOTATION") return AttributeSpec::NOTATION;
  if (typeString == "xs:NMTOKEN") return AttributeSpec::NMTOKEN;
  if (typeString == "xs:ID") return AttributeSpec::ID;
  if (typeString == "xs:IDREF") return AttributeSpec::IDREF;
  if (typeString == "xs:IDREFS") return AttributeSpec::IDREFS;

  return AttributeSpec::STRING;
}

void AttributeValidator::validateNames(Element& element,
      CharacterValidator& characterValidator) {
  DOMString elementName = element.getNodeName();
  ElementMap::iterator p = elementMap->find(elementName);
  if (p == elementMap->end()) return; // this is not a valid element

  int lineNo = element.getPosition();

  NamedNodeMap& attributes = element.getAttributes();
  size_t length = attributes.getLength();
  if (!length) return;

  for (size_t i = 0; i < length; i++) {
    Attr& attr = static_cast<Attr &>(attributes.item(i));
    DOMString attrName = attr.getLocalName();
    if (!attrName.empty()) {
      DOMString prefix = attr.getPrefix();
      characterValidator.checkValidName(prefix, lineNo);
      characterValidator.checkValidAttributeName(elementName, attrName, lineNo);
    }
    else {
      attrName = attr.getNodeName();
      characterValidator.checkValidAttributeName(elementName, attrName, lineNo);
    }
  }
}

void AttributeValidator::validateSpecified(Element& element) {
  NamedNodeMap attributes = element.getAttributes();
  size_t length = attributes.getLength();
  if (!length) return;

  DOMString elementName = element.getNodeName();
  ElementMap::iterator p = elementMap->find(elementName);
  if (p == elementMap->end()) {
    if (length) {
      int lineNo = element.getPosition();
      for (size_t i = 0; i < length; i++) {
        Node& attr = attributes.item(i);
        if (attr.getNodeName().startsWith("xml")) continue;

        *writer << "L:" << toString(lineNo) << "\t" << element.getNodeName();
        *writer << " has undefined ATTRIBUTE " << attr << wrt::endl;
      }
    }
    return;
  }

  AttributeMap map = p->second;
  for (size_t i = 0; i < length; i++) {
    Node& node = attributes.item(i);
    Attr &attr = static_cast<Attr &>(node);
    DOMString name = attr.getNodeName();
    if (name.startsWith("xml")) continue;

    AttributeMap::iterator q = map.find(name);
    if (q == map.end()) {
      int lineNo = element.getPosition();
      *writer << "L:" << toString(lineNo) << "\tUNDEFINED ATTRIBUTE [" << name;
      *writer << "] in element " << elementName << wrt::endl;
    }
  }
}

void AttributeValidator::validateValues(Element& element,
      CharacterValidator& characterValidator) {

  NamedNodeMap attributes = element.getAttributes();
  if (!attributes.isValid()) return;  // no attributes to validate

  DOMString elementName = element.getNodeName();
  ElementMap::iterator p = elementMap->find(elementName);
  if (p == elementMap->end()) return; // no matching element with attributes

  int lineNo = element.getPosition();
  AttributeMap map = p->second;

  size_t length = attributes.getLength();
  for (size_t i = 0; i < length; i++) {
    DOMString attrName = attributes.item(i).getNodeName();
    AttributeMap::iterator ap = map.find(attrName);
    if (ap == map.end()) continue;

    DOMString attrValue = attributes.item(i).getValue();
    AttributeSpec spec = ap->second;

    if (!spec.enumeration.empty())
      validateEnumeration(elementName, spec.enumeration, attrName, attrValue, lineNo);
    else
      validateType(element, spec, attrValue, characterValidator);
  }
}

void AttributeValidator::validateEnumeration(DOMString& elementName,
      Enumeration& enumeration, DOMString& attrName, DOMString& attrValue, int lineNo) {
  Enumeration::iterator p2 = enumeration.find(attrValue);
  if (p2 == enumeration.end()) {
    *writer << "L:" << toString(lineNo) << "\tATTRIBUTE [" << attrName;
    *writer  << "] in element " << elementName;
    *writer  << " has value [" << attrValue << "],\n\tand not [";
    outputSet(enumeration);
  }
}

void AttributeValidator::outputSet(Enumeration& enumeration) {
  size_t size = enumeration.size(), count = 1;
  Enumeration::iterator p = enumeration.begin();
  while (p != enumeration.end()) {
    *writer << *p;
    if (count < size - 1) *writer << ", ";
    else if (count < size) *writer << " or ";
    p++;
    count++;
  }
  *writer << "]" << wrt::endl;
}

void AttributeValidator::validateType(Element& element, AttributeSpec& spec,
      DOMString& attrValue, CharacterValidator& characterValidator) {
  DOMString elementName = element.getNodeName();
  int lineNo = element.getPosition();
  int type = spec.type;
  if (type == AttributeSpec::NMTOKEN || type == AttributeSpec::ID)
    characterValidator.checkValidNameToken(elementName, attrValue, lineNo);
  else
    characterValidator.checkValidAttributeValue(elementName, attrValue, lineNo);
}

void AttributeValidator::validateComplete(Element& element) {
  DOMString elementName = element.getNodeName();
  ElementMap::iterator p = elementMap->find(elementName);
  if (p == elementMap->end()) return;

  AttributeMap map = p->second;
  AttributeMap::iterator p2 = map.begin();

  while (p2 != map.end()) {
    AttributeSpec spec = p2->second;
    if (spec.required) {
      Attr& attr = element.getAttributeNode(p2->first);
      if (!attr.isValid())
        *writer << "L:" << element.getPosition() << "\tATTRIBUTE [" << p2->first\
         << "] MISSING in element " << elementName << wrt::endl;
    }
    p2++;
  }
}

size_t AttributeValidator::getSize() {
  size_t size = sizeof(*this);

  ElementMap::const_iterator p = elementMap->begin();
  while (p != elementMap->end()) {
    size += p->first.getSize();
    AttributeMap::const_iterator amp = p->second.begin();
    while (amp != p->second.end()) {
      size += amp->first.getSize();
      size += amp->second.getSize();
      ++amp;
    }
    ++p;
  }
  return size;
}

}
