/***************************************************************************
                          configurationmanager.cpp  -  description
                             -------------------
    begin                : Tue Mar 29 2005
    copyright            : (C) 2005 by John Redmond
    email                : john_redmond@optusnet.com.au
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "document.h"
#include "element.h"
#include "configurationmanager.h"
#include "configurationset.h"
#include "keyspecs.h"
#include "xslutility.h"
#include "stringreader.h"
#include "tokenstream.h"
#include "liststream.h"
#include "nodestream.h"
#include "attributestream.h"


namespace limpid {

ConfigurationManager::ConfigurationManager(Element& configureElement, bool setDefs):
  configureElement_(configureElement), setDefaults_(setDefs) {}

void ConfigurationManager::configure(ConfigurationSet& configurationSet,
      bool useOutputElement) {
  if (setDefaults_) {
    setDefaults(configurationSet);
    setKeys(configurationSet);
    setNamespaces(configurationSet);
  }

  if (useOutputElement) configureOutput(configurationSet); // use children of passed element
  else setOutput(configureElement_, configurationSet); // use attrs on passed element
}

void ConfigurationManager::configureOutput(ConfigurationSet& configurationSet) {
  ListStream listStream(configureElement_);
  while (true) {
    Node& node = listStream.nextNode();
    if (!node.isValid()) break;

    if (!XSLUtility::isXSLElement(node)) continue;

    Element& element = static_cast<Element &>(node);
    if (XSLUtility::getNodeName(element) != "output") continue;
    setOutput(element, configurationSet);
  }

}
void ConfigurationManager::setOutput(Element& outputElement,
      ConfigurationSet& configurationSet) {
  DOMString version = outputElement.getAttribute("version");
  if (!version.empty()) configurationSet.version = version;

  DOMString method = outputElement.getAttribute("method");
  if (!method.empty()) configurationSet.method = method;

  DOMString encoding = outputElement.getAttribute("encoding");
  if (!encoding.empty()) configurationSet.encoding = encoding;

  DOMString omit_xml_declaration =
    outputElement.getAttribute("omit-xml-declaration");
  if (!omit_xml_declaration.empty())
    configurationSet.omit_xml_declaration =
      (omit_xml_declaration == "yes" ? true : false);

  DOMString standalone = outputElement.getAttribute("standalone");
  if (!standalone.empty())
    configurationSet.standalone = (standalone == "yes" ? true : false);

  DOMString doctype_public = outputElement.getAttribute("doctype-public");
  if (!doctype_public.empty()) configurationSet.doctype_public = doctype_public;

  DOMString doctype_system = outputElement.getAttribute("doctype-system");
  if (!doctype_system.empty()) configurationSet.doctype_system = doctype_system;

  DOMString indent = outputElement.getAttribute("indent");
  if (!indent.empty())
    configurationSet.indent = (indent == "no" ? false : true);

  DOMString media_type = outputElement.getAttribute("media-type");
  if (!media_type.empty()) configurationSet.media_type = media_type;

  DOMString cdata_section_elements =
    outputElement.getAttribute("cdata-section-elements");
  if (cdata_section_elements.size())
    configurationSet.cdata_section_elements = cdata_section_elements;

  // confirm methods and versions
  if (configurationSet.method.endsWith("xml")) {
    configurationSet.method = "xml";
    version = "2.0";
    configurationSet.media_type = "text/xml";
    configurationSet.escape_text = true;
    configurationSet.escape_attributes = false;
  }
  else if (configurationSet.method.endsWith("xhtml")) {
    configurationSet.method = "xhtml";
    version = "1.0";
    configurationSet.media_type = "text/xhtml";
    configurationSet.escape_text = true;
    configurationSet.escape_attributes = true;
  }
  else if (configurationSet.method.endsWith("html")) {
    configurationSet.method = "html";
    version = "4.0";
    configurationSet.media_type = "text/html";
    configurationSet.escape_text = true;
    configurationSet.escape_attributes = true;
  }
  else if (configurationSet.method.endsWith("text")) {
    configurationSet.method = "text";
    configurationSet.media_type = "text";
    configurationSet.escape_text = false;
  }
}

void ConfigurationManager::setDefaults(ConfigurationSet& configurationSet) {
  configurationSet.version = "1.0";
  configurationSet.method = "xml";
  configurationSet.encoding = "utf-8";
  configurationSet.omit_xml_declaration = false;
  configurationSet.standalone = true;
  configurationSet.indent = true;
  configurationSet.escape_text = true;
  configurationSet.escape_attributes = true;
  configurationSet.building = true;
}

void ConfigurationManager::setKeys(ConfigurationSet& configurationSet) {
  NodeStream stream(configureElement_);

  while (true) {
    Node& node = stream.nextNode();
    if (!node.isValid()) break;

    if (!XSLUtility::isXSLElement(node)) continue;
    Element& element = static_cast<Element &>(node);
    if (XSLUtility::getNodeName(element) != "key") continue;

    DOMString keyName = element.getAttribute("name");
    configurationSet.keyMap.erase(keyName); // there may be an old version there
    
    KeySpecs keySpecs(element.getAttribute("match"), element.getAttribute("use"));
    configurationSet.keyMap.insert(make_pair(keyName, keySpecs));
  }
}

void ConfigurationManager::setNamespaces(ConfigurationSet& configurationSet) {
  ListStream stream(configureElement_);

  while (true) {
    Node& node = stream.nextNode();
    if (!node.isValid()) break;

    if (!XSLUtility::isXSLElement(node)) continue;
    Element& element = static_cast<Element &>(node);
    if (XSLUtility::getNodeName(element) != "namespace-alias") continue;

    DOMString stylesheetName = "xmlns", resultName = "xmlns";
    DOMString aliasName = element.getAttribute("stylesheet-prefix");
    DOMString modelName = element.getAttribute("result-prefix");

    if (!aliasName.empty()) {
      stylesheetName += ":";
      stylesheetName += aliasName;
    }

    if (!modelName.empty()) {
      resultName += ":";
      resultName += modelName;
    }

    DOMString copiedURI = configureElement_.getAttribute(resultName);
    configurationSet.namespaceMap.insert(make_pair(stylesheetName, copiedURI));
  }
}

}

