/***************************************************************************
                          contentvalidator.cpp  -  description
                             -------------------
    begin                : Sat Dec 6 2003
    copyright            : (C) 2003 by John Redmond
    email                : john_redmond@optusnet.com.au
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "contentvalidator.h"
#include "xmlwriter.h"
#include "traverser.h"
#include "writer.h"
#include "nodestream.h"
#include "sequencepoint.h"
#include "choicepoint.h"
#include "pointtracer.h"

namespace limpid {

ContentValidator::ContentValidator():
  writer_(new Writer), childrenMap_(new ChildrenMap), matchMap_(new MatchMap), typeMap_(new TypeMap), simple_(false) {}

void ContentValidator::setWriter(Writer& writer) {
  writer_ = new Writer(writer);
}

void ContentValidator::setup(Document& document) {
  setTypes(document);

  Element& documentElement = document.getDocumentElement();
  ListStream stream(documentElement);

  while (true) {
    Node& node = stream.nextNode();
    if (!node.isValid()) break;

    if (!isXSDElement(node)) continue;
    Element& element = static_cast<Element &>(node);

    DOMString localName = element.getLocalName();
    if (localName == "complexType") {/// || localName == "element") {
      addContentPoint(element);
      addChildrenList(element);
    }
  }
}

void ContentValidator::setTypes(Document& document) {
  NodeStream nodeStream(document.getDocumentElement(), false, false);

  while (true) {
    Node& node = nodeStream.nextNode();
    if (!node.isValid()) break;
    if (!isXSDElement(node)) continue;

    Element& element = static_cast<Element &>(node);
    if (element.getLocalName() != "element") continue;

    typeMap_->insert(make_pair(element.getAttribute("name"), element.getAttribute("type")));
  }
}

void ContentValidator::addContentPoint(Element& element) {
  DOMString key = element.getAttribute("name");
  Node& repeatNode = element.getFirstChild();
  if (!repeatNode.isValid()) return;

  if (!isXSDElement(repeatNode)) return;
  Element repeatElement = static_cast<Element &>(repeatNode);

  DOMString repeatName = repeatElement.getLocalName();

  if (repeatName == "sequence") {
    SequencePoint sequencePoint(repeatElement, key);
    matchMap_->insert(make_pair(key, &sequencePoint.copy()));
  }
  else if (repeatName == "choice") {
    ChoicePoint choicePoint(repeatElement, key);
    matchMap_->insert(make_pair(key, &choicePoint.copy()));
  }
}

void ContentValidator::addChildrenList(Element& element) {
  DOMString parentName = element.getAttribute("name");
  String childrenNames;
  NodeStream nodeStream(element, false, false);

  while (true) {
    Node& node = nodeStream.nextNode();
    if (!node.isValid()) break;

    if (!isXSDElement(element)) continue;
    Element& element = static_cast<Element &>(node);

    DOMString elementName = element.getAttribute("name");
    if (elementName.empty())
      elementName = element.getAttribute("type");

    childrenNames += " ";
    childrenNames += elementName;
  }

  childrenMap_->insert(make_pair(parentName, childrenNames.normalizeWhitespace()));
}

void ContentValidator::validate(Element& element) {
  if (simple_) {
    validateChildren(element);
    return;
  }

  DOMString typeName = (*typeMap_)[element.getNodeName()];
  MatchMap::iterator p = matchMap_->find(typeName);//element.getNodeName());
  if (p != matchMap_->end()) {
    MatchPoint& matchPoint = *p->second;
    PointTracer().trace(element, matchPoint, *writer_);
  }
}

void ContentValidator::validateChildren(Element& element) {
  if (!element.hasNonTextChildren()) return;

  DOMString elementName = element.getNodeName();
  DOMString typeName = (*typeMap_)[elementName];
  DOMString& childNames = (*childrenMap_)[typeName];
  ListStream stream(element);

  while (true) {
    Node& node = stream.nextNode();
    if (!node.isValid()) break;

    if (node.getNodeType() == Node::ELEMENT_NODE &&
      childNames.indexOf(node.getNodeName()) == -1)
      *writer_  << "L:" << node.getPosition() << ": " << node.getNodeName() << " ILLEGAL CHILD OF " << elementName << " (" << childNames << "  permitted)" << wrt::endl;
  }
}

size_t ContentValidator::getSize() {
  size_t size = sizeof(*this);
  return size;
}

bool ContentValidator::isXSDElement(Node& node) {
  if (node.getNodeType() != Node::ELEMENT_NODE) return false;
  Element& element = static_cast<Element &>(node);
  return element.getNamespaceURI() == "http://www.w3c.org/2001/XMLSchema";
}

}
