/***************************************************************************
 *   Copyright (C) 2007 by John Redmond   *
 *   john_redmond@optusnet.com.au   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "cssparser.h"
#include "element.h"
#include "comment.h"
#include "writer.h"
#include "argparser.h"
#include "domutils.h"
#include "domwriter.h"
#include "stringreader.h"
#include "tokenstream.h"

namespace limpid {

CSSParser::CSSParser(const DOMString& fileName): endNode_(new EndNode("END")),
    reader_(new Reader(fileName)), lastElement_(0), rootName_("STYLES"),
    state_(START), lineNo_(1) {}

CSSParser::CSSParser(CharacterReader& reader): endNode_(new EndNode("END")),
    reader_(&reader.copy()), lastElement_(0), rootName_("STYLES"),
    state_(START), lineNo_(1) {}

Node& CSSParser::nextNode() {
  while (true) {
    int oldState = state_;
    state_ = getState();

    // allow a start, even it it turns out to have no data
    if (state_ == START) {
      pushElement(rootName_, true);
      node_->setPosition(lineNo_);
      state_ = STARTED;
      return *node_;
    }

    if (state_ == DONE)
      return Node::failNode;

    if (state_ == ATSTART) {
      pushElement("AT", true);
      getChar();  // swallow the '@'

      DOMString envString = getDelimitedString(" \n\r\t");  // eg 'media'
      DOMUtils::addLeaf("ENV", envString, *node_);

      skipWhitespace();
      DOMString mediaName = getDelimitedString("{").trim();
      DOMUtils::addLeaf("TYPE", mediaName, *node_);
      return *node_;
    }

    if (state_ == ATEND) {
      node_ = new EndNode("AT");

      getChar(); // swallow the "}"
      return *node_;
    }

    if (state_ == FINISH) {
      node_ = new EndNode(rootName_);
      return *node_;
    }

    if (state_ == DONE) {
      return Node::failNode;
    }

    if (state_ == COMMENTSTART) {
      DOMString commentString = getCommentString();
      node_ = new Comment(commentString.trim());
      node_->setPosition(lineNo_);
      return *node_;
    }

    if (state_ == STYLESTART) {
      pushElement("STYLE", true);
      node_->setPosition(lineNo_);
      setTargets();
      setSpecs();
      return *node_;
   }

    if (state_ == STYLEEND) {
      node_ = new EndNode("STYLE");
      return *node_;
   }
 }
}

void CSSParser::pushElement (const DOMString& name, bool isParent) {
  Element* element;

  tagStack_.push_back(name);
  tagName_ = name;
  element = new Element(name);

  DOMString nameCopy = name;
  if (nameCopy.toLowerCase() == "div")
    element->setAttribute("keep", "true");

  lastElement_ = element;
  node_ = element;
  if (isParent)
    node_->setIsParent(true);
}

 int CSSParser::getState() {
   // state transitions:
   if (state_ == FINISH) return DONE;
   if (state_ == START ||  state_ == DONE)
     return state_; // no change

   if (state_ == STARTED) return ANY;
   if (state_ == COMMENTSTART) return ANY;
   if (state_ == STYLESTART) return STYLEEND;
   if (state_ == STYLEEND) return ANY;
   if (state_ == ATSTART) return ANY;
   if (state_ == ERROR) return FINISH;

   skipWhitespace();

   Char c = peekChar();
   if (c == FAILCHAR) return FINISH;

   // state determinations:
   if (state_ == ANY) {
     if (c == '}') return ATEND;
     if (c == '/') return COMMENTSTART;
     if (c == '@') return ATSTART;
     return STYLESTART;
   }

   // it must be an error
   else
     return ERROR;
}

void CSSParser::setTargets() {
  DOMString targetString = getDelimitedString("{");
  ArgParser parser(targetString, ',');
  int index = 0;

  while (true) {
    DOMString arg = parser.getArg(index++);
    if (arg.empty()) break;

    Element targetElement("TAG");
    node_->appendChild(targetElement);

    Text targetText(arg.normalizeWhitespace());
    targetElement.appendChild(targetText);
  }
}

void CSSParser::setSpecs() {
  DOMString specString = getDelimitedString("}");
  ArgParser parser(specString, ';');
  int index = 0;

  while (true) {
    DOMString arg = parser.getArg(index++);
    if (arg.empty()) break;

    int colonIndex = arg.indexOf(':');
    if (colonIndex == -1) continue; // error

    DOMString specName = arg.substr(0, colonIndex);
    Element specElement("SPEC");
    specElement.setAttribute("property", specName.trim());
    node_->appendChild(specElement);

    DOMString specValue = arg.substr(colonIndex + 1);
    specValue.trim();
    int specLength = specValue.getLength();

    if (specLength >= 2 && specValue[0] == '"' && specValue[specLength-1] == '"')
      specValue = specValue.substr(1, specLength - 2);

    Text specText(specValue);
    specElement.appendChild(specText);
  }

}

DOMString CSSParser::getCommentString() {
  bool hasAsterisk = false;
  DOMString commentString;

  while (true) {
    Char c = reader_->nextChar();
    if (c == '\n') ++lineNo_;
    if (c == FAILCHAR)
      break;

    commentString += c;

    if (hasAsterisk && c == '/')
      break;

    if (c == '*')
      hasAsterisk = true;
  }

  int commentLength = commentString.getLength();
  commentString = commentString.substr(2, commentLength - 4);

  return commentString.trim();
}

void CSSParser::skipWhitespace() {
  while (true) {
    Char c = reader_->nextChar();
    if (c == '\n') ++lineNo_;
    if (c == FAILCHAR) return;

    if (c != 32 & c != 13 & c != 10) break;
  }
  reader_->unget();
}

DOMString CSSParser::getLine() {
  DOMString resultString;

  while (true) {
    Char c = reader_->nextChar();
    if (c == '\n') ++lineNo_;
    if (c == FAILCHAR || c == '\n')
      break;
    resultString += c;
  }
  ++lineNo_;
  return resultString;
}

DOMString CSSParser::getDelimitedString(const DOMString& delimiters) {
  DOMString resultString;
  Char c;

  while (true) {
    c = reader_->nextChar();
    if (c == FAILCHAR) break;
    if (c == '\n') ++lineNo_;
    if (delimiters.contains(c))
      break;

    resultString += c;
  }
  return resultString;
}

Char CSSParser::peekChar() {
  Char c = reader_->nextChar();
  if (c != FAILCHAR) reader_->unget();
  return c;
}

Char CSSParser::getChar() {
  return(reader_->nextChar());
}

DOMString CSSParser::normalizePath(Element& parent, DOMString& sourceString) {
  bool isAttr = false;
  DOMString attrMarker;

  sourceString.normalizeWhitespace();
  sourceString.change(" +", "+");
  sourceString.change("+ ", "+");
  sourceString.change(" >", ">");
  sourceString.change("> ", ">");
  sourceString.change(" .", ".");
  sourceString.change(". ", ".");
  sourceString.change(" #", "#");
  sourceString.change("# ", "#");

  DOMString resultString;
  StringReader sp(sourceString);
  TokenStream ts(sp);
  ts.setDelimiters(" +>.#[]");
  ts.setWhitespaceChars("");

  wrt::Errors << wrt::endl << sourceString << wrt::endl;

  while (true) {
    DOMString token = ts.nextToken();
    if (token.empty()) break;

    if (isAttr) {
      resultString += token;
      if (attrMarker == "[")
        ts.nextToken(); // discard ']'
      resultString += "]";

      isAttr = false;
      continue;
    }

///    wrt::Errors << token << " ";
    if (token == " ") resultString += "//";
    else if (token == ">") resultString += "/";
    else if (token == ".") {
      isAttr = true;
      attrMarker = token;
      resultString += "[";
    }
    else if (token == "#") {
      isAttr = true;
      attrMarker = token;
      resultString += "[";
    }

    else resultString += token;
  }

//  parent.setAttribute("path", resultString);
  wrt::Errors << wrt::endl << "********** RES: " << resultString << wrt::endl;

  return resultString;
}

}
