/***************************************************************************
 *   Copyright (C) 2007 by John Redmond   *
 *   john_redmond@optusnet.com.au   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "csswriter.h"
#include "endnode.h"
#include "liststream.h"
#include "domutils.h"
#include "writer.h"

namespace limpid {

CSSWriter::CSSWriter() {}

CSSWriter::CSSWriter(const DOMString& destString):
  writer_(new Writer(destString)) {}

CSSWriter::CSSWriter(CharacterWriter& characterWriter):
    writer_(&(characterWriter.copy())) {}

void CSSWriter::update(const DOMString& destString) {
  DOMString outputString = destString;
  outputString += ".css";
  writer_ = new Writer(outputString);
}

Node& CSSWriter::process(Node& node) {
  int type = node.getNodeType();

  if (type == Node::ELEMENT_NODE) processElement(node);
  else if (type == Node::COMMENT_NODE) processComment(node);
  else if (type == Node::END_NODE) processEndNode(node);

  return node;
}

void CSSWriter::processComment(Node& node) {
  DOMString outputString = "/* ";
  outputString += node.getValue();
  outputString += " */";

  *writer_ << wrt::endl << outputString << wrt::endl;
}

void CSSWriter::processElement(Node& node) {
  if (node.getNodeName() == "AT")
    processAtStart(node);
}

void CSSWriter::processEndNode(Node& node) {
  Node& startNode = static_cast<EndNode &>(node).getStartNode();
  DOMString nodeName = startNode.getNodeName();

  if (nodeName == "STYLE") processStyleStart(startNode);
  if (nodeName == "AT") processAtEnd();
}

void CSSWriter::processStyleStart(Node& node) {
  bool hasTag = false, hasSpec = false;
  DOMString tagString, definitionString = " { ";

  ListStream childStream(node);
  while (true) {
    Node& node = childStream.nextNode();
    if (!node.isValid()) break;

    DOMString nodeName = node.getNodeName();
    if (nodeName == "TAG") {
      if (hasTag)
        tagString += ", ";
      else
        hasTag = true;

      tagString += node.getValue();
    }
    else {
      if (hasSpec)
        definitionString += "; ";
      else
        hasSpec = true;

      definitionString += static_cast<Element &>(node).getAttribute("property");//getNodeName();
      definitionString += ": ";
      definitionString += node.getValue();
    }
  }

  definitionString += " }";

  *writer_ << tagString;
  *writer_ << definitionString << wrt::endl;
}

void CSSWriter::processAtStart(Node& node) {
  DOMString envString = DOMUtils::getChildValue("ENV", node);
  DOMString typeString = DOMUtils::getChildValue("TYPE", node);
  *writer_ << "@" << envString << " " << typeString << " {" << wrt::endl;
}

void CSSWriter::processAtEnd() {
  *writer_ << "}" << wrt::endl;
}

}
