/***************************************************************************
                          csvparser.cpp  -  description
                             -------------------
    begin                : Wed Mar 29 2006
    copyright            : (C) 2006 by John Redmond
    email                : john_redmond@optusnet.com.au
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "csvparser.h"
#include "node.h"
#include "element.h"
#include "text.h"
#include "liststream.h"
#include "writer.h"
#include "stringreader.h"

#include "domwriter.h"
#include "converter.h"

namespace limpid {

CSVParser::CSVParser(const DOMString& fileName, int delimiterChar, bool useLabels):
    labels_(new StringVector),
    reader_(new Reader(fileName)), tis_(new TokenStream(*reader_)),
    delimiterChar_(delimiterChar), lineNo_(1), isStarted_(false), useLabels_(useLabels),
    hasLabels_(false), nodeSent_(false) {
  configure();
}

CSVParser::CSVParser(const Reader& charStream, int delimiterChar, bool useLabels):
    labels_(new StringVector),
    reader_(new Reader(charStream)), tis_(new TokenStream(*reader_)),
    delimiterChar_(delimiterChar), lineNo_(1), isStarted_(false), useLabels_(useLabels),
    hasLabels_(false), nodeSent_(false) {
  configure();
}

Node& CSVParser::nextNode() {
  if (nodeSent_) {
    nodeSent_ = false;
    return endNode_; // this will trigger denesting in an XML builder
  }
  
  if (!isStarted_) {
    isStarted_ = true;
    node_ = new Element("SHEET"); // this is the document element of the new document
    node_->addChildNodes();
    node_->setPosition(0);
    
    return *node_;
  }
  
  if (!setLine()) // the data stream is finished
    return Node::failNode;

  DOMString rowName = "ENTRY";
  bool isLabelRow = false;
  
  parseLine(rowName);
  if (isLabelRow)
    setLabels();
  node_->setPosition(lineNo_);

  nodeSent_ = true;
  return *node_;
}

// create a row and add the column elements to it
void CSVParser::parseLine(DOMString& rowName) {
  static DOMString delimiterString = delimiterChar_;
  static DOMString endString = "\n";
  
  node_ = new Element(rowName);
  
  StringReader stringReader(line_);
  TokenStream tis(stringReader);
  tis.setDelimiters(delimiterString_);
  tis.setWhitespaceChars("");

  size_t index = 0;
  DOMString value;
  
  while (tis.isValid()) {
    DOMString token = tis.nextToken();

    if (token == delimiterString || token == endString) { // this triggers setting of the value
      setColumn(*node_, value, index);
      if (token == "\n") break;
      ++index; // bump if continuing
    }
    else
      value = token;
  }
}

// add a column with data to the entry element
void CSVParser::setColumn(Node& entryElement, DOMString& value, size_t index) {
  DOMString nodeName = getColumnLabel(index);
  Element col(nodeName);
  entryElement.appendChild(col);

  // add data only if it is not empty
  if (value.getLength()) {
//    value.escapeDelimiters().trim();
    Text valueText(value);
    col.appendChild(valueText);
  }
}

// establish a line to be parsed
bool CSVParser::setLine() {
  while (true) {
    if (!tis_->isValid()) return false;

    line_ = tis_->nextToken();
    ++lineNo_;
    if (!line_.empty()) {
      line_ += '\n';
      return true;
    }
  }
}

// setup a list of labels to be applied to each of the entry rows
void CSVParser::setLabels() {
  StringReader stringReader(line_);
  TokenStream tis(stringReader);
  tis.setDelimiters(",\n");
  tis.setWhitespaceChars("");

  while (tis.isValid()) {
    DOMString token = tis.nextToken();
    if (token == ",") continue;
    
    labels_->push_back(token.trim());
  }
}

// get an appropriate label for the index
DOMString CSVParser::getColumnLabel(size_t index) {
  if (useLabels_ && index < labels_->size()) // should not have overflow here
    return (*labels_)[index];
  else {
    DOMString nodeName = "C";
    return nodeName += Converter::toString(index + 1);
  }
}

void CSVParser::configure() {
  delimiterString_ += Char(delimiterChar_);
  delimiterString_ += '\n';

  tis_->setWhitespaceChars("\n");
  tis_->setDelimiters("");
}

}
