/***************************************************************************
                          decimalformatter.cpp  -  description
                             -------------------
    begin                : Mon May 2 2005
    copyright            : (C) 2005 by John Redmond
    email                : john_redmond@optusnet.com.au
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <sstream>
#include <cmath>
#include "decimalformatter.h"
#include "decimalformat.h"
#include "writer.h"

namespace limpid {

DecimalFormatter::DecimalFormatter(const DOMString& formatString) {
  configure(formatString);
}

DOMString DecimalFormatter::format(double number) {
  if (isinf(number)) return format_.infinity;
  if (isnan(number)) return format_.nan;
  if (number < 0) {
    isNegative_ = true;
    number = -number;
  }

  double integer;
  double fraction = modf(number, &integer);

  if (fraction == 0)
    return getIntegerString(integer); // this may not be the desired behaviour

  return getRealString(number);
}

DOMString DecimalFormatter::format(double number, const DecimalFormat& decimalFormat) {
  format_ = decimalFormat;
  return format(number);
}

DOMString DecimalFormatter::getIntegerString(double integer) {
  DOMString integerString = getSimpleString(integer);
  DOMString formattedString =
    separateGroups(formatSimpleString(integerString, leftWidth_, true));

  if (!isNegative_) return formattedString; // no minus sign required

  DOMString resultString = format_.minusSign;
  resultString += formattedString;
  return resultString;
}

DOMString DecimalFormatter::getRealString(double number) {
  double integer;
  double fraction = modf(forceRounding(number), &integer); // repeat this process

  DOMString fractionString = getSimpleString(fraction);

  DOMString realString;
  if (isNegative_) realString += format_.minusSign;

  realString += formatSimpleString(getSimpleString(integer), leftWidth_, true);
  realString += format_.decimalSeparator;
  realString += formatSimpleString(fractionString, rightWidth_, false);

  return realString;
}

DOMString DecimalFormatter::getSimpleString(double number) {
  std::ostringstream output;
  output << number;

  DOMString simpleString = output.str().c_str();
  size_t pointPos = simpleString.indexOf('.');
  if (pointPos == simpleString.npos) return simpleString;

  return simpleString.substr(pointPos + 1, simpleString.size() - pointPos - 1);
}

double DecimalFormatter::forceRounding(double value) {
  if (!rightWidth_) return value;

  int exponent = -rightWidth_ - 1;
  double rounder = pow(10, exponent);
  return value + 5 * rounder;
}

DOMString DecimalFormatter::formatSimpleString(const DOMString& simpleString,
    size_t width, bool atStart) {
  size_t length = simpleString.size();

  if (length == width) return simpleString;   // nothing to do
  if (atStart && simpleString.size() >= width) return simpleString; // nothing to do
  // here we have a simple truncation, as the rounding has aleady been done
  if (!atStart && simpleString.size() > width) return simpleString.substr(0, width);

  Char zeroDigit = format_.zeroDigit[0]; // this will be used for padding
  DOMString fillerString(width - length, zeroDigit);
  DOMString resultString;

  if (atStart) {
    resultString += fillerString;
    resultString += simpleString;
  }
  else {
    resultString += simpleString;
    resultString += fillerString;
  }

  return resultString;
}

// this function is used only if there is no fractional part
DOMString DecimalFormatter::separateGroups(const DOMString& numberString) {
  DOMString groupedString;

  size_t length = numberString.size();
  if (length <= 3) return numberString;

  size_t start = 0, end = length % 3;

  while (true) {
    if (start >= length) break;

    groupedString += numberString.substr(start, end - start);
    if (end < length)
      groupedString += format_.groupingSeparator;

    start = end;
    end += 3;

    if (end > length) end = length; // should not happen?
  }

  return groupedString;

}

void DecimalFormatter::configure(const DOMString& formatString) {
  isNegative_ = false;
  size_t pointPos;

  pointPos = formatString.indexOf(".");
  if (pointPos == formatString.npos)
    pointPos = formatString.indexOf(",");

  if (pointPos != formatString.npos) {
    leftWidth_ = pointPos;
    rightWidth_ = formatString.size() - pointPos - 1;
  }
  else {
    leftWidth_ = formatString.size();
    rightWidth_ = 0;
  }
}

}
