/***************************************************************************
                          document.cpp  -  description
                             -------------------
    begin                : Tue Mar 18 2003
    copyright            : (C) 2003 by John Redmond
    email                : john_redmond@optusnet.com.au
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "document.h"
#include "documenttype.h"
#include "element.h"
#include "text.h"
#include "comment.h"
#include "cdatasection.h"
#include "processinginstruction.h"
#include "xmldeclaration.h"
#include "attr.h"
#include "xmldeclaration.h"
#include "systemmanager.h"
#include "nodestream.h"
#include "liststream.h"

namespace limpid {

Document::Document(): ParentNode(), previousCopy(0), nextCopy(0) {
  try {
    nodeContent_ = new DocumentContent(*this);
    DocumentContent *contentPtr = static_cast<DocumentContent *>(nodeContent_);
    contentPtr->ownerNode_ = this;
  }
  catch (std::bad_alloc) {
    delete nodeContent_;
    throw;
  }
}

Document::Document(const DOMString& namespaceURI, const DOMString& name):
      ParentNode(), previousCopy(0), nextCopy(0) {
  try {
    nodeContent_ = new DocumentContent(namespaceURI, name, *this);
    DocumentContent *contentPtr = static_cast<DocumentContent *>(nodeContent_);
    contentPtr->ownerNode_ = this;
  }
  catch (std::bad_alloc) {
    delete nodeContent_;
    throw;
  }
}

Document::Document(const Document& refNode): ParentNode(refNode) {
  // insert new document into linked list after refNode
  linkAfterDocument(refNode);
}

Document::~Document() {
  unlinkThisDocument();
}

Document& Document::operator=(const Document& refNode) {
  unlinkThisDocument();
  replaceContent(refNode);
  linkAfterDocument(refNode);
  return *this;
}

Document& Document::getOwnerDocument() const {
  return const_cast<Document &>(*this);
}

Element& Document::getDocumentElement() {
  ListStream stream(*this);
  while (true) {
    Node& node = stream.nextNode();
    if (!node.isValid()) break;

    if (node.getNodeType() == ELEMENT_NODE)
      return static_cast<Element &>(node);
  }

  return Element::failElement;
}

DocumentType& Document::getDoctype() {
  ListStream stream(*this);
  while (true) {
    Node& node = stream.nextNode();
    if (!node.isValid()) break;

    if (node.getNodeType() == DOCUMENT_TYPE_NODE)
      return static_cast<DocumentType &>(node);
  }

  return DocumentType::getDefaultDocumentType();
}

XMLDeclaration& Document::getXMLDeclaration() {
  ListStream stream(*this);
  while (true) {
    Node& node = stream.nextNode();
    if (!node.isValid()) break;

    if (node.getNodeType() == XML_DECLARATION_NODE)
      return static_cast<XMLDeclaration &>(node);
  }

  return XMLDeclaration::failDeclaration;
}

DOMString Document::getNamespaceURI() const {
  DocumentContent *contentPtr = static_cast<DocumentContent *>(nodeContent_);
  return contentPtr->namespaceURI_;
}

Node& Document::appendChild(Node& childNode) {
  Node& returnNode = ParentNode::appendChild(childNode);
  if (!returnNode.isValid()) return returnNode;

  DocumentContent *contentPtr = static_cast<DocumentContent *>(nodeContent_);
  returnNode.setDocumentContent(*contentPtr);
  contentPtr->ownerNode_ = this;

  return returnNode;
}

NodeList Document::getElementsByTagName(const DOMString& name) {
   return getElementsByTagName(name, *this);
}

NodeList Document::getElementsByTagName(const DOMString& name, const Node& root) {
  NodeList resultList;
  NodeStream stream(root, false);
  DOMString refName = name;

  while (true) {
    Node& node = stream.nextNode();
    if (!node.isValid()) break;

    if (node.getNodeType() != ELEMENT_NODE) continue;
    if (refName != "*" && node.getNodeName() != refName) continue;

    resultList.addNode(node);
  }
  return resultList;
}

NodeList Document::getElementsByTagNameNS(const DOMString& namespaceURI, const DOMString& name) {
   return getElementsByTagNameNS(namespaceURI, name, *this);
}

NodeList Document::getElementsByTagNameNS(const DOMString& namespaceURI,
      const DOMString& name, const Node& root) {
  DOMString refName = name;
  NodeList resultList;
  NodeStream stream(root, false);

  while (true) {
    Node& node = stream.nextNode();
    if (!node.isValid()) break;

    if (node.getNodeType() != ELEMENT_NODE) continue;
    if (refName != "*" && node.getLocalName() != refName) continue;

    if (node.getNamespaceURI() == namespaceURI)
      resultList.addNode(node);
  }
  return resultList;
}

Element& Document::getElementById(const DOMString& elementId) {
  return getElementById(elementId, *this);
}

Element& Document::getElementById(const DOMString& elementId, const Node& root) {
  if (root.getOwnerDocument() != *this) return Element::failElement;

  const StringMap idMap = getDoctype().getIDMap();
  NodeStream stream(root, false, false);

  while (true) {
    Node& node = stream.nextNode();
    if (!node.isValid()) break;
    if (node.getNodeType() != Node::ELEMENT_NODE) continue;
    Element &element = static_cast<Element &>(node);
    DOMString elementName = element.getNodeName();
    StringMap::const_iterator p = idMap.find(elementName);
    if (p == idMap.end()) continue;

    const DOMString attrName = p->second;
    if (element.getAttribute(attrName) == elementId) return element;
  }

  return Element::failElement;
}

DOMString Document::getIDAttrName(const DOMString& elementName) {
  const StringMap idMap = getDoctype().getIDMap();
  StringMap::const_iterator p = idMap.find(elementName);
  if (p == idMap.end()) return "";
  return p->second;
}

Node& Document::importNode(Node& node, bool deep) {
  DocumentContent *contentPtr = static_cast<DocumentContent *>(nodeContent_);
  if (deep) node.setDocumentContent(*contentPtr);
  else node.nodeContent_->documentContent_ = contentPtr;

  return node;
}

Element Document::createElement(const DOMString &tagName) {
  Element element(tagName);
  setDocumentContent(static_cast<DocumentContent &>(*nodeContent_));
  return element;
}

Element Document::createElementNS(const DOMString &namespaceURI, const DOMString &tagName) {
  Element element(namespaceURI, tagName);
  setDocumentContent(static_cast<DocumentContent &>(*nodeContent_));
  return element;
}

Text Document::createTextNode(const DOMString &data) {
  Text text(data);
  setDocumentContent(static_cast<DocumentContent &>(*nodeContent_));
  return text;
}

Comment Document::createComment(const DOMString &data) {
  Comment comment(data);
  setDocumentContent(static_cast<DocumentContent &>(*nodeContent_));
  return comment;
}

CDATASection Document::createCDATASection(const DOMString &data) {
  CDATASection cdataSection(data);
  setDocumentContent(static_cast<DocumentContent &>(*nodeContent_));
  return cdataSection;
}

ProcessingInstruction Document::createProcessingInstruction(const DOMString &target,
                                                            const DOMString &data) {
  ProcessingInstruction pi(target, data);
  setDocumentContent(static_cast<DocumentContent &>(*nodeContent_));
  return pi;
}

Attr Document::createAttribute(const DOMString &name) {
  Attr attr(name);
  setDocumentContent(static_cast<DocumentContent &>(*nodeContent_));
  return attr;
}

Attr Document::createAttributeNS(const DOMString& namespaceURI, const DOMString& name) {
  Attr attr(namespaceURI, name);
  setDocumentContent(static_cast<DocumentContent &>(*nodeContent_));
  return attr;
}

DOMString& Document::resolveEntities(DOMString& string) {
  return getDoctype().resolveEntities(string);
}

void Document::connect() {
  DocumentContent *contentPtr = static_cast<DocumentContent *>(nodeContent_);
  contentPtr->ownerNode_ = this;
}

Node& Document::cloneNode(bool deep) const {
  Document *clone = new Document();
  DocumentContent *cloneContent = static_cast<DocumentContent *>(clone->nodeContent_);
  cloneContent->controlDocument_ = clone;

  DocumentContent *contentPtr = static_cast<DocumentContent *>(nodeContent_);
  if (deep && contentPtr->childNodes_) {
    size_t length = contentPtr->childNodes_->getLength();
    for (size_t i = 0; i < length; i++) {
      Node& child = contentPtr->childNodes_->item(i).cloneNode(deep);
      clone->appendChild(child);
    }
  }

  const_cast<Document &>(*this).cacheNode(*clone);
  return *clone;
}

void Document::cacheNode(Node& node) {
  DocumentContent *contentPtr = static_cast<DocumentContent *>(nodeContent_);
  if (node.isValid() && contentPtr->cachedNode_) {
    delete contentPtr->cachedNode_;
    contentPtr->cachedNode_ = &node;
  }
}

void Document::linkAfterDocument(const Document& refNode) {
  nextCopy = refNode.nextCopy;
  const_cast<Document &>(refNode).nextCopy = this;
  previousCopy = const_cast<Document *>(&refNode);
}

void Document::unlinkThisDocument() {
  DocumentContent *contentPtr = static_cast<DocumentContent *>(nodeContent_);
  if (contentPtr->refCount_ == 1) return; //nothing to do

  if (contentPtr->controlDocument_ == this) {
    Document *controlDocument = 0;
    if (previousCopy) controlDocument = previousCopy;
    else controlDocument = nextCopy; // this may be null

    if (controlDocument) { // this is true if we are not unlinking the last copy
      contentPtr->ownerNode_ = controlDocument;
      contentPtr->controlDocument_ = controlDocument;
    }
  }

  if (previousCopy) previousCopy->nextCopy = nextCopy;
  if (nextCopy) nextCopy->previousCopy = previousCopy;
}

Document Document::failDocument("", "#fail");

}

