/***************************************************************************
                          documentmerger.cpp  -  description
                             -------------------
    begin                : Sun Feb 6 2005
    copyright            : (C) 2005 by John Redmond
    email                : john_redmond@optusnet.com.au
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "documentmerger.h"
#include "liststream.h"
#include "url.h"
#include "writer.h"
#include "nodestream.h"
#include "document.h"
#include "element.h"
#include "attr.h"
#include "dombuilder.h"
#include "timer.h"
#include "xslutility.h"
#include "domwriter.h"

namespace limpid {

DocumentMerger::DocumentMerger(bool includeWhitespace):
      includeWhitespace_(includeWhitespace) {
}

Document DocumentMerger::getDocument(const DOMString& urlString) {
  URL url(urlString);
  return getDocument(url);
}

Document DocumentMerger::getDocument(const URL& url) {
  return getMergedDocument(url);
}

Document DocumentMerger::getMergedDocument(const URL& url) {
  DOMBuilder builder(url);
  builder.setProcessingXSLT(includeWhitespace_);
  Document document = builder.build();

  doMerges(document, url, true);  // includes
  doMerges(document, url, false); // imports

  Element& documentElement = document.getDocumentElement();
  if (documentElement.isValid() &&
      !documentElement.getAttributeNode("xml:base").isValid())
    documentElement.setAttribute("xml:base", url.getBasePath());  // may need this path later

  return document;
}

Document DocumentMerger::doMerges(Document& parentDocument, const URL& parentURL,
      bool doInclude) {
  DOMString nodeName = "include";
  if (!doInclude)
    nodeName = "import";

  NodeList matchList = collectByName(parentDocument, nodeName);
  mergeContent(matchList, parentURL, doInclude);

  return parentDocument;
}

void DocumentMerger::mergeContent(NodeList& matchList, const URL& parentURL, bool doInclude) {
  ListStream stream(matchList);
  stream.setForward(doInclude);

  while (true) {
    Node& node = stream.nextNode();//matchList.item(index);
    if (!node.isValid()) break;

    DOMString hrefString = static_cast<Element &>(node).getAttribute("href");
    if (hrefString.empty())
      wrt::Errors << "Missing URL in " << node << wrt::endl;
    else {
      URL childURL(parentURL, hrefString);
      Document mergeDocument = getMergedDocument(childURL); // recurse
      if (doInclude) mergeInclude(node, mergeDocument);
      else mergeImport(node, mergeDocument);
    }
  }
}

// include REPLACES a reference node in the includING document with
// the CHILDREN of the document element of the importED document
void DocumentMerger::mergeInclude(Node& refNode, Document& mergedDocument) {
  Element documentElement = mergedDocument.getDocumentElement();
  if (!documentElement.isValid()) return;

  Node& parentNode = refNode.getParentNode();
  if (!parentNode.isValid()) return;

  Document& hostDocument = refNode.getOwnerDocument();
  ListStream stream(documentElement);

  while (true) {
    Node& node = stream.nextNode();
    if (!node.isValid()) break;

    hostDocument.importNode(node, false); // shallow import
    parentNode.insertBefore(node, refNode);
  }

  removeNode(refNode); // refNode is therefore replaced
}

// import REPLACES a reference node in the includING document with
// the document element of the importED document, inserted at the start
// of the document element.
void DocumentMerger::mergeImport(Node& refNode, Document& mergedDocument) {
  Document& hostDocument = refNode.getOwnerDocument();
  Element& hostDocumentElement = hostDocument.getDocumentElement();

  Element mergedElement = mergedDocument.getDocumentElement();
  hostDocument.importNode(mergedElement, false); // shallow import

  if (mergedElement.isValid()) {
    if (hostDocumentElement.hasChildNodes())
      hostDocumentElement.insertBefore(mergedElement,
          hostDocumentElement.getFirstChild());
    else
      hostDocumentElement.appendChild(mergedElement);
  }
  else
    wrt::Errors << "Cannot replace " << refNode << " with " << mergedElement << wrt::endl;

    removeNode(refNode);
}

NodeList DocumentMerger::collectByName(Document& parentDocument, DOMString& nodeName) {
  NodeList resultList;
  ListStream stream(parentDocument.getDocumentElement());

  while (true) {
    Node& node = stream.nextNode();
    if (!node.isValid()) break;

    Element& element = static_cast<Element &>(node);

    if (XSLUtility::isXSLElement(element) &&
        XSLUtility::getNodeName(element) == nodeName)
      resultList.addNode(node);
  }

  return resultList;
}

void DocumentMerger::removeNode(Node& node) {
  Node& parentNode = node.getParentNode();
  if (parentNode.isValid())
    parentNode.removeChild(node);
}

}
