/***************************************************************************
                          documenttemplate.cpp  -  description
                             -------------------
    begin                : Thu Apr 21 2005
    copyright            : (C) 2005 by John Redmond
    email                : john_redmond@optusnet.com.au
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "documenttemplate.h"
#include "compoundexpr.h"
#include "xpathfactory.h"
#include "url.h"
#include "dombuilder.h"
#include "writer.h"
#include "xpathstring.h"

namespace limpid {

DocumentTemplate::DocumentTemplate(const DOMString& matchValue,
      const DOMString& defaultBase):
      Template(matchValue), defaultBase_(defaultBase), multi_(false), relative_(false) {
  configure(matchValue);
}

void DocumentTemplate::configure(const DOMString& matchValue) {
  Expr expr = XPathFactory::getExpr(matchValue, false);
  CompoundExpr& content = static_cast<CompoundExpr &>(expr.getContent());

  size_t length = content.getLength();
  if (!length) {
    wrt::Errors << "DOCUMENT TEMPLATE ERROR: empty expression" << wrt::endl;
  }

  firstExpr_ = content.getItem(0);
  if (firstExpr_.isNodeSet())
      multi_ = true;
  if (length > 1) {
    secondExpr_ = content.getItem(1);
    relative_ = true;
  }
}

void DocumentTemplate::setResources(XSLResources& resources) {
  resources_ = resources;
}

bool DocumentTemplate::process(XSLContext& parentContext) {
  context_.setParent(parentContext);
  context_.setCurrentNode(parentContext.getCurrentNode());

  if (!multi_) {
    if (!relative_) processOne(parentContext);
    else processTwo(parentContext);
  }
  else processMulti(parentContext);

  return true;
}

bool DocumentTemplate::processNode(XSLContext& parentContext, Node&) {
  return process(parentContext);
}

void DocumentTemplate::processOne(XSLContext& parentContext) {
  DOMString pathString = firstExpr_.getString(parentContext);
  processDocument(getReader(pathString, parentContext), pathString);
}

void DocumentTemplate::processTwo(XSLContext& parentContext) {
  DOMString partString = firstExpr_.getString(parentContext);
  DOMString baseString = secondExpr_.getString(parentContext);

  processDocument(getReader(baseString, partString), partString);
}

void DocumentTemplate::processMulti(XSLContext& parentContext) {
  NodeSet nodeSet = firstExpr_.getNodeSet(parentContext);
  size_t length = nodeSet.getLength();

  for (size_t index = 0; index < length; ++index) {
    DOMString pathString = nodeSet.getItem(index).getValue();
    if (!relative_ || !secondExpr_.isValid())
      processDocument(getReader(pathString, parentContext), pathString);
    else {
      DOMString basePath = secondExpr_.getString(parentContext);
      processDocument(getReader(basePath, pathString), pathString);
    }
  }
}

void DocumentTemplate::processDocument(const Reader& reader, const DOMString& pathString) {
  if (!const_cast<Reader &>(reader).isValid()) {
    wrt::Errors << "No good input for " << pathString << wrt::endl;
    return;
  }
  Document document = DOMBuilder(reader).build();
  context_.setNode(document);

//  selectionTemplate_.process(context_);
}
/*
Document DocumentTemplate::getDocument(const URL& url) {
  DOMBuilder builder(url);
  return builder.build();
}
*/
Reader DocumentTemplate::getReader(const DOMString& pathString,
      XSLContext& parentContext) {
  Reader reader;
try {
  reader = getReader(getBaseURLString(parentContext.getNode()), pathString);
  if (reader.isValid()) return reader;

  return getReader(defaultBase_, pathString);
} catch (...) {
  wrt::Errors << "DocumentTemplate FAIL " << pathString << wrt::endl;
}
  return reader;

}

Reader DocumentTemplate::getReader(const DOMString& baseString, const DOMString& partString) {
  return Reader(getFullPath(baseString, partString));
}

DOMString DocumentTemplate::getBaseURLString(const Node& node) {
  return
    node.getOwnerDocument().getDocumentElement().getAttribute("xml:base");
}

DOMString DocumentTemplate::getFullPath(const DOMString& basePath, const DOMString& relPath) {
  DOMString fullPath = basePath;
  if (!basePath.endsWith("/"))
    fullPath += "/";
  fullPath += relPath;
  return fullPath;
}

}


