/***************************************************************************
                          documenttype.cpp  -  description
                             -------------------
    begin                : Tue Mar 18 2003
    copyright            : (C) 2003 by John Redmond
    email                : john_redmond@optusnet.com.au
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "documenttype.h"
#include "element.h"
#include "entity.h"
#include "notation.h"
#include "nodelist.h"
#include "nodestream.h"
#include "liststream.h"
#include "stringreader.h"
#include "reader.h"
#include "writer.h"
#include "dombuilder.h"
#include "domwriter.h"

namespace limpid {

DocumentType::DocumentType(const DOMString &name,
      const DOMString &pid, const DOMString &sid):
  Node(BASE_NODE) {
  try {
    nodeContent_ = new DocumentTypeContent(name, pid, sid);
  }
  catch (std::bad_alloc) {
    delete nodeContent_;
    throw;
  }
}

DocumentType::DocumentType(const DocumentType& refNode): Node(refNode) {
}

Node& DocumentType::operator=(const DocumentType& refNode) {
  replaceContent(refNode);
  return *this;
}

DOMString DocumentType::getName() const {
  DocumentTypeContent *contentPtr = static_cast<DocumentTypeContent *>(nodeContent_);
  return contentPtr->nodeName_;
}
DOMString DocumentType::getPublicId() const {
  DocumentTypeContent *contentPtr = static_cast<DocumentTypeContent *>(nodeContent_);
  return contentPtr->publicId_;
}

DOMString DocumentType::getSystemId() const {
  DocumentTypeContent *contentPtr = static_cast<DocumentTypeContent *>(nodeContent_);
  return contentPtr->systemId_;
}

DOMString DocumentType::getInternalSubset() const {
  DocumentTypeContent *contentPtr = static_cast<DocumentTypeContent *>(nodeContent_);
  return contentPtr->internalSubset_;
}

void DocumentType::setInternalSubset(DOMString& is) {
  DocumentTypeContent *contentPtr = static_cast<DocumentTypeContent *>(nodeContent_);
  contentPtr->internalSubset_ = is.copyChars();
}

NamedNodeMap& DocumentType::getUnparsedEntities() {
  DocumentTypeContent *contentPtr = static_cast<DocumentTypeContent *>(nodeContent_);
  if (!contentPtr->constrained_) configure();
  return *contentPtr->unparsedEntities_;
}

const NamedNodeMap& DocumentType::getNotations() {
  DocumentTypeContent *contentPtr = static_cast<DocumentTypeContent *>(nodeContent_);
  if (!contentPtr->constrained_) configure();
  return *contentPtr->notations_;
}

const NamedNodeMap& DocumentType::getEntities() {
  DocumentTypeContent *contentPtr = static_cast<DocumentTypeContent *>(nodeContent_);
  if (!contentPtr->constrained_) configure();
  return *contentPtr->unparsedEntities_;
}

const StringMap& DocumentType::getIDMap() {
  DocumentTypeContent *contentPtr = static_cast<DocumentTypeContent *>(nodeContent_);
  if (!contentPtr->constrained_) configure();
  return *contentPtr->idMap_;
}

const StringMap& DocumentType::getEntityMap() {
  DocumentTypeContent *contentPtr = static_cast<DocumentTypeContent *>(nodeContent_);
  if (!contentPtr->constrained_) configure();
  return *contentPtr->entityMap_;
}

void DocumentType::configure() {
  DocumentTypeContent *contentPtr = static_cast<DocumentTypeContent *>(nodeContent_);
  if (!contentPtr->constrained_) {
    processConstraints();
    contentPtr->constrained_ = true;
  }
}

DOMString DocumentType::getNodeName() const {
  DocumentTypeContent *contentPtr = static_cast<DocumentTypeContent *>(nodeContent_);
  return contentPtr->nodeName_;
}

void DocumentType::setSourceURL(URL& url) {
  DocumentTypeContent *contentPtr = static_cast<DocumentTypeContent *>(nodeContent_);
  contentPtr->sourceURL_ = new URL(url);
}

void DocumentType::processConstraints() {
  DocumentTypeContent *contentPtr = static_cast<DocumentTypeContent *>(nodeContent_);
  DOMString internalSubset = contentPtr->internalSubset_;
  if (!internalSubset.empty()) {
    StringReader internalStream(internalSubset);
    DTDConverter converter(internalStream);
    Document doc = converter.process();
    applySchema(doc);
  }

  // process from the SYSTEM URL; PUBLIC URLs are not yet supported
  DOMString sourceString = contentPtr->systemId_;
  if (!sourceString.empty()) {
    URL sourceURL = *contentPtr->sourceURL_;
    URL externalURL(sourceURL, sourceString);
    if (externalURL.isValid())
      applyExternalSource(externalURL);
  }
}

void DocumentType::applyExternalSource(URL& externalURL) {
  DOMString fullPath = externalURL.getPath();
  if (fullPath.endsWith(".dtd")) { // a DTD file is identified
    Reader reader(externalURL);
    DTDConverter converter(reader);//fullPath);
    Document& doc = converter.process();
    applySchema(doc);
  }
  else if (fullPath.endsWith(".xsd")) {
    DOMBuilder builder(fullPath); // we are working with a Schema
    Document doc = builder.build();
    applySchema(doc);
  }
  // else nothing
}

void DocumentType::applySchema(Document& doc) {
  DocumentTypeContent *contentPtr = static_cast<DocumentTypeContent *>(nodeContent_);
///  contentPtr->validationProcessor_->applySchema(doc);
  setEntities(doc);
  setIDs(doc);
}

void DocumentType::setEntities(Document& sd) {
  DocumentTypeContent *contentPtr = static_cast<DocumentTypeContent *>(nodeContent_);
  NamedNodeMap *unparsedEntities = contentPtr->unparsedEntities_;
  NamedNodeMap *notations = contentPtr->notations_;
  StringMap *entityMap = contentPtr->entityMap_;

  NodeList& children = sd.getDocumentElement().getChildNodes();
  size_t length = children.getLength();
  ListStream stream(sd.getDocumentElement());

  while (true) {
    Node& node = stream.nextNode();
    if (!node.isValid()) break;

    if (!isXSDElement(node)) continue;

    if (node.getLocalName() == "entity") {
      Element& element = static_cast<Element &>(node);
      DOMString entityName = element.getAttribute("name");
      if (!entityName.size()) continue; // cannot work with this

      DOMString value = element.getAttribute("value");
      if (value.size())
        entityMap->insert(make_pair(entityName, value));
      else {
        // it is an unparsed entity
        DOMString publicId = element.getAttribute("public");
        DOMString systemId = element.getAttribute("system");
        DOMString notationName = element.getAttribute("notation");
        Entity entity(entityName, publicId, systemId, notationName);
        unparsedEntities->setNamedItem(entity);
      }
    }
    else if (node.getLocalName() == "notation") {
      Element& element = static_cast<Element &>(node);
      DOMString notationName = element.getAttribute("name");
      DOMString publicId = element.getAttribute("public");
      DOMString systemId = element.getAttribute("system");
      Notation notation(notationName, publicId, systemId);
      notations->setNamedItem(notation);
    }
  }
}

void DocumentType::setIDs(Document& sd) {
  DocumentTypeContent *contentPtr = static_cast<DocumentTypeContent *>(nodeContent_);
  StringMap *idMap = contentPtr->idMap_;

  NodeStream stream(sd, false, false);  // no end nodes and do not include document

  while (true) {
    Node& node = stream.nextNode();
    if (!node.isValid()) break;

    if (!isXSDElement(node)) continue;
    DOMString nodeName = node.getLocalName();

    if (nodeName != "attribute") continue;

    Element& attrElement = static_cast<Element &>(node);
    DOMString attrType = attrElement.getAttribute("type");
    if (attrType != "xs:ID" && attrType != "ID") continue;

    DOMString attrName = attrElement.getAttribute("name");
    Element& elementElement = static_cast<Element &>(node.getParentNode());
    DOMString elementName = elementElement.getAttribute("name");
    if (!elementName.size()) continue;  // should never happen

    idMap->insert(make_pair(elementName, attrName));
  }
}

DOMString& DocumentType::resolveEntities(DOMString& string) {
  DocumentTypeContent *contentPtr = static_cast<DocumentTypeContent *>(nodeContent_);
///  if (!contentPtr->constrained_) configure();
  StringMap *map = contentPtr->entityMap_;
  if (!map->size()) return string;

  StringMap::iterator p = map->begin();

  while (true) {
    if (p == map->end()) break;

    string.resolveEntities(); // the standard entities
    replaceReference(string, p->first, p->second);
    ++p;
  }
  return string;
}

void DocumentType::replaceReference(DOMString& string,
    const DOMString& key, const DOMString& value) {
  // use a loop to handle multiple references
  while (true) {
    size_t hitIndex = string.indexOf(key);
    if (hitIndex == string.npos) return;

    size_t keySize = key.size();

    if (!hitIndex ||
      string[hitIndex - 1] != '&' ||
      hitIndex + keySize >= string.size() ||
      string[hitIndex + keySize + 1] != ';') return;

    string.remove(hitIndex - 1, key.size() + 2);
    string.insertString(hitIndex,value);
  }
}

DOMString DocumentType::getDisplayString() const {
  DocumentTypeContent *contentPtr = static_cast<DocumentTypeContent *>(nodeContent_);

  DOMString s("<!DOCTYPE ");
  s += contentPtr->nodeName_;

  if (contentPtr->publicId_) {
    s += " PUBLIC \"";
    s += contentPtr->publicId_;
    s += "\" \"";

    if (contentPtr->systemId_) { // there really should be one!
      s += contentPtr->systemId_;
      s += "\"";
    }
  }

  else if (contentPtr->systemId_) {
    s += " SYSTEM \"";
    s += contentPtr->systemId_;
    s += "\"";
  }

  if (contentPtr->internalSubset_) {
    s += " [";
    s += contentPtr->internalSubset_;
    s += "]";
  }

  s += ">";

  return s;
}

DocumentType& DocumentType::getDefaultDocumentType() {
  static DocumentType defaultDocumentType("default", "", "");
  return defaultDocumentType;
}

Node& DocumentType::cloneNode(bool) const {
  DocumentTypeContent *contentPtr = static_cast<DocumentTypeContent *>(nodeContent_);

  DOMString nodeName = contentPtr->nodeName_;
  DOMString publicId = contentPtr->publicId_;
  DOMString systemId = contentPtr->systemId_;
  DOMString internalSubset = "";
  DocumentType *clone = new DocumentType(nodeName, publicId, systemId);
  clone->nodeContent_->documentContent_ = contentPtr->documentContent_;

  getOwnerDocument().cacheNode(*clone);
  return *clone;
}

bool DocumentType::isXSDElement(Node& node) {
  if (node.getNodeType() != Node::ELEMENT_NODE) return false;
  Element& element = static_cast<Element &>(node);
  return element.getNamespaceURI() == "http://www.w3c.org/2001/XMLSchema";
}

}
