/***************************************************************************
                          domstring.cpp  -  description
                             -------------------
    begin                : Wed Jul 2 2003
    copyright            : (C) 2003 by John Redmond
    email                : john_redmond@optusnet.com.au
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <iostream>
#include <algorithm>
#include <map>
#include <ctype.h>
#include "domstring.h"
#include "node.h"
#include "writer.h"

using namespace std;

namespace limpid {
bool DOMString::operator==(const char* s) const {
  size_t len = strlen(s);
  if (size() != len) return false;
  return equal(begin(), end(), s);
}

bool DOMString::operator!=(const char* s) const {
  size_t len = strlen(s);
  if (size() != len) return true;
  return !equal(begin(), end(), s);
}

const char* DOMString::getCString() const {
  size_t len = size();
  if (!len) return 0; // null pointer for empty string

  char *cstring = new char[len + 1];
  char *cstart = cstring;

  DOMString::const_iterator p = begin();
  while (p != end()) *cstart++ = static_cast<char>(*p++);
  *cstart = 0;

  return cstring; // may be a null pointer if no heap space
}

size_t DOMString::indexOf(const DOMString& ss, size_t startPos) const {
  return find(ss, startPos);
}

size_t DOMString::lastIndexOf(const DOMString& ss, size_t startPos) const {
  return rfind(ss, startPos);
}

bool DOMString::contains(const DOMString& ss) const {
  if (size() < ss.size()) return false;
  DOMString::size_type pos = find(ss);
  return pos != npos;
}

bool DOMString::startsWith(const DOMString& ss) const {
  size_t mainLength = size();
  if (!mainLength) return false;

  size_t subLength = ss.size();
  if (mainLength < subLength) return false;

  return compare(0, subLength, ss) == 0;
}

bool DOMString::endsWith(const DOMString& ss) const {
  size_t mainLength = size();
  if (!mainLength) return false;

  size_t subLength = ss.size();
  if (mainLength < subLength) return false;

  return compare(mainLength-subLength, subLength, ss) == 0;
}

DOMString DOMString::substring(size_t start, size_t len) const {
  size_t limit = start + len;
  size_t total = size();
  if (limit > total) limit = total;

  DOMString result;
  if (!len) return result;

//  result.reserve(len);
  std::copy(begin() + start, begin() + limit, back_inserter(result));
  return result;
}

DOMString& DOMString::trim() {
  size_t startIndex = 0, endIndex = size();
  if (!endIndex) return *this;

  while (startIndex < endIndex && isWhitespace(at(startIndex)))
    ++startIndex;
  while (--endIndex)
    if (!isWhitespace(at(endIndex))) break;

  ++endIndex;

  DOMString result = substr(startIndex, endIndex - startIndex);
  swap(result);

  return *this;
}

DOMString& DOMString::normalizeWhitespace() {
  trim();
  return normalizeGaps();
}

inline DOMString& DOMString::normalizeGaps() {
  bool inGap = false;
  iterator p = begin();

  while (p != end()) {
    if (iswspace(*p)) {
      if (inGap) erase(p);
      else  {
        *p++ = ' ';
        inGap = true;
      }
    }
    else {
      ++p;
      inGap = false;
    }
  }
  return *this;
}

DOMString& DOMString::change(const DOMString& oldString, const DOMString& newString) {
  size_t startIndex = 0;
  DOMString resultString;

  while (true) {
    size_t pos = find(oldString, startIndex);
    if (pos == npos) {
      resultString += *this;
      break;  // nothing to replace
    }

    resultString += substr(0, pos);
    resultString += newString;
    erase(0, pos + oldString.size());
  }

  swap(resultString);

  return *this;
}

DOMString& DOMString::change(size_t offset, size_t count, const DOMString& newString) {
  if ((offset > size()) || ((offset + count) > size()))  append(newString);
  else
    replace(offset, count, newString);

  return *this;
}

DOMString& DOMString::insertString(size_t pos, const DOMString& newString) {
  if (pos >= size()) append(newString);
  else
    insert(begin() + pos, newString.begin(), newString.end());

  return *this;
}

DOMString& DOMString::remove(size_t offset, size_t count) {
  erase(offset, count);
  return *this;
}

DOMString& DOMString::toLowerCase() {
  transform(begin(), end(), begin(), ::tolower);
  return *this;
}

DOMString& DOMString::toUpperCase() {
  transform(begin(), end(), begin(), ::toupper);
  return *this;
}

DOMString& DOMString::capitalize() {
  if (!empty()) at(0) = toupper(at(0));
  return *this;
}

DOMString& DOMString::reverse() {
  std::reverse(begin(), end());
  return *this;
}

DOMString DOMString::getDelimitedString(size_t startPos) {
  Char delimiter = at(startPos);
  Char openingDelimiter = delimiter;

  if (delimiter == '(') delimiter = ')';
  else if (delimiter == '[') delimiter = ']';
  else if (delimiter == '{') delimiter = '}';

  size_t limit = size();
  size_t pos = startPos, nestings = 0;

  for (size_t index = startPos; index < limit; ++index) {
    if (at(index) == openingDelimiter) ++nestings;
    else if (at(index) == delimiter) --nestings;
    ++pos;
    if (!nestings) return substring(startPos, pos - startPos);
  }

  // we reach here only if we could not find a string with balanced delimiters
  return substring(startPos, limit - startPos);
}

int DOMString::toInt(int base) const {
  bool extended = (base > 10);
  int result = 0;

  const_iterator p = begin();
  while (p != end()) {
    int c = toupper(*p++);
    if (isdigit(c)) result = result * base + c - '0';
    else if (extended && isxdigit(c)) {
      int extra = c - 'A' + 10;
      if (extra < base) result = result * base  + extra;
      else return ConversionError;
    }
    else return ConversionError;
  }
  return result;
}

Char *DOMString::copyChars() const {
  if (empty()) return 0; // null pointer

  Char *cpy = new Char[size() + 1];
  cpy[this->copy(cpy, this->npos)] = 0;

  return cpy; // may be a null pointer if no heap space
}

DOMString &DOMString::appendASCII(const char *str) {
  if (!str) return *this;

  while (*str) push_back((Char)*str++);
  return *this;
}

DOMString& DOMString::escapeBra() {
  static DOMString lt_lit = "<";
  static DOMString lt = "&lt;";

  while (true) {
    size_t index = indexOf(lt_lit);
    if (index == npos) break;

    DOMString::iterator pos = begin() + index;
    change(index, 1, lt);
  }

  return *this;
}

DOMString& DOMString::escapeDelimiters() {
  static DOMString apos = "&apos;";
  static DOMString quot = "&quot;";

  while (true) {
    size_t index = indexOf("'");
    if (index == npos) break;

    DOMString::iterator pos = begin() + index;
    replace(pos, pos + 1, apos.begin(), apos.end());
  }

  while (true) {
    size_t index = indexOf("\"");
    if (index == npos) break;

    DOMString::iterator pos = begin() + index;
    replace(pos, pos + 1, quot.begin(), quot.end());
  }

  return *this;
}

DOMString& DOMString::escapeTex() {
  resolveEntities(true);  // need this for converting HTML to TeX
  // Latex is very fussy about blank lines and (invisible) carriage return characters
  change("\r\n\r\n", "\r\n");
  change("\n\n", "\n");

  static DOMString escapedChars ="\\&$%{}#^";
  const Char insertedChar = '\\';
  size_t length = size();
  DOMString newString;

  for (size_t index = 0; index < size(); ++index) {
    Char c = (*this)[index];

    if (c == '[') {
      newString += "{\\lbrack}";
      continue;
    }

    else if (c == ']') {
      newString += "{\\rbrack}";
      continue;
    }

    if (c == '%') {
      newString += "{\\%}";
      continue;
    }

    if (c == '{') {
      newString += "{\\textbraceleft}";
      continue;
    }

    else if (c == '}') {
      newString += "{\\textbraceright}";
      continue;
    }

    else if (c == '<') {
      newString += "{\\textless}";
      continue;
    }

    else if (c == '>') {
      newString += "{\\textgreater}";
      continue;
    }

    else if (c == '_') {
      newString += "{\\textunderscore}";
      continue;
    }

    else if (c == '|') {
      newString += "{\\textbar}";
      continue;
    }

    else if (c == '\\') {
      newString += "\\textbackslash ";
      continue;
    }

    else if (c == '"') {
      newString += "''";
      continue;
    }

    else if (escapedChars.find(c) != npos)
      newString.push_back('\\'); // add escape char

    // finally, add the char
    newString.push_back(c);
  }

  trim();
  newString.change("LaTeX", "\\LaTeX\\/");
  swap(newString);

  return *this;
}

DOMString& DOMString::resolveBra() {
  static DOMString ltString = "&lt;";
  static DOMString lt = "<";
  return change(ltString, lt);
}

DOMString& DOMString::resolveEntities(bool includeLT) {
  if (!size()) return *this;

  while (doResolve(*this, includeLT) || doExtendedChar(*this)) {}
  return *this;
}

bool DOMString::doResolve(DOMString& string, bool includeLT) {
  static DOMString ampString = "&amp;";
  static DOMString ltString = "&lt;";
  static DOMString gtString = "&gt;";
  static DOMString aposString = "&apos;";
  static DOMString quotString = "&quot;";
  static DOMString nbspString = "&nbsp;";

  size_t index;

  index = string.indexOf(gtString);
  if (index != string.npos) {
    string.replace(index, 4, 1, Char('>'));
    return true;
  }

  index = string.indexOf(aposString);
  if (index != string.npos) {
    string.replace(index, 6, 1, Char('\''));
    return true;
  }

  index = string.indexOf(quotString);
  if (index != string.npos) {
    string.replace(index, 6, 1, Char('"'));
    return true;
  }

  if (includeLT) {
    index = string.indexOf(ltString);
    if (index != string.npos) {
      string.replace(index, 4, 1, Char('<'));
      return true;
    }

  index = string.indexOf(ampString);
  if (index != string.npos) {
    string.replace(index, 5, 1, Char('&'));
    return true;
  }

  index = string.indexOf(nbspString);
  if (index != string.npos) {
    string.replace(index, 5, 1, Char(' '));
    return true;
  }
  }

  return false;
}

bool DOMString::doExtendedChar(DOMString& string) {
  size_t index = string.indexOf("&#");
  if (index == string.npos) return false;

  size_t endIndex = string.indexOf(';', index + 1);
  if (endIndex == string.npos) return false;

  DOMString charString = string.substr(index + 1, endIndex - index - 1);
  Char c = Char(getExtendedChar(charString)); // this may unmask an amp
  string.replace(index, endIndex - index + 1, 1, c);

  return true;
}

inline int DOMString::getExtendedChar(DOMString& entityName) {
  if (entityName.size() < 2|| entityName[0] != '#') return 0;

  if (entityName[1] == 'x' || entityName[1] == 'X') {
    entityName.erase(0, 2);
    return entityName.toInt(16);
  }
  else {
    entityName.erase(0, 1);
    return entityName.toInt(10);
  }
}

DOMString DOMString::failString;

const DOMString operator+(const DOMString& lhs, const DOMString& rhs) {
  return DOMString(lhs) += rhs; }
#ifndef _NARROW_
const DOMString operator+(const DOMString& lhs, const Char* rhs) {
  return DOMString(lhs) += rhs; }
#endif
const DOMString operator+(const DOMString& lhs, const char* rhs) {
  return DOMString(lhs) += rhs; }

}
