/***************************************************************************
 *   Copyright (C) 2007 by root   *
 *   john_redmond@optusnet.com.au   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "domutils.h"
#include "element.h"
#include "text.h"
#include "comment.h"
#include "domstring.h"
#include "liststream.h"
#include "writer.h"

namespace limpid {

Node& DOMUtils::getChild(const DOMString& name, Node& parentNode) {
  ListStream stream(parentNode);

  while (true) {
    Node& node = stream.nextNode();
    if (!node.isValid()) break;

    if (node.getNodeName() == name)
      return node;

  }

  return Node::failNode;
}

NodeList DOMUtils::getChildren(const DOMString& name, Node& parentNode) {
  ListStream stream(parentNode);
  NodeList resultList;

  while (true) {
    Node& node = stream.nextNode();
    if (!node.isValid()) break;

    if (node.getNodeName() == name)
      resultList.addNode(node);

  }

  return resultList;
}

int DOMUtils::getParentPosition(const Node& node) { // position in the parent's node list
  Node& parentNode = node.getParentNode();
  if (!parentNode.isValid()) return -1;

  NodeList nodeList = node.getChildNodes();
  for (int index = 0; index < nodeList.getLength(); index++)
    if (nodeList.item(index) == node) return index;

  return -1; // impossible? 
}

Element& DOMUtils::addElement(Element& parentNode, const DOMString& elementName) {
  Element childElement(elementName);
  parentNode.appendChild(childElement);
  return static_cast<Element &>(parentNode.getLastChild());
}

Element& DOMUtils::addLeaf(const DOMString& name, const DOMString& value, Node& parentNode) {
  if (value.empty()) return Element::failElement;  // do not add an empty leaf
  Element nameElement(name);
  parentNode.appendChild(nameElement);
  Text nameText(value);
  nameElement.appendChild(nameText);
  return static_cast<Element &>(parentNode.getLastChild());
}

void DOMUtils::addText(Element& parentElement, const DOMString& textString) {
  Text text(textString);
  parentElement.appendChild(text);
} 

void DOMUtils::addComment(Element& parentElement, const DOMString& commentString) {
  Comment comment(commentString);
  parentElement.appendChild(comment);
}

void DOMUtils::moveContent(Element& sourceElement, Element& destElement) {
  while (true) {
    Node& node = sourceElement.getFirstChild();
    if (!node.isValid()) break;

    destElement.appendChild(node);
  }
}

void DOMUtils::spillContent(Node& node) {
  Node& parentNode = node.getParentNode();
  if (!parentNode.isValid()) return;  //cannot proceed

  while (true) {
    Node& childNode = node.getFirstChild();
    if (!childNode.isValid()) break;

    parentNode.insertBefore(childNode, node);
  }

  parentNode.removeChild(node);
}


bool DOMUtils::isLeaf(const Node& node) {
  return node.getNodeType() == Node::ELEMENT_NODE &&
    node.getChildNodes().getLength() == 1 &&
    node.getFirstChild().getNodeType() == Node::TEXT_NODE;
}

bool DOMUtils::hasText(const Node& parentNode) {
  if (parentNode.getNodeType() != Node::ELEMENT_NODE) return false;

  ListStream childStream(const_cast<Node &>(parentNode));

  while (true) {
    Node& node = childStream.nextNode();
    if (!node.isValid()) return false;

    if (node.getNodeType() == Node::TEXT_NODE) return true;
  }
}

bool DOMUtils::hasMixedContent(Node& parentNode) {
  if (parentNode.getNodeType() != Node::ELEMENT_NODE) return false;
 
  bool hasText = false, hasNonText = false;
  ListStream childStream(parentNode);

  while (true) {
    Node& node = childStream.nextNode();
    if (!node.isValid()) return false;

    int nodeType = node.getNodeType();
    if (nodeType == Node::TEXT_NODE) hasText = true;
    else if (nodeType == Node::ELEMENT_NODE) hasNonText = true;

    if (hasText && hasNonText) return true;
  }
}

int DOMUtils::getDepth(Node& node) {
  int depth = 1;
  Node* activeNode = &node;
  while (true) {
    activeNode = &activeNode->getParentNode();
    if (activeNode->getNodeType() != Node::ELEMENT_NODE) break;

    ++depth;
  }
  return depth;
}

int DOMUtils::getPosition(Node& node) {
  int position = 0;
  Node* activeNode = &node;
  while (true) {
    activeNode = &activeNode->getPreviousSibling();
    if (!activeNode->isValid()) break;

    ++position;
  }
  return position;
}

bool DOMUtils::isXHTMLNode(Node& node) {
  if (node.getNodeType() != Node::ELEMENT_NODE) return false;
  return isXHTMLName(node.getNodeName());
}

bool DOMUtils::isXHTMLName(const DOMString& nodeName) {
  static DOMString xhtmlNames = "html head body h1 h2 h3 h4 h5 h6 em emph bold div p br ol ul li dl dt dd table header tr td a pre";
  return xhtmlNames.indexOf(nodeName) != -1;
}
}
