/***************************************************************************
 *   Copyright (C) 2007 by John Redmond   *
 *   john_redmond@optusnet.com.au   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "dotparser.h"
#include "attr.h"
#include "comment.h"
#include "processinginstruction.h"
#include "xmldeclaration.h"
#include "documenttype.h"
#include "cdatasection.h"
#include "writer.h"
#include "stringreader.h"
#include "tokenstream.h"

namespace limpid {

DotParser::DotParser(const DOMString& fileName): endNode_(new EndNode("END")),
  reader_(new Reader(fileName)), lastElement_(0), rootName_("root"),
  blockDepth_(0), currentDepth_(0), state_(START), lineNo_(1) {}

DotParser::DotParser(const Reader& reader): endNode_(new EndNode("END")),
  reader_(new Reader(reader)), lastElement_(0), rootName_("root"),
  blockDepth_(0), currentDepth_(0), state_(START), lineNo_(1) {}

Node& DotParser::nextNode() {
  while (true) {
    skipWhitespace();

    // deal with the termination conditions first:
    if (state_ == DONE) {
      return Node::failNode;
    }

    if (state_ == FINISH) {
      if (!tagStack_.empty()) {
        state_ = DENEST;
        continue;
     }
     else {
       state_ = DONE;
       continue;
     }
   }

   if (state_ == PARTIALDENEST) {
     if (nestingStack_.size() && tagStack_.size() && currentDepth_ <= nestingStack_.back()) { 
       node_ = new EndNode(tagStack_.back());
       popStacks();

       if (nestingStack_.empty() || nestingStack_.back() <= currentDepth_)
         state_ = BLOCKREADY;

       return *node_;
     }
     else {
       nestingStack_.push_back(currentDepth_);
       state_ = BLOCKREADY;
       continue;
     }
   }

   if (state_ == DENEST) {
     if (tagStack_.empty()) {
      state_ = DONE;
      continue;
     }
     if (tagStack_.size()) {
       node_ = new EndNode(tagStack_.back());
       tagStack_.pop_back();
       node_->setPosition(lineNo_);
       return *node_; // do not change state
     }
     else {
       state_ = ANY;
       continue;
     }
   }

    // allow a start, even it it turns out to have no data
   if (state_ == START) {
     pushElement(rootName_, true);
     root_ = lastElement_;
     state_ = ANY;
     node_->setPosition(lineNo_);
     return *node_;
   }

   if (peekChar() == FAILCHAR) { // when input is exhausted, unpack the nesting
     state_ = FINISH;
     continue;
   }

   if (state_ == ANY) {
     setAny();
     continue;
   }

   if (state_ == BLOCKSTART) {
     state_ = PARTIALDENEST;
     continue;
   }

   if (state_ == TAGSTART) {
     setTagStart();
     node_->setPosition(lineNo_);
     return *node_;
   }

   if (state_ == ATTRIBUTESTART) {
     setAttribute();
     continue;
   }

   if (state_ == TEXTSTART) {
     activeString_ = getTextString();
     node_ = new Text(activeString_);
     state_ = ANY;
     node_->setPosition(lineNo_);
     return *node_;
   }

   if (state_ == COMMENTSTART) {
     DOMString commentString = getHashString().trim();
     node_ = new Comment(commentString);
     state_ = ANY;
     node_->setPosition(lineNo_);
     return *node_;
   }

   if (state_ == PISTART) {
     DOMString piName = getWord().trim();
     if (piName.toLowerCase() == "xml") {
       state_ = XMLDECLARATIONSTART;
       continue;
     }

     DOMString piString = getHashString();
     node_ = new ProcessingInstruction(piName, piString);
     state_ = ANY;
     node_->setPosition(lineNo_);
     return *node_;
   }

   if (state_ == XMLDECLARATIONSTART) {
     setXMLDeclaration();
     state_ = ANY;
     continue;
   }

   if (state_ == DOCUMENTTYPESTART) {
     setDocumentType();
     state_ = ANY;
     continue;
   }

   if (state_ == CDATASTART) {
     DOMString cdataString = getHashString();
     node_ = new CDATASection(cdataString);
     state_ = ANY;
     node_->setPosition(lineNo_);
     return *node_;
   }

   if (state_ == BLOCKREADY) {
     setBlock(); 
     node_->setPosition(lineNo_);
     return *node_;
   }

   if (state_ == HEADINGSTART) {
     if (blockName_.empty() || blockName_.startsWith("(")) { // no heading after the dots
       state_ = ANY;
       continue;
     }
     DOMString headingName = "h";
     if (!nestingStack_.empty())
      headingName += Converter::toString(nestingStack_.back());
     else
      headingName += "6";

     pushElement(headingName, true);
     state_ = HEADINGCONTENT;
     node_->setPosition(lineNo_);
     return *node_;
   }

   if (state_ == DEFAULTSTART) {
     pushElement("p", true);
     state_ = CONTENT;
     return *node_;
   }

   if (state_ == HEADINGCONTENT) {
     node_ = new Text(blockName_);
     state_ = CONTENTEND;
     node_->setPosition(lineNo_);
     return *node_;
   }

   if (state_ == CONTENT) {
     setContent();
     node_->setPosition(lineNo_);
     return *node_;
   }

   if (state_ == EMPTY) {
     state_ = ANY;
     if (tagStack_.size())
     tagStack_.pop_back();
     continue;
   }

   if (state_ == CONTENTEND) {
     node_ = new EndNode(tagStack_.back());
     state_ = ANY;
     if (tagStack_.size())
       tagStack_.pop_back();
     node_->setPosition(lineNo_);
     return *node_;
   }

   if (state_ == TAGEND) {
     node_ = new EndNode(tagStack_.back());
     getLine(); // swallow the rest of the line
     state_ = ANY;
     if (tagStack_.size())
       tagStack_.pop_back();
     node_->setPosition(lineNo_);
     return *node_;
   }
 }
}

void DotParser::setAny() {
   Char c = peekChar();

   if (c == '.') {
     activeString_ = getDotString();
     currentDepth_ = activeString_.getLength();
     blockName_ = getLine().normalizeWhitespace();
     state_ = BLOCKSTART;
   }

   else if (c == '_') {
     DOMString tagString = getUSString();
     if (tagString.getLength() == 1)
      state_ = TAGSTART;
     else
      state_ = TAGEND;
   }

   else if (c == '=') {
     reader_->nextChar(); // swallow the char
     state_ = TEXTSTART;
   }

   else if (c == '#') {
     reader_->nextChar(); // swallow the char
     state_ = COMMENTSTART;
   }

   else if (c == '@') {
     reader_->nextChar(); // swallow the char
     state_ = ATTRIBUTESTART;
   }

   else if (c == '?') {
     reader_->nextChar(); // swallow the char
     state_ = PISTART;
   }

   else if (c == '[') {
     reader_->nextChar(); // swallow the char
     state_ = CDATASTART;
   }

   else if (c == '!') {
     reader_->nextChar(); // swallow the char
     state_ = DOCUMENTTYPESTART;
   }

   else {
     state_ = DEFAULTSTART;
   }
}

void DotParser::setXMLDeclaration() {
  DOMString decString = getLine();
  StringReader stringReader(decString);
  TokenStream tis(stringReader);
  tis.setDelimiters("");
  tis.setWhitespaceChars("|");

  DOMString versionString = tis.nextToken();
  DOMString encodingString = tis.nextToken();
  DOMString standaloneString = tis.nextToken();

  XMLDeclaration xmlDeclaration(versionString, encodingString, standaloneString);
  root_->getParentNode().insertBefore(xmlDeclaration, *root_);
}

void DotParser::setDocumentType() {
  DOMString dtString = getHashString();
  StringReader stringReader(dtString);
  TokenStream tis(stringReader);
  tis.setDelimiters("");
  tis.setWhitespaceChars("|");

  DOMString nameString = tis.nextToken();
  DOMString publicString = tis.nextToken();
  DOMString privateString = tis.nextToken();
  DOMString subsetString = tis.nextToken();

  DocumentType docType(nameString, publicString, privateString);//, "four");
  if (!subsetString.empty())
    docType.setInternalSubset(subsetString);
  root_->getParentNode().insertBefore(docType, *root_);
}

void DotParser::setBlock() {
DOMString attrString = blockName_;
  DOMString attributeName = "class";
  bool hasHeading = false;

  if (attrString.startsWith("(")) {
    attrString = attrString.substr(1, attrString.getLength() - 2);
  }
  else if (attrString.startsWith("[")) {
    attributeName = "id";
    attrString = attrString.substr(1, attrString.getLength() - 2);
  }
  else {
    attrString = "L";
    attrString += Converter::toString(currentDepth_);
    hasHeading = true;
  }

  pushElement("div", attributeName, attrString);
  nestingStack_.push_back(currentDepth_);

  if (!blockName_.empty() && hasHeading)
  state_ = HEADINGSTART;
  else
  state_ = ANY;
}

void DotParser::setContent() {
  if (tagStack_.back() == "pre" || tagStack_.back() == "script") {
    activeString_ = getHashString();
    activeString_.change("<", "&lt;");
    activeString_.change("||", "\n");
  }
  else {
    activeString_ = getTextString();
  }

  node_ = new Text(activeString_);
  state_ = CONTENTEND;
}

void DotParser::setTagStart() {
  DOMString tagName = getDelimitedString(" =/\n").trim();
  Char peekedChar = peekChar();

  bool isParent = (peekedChar != '/');
  if (tagName.startsWith("(")) {
    tagName = tagName.substr(1, tagName.getLength() - 2);
    pushElement("div", "class", tagName);
    state_ = ANY;
  }
  else if (tagName.startsWith("[")) {
    tagName = tagName.substr(1, tagName.getLength() - 2);
    pushElement("div", "id", tagName);
    state_ = ANY;
  }
  else
  pushElement(tagName, isParent);

  if (peekedChar == '=') {
  reader_->nextChar();  // drop the '='
  state_ = CONTENT;
  }
  else if (peekedChar == '/') {
    reader_->nextChar();  // drop the '/'
    state_ = EMPTY;
  }
  else state_ = ANY; // expect some child nodes
}

void DotParser::setAttribute() {
  DOMString attrName = getDelimitedString(" =").trim();
  DOMString attrValue = "true"; 
  Char c = reader_->nextChar();

  if (c == '=') {
    DOMString valueString = getAttributeString();
    if (!valueString.empty())
      attrValue = valueString; // we want a non-empty attribute value
  }

  if (lastElement_ != 0)
    lastElement_->setAttribute(attrName, attrValue);

   state_ = ANY;
}

void DotParser::pushElement (const DOMString& name, bool isParent) {
  Element* element;

  tagStack_.push_back(name);
  tagName_ = name;
  element = new Element(name);

  DOMString nameCopy = name;
  if (nameCopy.toLowerCase() == "div")
    element->setAttribute("keep", "true");

  lastElement_ = element;
  node_ = element;
  if (isParent)
    node_->setIsParent(true);
}

void DotParser::pushElement (const DOMString& name, const DOMString& attrName, const DOMString& attrValue) {
  tagStack_.push_back(name);
//  tagName_ = name;
  Element* element = new Element(name);//tagName_);
  element->setAttribute(attrName, attrValue);
  node_ = element;
  lastElement_ = element;
  node_->setIsParent(true);
}

DOMString DotParser::getDotString() {
  DOMString resultString;

  while (true) {
    Char c = reader_->nextChar();
    if (c == FAILCHAR) break;
    if (c != '.') {
      reader_->unget();
      break;
    }
    resultString += c;
  }
  return resultString;
}

DOMString DotParser::getUSString() {
  DOMString resultString;

  while (true) {
    Char c = reader_->nextChar();
    if (c == FAILCHAR) break;
    if (c != '_') {
      reader_->unget();
      break;
    }
    resultString += c;
  }
  return resultString;
}

DOMString DotParser::getWord() {
  DOMString resultString;

  while (true) {
    Char c = reader_->nextChar();
    if (c == FAILCHAR || c == '\n' || c == ' ' || c == '\t')
      break;
    resultString += c;
  }

  return resultString;
}

DOMString DotParser::getLine() {
  DOMString resultString;

  while (true) {
    Char c = reader_->nextChar();
    if (c == FAILCHAR || c == '\n')
      break;
    resultString += c;
  }
  ++lineNo_;
  return resultString;
}

DOMString DotParser::getTextString() {
  DOMString textString = getLine(); // no trimming for mixed content

  textString.change("&#92;n", "\n");
  textString.change("||", "\n");
  textString.change("&#10;", "\n");
  textString.change(" & ", " &amp; ");
  textString.change("&#46;", ".");

  return textString;
}


DOMString DotParser::getAttributeString() {
  return getLine().change("'", "&apos;").change("\"", "&quot;").change("<", "&lt;");
}

DOMString DotParser::getHashString() {
  DOMString resultString;
  bool hashAlready = false, isHash;

  while (true) {
    Char c = reader_->nextChar();
    if (c == FAILCHAR) break;
    if (c == '\n') ++lineNo_;

    isHash = c == '#';

    if (hashAlready && isHash)
      break;

    if (hashAlready && !isHash)
      resultString += "#";  // catch up

    if (!isHash)
      resultString += c; // finally, add the char

    hashAlready = isHash; // update the flag
  }

  getLine(); // swallow the rest of the line

  return resultString;
}

void DotParser::popStacks() {
  tagStack_.pop_back();
  nestingStack_.pop_back();
}

void DotParser::skipWhitespace() {
  while (true) {
    Char c = reader_->nextChar();
    if (c == FAILCHAR) return;

    if (c != 32 & c != 13 & c != 10) break;
  }
  reader_->unget();
}

DOMString DotParser::getDelimitedString(const DOMString& delimiters) {
  DOMString resultString;
  Char c;

  while (true) {
    c = reader_->nextChar();
    if (c == FAILCHAR) break;
    if (c == '\n') ++lineNo_;
    if (c == '\n' || delimiters.contains(c)) {
      reader_->unget();
      break;
    }
    resultString += c;
  }
  return resultString;
}

Char DotParser::peekChar() {
  Char c = reader_->nextChar();
  if (c != FAILCHAR) reader_->unget();
  return c;
}

}
