/***************************************************************************
 *   Copyright (C) 2007 by root   *
 *   john_redmond@optusnet.com.au   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "dotwriter.h"
#include "cdatasection.h"
#include "comment.h"
#include "documenttype.h"
#include "element.h"
#include "endnode.h"
#include "processinginstruction.h"
#include "text.h"
#include "xmldeclaration.h"
#include "attributestream.h"
#include "converter.h"
#include "domwriter.h"
#include "stringwriter.h"
#include "nodestream.h"

namespace limpid {

DotWriter::DotWriter(): writer_(new Writer), depth_(0), blockedNode_(0), rootName_("root"), started_(false) {}

DotWriter::DotWriter(const DOMString& destString):
  writer_(new Writer(destString)), depth_(0), blockedNode_(0), rootName_("root"), started_(false) {}
  
DotWriter::DotWriter(CharacterWriter& characterWriter):
    writer_(&(characterWriter.copy())), started_(false) {}

void DotWriter::update(const DOMString& destString) {
  DOMString outputString = destString;
  outputString += ".txt";
  writer_ = new Writer(outputString);
}

Node& DotWriter::process(Node& node) {
  if (blockedNode_) {
    unblockProcessing(node);
    return node;
  }

  int type = node.getNodeType();

  if (type == Node::DOCUMENT_NODE) return node; // no action
  if (type == Node::ELEMENT_NODE) processElement(node);
  if (type == Node::END_NODE) processEndNode(node);
  if (type == Node::TEXT_NODE) processText(node);
  if (type == Node::CDATA_SECTION_NODE) processCDATASection(node);
  if (type == Node::XML_DECLARATION_NODE) processXMLDeclaration(node);
  if (type == Node::PROCESSING_INSTRUCTION_NODE) processProcessingInstruction(node);
  if (type == Node::DOCUMENT_TYPE_NODE) processDocumentType(node);
  if (type == Node::COMMENT_NODE) processComment(node);

  started_ = true;

  return node;
}

void DotWriter::processElement(Node& node) {
  DOMString nodeName = node.getNodeName();
   if (nodeName == rootName_ || nodeName == "page") return;

  bool hasChildren = node.getChildNodes().getLength();
  bool hasAttributes = node.getAttributes().getLength();

// not that, if text content is to be generated here,
// blockedNode_ is updated to prevent descendant nodes from being processed.

  if (!hasChildren) {  // an empty element
    *writer_  << "_" << nodeName << "/" << wrt::endl;
    return;
  }

  if (isLeaf(node)) { // we have plain text
    blockedNode_ = &node; // block downward processing
    DOMString valueString = node.getRawValue();
    if (nodeName.startsWith("."))
      *writer_ << wrt::endl << nodeName; 
    else if (nodeName != "p")
      *writer_ << "_" << nodeName << "=";

    processAttributes(node);

    if (nodeName == "pre" || nodeName == "script")
      *writer_ << escapePreString(valueString) << "##" << wrt::endl; // content delimiter
    else
      *writer_ << valueString.resolveEntities();//escapeString(valueString);

    *writer_ << wrt::endl; // line termination

    return;
  }

  if (hasMixedContent(node)) {
    blockedNode_ = &node; // block downward processing

    if (nodeName.startsWith("."))
      *writer_ << nodeName; 
    else if (nodeName != "p")
      *writer_ << "_" << nodeName << "=";

    processAttributes(node);

    DOMString mixedString = getMixedString(node);
    *writer_ << escapeString(mixedString.resolveEntities()) << wrt::endl;//escapeString(mixedString);

    return;
  }

// there must be non-text children
    if (started_)
      *writer_ << wrt::endl;

    *writer_ << "_" << nodeName << wrt::endl;

    processAttributes(node);
}

void DotWriter::processEndNode(Node& node) {
  Node& startNode = static_cast<EndNode &>(node).getStartNode();
  DOMString startName = startNode.getNodeName();
  if (startName == rootName_ || startName == "page") return;

  if (startNode.getChildNodes().getLength())
    *writer_ << "__ " << startNode.getNodeName();
  if (startName == "div") {
    Element& startElement = static_cast<Element &>(startNode);
    DOMString className = startElement.getAttribute("class");
    if (!className.empty())
      *writer_ << "(" << className << ")";
    else
      *writer_ << "[" << startElement.getAttribute("id") << "]";
  }

  *writer_ << wrt::endl;// << wrt::endl;
}

void DotWriter::processText(Node& node) {
  DOMString rawValue = node.getRawValue();
  *writer_<< "=" << escapeString(rawValue) << wrt::endl;
}

void DotWriter::processCDATASection(Node& node) {
  *writer_ << "!";
  *writer_ << node.getNodeValue() << "##" << wrt::endl;
}

void DotWriter::processXMLDeclaration(Node& node) {
  *writer_ << "?xml ";

  XMLDeclaration& xmlDeclaration = static_cast<XMLDeclaration &>(node);
  *writer_<< xmlDeclaration.getVersion() << "|" << xmlDeclaration.getEncoding()\
    << "|" << xmlDeclaration.getStandAlone() << wrt::endl;
}

void DotWriter::processProcessingInstruction(Node& node) {
  *writer_ << "?";

  ProcessingInstruction& instruction = static_cast<ProcessingInstruction &>(node);
  DOMString dataString = instruction.getData();
  *writer_ << instruction.getTarget() << "=" << escapeString(dataString) << "##" << wrt::endl;
}

void DotWriter::processDocumentType(Node& node) {
  *writer_ << "!";

  DocumentType& doctype = static_cast<DocumentType &>(node);
  DOMString internalSubset = doctype.getInternalSubset();
  escapeString(internalSubset);

  *writer_ << doctype.getName() << "|" << doctype.getPublicId() << "|"\
  << doctype.getSystemId() << "|" << internalSubset << "##" << wrt::endl;
}

void DotWriter::processComment(Node& node) {
  DOMString valueString = node.getRawValue();
  *writer_ << "\n#";
  *writer_ << valueString << "##" << wrt::endl;
}

void DotWriter::processAttributes(Node& node) {
  AttributeStream stream(node);

  while (true) {
    Node& attrNode = stream.nextNode();
    if (!attrNode.isValid()) break;

    *writer_ << "@" << attrNode.getNodeName() << "=" << attrNode.getRawValue() << wrt::endl;
  }
}

void DotWriter::unblockProcessing(Node& node) {
  if (node.getNodeType() != Node::END_NODE) return;

  Node& startNode = static_cast<EndNode &>(node).getStartNode();
  if (blockedNode_ == &startNode)
    blockedNode_ = 0;
}

DOMString& DotWriter::escapeString(DOMString& valueString) {
  valueString.change("\r", "");
  valueString.change("\n", "||");
  if (valueString[0] == '.')
    valueString.change(".", "&#46;");
  return valueString.change("##", "&#35;#"); 
}

DOMString& DotWriter::escapePreString(DOMString& valueString) {
  valueString.change("&lt;", "<");
  valueString.change("||", "\n");
  return valueString.change("##", "&#35;#"); 
}

DOMString DotWriter::getMixedString(Node& parentNode) {
  StringWriter writer;
  XMLWriter xmlWriter(writer);

  NodeStream nodeStream(parentNode, true, false); // exclude node itself
  while (true) {
    Node& node = nodeStream.nextNode();
    if (!node.isValid()) break;

    xmlWriter.process(node);
  }

  return writer.getString();
}

}
