/***************************************************************************
                          dtdconverter.cpp  -  description
                             -------------------
    begin                : Wed Sep 24 2003
    copyright            : (C) 2003 by John Redmond
    email                : john_redmond@optusnet.com.au
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "dtdconverter.h"
#include "characterreader.h"
#include "xmldeclaration.h"
#include "namednodemap.h"
#include "attr.h"
#include "text.h"
#include "cdatasection.h"
#include "nodelist.h"
#include "nodestream.h"
#include "tokenstream.h"
#include "url.h"
#include "reader.h"
#include "stringreader.h"
#include "nodestream.h"
#include "liststream.h"
#include "systemmanager.h"
#include "traverser.h"
#include "domwriter.h"
#include "converter.h"

namespace limpid {

DTDConverter::DTDConverter(const URL& url): input(TokenStream(url)) {
  initialize();
}

DTDConverter::DTDConverter(CharacterReader& cstream):
      input(TokenStream(cstream)) {
  initialize();
}

Document& DTDConverter::process() {
  Node::systemManager.setResponse(SystemManager::WARN);
  collectInputLists();
  mapLists();
  buildElements();
  updateReferences();
  addEntryElement();

  return document;
}

Document& DTDConverter::getDocument() {
  return document;
}

void DTDConverter::processTokens() {
  while (true) {
    DOMString token = peekTokenSimple();
    if (token.empty()) break;

    if (token == "<") {
      bumpInput();
      bumpInput(); // the '!'
      DOMString type = nextToken();
/*
      if (type.empty()) return;
      if (type == "ELEMENT") createElement();
      else if (type == "ATTLIST") createElementAttrs();
      else if (type == "ENTITY") createEntity();
      else if (type == "NOTATION") createNotation();
      else if (type.startsWith("--")) createDocumentation(type);
      else swallow();
 */
   }

    else {
      wrt::Errors << "BAD TOKEN: " << token << wrt::endl;
      bumpInput();
    }
  }
}

void DTDConverter::setAttrNodeDefault(Element& attributeElement) {
  DOMString token = peekToken();
  if (token == "\"" || token == "'") {
    token = getDelimitedString();
    attributeElement.setAttribute("default", token);
  }
}

void DTDConverter::setAttrNodeStatus(Element& attributeElement) {
  DOMString token = peekToken();
  if (token == "#") {
    bumpInput();
    token = nextToken();
  }
  else token = "IMPLIED";

  if (!token.empty()) {
    DOMString attrValue = "required";
    if (token == "FIXED") attrValue = "fixed";
    else if (token == "IMPLIED") attrValue = "optional";

    attributeElement.setAttribute("use", attrValue);
  }
}

void DTDConverter::processAttrClosure(Element& childAttrElement) {
  Element restriction("xsd:restriction");
  childAttrElement.appendChild(restriction);

  DOMString type = childAttrElement.getAttribute("type");
  if (!type.empty()) {
    restriction.setAttribute("base", type);
    childAttrElement.removeAttribute("type");
  }
  else
    restriction.setAttribute("base", "xsd:string");

  bumpInput(); // throw away the "("

  while (true) {
    DOMString token = nextToken();
    if (token.empty() || token == ")") {
      break;
    }
    if (token == "|") continue;
    else {
      Element enumeration("xsd:enumeration");
      enumeration.setAttribute("value", token);
      restriction.appendChild(enumeration);
    }
  }
}

void DTDConverter::setAttributeNodeType(Element& attributeElement) {
  DOMString token = peekToken();
// is it an attribute type?
  if (token.empty() || token =="CDATA" ||
        token == ")" || token == "'" || token == "\"" ||
        token == "#" || token == ">") {
    if (token == "CDATA")
      bumpInput();
    attributeElement.setAttribute("type", "xsd:string");
  }
  else if (token == "NOTATION" || token =="ENTITY" ||
                  token == "ENTITIES" || token == "NMTOKEN" ||
                  token == "NMTOKENS" || token == "IDREF" ||
                  token == "IDREFS" || token == "ID") {
    bumpInput();
    DOMString typeName = "xsd:";
    typeName += token;

    attributeElement.setAttribute("type", typeName);
  }
  else
    attributeElement.setAttribute("type", "xsd:string");
}

void DTDConverter::createEntity() {
  DOMString name = nextTokenSimple();
  if (name.empty()) return;
  DOMString token;

  if (name == "%") {
    setParameterEntity();
  }
  else {
    Element entry("xsd:entity");
    entry.setAttribute("name", name);
    document.getDocumentElement().appendChild(entry);

    token = peekToken(); // this may be " or ' or SYSTEM

    if (token == "\"" || token == "'") {
      DOMString value = getDelimitedString();
      if (!value.empty())
        entry.setAttribute("value", value);
    }
    else {
      addIDs(entry);
      addNDATA(entry);
    }
  }
  input.alignOn(">");
  bumpInput();
}

void DTDConverter::setParameterEntity() {
  DOMString name = nextToken();
  if (name.empty() || name == ">") return; // nothing to do

  DOMString token = peekToken();
  if (token.empty() || token == ">") return; // nothing there

  if (token == "\"" || token == "'") {
    DOMString value = getDelimitedString();
    parameterMap.insert(make_pair(name, value));
  }
}

void DTDConverter::addIDs(Element& entityElement) {
  DOMString type = peekToken();
  if (type != "PUBLIC" && type != "SYSTEM") {
    input.alignOn(">");
    return;
  }
  else bumpInput();

  if (type == "PUBLIC") {
    DOMString pubString = getDelimitedString();
    entityElement.setAttribute("public", pubString);
  }

  DOMString sysString = getDelimitedString(); // get the system ID string
  entityElement.setAttribute("system", sysString);
}

void DTDConverter::addNDATA(Element& entityElement) {
  DOMString ndata = peekToken();
  if (ndata == "NDATA") {
    bumpInput();
    DOMString notation = peekToken();
    if (!notation.empty() && notation != ">") {
      bumpInput();
      entityElement.setAttribute("notation", notation);
    }
  }
}

void DTDConverter::createNotation() {
  DOMString name = nextToken();
  Element entry("xsd:notation");
  entry.setAttribute("name", name);
  addIDs(entry);
  document.getDocumentElement().appendChild(entry);
  input.alignOn(">");
  bumpInput();
}

void DTDConverter::createDocumentation(DOMString s) {
  size_t length = s.size();
  if (length >= 2) s = s.substring(2, length);

  DOMString delimiterString = "-->";
  s += input.getDelimitedString(delimiterString);
  s.normalizeWhitespace();

  Element entry("xsd:annotation");
  Element documentation("xsd:documentation");
  entry.appendChild(documentation);
  documentation.setAttribute("xml:lang", "en");
  Text text(s);
  documentation.appendChild(text);
  document.getDocumentElement().appendChild(entry);
}

DOMString DTDConverter::getParameter(DOMString& key) {
  ParameterMap::iterator p = parameterMap.find(key);
  if (p != parameterMap.end())
    return p->second;
  else
    return "unknown key";
}

Node& DTDConverter::getAppendPoint(Node& entry) {
  DOMString schemaNamespace = "http://www.w3c.org/2001/XMLSchema";
  Node& firstChild = entry.getFirstChild();
  DOMString localName = firstChild.getLocalName();
  if ((localName == "complexContent" || localName == "simpleContent") &&
      firstChild.getNamespaceURI() == schemaNamespace) {
    if (firstChild.getFirstChild().getNamespaceURI() == schemaNamespace)
      return firstChild.getFirstChild();
  }
  return entry; // append to entry itself
}

void DTDConverter::removeEmptyTypes() {
  Element& root = document.getDocumentElement();

  NodeList& rootChildren = root.getChildNodes();
  ListStream stream(rootChildren);
  NodeList emptyTypes;

  while (true) {
    Node& node = stream.nextNode();
    if (!node.isValid()) break;

    DOMString localName = node.getLocalName();

    if ((localName != "simpleType" && localName != "complexType") ||
        node.getNamespaceURI() != "http://www.w3c.org/2001/XMLSchema")
      continue;

    if (localName == "simpleType" && !node.getFirstChild().hasChildNodes())
      emptyTypes.addNode(node);
    else if (!node.hasChildNodes())
      emptyTypes.addNode(node);
  }

  for (size_t i = 0; i < emptyTypes.getLength(); i++)
    root.removeChild(emptyTypes.item(i));
}

DOMString DTDConverter::findOrigin() {
  Element& root = document.getDocumentElement();

  ListStream stream(root);
  NameSet elements;

  while (true) {
    Node& node = stream.nextNode(); // look at just the global entries
    if (!node.isValid()) break;

    if (node.getLocalName() != "complexType" ||
        node.getNamespaceURI() != "http://www.w3c.org/2001/XMLSchema")
      continue;

    Element elementElement = static_cast<Element &>(node);
    DOMString refName = elementElement.getAttribute("name");
    if (refName.empty()) continue;

    elements.insert(refName);
  }

  NodeStream globalStream(root);

  while (true) {
    Node& node = globalStream.nextNode();  // look below the global entries
    if (!node.isValid()) break;

    if (node.getLocalName() != "element" ||
        node.getNamespaceURI() != "http://www.w3c.org/2001/XMLSchema")
      continue;

    DOMString elementName = (static_cast<Element &>(node)).getAttribute("type");
    NameSet::iterator p = elements.find(elementName);
    if (p != elements.end())
      elements.erase(p);
  }

  if (elements.size() == 1)  // there should be only one top-level element defined
    return *elements.begin();
  else return "";
}

void DTDConverter::copyAttributes(Element& source, Element& dest) {
  NamedNodeMap &attrAttributes = source.getAttributes();
  size_t attrCount = attrAttributes.getLength();
  if (!attrCount) return; // no attributes!!

  for (size_t i = 0; i < attrCount; i++) {
    Attr &attr = static_cast<Attr &>(attrAttributes.item(i));
    Attr attrClone = static_cast<Attr &>(attr.cloneNode(true));
    dest.setAttributeNode(attrClone);
  }
}

void DTDConverter::copyChildren(Element& source, Element& dest) {
  NodeList &children = source.getChildNodes();
  size_t childCount = children.getLength();
  if (!childCount) return; // no attributes!!

  for (size_t i = 0; i < childCount; i++)
    dest.appendChild(children.item(i).cloneNode(true));
}

void DTDConverter::initialize() {
  input.setProcessDTD();
  XMLDeclaration pi("1.0", "UTF-8", "yes");
  document.appendChild(pi);
  Element root("xsd:schema");
  document.appendChild(root);
  root.setAttribute("xmlns:xsd", "http://www.w3c.org/2001/XMLSchema");
  root.setAttribute("xmlns", "http://limpid.namespace");
}


DOMString& DTDConverter::nextToken() {
  DOMString& result = input.nextToken();
  if (!result.empty()) {
    if (result == "%") {
      getStringInput();
      result = input.nextToken();
    }
    return result;
  }

  if (!inputStack.empty()) {
    input = inputStack.back();
    inputStack.pop_back();
    return nextToken();
  }
  else
    return result;
}

DOMString& DTDConverter::peekToken() {
  DOMString& result = input.peekToken();
  if (!result.empty()) {
    if (result == "%") {
      bumpInput();
      getStringInput();
      result = input.peekToken();
    }
    return result;
  }

  if (!inputStack.empty()) {
    input = inputStack.back();
    inputStack.pop_back();
    return peekToken();
  }
  else
    return result;  // empty string
}

DOMString& DTDConverter::nextTokenSimple() {
  DOMString& result = input.nextToken();
  if (!result.empty())
    return result;

  if (!inputStack.empty()) {
    input = inputStack.back();
    inputStack.pop_back();
    return nextToken();
  }
  else
    return result;
}

DOMString& DTDConverter::peekTokenSimple() {
  DOMString& result = input.peekToken();
  if (!result.empty()) return result;

  if (!inputStack.empty()) {
    input = inputStack.back();
    inputStack.pop_back();
    return peekToken();
  }
  else {
    return result;
  }
}

DOMString DTDConverter::getDelimitedString() {
  DOMString token = nextToken();
  DOMString s = input.getDelimitedString(token[0]);
  return s;
}

void DTDConverter::redirectInput() {
  DOMString source = nextToken();
  DOMString semicolon = nextToken();  // the ';'
  if (semicolon != ";") return; // error!!

  ParameterMap::iterator p = parameterMap.find(source);
  if (p != parameterMap.end())
    source = p->second;

  inputStack.push_back(input);
  URL inputURL = input.getHandler().getURL();
  URL newURL(inputURL, source);
  Reader newStream(newURL);

  input = TokenStream(newStream);
  input.setProcessDTD();
}

void DTDConverter::getStringInput() {
  DOMString token = nextToken();
  bumpInput();  // the ';'

  ParameterMap::iterator p = parameterMap.find(token);
  if (p == parameterMap.end())
    return; // cannot proceed

  DOMString tokens = p->second;
  inputStack.push_back(input);
  StringReader newStream(tokens);

  input = TokenStream(newStream);
  input.setProcessDTD();
}

void DTDConverter::collectInputLists() {
  Element& documentElement = document.getDocumentElement();
  StringList resultList;

  while (true) {
    DOMString token = peekToken();
    if (token.empty()) break;
    resultList.clear();

    getTokenList(resultList);
    inputList.push_back(resultList);
  }
}

void DTDConverter::mapLists() {
  Element& documentElement = document.getDocumentElement();

  SuperStringList::iterator p = inputList.begin();
  while (p != inputList.end()) {
    StringList resultList = *p++;

    if (resultList.empty()) continue;

    resultList.pop_front(); // the !
    if (resultList.empty()) continue;

    DOMString listType = getNext(resultList);
    if (resultList.empty()) continue;

    DOMString listName = getNext(resultList);
    if (resultList.empty()) continue;

    if (resultList.back() == "(")
      resultList.pop_front();
    if (resultList.back() == ")")
      resultList.pop_back();

    if (listType == "ELEMENT" && !resultList.empty()) {
      elementList.push_back(listName);
      elementSet.insert(listName);
      childSet.insert(listName);
      elementMap[listName] = resultList;
    }
    else {
      attributeMap[listName] = resultList;
      attributeSet.insert(listName);
    }
  }
}

void DTDConverter::buildElements() {
  Element& documentElement = document.getDocumentElement();
  StringList resultList;

  StringList::iterator p = elementList.begin();
  while (p != elementList.end()) {
    DOMString elementName = *p++;
    DOMString elementType = getTypeName(elementName);

    StringList tokenList = elementMap[elementName];
    Element elementElement("xsd:complexType");
    elementElement.setAttribute("name", getTypeName(elementName));

    if (addRow(tokenList, elementName, elementElement))
      documentElement.appendChild(elementElement);
    else
      leafSet.insert(elementName);
  }
}

void DTDConverter::buildRows() {
  Element& documentElement = document.getDocumentElement();
  StringList resultList;

  while (true) {
    DOMString token = peekToken();
    if (token.empty()) break;
    resultList.clear();
    getTokenList(resultList);

    resultList.pop_front(); // the !
    DOMString typeString = getNext(resultList);
    if (typeString == "ELEMENT")
      typeString = "xsd:complexType";

    Element typeElement(typeString);
    documentElement.appendChild(typeElement);
    addRow(resultList, typeString, typeElement);
  }
}

void DTDConverter::addEntryElement() {
  if (childSet.size()) {
    StringSet::iterator p = childSet.begin();
    DOMString elementName = *p;
    DOMString elementType = getTypeName(elementName);

    Element entryElement("xsd:element");
    entryElement.setAttribute("name", elementName);
    entryElement.setAttribute("type", elementType);

    Element& documentElement = document.getDocumentElement();
    documentElement.insertBefore(entryElement, documentElement.getFirstChild());
  }
}


bool DTDConverter::addRow(StringList& tokenList, DOMString& typeString, Element& parent) {
  DOMString rowName = getNext(tokenList);

  int childCount = addClosure(tokenList, parent, parent); // result is not used

  StringList attributeTokens = attributeMap[typeString];
  addAttributes(attributeTokens, parent);

  return parent.getChildNodes().getLength() >= 1 || parent.getFirstChild().getChildNodes().getLength() > 1;
}

int DTDConverter::addClosure(StringList& tokenList, Element& parent, Element& rootElement) {
  int childCount = 0;

  Element closureElement("xsd:sequence");

  if (tokenList.front() == "(") tokenList.pop_front();
  if (tokenList.back() == ")") tokenList.pop_back();

  DOMString typeString = "sequence";
  bool isSequence = true;

  while (true) {
    if (tokenList.empty()) break;

    DOMString token = getNext(tokenList);

    if (token == "(")
      childCount += addClosure(tokenList, closureElement, rootElement);

    else if (token == ")")
      break;

    else if (token == "+" || token == "*") {
      Element& lastChild = static_cast<Element &>(closureElement.getLastChild());
      if (token == "*")
        lastChild.setAttribute("minOccurs", "0");
      lastChild.setAttribute("maxOccurs", "unbounded");
    }

    else if (token == "|") isSequence = false;
    else if (token == "," ) continue;

    else if (token == "#") {
      DOMString peekToken = peekNext(tokenList);
      if (peekToken == "PCDATA" || peekToken == "cdata-section") {
        getNext(tokenList);
        rootElement.setAttribute("mixed", "true");
      }
      continue;
    }
    else {
      addItem(token, closureElement, tokenList);
      childCount += 1;
    }
  }

  if (!childCount) return 0; // nothing to add

  if (isSequence) {
    parent.appendChild(closureElement);
    addCount(closureElement, tokenList);
  }
  else {
    Element choiceElement("xsd:choice");
    moveContent(closureElement, choiceElement);
    addCount(choiceElement, tokenList);
    parent.appendChild(choiceElement);
  }

  return childCount;
}

bool DTDConverter::addAttributes(StringList& tokenList, Element& parent) {
  DOMString nameString, typeString, defaultString;
  bool hasName = false;
  while (!tokenList.empty()) {
    nameString = getNext(tokenList);
    hasName = true;
    if (!tokenList.empty())
      typeString = getNext(tokenList);

    if (peekNext(tokenList) == "\"" || peekNext(tokenList) == "'") {
      getNext(tokenList);
      defaultString = getNext(tokenList);
      getNext(tokenList);
      static_cast<Element &>(parent.getLastChild()).setAttribute("default", defaultString);
    }

    Element attributeElement("xsd:attribute");
    attributeElement.setAttribute("name", nameString);
    parent.appendChild(attributeElement);
  }
  return true;
}

bool DTDConverter::addItem(DOMString& itemName, Element& parent, StringList& tokenList) {
  Element itemElement("xsd:element");
  itemElement.setAttribute("name", itemName);
  itemElement.setAttribute("type", getTypeName(itemName));
  addCount(itemElement, tokenList);
  parent.appendChild(itemElement);

  StringSet::iterator p = childSet.find(itemName);
  if (p != childSet.end())
    childSet.erase(p);

  return true;
}

void DTDConverter::updateReferences() {
  NodeStream nodeStream(document.getDocumentElement(), false, false);

  while (true) {
    Node& node = nodeStream.nextNode();
    if (!node.isValid()) break;

    if (node.getNodeType() != Node::ELEMENT_NODE) continue;
    if (node.getNodeName() != "xsd:element") continue;

    Element& element = static_cast<Element &>(node);
    DOMString tagName = element.getAttribute("name");
    if (leafSet.find(tagName) != leafSet.end())
      element.setAttribute("type", "xsd:string");
  }
}

void DTDConverter::swallow() {
  while (true) {
    DOMString token = nextToken();
    if (token.empty() || token == ">") return;
  }
}

void DTDConverter::getTokenList(StringList& resultList) {

  while (true) {
    DOMString token = nextToken();
    if (token.empty()) break;

    if (token == ">") break;
    else if (token != "<") resultList.push_back(token);
  }
}

void DTDConverter::displayList(StringList& resultList) {
  StringList::iterator p = resultList.begin();
  while (p != resultList.end()) {
    wrt::Console << *p++ << " ";
  }
    wrt::Console << wrt::endl;
}

DOMString DTDConverter::peekNext(StringList& tokenList) {
  if (tokenList.empty()) return "";
  else return tokenList.front();
}

DOMString DTDConverter::getNext(StringList& tokenList) {
  if (tokenList.empty()) return "";
  DOMString token = tokenList.front();
  tokenList.pop_front();
  return token;
}

DOMString DTDConverter::getTypeName(const DOMString& name) {
  DOMString type = name;
  type.toLowerCase().capitalize() += "Type";
  return type;
}

void DTDConverter::addCount(Element& element, StringList& tokenList) {
  DOMString countToken = peekNext(tokenList);
  if (countToken == "*" || countToken == "+" || countToken == "?") {
    getNext(tokenList);  // swallow the token
    if (countToken == "*" || countToken == "?")
      element.setAttribute("minOccurs", "0");
    if (countToken == "?")
      element.setAttribute("maxOccurs", "1");
    else
      element.setAttribute("maxOccurs", "unbounded");
  }
}

}
