/***************************************************************************
                          element.cpp  -  description
                             -------------------
    begin                : Tue Mar 18 2003
    copyright            : (C) 2003 by John Redmond
    email                : john_redmond@optusnet.com.au
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <string>
#include <list>
#include <iostream>
#include "element.h"
#include "document.h"
#include "attr.h"
#include "namednodemap.h"
#include "elementcontent.h"
#include "systemmanager.h"
#include "writer.h"

namespace limpid {

Element::Element(): ParentNode() {
  try {
    nodeContent_ = new ElementContent("", *this);
  }
  catch (std::bad_alloc) {
    delete nodeContent_;
    throw;
  }
}

Element::Element(const DOMString& name): ParentNode() {
  try {
    nodeContent_ = new ElementContent(name, *this);
  }
  catch (std::bad_alloc) {
    delete nodeContent_;
    throw;
  }
}

Element::Element(const DOMString& namespaceURI, const DOMString& name): ParentNode() {
  try {
    nodeContent_ = new ElementContent(name, *this);
    setNamespaceURI(namespaceURI);
  }
  catch (std::bad_alloc) {
    delete nodeContent_;
    throw;
  }
}

Element::Element(const Element& refNode): ParentNode(refNode)  {
}

Element& Element::operator=(const Element& refNode) {
  replaceContent(refNode);
  return *this;
}

Attr& Element::getAttributeNode(const DOMString& name) const {
  ElementContent *contentPtr = static_cast<ElementContent *>(nodeContent_);
  if (contentPtr->attributes_)
    return (Attr &)contentPtr->attributes_->getNamedItem(name);
  else return Attr::failAttr;
}

Attr& Element::getAttributeNodeNS(const DOMString& namespaceURI,
                                      const DOMString& name) const {
  ElementContent *contentPtr = static_cast<ElementContent *>(nodeContent_);
  if (contentPtr->attributes_)
    return (Attr &)contentPtr->attributes_->getNamedItemNS(namespaceURI, name);
  else return Attr::failAttr;
}

DOMString Element::getAttribute(const DOMString& name) const {
  ElementContent *contentPtr = static_cast<ElementContent *>(nodeContent_);
  if (contentPtr->attributes_) {
    Node& attr = (Attr &)contentPtr->attributes_->getNamedItem(name);
    if (!attr.isValid()) return "";
    else return static_cast<Attr &>(attr).getValue();
  }
  return "";
}

void Element::setAttribute(const DOMString& name, const DOMString& value) {
  Attr attr(name);
  attr.setValue(value);
  setAttributeNode(attr);
}

void Element::removeAttribute(const DOMString& name) {
  if (!systemManager.checkModificationPermission(*this)) return;

  ElementContent *contentPtr = static_cast<ElementContent *>(nodeContent_);
  if (contentPtr->attributes_) {
    Attr& resultAttr = (Attr &)contentPtr->attributes_->removeNamedItem(name);
    if (resultAttr.isValid())
      getOwnerDocument().cacheNode(resultAttr);
  }
}

Attr& Element::setAttributeNode(Attr& attr) {
  ElementContent *contentPtr = static_cast<ElementContent *>(nodeContent_);
  if (!contentPtr->attributes_)
    contentPtr->attributes_ = new NamedNodeMap(*this);

  attr.setSpecified(true);
  Attr& resultAttr = (Attr &)contentPtr->attributes_->setNamedItem(attr);
  if (!resultAttr.isValid()) return resultAttr;

  AttributeContent *resultPtr = static_cast<AttributeContent *>(resultAttr.nodeContent_);
  resultPtr->documentContent_ = contentPtr->documentContent_;

  return resultAttr;
}

Attr& Element::setAttributeNodeNS(Attr& attr) {
  ElementContent *contentPtr = static_cast<ElementContent *>(nodeContent_);
  if (!contentPtr->attributes_)
    contentPtr->attributes_ = new NamedNodeMap(*this);
  if (!attr.nodeContent_->parentContent_)
    attr.nodeContent_->parentContent_ = contentPtr;

  attr.setSpecified(true);
  Attr& resultAttr = (Attr &)contentPtr->attributes_->setNamedItemNS(attr);
  if (!resultAttr.isValid()) return resultAttr;

  AttributeContent *resultPtr = static_cast<AttributeContent *>(resultAttr.nodeContent_);
  resultPtr->documentContent_ = contentPtr->documentContent_;
  if (nodeContent_->documentContent_)
    resultAttr.setDocumentContent(*nodeContent_->documentContent_);

  return resultAttr;
}

Attr& Element::removeAttributeNode(Attr& attr) {
  if (!systemManager.checkModificationPermission(*this)) return Attr::failAttr;

  ElementContent *contentPtr = static_cast<ElementContent *>(nodeContent_);
  if (contentPtr->attributes_) {
    Attr& resultAttr = (Attr &)contentPtr->attributes_->removeSpecifiedItem(attr);
    if (resultAttr.isValid())
      getOwnerDocument().cacheNode(resultAttr);
    else
      systemManager.processNotFoundError(attr);

    return resultAttr;
  }
  return Attr::failAttr;
}

DOMString Element::getAttributeNS(const DOMString& namespaceURI,
                                  const DOMString& name) const {
  ElementContent *contentPtr = static_cast<ElementContent *>(nodeContent_);
  if (contentPtr->attributes_) {
    Node& attr = (Attr &)contentPtr->attributes_->getNamedItemNS(namespaceURI, name);
    if (!attr.isValid()) return DOMString::failString;
    else return ((Attr &)attr).getValue();
  }
  return "";
}

void Element::setAttributeNS(const DOMString& namespaceURI,
    const DOMString& name, const DOMString& value) {
  Attr attr(namespaceURI, name);
  attr.setValue(value);
  setAttributeNodeNS(attr);
}

void Element::removeAttributeNS(const DOMString& namespaceURI, const DOMString& name) {
  if (!systemManager.checkModificationPermission(*this)) return;

  ElementContent *contentPtr = static_cast<ElementContent *>(nodeContent_);
  if (contentPtr->attributes_) {
    Attr& resultAttr =
      (Attr &)contentPtr->attributes_->removeNamedItemNS(namespaceURI, name);
    if (resultAttr.isValid())
      getOwnerDocument().cacheNode(resultAttr);
  }
}

NamedNodeMap& Element::getAttributes() const {
  ElementContent *contentPtr = static_cast<ElementContent *>(nodeContent_);
  if (contentPtr->attributes_) return *contentPtr->attributes_;
  else return NamedNodeMap::defaultAttributes;
}

DOMString Element::getDisplayString() const {
  DOMString s("<");
  s += getNodeName();
  s += getAttributeString();
//  if (!getChildNodes().getLength()) s += "/";
  if (!getChildNodes().getLength()) s += "/";
  s += ">";

  return s;
}

DOMString Element::getAttributeString() const {
  DOMString result;
  ElementContent *contentPtr = static_cast<ElementContent *>(nodeContent_);
  if (!contentPtr->attributes_) return result;

  int length = contentPtr->attributes_->getLength();
  for (int i = 0; i < length; i++) {
    result += " ";
    Node& attr = contentPtr->attributes_->item(i);
    result += attr.getDisplayString();
  }
  return result;
}

DOMString Element::getRawValue() const {
  DOMString content;
  ElementContent *contentPtr = static_cast<ElementContent *>(nodeContent_);
  if (!contentPtr->childNodes_) return content;

  int childCount = contentPtr->childNodes_->getLength();
  if (childCount > 0) {
  for (int i = 0; i < childCount; i++) {
    Node& node = contentPtr->childNodes_->item(i);//i);
    short type = node.getNodeType();
    if (type == Node::TEXT_NODE || type != Node::ENTITY_REFERENCE_NODE)
      content += node.getNodeValue();
    else if (type == Node::CDATA_SECTION_NODE)
      content += node.getNodeValue();
    }
  }

  return content;
}

DOMString Element::getValue() const {
  DOMString rawValue = getRawValue();
  return getOwnerDocument().resolveEntities(rawValue);
}

Node& Element::cloneNode(bool deep) const {
  DOMString name = getNodeName();
  Element *clone = new Element(name);
  ElementContent *contentPtr = static_cast<ElementContent *>(nodeContent_);

  if (contentPtr->attributes_) { // must copy, even if deep == false
    size_t length = contentPtr->attributes_->getLength();
    for (size_t i = 0; i < length; i++) {
      Attr& attr =
        (Attr &)contentPtr->attributes_->item(i).cloneNode(true);
      clone->setAttributeNode(attr);
    }
  }

  if (deep && contentPtr->childNodes_) {
    size_t length = contentPtr->childNodes_->getLength();
    for (size_t i = 0; i < length; i++) {
      Node& child = contentPtr->childNodes_->item(i).cloneNode(true);
      clone->appendChild(child);
    }
  }

  getOwnerDocument().cacheNode(*clone);
  return *clone;
}

NodeList Element::getElementsByTagName(const DOMString& name) const {
   return getOwnerDocument().getElementsByTagName(name, *this);
}

NodeList Element::getElementsByTagNameNS(const DOMString& namespaceURI,
    const DOMString& localName) const {
   return getOwnerDocument().getElementsByTagNameNS(namespaceURI, localName, *this);
}

DOMString Element::getNamespaceURI() const {
  ParentContent *contentPtr = static_cast<ParentContent *>(nodeContent_);
  DOMString attrName = "xmlns";
  if (contentPtr->prefix_) {
    attrName += ":";
    attrName += contentPtr->prefix_;
  }
  return findNamespaceURI(attrName);
}

void Element::setDocumentContent(DocumentContent& documentContent) {
  ParentNode::setDocumentContent(documentContent);

  ElementContent *contentPtr = static_cast<ElementContent *>(nodeContent_);
  if (!contentPtr->attributes_) return;

  NamedNodeMap & attributes_ = *contentPtr->attributes_;
  size_t length = attributes_.getLength();
  for (size_t i = 0; i < length; i++)
    attributes_.item(i).setDocumentContent(documentContent);

}

DOMString Element::findNamespaceURI(DOMString& attrName) const {
  const Node* node = this;
  while (true) {
    DOMString value = static_cast<const Element *>(node)->getAttribute(attrName);
    if (value.size()) return value;

    node = &node->getParentNode();
    if (!node->isValid() || node->getNodeType() != ELEMENT_NODE) break;
  }
  return "";
}

void Element::setNamespaceURI(const DOMString& namespaceURI) {
  ParentContent *contentPtr = static_cast<ParentContent *>(nodeContent_);
  DOMString attrName = "xmlns";
  if (contentPtr->prefix_) {
    attrName += ":";
    attrName += getPrefix();
  }
  setAttribute(attrName, namespaceURI);
}

void Element::clearLinks() {
  ElementContent *contentPtr = static_cast<ElementContent *>(nodeContent_);
  nodeContent_->parentContent_ = 0;
  nodeContent_->previousSibling_ = 0;
  nodeContent_->nextSibling_ = 0;

  if (this == contentPtr->ownerNode_) {  // we have removed a control node
    contentPtr->ownerNode_ = 0;
    if (contentPtr->childNodes_)
      contentPtr->childNodes_->parentContent_ = 0;
    if (contentPtr->attributes_)
      contentPtr->attributes_->elementContent_ = 0;
  }
}

Element Element::failElement("fail");

}
