/***************************************************************************
                          elementtemplate.cpp  -  description
                             -------------------
    begin                : Fri Jul 1 2005
    copyright            : (C) 2005 by John Redmond
    email                : john_redmond@optusnet.com.au
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "elementtemplate.h"
#include "xslcontext.h"
#include "element.h"
#include "endnode.h"
#include "attr.h"
#include "text.h"

#include "liststream.h"

namespace limpid {

ElementTemplate::ElementTemplate(const DOMString& name, Expr& nameExpr,
      bool hasChildren, bool isHTML): Template(name), nameExpr_(nameExpr),
      hasChildren_(hasChildren), isHTML_(isHTML) {}


bool ElementTemplate::process(XSLContext& parentContext) {
  DOMString elementName = nameExpr_.eval(parentContext).getString(parentContext);

  if (!templates_.getLength())
    return processEmptyElement(elementName, parentContext);

  return processParentElement(elementName, parentContext);
}

bool ElementTemplate::processNode(XSLContext& parentContext, Node& node) {
  DOMString elementName = nameExpr_.eval(parentContext).getString(parentContext);
  Element element(elementName);
  node.appendChild(element);

// set the attributes and append any child nodes
  attributes_.processNode(parentContext, element);
  return templates_.processNode(parentContext, element);
}

Element ElementTemplate::getElement(XSLContext& parentContext) {
  DOMString elementName = nameExpr_.eval(parentContext).getString(parentContext);
  Element element(elementName);

// set the attributes and append any child nodes
  attributes_.processNode(parentContext, element);
  templates_.processNode(parentContext, element);

  return element;
}

void ElementTemplate::addTemplate(Template& childTemplate) {
  if (childTemplate.isAttribute())
    attributes_.addTemplate(childTemplate);
  else
    templates_.addTemplate(childTemplate);
}

void ElementTemplate::outputAliases(CharacterWriter& writer) {
  size_t length = namespaceMap_.size();
  if (!length) return;

  StringMap::const_iterator p = namespaceMap_.begin();
  StringMap::const_iterator pend = namespaceMap_.end();

  while (p != pend) {
    writer << " " << p->first << "=\"" << p->second << "\"";
    ++p;
  }
}

void ElementTemplate::setAliases(Element& element) {
  size_t length = namespaceMap_.size();
  if (!length) return;

  StringMap::const_iterator p = namespaceMap_.begin();
  StringMap::const_iterator pend = namespaceMap_.end();

  while (p != pend) {
   element.setAttribute(p->first, p->second);
    ++p;
  }
}


bool ElementTemplate::isEmptyTag(DOMString tagName) {
  tagName.toLowerCase();

  static DOMString emptyNames =
    "br col frame hr img input isindex link meta param";

  return emptyNames.contains(tagName);
}

bool ElementTemplate::processEmptyElement(DOMString& elementName,
    XSLContext& parentContext) {
  CharacterWriter& writer = parentContext.getWriter();

  writeFormat(parentContext);

  nest(parentContext);

  writer << "<" << elementName;
  // output any attributes
  attributes_.process(parentContext);
  writer << "/>";

  deNest(parentContext);

  return true;
}

bool ElementTemplate::processParentElement(DOMString& elementName,
    XSLContext& parentContext) {

  writeFormat(parentContext);

  CharacterWriter& writer = parentContext.getWriter();
  writer << "<" << elementName;

  nest(parentContext);

// output any attributes
  attributes_.process(parentContext);
  writer << ">";

  // process the child templates
  templates_.process(parentContext);

  deNest(parentContext);
  /* we cannot determine whether the last child is really text because,
  where apply-templates is used, the dynamically generated content cannot be
  known at setup time.
  Bearing in mind that the main imperative is to avoid contaminating literal
  content with whitespace added by formatting, the safer course is to avoid
  formatting before an end node.
  A consequence is some piling up of end nodes, particularly in complex output.
  If good formatting is an imperative, it is best to use the building mode.
  */

// Simulate the output of an end node.
  writer << "</" << elementName << ">";
  writer.flush();

  return true;
}

}
