/***************************************************************************
                          generalcomparison.cpp  -  description
                             -------------------
    begin                : Thu Sep 15 2005
    copyright            : (C) 2005 by John Redmond
    email                : john_redmond@optusnet.com.au
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "generalcomparison.h"
#include "xpathstring.h"
#include "boolean.h"
#include "number.h"
#include "context.h"
#include "domwriter.h"

namespace limpid {

GeneralComparison::GeneralComparison(): logic_(true) {}

GeneralComparison::GeneralComparison(const DOMString& typeString): logic_(typeString == "=") {}

Boolean GeneralComparison::getBoolean(Context& context) {
  if (getLength() != 2) return false;  // should be impossible

  // this preliminary is necessary to evaluate any possible variables
  Expr firstExpr = components->at(0)->eval(context);
  Expr secondExpr = components->at(1)->eval(context);

  bool firstIsSet = firstExpr.isNodeSet();
  bool secondIsSet = secondExpr.isNodeSet();

  if (firstIsSet) {
    NodeSet firstSet = firstExpr.getNodeSet(context);
    if (!secondIsSet)
      return compareOneNodeSet(firstSet, secondExpr.getContent(), context);

    // otherwise we compare two node sets
    NodeSet secondSet = secondExpr.getNodeSet(context);

    // comparing two node sets can be ambiguous
    return compareNodeSets(firstSet, secondSet);
  }

  if (secondIsSet) {  // firstIsSet is false at this point
    NodeSet secondSet = firstExpr.getNodeSet(context);
    return compareOneNodeSet(secondSet, firstExpr.getContent(), context);
  }

  // otherwise it is a simple string comparison
  return comparePair(firstExpr.getContent(), secondExpr.getContent(), context);

}

bool GeneralComparison::comparePair(XPathExpression& first, XPathExpression& second,
      Context& context) {
  XPathString s0 = first.getString(context);
  XPathString s1 = second.getString(context);
  return compareStrings(s0, s1) == logic_;
}

bool GeneralComparison::compareOneNodeSet(const NodeSet& nodeSet,
      const XPathExpression& secondExpression, Context& context) {
  return matchSet(nodeSet, const_cast<XPathExpression &>(secondExpression).getString(context));
}

bool GeneralComparison::compareNodeSets(const NodeSet& firstSet, const NodeSet& secondSet) {
  size_t firstLength = firstSet.getLength();
  size_t secondLength = secondSet.getLength();

  // first, we do a comparison of string values
  for (size_t index = 0; index < secondLength; ++index) {
    XPathString secondString = secondSet.getItem(index).getValue();
    if (matchSet(firstSet, secondString) ) return true;
  }

  // but we MAY have a comparison of the single nodes in each set
  if (firstLength == 1 && secondLength == 1) {
    if (compareNodes(firstSet.item(0), secondSet.item(0)) == logic_) return true;
  }

  return false;
}

bool GeneralComparison::matchSet(const NodeSet& nodeSet, const XPathString& testString) {

  size_t setLength = nodeSet.getLength();

  for (size_t index = 0; index < setLength; ++index) {
    XPathString nodeString = nodeSet.getItem(index).getValue();
    if (testString.empty() && nodeString.empty()) continue; // no meaningful comparison
    if (compareStrings(nodeString, testString)  == logic_) return true;
  }
                                                                                            // otherwise, we do a comparison of string values
  return false;
}

}
