/***************************************************************************
                          graphnode.h  -  description
                             -------------------
    begin                : Tue Nov 25 2003
    copyright            : (C) 2003 by John Redmond
    email                : john_redmond@optusnet.com.au
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef GRAPHNODES_H
#define GRAPHNODES_H

#include <iostream>
#include <list>
#include <set>
#include <algorithm>
#include "copied.h"
#include "domstring.h"
#include "element.h"
#include "writer.h"
#include "pointer.h"

/**
  *@author John Redmond
  */
namespace limpid {
  class GraphNode;
    typedef std::list<GraphNode *> GraphNodeList;
    typedef std::set<GraphNode *> GraphNodeSet;
    typedef std::list<DOMString> NameList;

  class GraphNode: public Copied {
    friend class RepeatNode;
    friend class GraphNavigator;

  public:
    enum GraphType { ELEMENT, SEQUENCE, CHOICE, LOOP, SUCCESS, END, ERROR };

    GraphNode(const DOMString& name, GraphType graphType = ELEMENT);
    GraphNode(Element& element);

    static GraphNode failNode;
    static GraphNode endNode;

    GraphType getType() { return type; }
    DOMString& getInstanceName() { return instanceName; }
    virtual GraphNode& getContentStart() { return *this; }

    virtual void clearHits() { hits = 0; }
    virtual void clearChildHits() {}
    size_t getHits() { return hits; }

    GraphNodeList getAvailableNodes();
    GraphNode& find(DOMString& find, bool includeEnd = false);
    NameList getOptions();
    bool isValid () { return type != ERROR; }
    bool isEnd ();
    size_t getSize();
    GraphNode& copy() { return *(new GraphNode(*this)); }


  protected:
    GraphType type;
    DOMString instanceName;
    GraphNode *nextNode;  // this is volatile
    size_t minOccurs, maxOccurs, hits;

    bool isAvailable();
    bool wasAvailable();
    bool isRequired();
    bool wasRequired();

    virtual void clearContentHits() {}
    void addHit() { if (type != END) ++hits; }
    virtual void connect() {}

    void setType(Element& element);
    void setFrequency(Element& element);
    int getIntAttribute(Element& element, const char* attrName);

    virtual GraphNode& getNextNode(){ if (nextNode) return *nextNode; else return failNode; }
    virtual void setNextNode(GraphNode& node) { nextNode = &node; node.connect(); }
    virtual bool addAvailableNodes(GraphNodeList& resultList,
                                           GraphNodeSet& resultSet);
    virtual GraphNode& findNextNode(DOMString& name);
    virtual GraphNode& getPossibleNode() { return getNextNode(); }

  };

  class LoopNode : public GraphNode  {
    class RepeatNode;
    friend class RepeatNode;
  public:

    LoopNode();
//    LoopNode(RepeatNode& owner);
    RepeatNode *repeatNode;

  protected:
    RepeatNode& getRepeatNode() { return *repeatNode; }
    void setRepeatNode(RepeatNode& repeat) {repeatNode = &repeat; }
    bool addAvailableNodes(GraphNodeList& resultList, GraphNodeSet& resultSet);
    GraphNode& getPossibleNode();
    GraphNode& findNextNode(DOMString& name);
  };

  class RepeatNode : public GraphNode  {
  public:
    friend class LoopNode;

    RepeatNode(GraphType t);
    RepeatNode(Element &element);

    size_t getSize();
    void clearHits() { hits = 0; clearChildHits(); }
    void clearChildHits();
    GraphNode& getContentStart();
    RepeatNode& copy() { return *(new RepeatNode(*this)); }

  protected:
    Pointer<GraphNodeList> nodes;
    Pointer<LoopNode> loopNode;

    void addContentNodes(Element& root);
    void addContent(Element& schemaElement);
    void addContentNode(GraphNode& graphNode);
    GraphNode& findNextNode(DOMString& name);

    GraphNode& getNextNode();
    bool addAvailableNodes(GraphNodeList& resultList, GraphNodeSet& resultSet);
    GraphNode& getPossibleNode();
    LoopNode& getLoopNode() { return *loopNode; }
    void setNextNode(GraphNode& node) { loopNode->setNextNode(node); }
    void connect();
  };
}

#endif
