/***************************************************************************
 *   Copyright (C) 2007 by John  Redmond   *
 *   john_redmond@optusnet.com.au                                          *
 *                                                                      *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "htmlstyler.h"
#include "documentsummary.h"
#include "domstring.h"
#include "liststream.h"
#include "element.h"
#include "text.h"
#include "reader.h"
#include "stringwriter.h"

  namespace limpid {

Document& HTMLStyler::processDocument(Document& sourceDocument) {
  buildDocument(sourceDocument);
  return outputDocument_;
}

void HTMLStyler::buildDocument(Document& sourceDocument) {
  Document document;
  outputDocument_ = document; // make sure we start with an empty document
  Element documentElement = getDocumentElement();
  outputDocument_.appendChild(documentElement);

  addDocumentTemplateElement(documentElement);

  ListStream listStream(getChild("ELEMENTS", sourceDocument.getDocumentElement()));

  while (true) {
    Node& node = listStream.nextNode();
    if (!node.isValid()) break;

    if (node.getNodeName() == "ELEMENT")
     addDefinedTemplate(node, documentElement);
  }
}

void HTMLStyler::addDefinedTemplate(Node& node, Element& documentElement) {
  Element& element = static_cast<Element &>(node);

  DOMString matchString = getChild("MATCH", node).getValue();
  DOMString tagString = getChild("TAG", node).getValue();
  DOMString nameString = getChild("NAME", node).getValue();
  DOMString depthString = getChild("DEPTH", node).getValue();

  if (tagString.empty())
    tagString = "p";

  addTemplateElement(matchString, tagString, nameString, depthString, documentElement);
}

void HTMLStyler::addDocumentTemplateElement(Element& parentElement) {
  Element templateElement("xsl:template");
  parentElement.appendChild(templateElement);
  templateElement.setAttribute("match", "/");

  Element htmlElement("html");
  htmlElement.setAttribute("xmlns", "http://www.w3c.org/1999/xhtml");
  templateElement.appendChild(htmlElement);

  Element headElement("head");
  htmlElement.appendChild(headElement);

  if (expandingStyles_) {
    if (cssScreenPath_.empty() || cssPrintPath_.empty()) // use a single stylesheet, if one is defined
      expandStyle(headElement, "", cssScreenPath_, "text/css");
    else { // use separate stylesheets
      expandStyle(headElement, "screen", cssScreenPath_, "text/css");
      expandStyle(headElement, "print", cssPrintPath_, "text/css");
    }
  }
  else {
    if (cssScreenPath_.empty() || cssPrintPath_.empty()) // use a single stylesheet, if one is defined
      addLink(headElement, "", cssScreenPath_, "text/css");
    else { // use separate stylesheets
      addLink(headElement, "screen", cssScreenPath_, "text/css");
      addLink(headElement, "print", cssPrintPath_, "text/css");
    }
  }

  Element bodyElement("body");
  htmlElement.appendChild(bodyElement);

  Element applyElement("xsl:apply-templates");
  bodyElement.appendChild(applyElement);
}

void HTMLStyler::addTemplateElement(const DOMString& matchString, const DOMString& tagString, const DOMString& nameString, const DOMString& depthString, Element& parentElement) {
  Element templateElement("xsl:template");
  parentElement.appendChild(templateElement);
  templateElement.setAttribute("match", matchString);

  DOMString tagName = tagString;
  if (tagName == "root")
    tagName = "div";

  Element tagElement(tagName);
  templateElement.appendChild(tagElement);
  if (tagName == "div") {
    DOMString attributeString = "L";
    attributeString += depthString;

    if (tagString == "root")
      tagElement.setAttribute("id", attributeString);
    else
      tagElement.setAttribute("class", attributeString);

    tagElement.setAttribute("tag", nameString);
  }

  if (false)
    addTag(tagElement, matchString, tagName);

  Element applyElement("xsl:apply-templates");
  tagElement.appendChild(applyElement);
}

Element HTMLStyler::getDocumentElement() {
  Element documentElement("xsl:stylesheet");
  documentElement.setAttribute("xmlns:xsl", "http://www.w3.org/1999/XSL/Transform");
  documentElement.setAttribute("version", "1.1");

  Element outputElement("xsl:output");
  outputElement.setAttribute("method", "html");
  outputElement.setAttribute("doctype-public", "-//W3C//DTD XHTML 1.0 Transitional//EN");
  documentElement.appendChild(outputElement);

  return documentElement;
}

void HTMLStyler::addLink(Element& parentElement, const DOMString& use,
    const DOMString& path, const DOMString& type) {
  if (path.empty()) return; // no link is specified

  Element linkElement("link");
  parentElement.appendChild(linkElement);

  linkElement.setAttribute("rel", "stylesheet");
  if (!use.empty())
    linkElement.setAttribute("media", use);
  linkElement.setAttribute("href", path);
  linkElement.setAttribute("type", type);
}

void HTMLStyler::expandStyle(Element& parentElement, const DOMString& use,
    const DOMString& path, const DOMString& type) {
  if (path.empty()) return;

  Element linkElement("style");
  if (!use.empty())
    linkElement.setAttribute("media", use);
  linkElement.setAttribute("type", type);
  parentElement.appendChild(linkElement);

  Reader reader(path);
  StringWriter stringWriter;
  stringWriter.putChar('\n');

  while (true) {
    Char c = reader.nextChar();
    if (c == FAILCHAR) break;
    else stringWriter.putChar(c);
  }

  Text text(stringWriter.getString());
  linkElement.appendChild(text);
}

void HTMLStyler::addTag(Element& parentElement, const DOMString& matchString,
    const DOMString& tagName) {
  Element labelElement("xsl-text");
  DOMString textString = tagName;
  textString += "(";
  textString += matchString;
  textString += "): ";
  addText(labelElement, textString);
  parentElement.appendChild(labelElement);
}

}
