/***************************************************************************
 *   Copyright (C) 2007 by root   *
 *   john_redmond@optusnet.com.au   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "inspector.h"
#include "document.h"
#include "element.h"
#include "text.h"
#include "liststream.h"
#include "converter.h"

namespace limpid {

Document& Inspector::processDocument(Document& sourceDocument) {
  // set up the node summaries in the document summary
  documentSummary_.setup(sourceDocument);
  // configure the node summaries
  setTypes();
  // get an XML document based on these node summaries
  Document& specificationDocument = documentSummary_.getDocument();

  Element& documentElement = specificationDocument.getDocumentElement();

  // insert the default LaTeX styles into the XML document
  Node& definitionsElement = getChild("DEFINITIONS", documentElement);
  Element latexElement("LATEX");
  definitionsElement.appendChild(latexElement);

  appendFormat(latexElement);
  appendPackages(latexElement);
  setStyles(latexElement);

  return specificationDocument;
}

void Inspector::setTypes() {
  setRoot();
  setTitle();
  setSections();
  setTailHeadings();
  setHeadings();
  setSubheadings();
  setMixed();
  setGroups();
  setFeatures();
  setUntyped();
}

void Inspector::setRoot() {
  NodeSummary& rootSummary = documentSummary_.getSummary(0);
  rootSummary.tag = "root";
  rootSummary.type = "ROOT";
  rootSummary.style = "root";
}

void Inspector::setTitle() {
  // move through the summaries in document order
  for (size_t index = 0; index < documentSummary_.getPathCount(); ++index) {
    NodeSummary& nodeSummary = documentSummary_.getSummary(index);
    if (!nodeSummary.type.empty()) continue;
    if (nodeSummary.isLeaf && nodeSummary.alwaysFirst &&
        nodeSummary.totalCount == 1 && nodeSummary.selfMax == 1) {
      nodeSummary.tag = "h1";
      nodeSummary.type = "TITLE";
      nodeSummary.style = "atitlea";
      break;  // stop at one title
    }
  }
}

void Inspector::setStyles(Node& parentNode) {
  Element stylesElement("STYLES");
  parentNode.appendChild(stylesElement);

  setStyle("achead", "\\textbf{\\Large #1\\\\}", stylesElement);
  setStyle("alhead", "\\textbf{\\large #1}", stylesElement);
  setStyle("arhead", "\\textbf{\\large #1}", stylesElement);
  setStyle("atitlea", "\\begin{center}\\textbf{\\Large #1}\\end{center}", stylesElement);
  setStyle("atitle", "\\textbf{\\Large #1}", stylesElement);

  setStyle("asection", "\\textbf{\\Large\\centering #1}\\medskip", stylesElement);
  setStyle("asubsection", "\\textbf{\\large\\centering #1}\\medskip", stylesElement);
  setStyle("asubsubsection", "\\textbf{\\large\\centering #1}\\medskip", stylesElement);

  setStyle("aheadinga", "\\smallskip\\textbf{#1}", stylesElement);
  setStyle("aheadingb", "\\smallskip\\textbf{#1}", stylesElement);
  setStyle("aheadingc", "\\textbf{#1}", stylesElement);
  setStyle("aheadingd", "\\textbf{#1}", stylesElement);
  setStyle("aheadinge", "\\textbf{#1}", stylesElement);
  setStyle("aheadingf", "\\textbf{#1}", stylesElement);

  setStyle("asubheadinga", "\\textbf{#1}", stylesElement);
  setStyle("asubheadingb", "\\textbf{#1}", stylesElement);
  setStyle("asubheadingc", "\\textbf{#1}", stylesElement);
  setStyle("asubheadingd", "\\textbf{#1}", stylesElement);

  setStyle("atailheadinga", "\\emph{#1\\smallskip}", stylesElement);
  setStyle("atailheadingb", "\\emph{#1\\smallskip}", stylesElement);
  setStyle("atailheadingc", "\\emph{#1}", stylesElement);
  setStyle("atailheadingd", "\\emph{#1}", stylesElement);

  setStyle("afeaturea", "\\smallskip\\emph{#1}", stylesElement);
  setStyle("afeatureb", "\\emph{#1}", stylesElement);
  setStyle("afeaturec", "\\emph{#1}", stylesElement);
  setStyle("afeatured", "\\emph{#1}", stylesElement);
  setStyle("afeaturee", "\\emph{#1}", stylesElement);
  setStyle("afeaturef", "\\textbf{#1}", stylesElement);
  setStyle("afeatureg", "\\textbf{#1}", stylesElement);
  setStyle("afeatureh", "\\textsc{#1}", stylesElement);
  setStyle("afeaturei", "\\textsc{#1}", stylesElement);
  setStyle("afeaturej", "\\textsc{#1}", stylesElement);

  setStyle("amixeda", "{#1}", stylesElement);
  setStyle("amixedb", "{#1}", stylesElement);
  setStyle("averbatima", "\\begin{Verbatim}#1\\end{Verbatim}", stylesElement);
}

void Inspector::setStyle(const DOMString& key, const DOMString& preString, Node& parentNode) {
  Element styleElement("STYLE");
  parentNode.appendChild(styleElement);
  addLeaf("NAME", key, styleElement);
  addLeaf("PRE", preString, styleElement);
}

void Inspector::setStyle(const DOMString& key, const DOMString& preString, const DOMString& postString, Node& parentNode) {
  Element styleElement("STYLE");
  parentNode.appendChild(styleElement);
  addLeaf("NAME", key, styleElement);
  addLeaf("PRE", preString, styleElement);
  if (!postString.empty())
    addLeaf("POST", postString, styleElement);
}

void Inspector::setMixed() {
  for (size_t index = 0; index < documentSummary_.getPathCount(); ++index) {
    NodeSummary& nodeSummary = documentSummary_.getSummary(index);
    if (!nodeSummary.type.empty()) continue; // type is already set
    if (nodeSummary.hasText & nodeSummary.childNames.size()) {
      nodeSummary.type = "MIXED";
      nodeSummary.style = "amixeda";
    }
  }
}

void Inspector::setSections() {
  DOMString sectionPath, subsectionPath, subsubsectionPath;
  int level = 0, sectionDescendantCount = 0, subsectionDescendantCount = 0, subsubsectionDescendantCount = 0;


  // move through the summaries in document order
  for (size_t index = 0; index < documentSummary_.getPathCount(); ++index) {
    NodeSummary& nodeSummary = documentSummary_.getSummary(index);
    if (!nodeSummary.type.empty()) continue; // type is already set
    if (nodeSummary.childPaths.empty()) continue; // we need children

    // exclude a possible title (which is unique):
    if (nodeSummary.totalCount < 2) continue; // we want multiples

    NodeSummary& childSummary = documentSummary_.getSummary(nodeSummary.childPaths[0]);
//    if (!nodeSummary.type.empty()) continue;

    if (childSummary.isLeaf && childSummary.alwaysFirst && childSummary.selfMax == 1 && !childSummary.textBefore && !nodeSummary.textAfter) {
      int nodeDescendantCount = nodeSummary.descendantMax;

      if (nodeDescendantCount > sectionDescendantCount) {
        sectionPath = nodeSummary.path;
        sectionDescendantCount = nodeDescendantCount;
      }
      else if (nodeDescendantCount > subsectionDescendantCount) {
        subsectionPath = nodeSummary.path;
        subsectionDescendantCount = nodeDescendantCount;
      }
      else if (nodeDescendantCount > subsubsectionDescendantCount) {
        subsubsectionPath = nodeSummary.path;
        subsubsectionDescendantCount = nodeDescendantCount;
      }
    }
  }

  setSection(sectionPath, "SECTION");
  setSection(subsectionPath, "SUBSECTION");
//  setSection(subsubsectionPath, "SUBSUBSECTION");
}

void Inspector::setSection(const DOMString& pathString, const DOMString& sectionType) {
  if (pathString.empty()) return; // no such section
  DOMString tagType = sectionType;
  tagType += "_TAG";

  NodeSummary& nodeSummary = documentSummary_.getSummary(pathString);
  nodeSummary.tag = "div";
  nodeSummary.type = sectionType;

  NodeSummary& childSummary = documentSummary_.getSummary(nodeSummary.childPaths[0]);
  childSummary.type = tagType;

  if (sectionType == "SECTION") {
    childSummary.tag = "h2";
    childSummary.style = "asection";
  }
  else if (sectionType == "SUBSECTION") {
    childSummary.tag = "h3";
    childSummary.style = "asubsection";
  }
//  else
//    childSummary.style = "asubsubsection";
}

void Inspector::setHeadings() {
  Char suffix = 'a';

  for(size_t index = 0; index < documentSummary_.getPathCount(); ++index) {
    NodeSummary& nodeSummary = documentSummary_.getSummary(index);

    if (!nodeSummary.type.empty()) continue;// || nodeSummary.selfMax != 1) continue;
    if (nodeSummary.textBefore || nodeSummary.textAfter) continue;  // not in mixed content
    if (!nodeSummary.alwaysFirst) continue;

    DOMString parentPath = nodeSummary.parentPath;
    if (!documentSummary_.hasSummary(parentPath)) continue;  // should never happen

    NodeSummary& parentSummary = documentSummary_.getSummary(parentPath);
    if (parentSummary.hasTailHeading) continue; // there must be a tail summary

    setHeading(nodeSummary, suffix);
    if (suffix < 'd') ++suffix;
//    parentSummary.hasHeading = true; // mark the presence of a summary
  }
}

void Inspector::setSubheadings() {
  Char suffix = 'a';

  for(size_t index = 0; index < documentSummary_.getPathCount(); ++index) {
    NodeSummary& nodeSummary = documentSummary_.getSummary(index);

    if (!nodeSummary.type.empty()) continue;// || nodeSummary.selfMax != 1) continue;
    if (nodeSummary.textBefore || nodeSummary.textAfter) continue;  // not in mixed content
    if (!nodeSummary.alwaysFirst) continue;

    DOMString parentPath = nodeSummary.parentPath;
    if (!documentSummary_.hasSummary(parentPath)) continue;  // should never happen

    NodeSummary& parentSummary = documentSummary_.getSummary(parentPath);
    if (parentSummary.hasTailHeading) continue; // there must be a tail summary

    setHeading(nodeSummary, suffix);
    if (suffix < 'd') ++suffix;
//    parentSummary.hasHeading = true; // mark the presence of a summary
  }
}

void Inspector::setTailHeadings() {
  Char suffix = 'a';

  for(size_t index = 0; index < documentSummary_.getPathCount(); ++index) {
    NodeSummary& nodeSummary = documentSummary_.getSummary(index);

    if (!nodeSummary.type.empty()) continue;// || nodeSummary.selfMax != 1) continue;
    if (nodeSummary.textBefore || nodeSummary.textAfter) continue;  // not in mixed content
    if (!nodeSummary.alwaysLast) continue;

    DOMString parentPath = nodeSummary.parentPath;
    if (!documentSummary_.hasSummary(parentPath)) continue;  // should never happen
    documentSummary_.getSummary(parentPath).hasTailHeading = true;

    setTailHeading(nodeSummary, suffix);
    if (suffix < 'd') ++suffix;
  }
}

void Inspector::setHeading(NodeSummary& nodeSummary, Char suffix) {
  DOMString style = "aheading";
  style += suffix;
  nodeSummary.style = style;

  DOMString tag = "h";
  tag += Converter::toString(nodeSummary.depth - 1);
  nodeSummary.tag = tag;

  nodeSummary.type = "HEADING";
}

void Inspector::setSubheading(NodeSummary& nodeSummary, Char suffix) {
  DOMString style = "asubheading";
  style += suffix;
  nodeSummary.style = style;

  DOMString tag = "h";
  tag += Converter::toString(nodeSummary.depth - 1);
  nodeSummary.tag = tag;

  nodeSummary.type = "SUBHEADING";
}

void Inspector::setTailHeading(NodeSummary& nodeSummary, Char suffix) {
  DOMString style = "atailheading";
  style += suffix;
  nodeSummary.style = style;

  DOMString tag = "emph";
  nodeSummary.tag = tag;

  nodeSummary.type = "HEADING";
}

void Inspector::setGroups() {
  for (size_t index = 0; index < documentSummary_.getPathCount(); ++index) {
    NodeSummary& nodeSummary = documentSummary_.getSummary(index);
    if (!nodeSummary.type.empty()) continue;
    if (nodeSummary.isLeaf) continue;

    nodeSummary.type = "BLOCK";
    nodeSummary.tag = "div";
    nodeSummary.style = "block";
  }
}

void Inspector::setFeatures() {
  Char tag = 'a';

  for (size_t index = 0; index < documentSummary_.getPathCount(); ++index) {
    NodeSummary& nodeSummary = documentSummary_.getSummary(index);

    if (!nodeSummary.type.empty()) continue;

    if (nodeSummary.isLeaf && nodeSummary.isFeature) {
      setFeature(nodeSummary, tag++);
      continue;
    }

    if (!nodeSummary.isLeaf && !nodeSummary.hasText) continue;
//    if (nodeSummary.everLast) continue;
    if (nodeSummary.selfMax > 2) continue;

    setFeature(nodeSummary, tag);
    if (tag < 'j') ++tag;
  }
}

void Inspector::setFeature(NodeSummary& nodeSummary, Char tag) {
//  if (nodeSummary.selfMax > 2) return;

  nodeSummary.tag = "em";
  nodeSummary.type = "FEATURE";

  DOMString style = "afeature";
  style += tag;
  nodeSummary.style = style;
}

void Inspector::setUntyped() {
  for (size_t index = 0; index < documentSummary_.getPathCount(); ++index) {
    NodeSummary& nodeSummary = documentSummary_.getSummary(index);
    if (!nodeSummary.type.empty()) continue;

    nodeSummary.style = "default";
    nodeSummary.type = "DEFAULT";
  }
}

void Inspector::appendFormat(Node& parentNode) {
  Element formatElement("FORMAT");
  parentNode.appendChild(formatElement);
      addLeaf("CLASS", "article", formatElement);
  addLeaf("PAGE", "a4paper, 10pt", formatElement);
  addLeaf("COLUMNS", "2", formatElement);
  addLeaf("PAGE_NUMBERING", "arabic", formatElement);
  appendTOCHeaders(formatElement);
  appendBodyHeaders(formatElement);
}

void Inspector::appendTOCHeaders(Node& parentNode) {
  Element tocElement("TOC_HEADERS");
  parentNode.appendChild(tocElement);
  addLeaf("CENTER", "true", tocElement);
}

void Inspector::appendBodyHeaders(Node& parentNode) {
  Element bodyElement("BODY_HEADERS");
  parentNode.appendChild(bodyElement);
  addLeaf("LEFT", "true", bodyElement);
  addLeaf("RIGHT", "true", bodyElement);

}

void Inspector::appendHeader(const DOMString& name, const DOMString& style, Node& parentNode) {
  Element headerElement("HEADER");
  parentNode.appendChild(headerElement);
  addLeaf("PLACE", name, headerElement);
  addLeaf("STYLE", style, headerElement);
}

void Inspector::appendPackages(Node& parentNode) {
  Element packagesElement("PACKAGES");
  parentNode.appendChild(packagesElement);
  appendPackage("geometry", "a4paper, top=15mm, left=25mm, bottom=25mm, right=0mm, headheight=55pt, includehead, includemp", packagesElement);
  appendPackage("fancyhdr", "", packagesElement);
  appendPackage("multicol", "", packagesElement);
  appendPackage("fancyvrb", "", packagesElement);
}

void Inspector::appendPackage(const DOMString& name, const DOMString& value, Node& parentNode) {
  Element packageElement("PACKAGE");
  parentNode.appendChild(packageElement);

  addLeaf("NAME", name, packageElement);
  addLeaf("PARAMETERS", value, packageElement);
}

}
