/***************************************************************************
                          keymanager.cpp  -  description
                             -------------------
    begin                : Wed Aug 3 2005
    copyright            : (C) 2005 by John Redmond
    email                : john_redmond@optusnet.com.au
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "keymanager.h"
#include "keyspecs.h"
#include "stringreader.h"
#include "tokenstream.h"

namespace limpid {

KeyManager::KeyManager() {}

// where we have a key definition of the form:
// <xsl:key name='test' match='block' use='@title'/>
// and use it as:
// <xsl:apply-templates select='key(test, tone)'/>
// getKeyedString converts this the select path to:
// block[@title = tone]

DOMString KeyManager::getExprString(const ConfigurationSet& configurationSet,
    const DOMString& sourceString) {
  DOMString outputString = sourceString;

  while (true) {
    DOMString keyString = "key(";
    size_t matchPos = outputString.rfind(keyString);
    if (matchPos == outputString.npos) break; // no key expression found

    size_t inputLength = outputString.size();
    size_t endPos = getKeyEnd(outputString, keyString, matchPos);
    if (endPos == outputString.npos) break; // error: no closing parenthesis for key(...)

    DOMString extractedString = outputString.substr(matchPos, inputLength - matchPos);
    DOMString keyedString = getKeyedExprString(configurationSet, extractedString);

    DOMString newString;
    if (matchPos)
      newString += outputString.substr(0, matchPos);  // before key(...)
    newString += keyedString; // replacement for key(...)

    if (endPos < inputLength)
      newString += outputString.substr(endPos, inputLength - endPos); // after key(...)

    outputString = newString; // ready to repeat loop
  }

  return outputString;
}

size_t KeyManager::getKeyEnd(const DOMString& sourceString,
    const DOMString& keyString, size_t startIndex) {
  size_t index = startIndex + keyString.size();
  int depth = 1;

  while (depth) {  // the parentheses had better be balanced here!
    Char c = sourceString[index++];
    if (c == '(') ++depth;
    else if (c == ')') --depth;
  }

  return index;
}

DOMString KeyManager::getKeyedExprString(const ConfigurationSet& configurationSet,
    const DOMString& keyString) {
  size_t commaPos = keyString.indexOf(',');
  if (commaPos == keyString.npos) {
    wrt::Errors << "Missing comma in key string: " << keyString << wrt::endl;
    exit(1);
    return keyString; // fatal error, cannot continue
  }

  DOMString lhs = keyString.substr(0, commaPos);
  lhs = lhs.substr(4, lhs.size() - 4);  // remove "key("
  lhs.trim(); // must remove whitespace
  lhs = lhs.substr(1, lhs.size() - 2); // remove quotes

  DOMString rhs = keyString.substr(commaPos + 1, keyString.size() - commaPos - 2);
  rhs.trim();  // here we have removed the comma, closing parenthesis and wihtespace

  KeyMap::const_iterator p = configurationSet.keyMap.find(lhs);
  if (p == configurationSet.keyMap.end()) {
    wrt::Errors << "Key name not found: " << lhs << wrt::endl;
    return keyString; // a fatal error
  }

  DOMString second = p->second.match;
  KeySpecs keySpecs = p->second;

  DOMString exprString = keySpecs.match;
  exprString += "[";
  exprString += keySpecs.use;
  exprString += " = ";
  exprString += rhs.trim();
  exprString += "]";

  return exprString;
}

}
