/***************************************************************************
//  *   Copyright (C) 2007 by John Redmond                                 *
 *   john_redmond@optusnet.com.au                                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "latexconfiguration.h"
#include "domutils.h"
#include "converter.h"
#include "getter.h"
#include "xmlparser.h"
#include "cssparser.h"
#include "liststream.h"
#include "nodestream.h"
#include "siblingstream.h"
#include "argparser.h"
#include "stringreader.h"
#include "matchedtemplate.h"

#include "xpathfactory.h"

namespace limpid {

LatexConfiguration::LatexConfiguration(): verbose_(false) {}

Document LatexConfiguration::getDefaultDocument() {
  DOMString specString = "document { class: article; format: portrait; font: palatino; text-align: left; toc-columns: 0; body-columns: 1; page-numbering: arabic; numbering-depth: 2; }\
\
package { select: array, fancyvrb, color, relsize, multicol, tabulary, url; }\
package { select: fancyhdr, graphicx, caption, textcomp, float, alltt;}\
\
geometry { param: a4paper, top=25mm, left=30mm, bottom=40mm, right=0mm, headheight=55pt, includehead, includemp }\
lengths { tymin: 35pt; tymax: 0.4\\linewidth; extrarowheight: 4pt; parskip: \"0.5ex plus0.2ex minus0.1ex\"; }\
constants {   breaklevel: 2500; }\
floats { figure: h; table: h; }\
\
title { font-weight: bold; font-size: Large; }\
\
section { font-weight: bold; font-size: large; text-align: center; }\
subsection { font-weight: bold; font-size: large; text-align: center; }\
subsubsection { font-weight: bold; text-align: center; }\
subsubsubsection { font-weight: bold; }\
\
fonts { times: 9/10.5, ptm, m, n; \
 bookman: 10/11.5, pbk, m, n;\
 charter: 10/11.5, pch, m, n;\
 palatino: 10/11.5, ppl, m, n;\
 avantgarde: 9/10.5, pag, m, n;\
 chancery: 9/10.5, pzc, m, il; \
 courier: 9/11, pcr, m, n;\
 helvetica: 8/9.5, phv, m, n; }\
\
toc-headers, body-headers, headers { font-weight: bold; font-size: normal; }\
toc-headers { font-size: large; }\
\
/* hidden */\
head { display: none; }\
\
/* lists */\
ol { environment: enumerate; }\
ul { environment: itemize; }\
dl { environment: description; }\
li { text-before: \\item; } \
dt { text-left: \\item[; text-right: ];}\
dd { text-weight: normal; }\
\
/* special */\
pre { environment: Verbatim; }\
\
/* leaves */\
h1, h2, h3, h4 { text-weight: bold; font-size: large; }\
h1 { font-size: Large; }\
p { display: block; }\
emph, em { font-shape: italic; }\
\
";

  StringReader reader(specString);
  CSSParser parser(reader);
  Getter getter(parser);

  return getter.getDocument();
}

Document LatexConfiguration::getConfigurationDocument(const DOMString& configPath) {
  Reader reader(configPath);

  if (configPath.endsWith(".xml")) {
    XMLParser xmlParser(reader);
    Getter getter(xmlParser);
    return getter.getDocument();
  }
  else if (configPath.endsWith(".css")) {
    CSSParser cssParser(reader);
    Getter getter(cssParser);
    return getter.getDocument();
  }

  // else we just use the default document
  wrt::Console << "cannot configure texconverter with " << configPath << wrt::endl;
  wrt::Console << "the file name must end in '.xml' or '.css'" << wrt::endl;
  return Document::failDocument;
}

void LatexConfiguration::configure(const DOMString& configPath) {
  Document defaultDocument = getDefaultDocument();
  documentConfiguration_.configure(defaultDocument);

  if (configPath.empty())
    applyConfiguration(defaultDocument, true);  /// all we have are the defaults

  else {
    Document configDocument = getConfigurationDocument(configPath);
    documentConfiguration_.configure(configDocument);

    applyConfiguration(defaultDocument, documentConfiguration_.isXHTML());
    applyConfiguration(configDocument, true);
  }

  finalizeTagStyles();
}

void LatexConfiguration::applyConfiguration(Document& configDocument, bool force) {
  Element& documentElement = configDocument.getDocumentElement();
  NodeStream nodeStream(documentElement, false, false);

  while (true) {
    Node& node = nodeStream.nextNode();
    if (!node.isValid()) break;
    if (node.getNodeType() != Node::ELEMENT_NODE) continue;
    if (node.getNodeName() != "TAG") continue;

    Element& element = static_cast<Element &>(node);

    DOMString tagString = element.getValue();
    if (tagString.empty()) {
      continue;
    }

    if (tagString == "title" || tagString == "section" || tagString == "subsection" || tagString == "subsubsection") {
      configureSection(tagString, node);
      continue;
    }
    if (tagString == "fonts") {
      configureFonts(node);
      continue;
    }
    if (tagString == "lengths") {
      configureLengths(node);
      continue;
    }
    if (tagString == "constants") {
      configureConstants(node);
      continue;
    }
    if (tagString == "floats") {
      configureFloats(node);
      continue;
    }
    if (tagString == "toc-headers") {
      configureHeaders(node, tocHeadersMap_);
      continue;
    }
    if (tagString == "body-headers") {
      configureHeaders(node, bodyHeadersMap_);
      continue;
    }
    if (tagString == "body-headers-contentxxx") {
      configureHeadersContent(node, bodyHeadersMap_);///headersContentMap_);
      continue;
    }

    if (force) {
      Element& tagElement = static_cast<Element &>(node);
      configureTag(tagString, tagElement);
    }
  }
}

/// configuration group (private)

void LatexConfiguration::configureSection(DOMString& sectionName, Node& tagNode) {
  SelectorMap::iterator p = selectorMap_.find(sectionName);
  if (p == selectorMap_.end()) {
    SelectorSpec selectorSpec;
    selectorMap_[sectionName] = selectorSpec;
  }

  SelectorSpec& thisSpec = selectorMap_[sectionName];
  populateSelectorSpec(tagNode, thisSpec, sectionName);
}

void LatexConfiguration::configureFonts(Node& tagNode) {
  SiblingStream siblingStream(tagNode);

  while (true) {
    Node& node = siblingStream.nextNode();
    if (!node.isValid()) break;

    if (node.getNodeType() != Node::ELEMENT_NODE) continue;
    if (node.getNodeName() != "SPEC") continue;

    DOMString fontName = static_cast<Element &>(node).getAttribute("property").trim();
    configureFont(fontName, node);
  }
}

void LatexConfiguration::configureLengths(Node& tagNode) {
  SiblingStream siblingStream(tagNode);
  while (true) {
    Node& siblingNode = siblingStream.nextNode();
    if (!siblingNode.isValid()) break;

    if (siblingNode.getNodeName() != "SPEC") continue;

    Element& specElement = static_cast<Element &>(siblingNode);
    DOMString propertyName = specElement.getAttribute("property").trim();
    lengthMap_[propertyName] = specElement.getValue();
  }
}

void LatexConfiguration::configureConstants(Node& tagNode) {
  SiblingStream siblingStream(tagNode);
  while (true) {
    Node& siblingNode = siblingStream.nextNode();
    if (!siblingNode.isValid()) break;

    if (siblingNode.getNodeName() != "SPEC") continue;

    Element& specElement = static_cast<Element &>(siblingNode);
    DOMString propertyName = specElement.getAttribute("property").trim();
    constantMap_[propertyName] = specElement.getValue();
  }
}

void LatexConfiguration::configureFloats(Node& tagNode) {
  SiblingStream siblingStream(tagNode);
  while (true) {
    Node& siblingNode = siblingStream.nextNode();
    if (!siblingNode.isValid()) break;

    if (siblingNode.getNodeName() != "SPEC") continue;

    Element& specElement = static_cast<Element &>(siblingNode);
    DOMString propertyName = "float-";
    propertyName += specElement.getAttribute("property").trim();

    detailMap_[propertyName] = specElement.getValue();
  }
}

void LatexConfiguration::configureHeaders(Node& tagNode, StringMap& headersMap) {
  SiblingStream siblingStream(tagNode, true);
  DOMString headersType = tagNode.getValue();

  while (true) {
    Node& node = siblingStream.nextNode();
    if (!node.isValid()) break;

    if (node.getNodeName() != "SPEC") continue;
    DOMString propertyString = static_cast<Element &>(node).getAttribute("property");

    if (headersType == "toc-headers")
      tocHeadersMap_[propertyString] = node.getValue();
    else
      bodyHeadersMap_[propertyString] = node.getValue();
  }

  Element& tagElement = static_cast<Element &>(tagNode);
  configureTag(tagElement.getValue(), tagElement);
}

void LatexConfiguration::configureHeadersContent(Node& tagNode, StringMap& headersMap) {
  SiblingStream siblingStream(tagNode, true);
  DOMString headersType = tagNode.getValue();

  while (true) {
    Node& node = siblingStream.nextNode();
    if (!node.isValid()) break;

    if (node.getNodeName() != "SPEC") continue;

    DOMString propertyString = static_cast<Element &>(node).getAttribute("property");

    if (headersType == "toc-headers")
      tocHeadersMap_[propertyString] = node.getValue();
    else
      bodyHeadersMap_[propertyString] = node.getValue();

    headersMap[propertyString] = node.getValue();
  }
}

/// getXXX() group: public

const DOMString& LatexConfiguration::getPreambleString(Document& dataDocument) {
  preambleString_ = documentConfiguration_.getPreambleString(dataDocument, titleString_);
  preambleString_ += getFontDefinitions();
  preambleString_ += getTagDefinitions();
  preambleString_ += documentConfiguration_.getLengthsString();
  preambleString_ += "\n%end preamble string\n";

  return preambleString_;
}

DOMString LatexConfiguration::getDefinitionsString() {
  return documentConfiguration_.getDefinitionsString();
}

const DOMString& LatexConfiguration::getSectionsString() {
  if (sectionsString_.empty())
    sectionsString_ = documentConfiguration_.getSectionsString();
  return sectionsString_;
}

const DOMString& LatexConfiguration::getDocumentStartString(Document& dataDocument) {
  documentStartString_ = documentConfiguration_.getDocumentStartString();

  DOMString titleString = "\\tocheadersStyle ";
  DOMString defaultTitle = "UNTITLED"; // default value
  if (documentMap_["titlePath"].empty())
    titleString += defaultTitle;
  else {
    DOMString trueTitle = findTitle(documentMap_["titlePath"], dataDocument);
    if (trueTitle.empty())
      trueTitle = "NOT FOUND";

    titleString += trueTitle;
  }

  documentStartString_.change("**TITLE**", titleString);  // patch the title
  documentStartString_.change("**TOCCENTER**", titleString);  // patch the title

  documentStartString_.change("**TOCLEFT**", getHeaderString("left", tocHeadersMap_, false));
  documentStartString_.change("**TOCRIGHT**", getHeaderString("right", tocHeadersMap_, false));

  documentStartString_.change("**BODYCENTER**", getHeaderString("center", bodyHeadersMap_, true));
  documentStartString_.change("**BODYLEFT**", getHeaderString("left", bodyHeadersMap_, true));
  documentStartString_.change("**BODYRIGHT**", getHeaderString("right", bodyHeadersMap_, true));

  return documentStartString_;
}

const DOMString& LatexConfiguration::getTailString() {
  if (tailString_.empty())
    tailString_ = documentConfiguration_.getTailString();

  return tailString_;
}

// end of public interfaces

// higher-level internal configuration functions

void LatexConfiguration::configureTag(const DOMString& refName, Element& tagElement) {
  DOMString mapName = refName; // the refname here is in CSS format
  bool isXPath = refName.startsWith("xpath:");
  if (isXPath)
    mapName.change("xpath:", "");

  addSelector(mapName, isXPath);

  SelectorSpec& thisSpec = selectorMap_[mapName]; // note that this is a REFERENCE
  populateSelectorSpec(tagElement, thisSpec, mapName);

  pathRegistry_.addPath(tagElement, thisSpec.priority, isXPath);
}

void LatexConfiguration::configureFont(DOMString& fontName, Node& specNode) {
  FontMap::iterator p = fontMap_.find(fontName);
  if (p == fontMap_.end()) {
    FontSpec fontSpec;
    fontMap_[fontName] = fontSpec;
  }

  FontSpec& thisSpec = fontMap_[fontName];
  ArgParser parser(specNode.getValue(), ',');
  DOMString sizeSpec = parser.getArg(0).trim();
  thisSpec.selector = parser.getArg(1).trim();
  thisSpec.weight = parser.getArg(2).trim();
  thisSpec.shape = parser.getArg(3).trim();
  DOMString size, height;
  int slashPos = sizeSpec.indexOf("/");

  if (slashPos != sizeSpec.npos) {
    thisSpec.size = sizeSpec.substr(0, slashPos);
    thisSpec.height = sizeSpec.substr(slashPos + 1);
  }
  else {
    thisSpec.size = sizeSpec;
    thisSpec.height = Converter::toString(Converter::toDouble(sizeSpec) + 2);
  }
}

void LatexConfiguration::registerTagName(const DOMString& tagName, bool isXPath) {
  DOMString alphaName = tagName;
  alphaName = getPreStringName(alphaName);

  if (alphaName.empty()) return;  // empty name !!!

  DOMString preStringName;
  if  (tagName.startsWith("L") || tagName == "div")
    preStringName += "\n";

  SelectorSpec selectorSpec;
  selectorMap_[tagName] = selectorSpec;

  preStringName += "\\";
  preStringName += alphaName;
  preStringName += "Style ";

  if (!selectorSpec.specialField.empty()) {
    DOMString fieldName = "\\";
    fieldName += selectorSpec.specialField;
    preStringMap_[tagName] = fieldName; // here we specify title or section, etc
  }
  else
    preStringMap_[tagName] = preStringName; // here we use indirection
}

void LatexConfiguration::populateSelectorSpec(Node& tagNode, SelectorSpec& selectorSpec, const DOMString& refName) {
  SiblingStream siblingStream(tagNode);
  while (true) {
    Node& siblingNode = siblingStream.nextNode();
    if (!siblingNode.isValid()) break;

    if (siblingNode.getNodeName() == "SPEC") {
      DOMString environmentName, alignmentName;

      Element& specElement = static_cast<Element &>(siblingNode);

      DOMString tagLabel = specElement.getAttribute("property").trim();
      DOMString propertyName = getCamelBack(tagLabel);

      DOMString rawValueString = siblingNode.getValue();
      DOMString valueString = correct(rawValueString); // get the string right

      if (propertyName == "display") {
        valueString.trim();
        selectorSpec.display = rawValueString;
        if (valueString == "none")
          selectorSpec.priority = "1234";
        else if (valueString == "title") {
          DOMString titlePath = tagNode.getValue();
          documentMap_["titlePath"] = pathRegistry_.correctPath(titlePath);
        }
        else if (valueString == "section" || valueString == "subsection" || valueString == "subsection" || valueString == "subsubsubsection") {
          if (valueString == "section")
            selectorSpec.specialField = "\\section";
          else if (valueString == "subsection")
            selectorSpec.specialField = "\\subsection";
          else if (valueString == "subsubsection")
            selectorSpec.specialField = "\\subsubsection";
          else if (valueString == "subsubsubsection")
            selectorSpec.specialField = "\\subsubsubsection";
        }
      }
      else if (propertyName == "environment") {
        environmentName = rawValueString; //save for later
        selectorSpec.environment = rawValueString;
      }
      else if (propertyName == "width") { // this is block width, as a percentage of max width
        valueString.trim().change("%", "");
        widthMap_[refName] = valueString;
      }
      else if (propertyName == "blockAlign") { // this is block alignment: left, center or right (center is default)
        valueString.toLowerCase().trim();
        if (valueString == "left" || valueString == "right")
          selectorSpec.blockAlign = valueString;
        else
          selectorSpec.blockAlign = "center";
      }
      else if (propertyName == "border") { // this is block border
        if (rawValueString.trim().empty()) continue; // nothing to do

        if (rawValueString == "shadow" || rawValueString == "double" || rawValueString == "oval" || rawValueString == "Oval") {
          rawValueString += "box";
          borderMap_[refName] = rawValueString;
        }
        else if (rawValueString == "true" || rawValueString == "box")
          selectorSpec.display = "true";
        else
          selectorSpec.display = "true";
      }

      else if (propertyName == "color")
        selectorSpec.color = valueString;
      else if (propertyName == "textAlign") {
        alignmentName = rawValueString;
        selectorSpec.textAlign = correct(valueString);
      }
      else if (propertyName == "font")
        selectorSpec.font = valueString;
      else if (propertyName == "fontSize")
        selectorSpec.fontSize = valueString;
      else if (propertyName == "fontShape")
        selectorSpec.fontShape = correct(valueString);
      else if (propertyName == "fontWeight")
        selectorSpec.fontWeight = correct(valueString);

      else if (propertyName == "textBefore")
        selectorSpec.textBefore = valueString;
      else if (propertyName == "textAfter") {
        DOMString valueCopy = valueString;
        if (!valueCopy.trim().empty()) {
          selectorSpec.textAfter = " ";
          selectorSpec.textAfter += valueString;
        }
      }
      else if (propertyName == "textLeft")
        selectorSpec.textLeft = valueString;
      else if (propertyName == "textRight") {
        DOMString valueCopy = valueString;
        if (!valueCopy.trim().empty()) {
          selectorSpec.textRight = " ";
          selectorSpec.textRight += valueString;
        }
      }

      else if (propertyName == "priority") {
        selectorSpec.priority = valueString;
      }
    }
  }
}

void LatexConfiguration::finalizeTagStyles() {
  SelectorMap::iterator p = selectorMap_.begin();
  while (p != selectorMap_.end()) {
    DOMString tagName = p->first;
    SelectorSpec& selectorSpec = p->second;
    ++p;

    DOMString specString;
    if (selectorSpec.display == "block" && !verbose_) {
      addToSet(tagName, blockSet_);
    }

    if (selectorSpec.display == "text")
      addToSet(tagName, stringSet_);

    if (!selectorSpec.environment.empty())
      addToSet(tagName, environmentSet_);

    if (!selectorSpec.specialField.empty())
      sectionMap_[pathRegistry_.correctPath(tagName)] = selectorSpec.specialField;

    specString = getLeftString(selectorSpec);
    tagStyleMap_[tagName] = specString;

    DOMString postString = getRightString(selectorSpec);
    postStringMap_[tagName] = postString;

    if (specString.trim().empty() && postString.trim().empty()) {
      preStringMap_[tagName] = "";
      setUndefined(tagName); // belt and braces
    }
  }
}

DOMString LatexConfiguration::getLeftString(SelectorSpec& selectorSpec) {
  DOMString tagString;

  tagString += selectorSpec.textBefore;

  DOMString environmentString = selectorSpec.environment;
  if (!environmentString.empty()) {
    DOMString environmentBegin = "\\begin{**ENV**}";
    environmentBegin.change("**ENV**", environmentString);
    tagString += environmentBegin;
  }

  if (!selectorSpec.color.empty()) {
    tagString += "\\color{**COLOR**}";
    tagString.change("**COLOR**", selectorSpec.color);
  }

  appendWithBackslash(tagString, selectorSpec.textAlign);
  if (!selectorSpec.font.empty()) {
    DOMString fontString = selectorSpec.font;
    fontString += "font";
    appendWithBackslash(tagString, fontString);
  }
  appendWithBackslash(tagString, selectorSpec.fontSize);
  appendWithBackslash(tagString, selectorSpec.fontShape);
  appendWithBackslash(tagString, selectorSpec.fontWeight);

  if (!selectorSpec.textLeft.empty()) {
    tagString += " ";
    tagString += selectorSpec.textLeft;
  }

  return tagString;
}

DOMString LatexConfiguration::getRightString(SelectorSpec& selectorSpec) {
  DOMString tagString;

  if (!selectorSpec.textRight.empty()) {
    tagString += " ";
    tagString += selectorSpec.textRight;
  }

  DOMString environmentString = selectorSpec.environment;
  if (!environmentString.empty()) {
    DOMString environmentBegin = "\\end{**ENV**} ";
    environmentBegin.change("**ENV**", environmentString);
    tagString += environmentBegin;
  }

  if (!selectorSpec.textAfter.empty()) {
    tagString += " ";
    tagString += selectorSpec.textAfter;
  }

  if (!environmentString.empty())
    tagString += "\n\n"; // we need a blank line to show off an environment

  return tagString;
}

DOMString LatexConfiguration::getTagDefinitions() {
  DOMString tagString;

  StringMap::iterator p = tagStyleMap_.begin();
  while (p != tagStyleMap_.end()) {
    DOMString& specString = p->second;
    DOMString tagName = p->first;

    ++p;

    DOMString specCopy = specString;
    if (specCopy.trim().empty())
      continue;

    tagName = getPreStringName(tagName);
    tagName += "Style";

    DOMString patternString = "\\newcommand{\\**NAME**}{**SPEC**}\n";
    patternString.change("**NAME**", tagName);
    patternString.change("**SPEC**", specString);

    tagString += patternString;
  }

  return tagString;
}

DOMString LatexConfiguration::getPackageString() {
  DOMString packageString;

  StringSet::iterator p = packageSet_.begin();
  while (p != packageSet_.end()) {
    DOMString patternString = "\\usepackage{**NAME**}\n";
    patternString.change("**NAME**", *p);
    packageString += patternString;
    ++p;
  }

  return packageString;
}

DOMString LatexConfiguration::getFontDefinitions() {
  DOMString fontsString;

  FontMap::iterator p = fontMap_.begin();

  while (p != fontMap_.end()) {
    DOMString fontName = p->first;
    fontsString += getFontDefinition(p->first, p->second);
    ++p;
  }

  fontsString += "\\newcommand{\\defaultfont}{\\**DEFAULT**font}\n\n";
  DOMString defaultValue = documentConfiguration_.getDefaultFont();
  if (!defaultValue.empty())
    fontsString.change("**DEFAULT**", defaultValue);
  else
    fontsString.change("**DEFAULT**", "times");

  return fontsString;
}

DOMString LatexConfiguration::getFontDefinition(const DOMString& fontName, const FontSpec& fontSpec) {
  DOMString fontString = "\\newcommand{\\";
  fontString += fontName;
  fontString += "font}{\\fontsize{";
  fontString += fontSpec.size;
  fontString += "}{";
  fontString += fontSpec.height;
  fontString += "}\\usefont{T1}{";
  fontString += fontSpec.selector;
  fontString += "}{";
  fontString += fontSpec.weight;
  fontString += "}{";
  fontString += fontSpec.shape;
  fontString += "}}\n";

  return fontString;
}

void LatexConfiguration::addToSet(const DOMString& name, StringSet& stringSet) {
  DOMString correctName = name;
  correctName = pathRegistry_.correctPath(correctName);

  if (stringSet.find(correctName) == stringSet.end())
    stringSet.insert(correctName);
}

bool LatexConfiguration::isFontSpec(const DOMString& specString) {
  static DOMString matchString = "font, size, height, shape, weight, italic, other";
  return matchString.contains(specString);
}

bool LatexConfiguration::isSelectorSpec(const DOMString& specString) {
  static DOMString matchString = "environment, width, border, padding, color, textAlign, baseFont, fontSize, fontShape, fontWeight, textBefore, textAfter, text-left, text-right, marginLeft, marginRight";
  return matchString.contains(specString);
}

void LatexConfiguration::addSpec(DOMString& styleString, DOMString& specString) {
  if (specString.empty()) return;
  styleString += "\\";
  styleString += specString;
}

DOMString LatexConfiguration::getPreStringName(const DOMString& pathName) {
  // general cases
  DOMString tagName = pathName;

  tagName.change("/", "");
  tagName.change("[", "Attr");
  tagName.change("]", "");
  tagName.change("(", "");
  tagName.change(")", "");
  tagName.change("attr", "");
  tagName.change("=", "");
  tagName.change("'", "");
  tagName.change("\"", "");
  tagName.change(":", "");
  tagName.change("-", "");
  tagName.change("*", "");
  tagName.change("+", "plus");
  tagName.change(" ", "desc");
  tagName.change(">", "child");
  tagName.change(".", "");
  convertDigits(tagName); // get a Latex-friendly name

  return tagName;
}

DOMString LatexConfiguration::correct(DOMString& s) {
  if (s == "roman") return "rmfamily";
  if (s == "sans-serif") return "sffamily";
  if (s == "typewriter") return "ttfamily";

  if (s == "normal") return "mdseries";
  if (s == "bold") return "bfseries";

  if (s == "upright-shape") return "upshape";
  if (s == "italic-shape") return "itshape";
  if (s == "upright-shape") return "upshape";
  if (s == "slanted-shape") return "slshape";
  if (s == "italic-shape") return "itshape";
  if (s == "small-caps-shape") return "scshape";

  if (s == "slanted") return "slshape";
  if (s == "italic") return "itshape";
  if (s == "small-caps") return "scshape";
  if (s == "emphasised") return "em";
  if (s == "emphasized") return "em";


  if (s == "center") return "centering";
  if (s == "left") return "raggedright";
  if (s == "right") return "raggedleft";

  return s;
}

DOMString LatexConfiguration::getHeaderString(const DOMString& keyString, StringMap& map, bool isBody) {
  DOMString resultString;
  if (!isBody)
    resultString = "\\tocheadersStyle ";
  else
    resultString = "\\bodyheadersStyle ";

  //handle default cases where nothing is specified: center title for toc, otherwise left and right with counters
  bool isFound = map.find(keyString) != map.end();
  if (!isFound && !isBody) {
    if (keyString == "center") {
      resultString += titleString_;
      return resultString;
    }
    else return "";
  }

  if (!isFound && isBody) {
    if (keyString == "center") return "";
    if (keyString == "left") resultString +=  "\\leftmark";
    else resultString += "\\rightmark";
    return resultString;
  }

  // if an empty string is specified, use it without styling
  if (map[keyString].trim().empty()) return "";

  resultString += map[keyString];

  if (!isBody || keyString == "center") // center heading has no counter
    return resultString;

  // now append a counter, which is Roman by default
  DOMString numberingString;
  if (keyString == "left")
    numberingString = map["left-numbering"];
  else if (keyString == "right")
    numberingString =  map["right-numbering"];

  if (numberingString != "roman" && numberingString != "arabic")
    numberingString = "Roman"; // empty for Roman (the default)

  resultString += " \\";
  resultString += numberingString;
  resultString += "{";

  // finally the section, as required
  if (keyString == "left")
    resultString += "section}";
  else
    resultString += "subsection}";

  return resultString;
}

DOMString LatexConfiguration::findTitle(const DOMString& titlePath, Document& dataDocument) {
  MatchedTemplate matchTemplate(titlePath);

  NodeStream stream(dataDocument.getDocumentElement(), false, false); // exclude root, no end nodes

  while (true) {
    Node& node = stream.nextNode();
    if (!node.isValid()) break;

    if (node.getNodeType() != Node::ELEMENT_NODE)
      continue;

    if (matchTemplate.matchNode(node, true)) // full match
      return node.getValue();
  }

  return "NO TITLE";
}

void LatexConfiguration::addSelector(const DOMString& selectorName, bool isXPath) {
  SelectorMap::iterator p = selectorMap_.find(selectorName);
  if (p == selectorMap_.end()) {
    registerTagName(selectorName, isXPath);
    SelectorSpec selectorSpec;
    selectorMap_[selectorName] = selectorSpec;
    if (isXPath)
      selectorSpec.isXPath = "true";
  }
}

}
