/***************************************************************************
 *   Copyright (C) 2007 by John Redmond   *
 *   john_redmond@optusnet.com.au   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef LIMPIDLATEXCONFIGURATION_H
#define LIMPIDLATEXCONFIGURATION_H

#include "stringutils.h"
#include "getter.h"
#include "domutils.h"
#include "pathregistry.h"
#include "latexdocumentconfiguration.h"
#include <vector>
#include <map>

namespace limpid {

/**
	@author John Redmond <john_redmond@optusnet.com.au>
*/

class LatexConfiguration: public DOMUtils, public StringUtils, public Converter {
  public:
    LatexConfiguration();
    void configure(const DOMString& configPath);
    void setVerbose(bool verbose) { verbose_ = verbose; }

    // these public interfaces return constant references, while the internal ones return values
    const DOMString& getPreambleString(Document& document);
    DOMString getDefinitionsString(); // pass-through to document configuration
    const DOMString& getSectionsString();
    const DOMString& getDocumentStartString(Document& dataDocument);
    const DOMString& getTailString();

    bool isVerbose() { return documentConfiguration_.isVerbose(); }
    bool isBlock(const DOMString& tagName) { return blockSet_.find(tagName) != blockSet_.end(); }
    bool isText(const DOMString& tagName) { return stringSet_.find(tagName) != blockSet_.end(); }
    bool isEnvironment(const DOMString& tagName) { return environmentSet_.find(tagName) != blockSet_.end(); }
    bool isHidden(const DOMString& tagName) { return hiddenSet_.find(tagName) != hiddenSet_.end(); }
    bool isUndefined(const DOMString& tagName)  { return undefinedSet_.find(tagName) != undefinedSet_.end(); }
    void setUndefined(const DOMString& tagName) { addToSet(tagName, undefinedSet_); }

    DOMString getMatchedPath(Node& node) { return pathRegistry_.getPath(node); }
    DOMString getPrestring(const DOMString& key) { return preStringMap_[pathRegistry_.getUncorrectedPath(key)]; }
    DOMString getPoststring(const DOMString& key) { return postStringMap_[pathRegistry_.getUncorrectedPath(key)]; }
    DOMString getBorderString(const DOMString& key) { return borderMap_[key]; }
    DOMString getWidthString(const DOMString& key) { return widthMap_[key]; }
    DOMString getAlignmentString(const DOMString& key) { return alignMap_[key]; }
    DOMString getSectionString(const DOMString& key) { return sectionMap_[key]; }
    bool isSectionNode(const Node& node) { DOMString nodeName = node.getNodeName(); return nodeName == "h1x" || nodeName == "h2" || nodeName == "h3"; }


  private:
    struct FontSpec { DOMString size, height, selector, weight, shape; };

    struct SelectorSpec {
      DOMString specialField, display, environment, font, fontSize, fontShape, fontWeight, color, frame, blockAlign, textAlign, textBefore, textAfter, textLeft, textRight, marginTop, marginBottom, marginLeft, marginRight, width, border, priority, isXPath; };

    typedef std::map<DOMString, FontSpec> FontMap;
    typedef std::map<DOMString, SelectorSpec> SelectorMap;
    typedef std::map<DOMString, DOMString> StringMap;
    typedef std::set<DOMString> StringSet;

    bool verbose_;
    Document configDocument_;
    Getter configGetter_;
    StringSet packageSet_;
    PathRegistry pathRegistry_;
    LatexDocumentConfiguration documentConfiguration_;

    DOMString titleString_, titlePathString_, preambleString_, sectionsString_, documentStartString_, tailString_;

    FontMap fontMap_;
    SelectorMap selectorMap_;
    StringMap documentMap_, packageMap_, sectionMap_, tocHeadersMap_, bodyHeadersMap_, detailMap_, displayMap_, headersContentMap_;
    StringMap lengthMap_, constantMap_, numberingMap_, tagStyleMap_, preStringMap_, postStringMap_, borderMap_, widthMap_, alignMap_;

    StringSet emptySet_, sectionSet_, stringSet_, blockSet_, environmentSet_, hiddenSet_, undefinedSet_;

    // internal high-level configuration functions
    Document getConfigurationDocument(const DOMString& configPath);
    Document getDefaultDocument();
    void configure(Document& configDocument);
    void configureDocument(DOMString& tagString, Node& tagNode);
    void applyConfiguration(Document& configDocument, bool force);

    void configureDefaults();
    void configureHeaders(Node& node, StringMap& headersMap);
    void configureHeadersContent(Node& node, StringMap& headersMap);
    void configurePackage(Node& tagNode);
    void configureTag(const DOMString& refName, Element& tagElement);
    void configureSection(DOMString& sectionName, Node& tagNode);
    void configureFonts(Node& tagNode);
    void configureFont(DOMString& fontName, Node& specNode);
    void configureLengths(Node& tagNode);
    void configureConstants(Node& tagNode);
    void configureFloats(Node& tagNode);
    void configureNumbering(Node& tagNode);

    void setGeometry(Node& tagNode);
    void setEnvironmentDefinition(const DOMString& key, const DOMString& environmentName);

    // management of node specifications
    void registerTagName(const DOMString& tagName, bool isXPath);
    void populateSelectorSpec(Node& tagNode, SelectorSpec& selectorSpec, const DOMString& refName);
    void finalizeTagStyles();
    void setDocumentSpecs(Element& rootElement);
    void setSelectorSpecs(Element& rootElement);
    void setPathAttributes(Element& documentElement);
    void setDefinition(const DOMString& key, const DOMString& preString, const DOMString& postString = "}\n");
    void setBlockStyle(const DOMString& blockName, SelectorSpec& blockSpec);

    // internal string generating functions
    DOMString getLeftString(SelectorSpec& selectorSpec);
    DOMString getRightString(SelectorSpec& selectorSpec);
    DOMString getTOCString();
    DOMString getTagDefinitions();
    DOMString getPackageString();
    DOMString getMappedString(StringMap& sourceMap, const DOMString& patternString, bool isBody);
    DOMString getFancyMappedString(StringMap& sourceMap, bool isBody);
    DOMString getFontDefinitions();
    DOMString getFontDefinition(const DOMString& fontName, const FontSpec& fontSpec);

    // map handling
    void setMaps(Document& configDocument);
    void setSpecs(Node& refNode, StringMap& map);

    // set handling
    void addToSet(const DOMString& name, StringSet& stringSet);

    // queries
    bool isDocumentSpec(const DOMString& specString);
    bool isFontSpec(const DOMString& specString);
    bool isSelectorSpec(const DOMString& specString);

    // string handling
    void addSpec(DOMString& styleString, DOMString& specString);
    DOMString getPreStringName(const DOMString& pathName);
    DOMString correct(DOMString& s);
    DOMString getHeaderString(const DOMString& keyString, StringMap& map, bool isBody);

    DOMString findTitle(const DOMString& titlePath, Document& dataDocument);
    void addSelector(const DOMString& selectorName, bool isXPath);
};

}

#endif
