/***************************************************************************
 *   Copyright (C) 2007 by John Redmond   *
 *   john_redmond@optusnet.com.au
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "latexdocumentconfiguration.h"
#include "nodestream.h"
#include "siblingstream.h"
#include "liststream.h"
#include "argparser.h"
#include "domwriter.h"

namespace limpid {

void LatexDocumentConfiguration::configure(Document& configDocument) {
  setMaps(configDocument);

  Element& documentElement = configDocument.getDocumentElement();
  NodeStream nodeStream(documentElement, false, false);

  while (true) {
    Node& node = nodeStream.nextNode();
    if (!node.isValid()) break;
    if (node.getNodeType() != Node::ELEMENT_NODE) continue;
    if (node.getNodeName() != "TAG") continue;

    Element& element = static_cast<Element &>(node);

    DOMString tagString = element.getValue();
    if (tagString.empty()) {
      continue;
    }
    if (tagString == "document")
      configureDocument(tagString, node);
    else if (tagString == "geometry")
      documentMap_["geometry"] = node.getNextSibling().getValue();
    else if (tagString == "package")
      configurePackage(node);
    else if (tagString == "lengths")
      configureLengths(node);
  }
}

void LatexDocumentConfiguration::configureDocument(DOMString& tagString, Node& tagNode) {
  SiblingStream siblingStream(tagNode);
  while (true) {
    Node& siblingNode = siblingStream.nextNode();
    if (!siblingNode.isValid()) break;

    if (siblingNode.getNodeType() != Node::ELEMENT_NODE) continue;

    if (siblingNode.getNodeName() == "SPEC") {
      DOMString propertyString = static_cast<Element &>(siblingNode).getAttribute("property");
      if (propertyString.trim().empty()) continue;

      propertyString = getCamelBack(propertyString);

      DOMString valueString = siblingNode.getValue();
      documentMap_[propertyString] = valueString;
    }
  }
}

void LatexDocumentConfiguration::configurePackage(Node& node) {
  SiblingStream siblingStream(node, true);

  while (true) {
    Node& specNode = siblingStream.nextNode();
    if (!specNode.isValid()) break;

    if (specNode.getNodeType() != Node::ELEMENT_NODE) continue;
    if (specNode.getNodeName() != "SPEC") continue;

    ArgParser parser(specNode.getValue(), ',');
    int argIndex = 0;

    while (true) {
      DOMString argString = parser.getArg(argIndex++).trim();
      if (argString.empty()) break;

      if (packageSet_.find(argString) == packageSet_.end())
        packageSet_.insert(argString);
    }
  }
}

void LatexDocumentConfiguration::configureLengths(Node& node) {
  SiblingStream siblingStream(node, true);

  while (true) {
    Node& specNode = siblingStream.nextNode();
    if (!specNode.isValid()) break;

    if (specNode.getNodeType() != Node::ELEMENT_NODE) continue;
    if (specNode.getNodeName() != "SPEC") continue;

    Element lengthElement = static_cast<Element &>(specNode);
    DOMString propertyName = lengthElement.getAttribute("property");
    DOMString propertyValue = lengthElement.getValue();

    lengthMap_[propertyName] = propertyValue;
  }
}

void LatexDocumentConfiguration::setDocumentSpecs(Element& documentElement) {
  NodeStream nodeStream(documentElement, false, false);

  while (true) {
    Node& node = nodeStream.nextNode();
    if (!node.isValid()) break;

    if (node.getNodeType() != Node::ELEMENT_NODE) continue;
    if (node.getNodeName() == "TAG" && node.getValue() == "document") {
      SiblingStream siblingStream(node);
      while (true) {
        Node& siblingNode = siblingStream.nextNode();
        if (!siblingNode.isValid()) break;

        if (siblingNode.getNodeName() == "SPEC") {
          DOMString propertyString = static_cast<Element &>(siblingNode).getAttribute("property");
          if (propertyString.trim().empty()) continue;

          propertyString = getCamelBack(propertyString);

          DOMString valueString = siblingNode.getValue().trim();
          documentMap_[propertyString] = valueString;
        }
      }
    }
  }
}

DOMString LatexDocumentConfiguration::getPreambleString(Document& dataDocument, const DOMString& titlePath) {
  DOMString preambleString;

  preambleString = "\\documentclass";
  DOMString formatString = documentMap_["format"];
  DOMString fontSizeString = documentMap_["font-size"];

  // add document details if required
  if (formatString == "landscape" || !fontSizeString.empty()) {
    preambleString += "[";
    if (!formatString.empty() && !fontSizeString.empty()) {
      preambleString += formatString;
      preambleString += ", ";
      preambleString += fontSizeString;
      preambleString += "]";
    }
    else if (preambleString.empty()) {
      preambleString += fontSizeString;
      preambleString += "]";
    }
    else if (fontSizeString.empty()) {
      preambleString += formatString;
      preambleString += "]";
    }
  }

  preambleString += "{**CLASSSTRING**}";

  // now state the document class
  DOMString classString = documentMap_["class"];
  if (classString.empty()) classString = "article";
  preambleString.change("**CLASSSTRING**", classString);
  preambleString += "\n";

  DOMString geometryString = documentMap_["geometry"].trim();
  if (!geometryString.empty()) {
    preambleString += "\\usepackage[**GEOMETRYSTRING**]{geometry}\n";
    preambleString.change("**GEOMETRYSTRING**", geometryString);
  }

  preambleString += getPackageString();
  preambleString += "\\usepackage{fancyhdr}\n\\pagestyle{fancy}\n";

  // generally, we want arabic page numbering. roman or Roman can cause overflow problems with the TOC
  DOMString numberingString = "\n\\pagenumbering{**NUMBERING**}\n\n";
  numberingString.change("**NUMBERING**", documentMap_["pageNumbering"]);
  preambleString += numberingString;

  return preambleString;
}

DOMString LatexDocumentConfiguration::getPackageString() {
  DOMString resultString;

  StringSet::iterator p = packageSet_.begin();
  while (p != packageSet_.end()) {
    DOMString patternString = "\\usepackage{**NAME**}\n";
    patternString.change("**NAME**", *p);
    resultString += patternString;
    ++p;
  }

  return resultString;
}

DOMString LatexDocumentConfiguration::getLengthsString() {
  DOMString resultString = "\n% specify lengths\n";

  StringMap::iterator p = lengthMap_.begin();
  while (p != lengthMap_.end()) {
    DOMString lengthName = p->first;

    if (lengthName.endsWith("-new")) { // need to defne a new length
      lengthName.change("-new", "");
      DOMString newPatternString = "\\newlength{\\**KEY**}";
      newPatternString.change("**KEY**", lengthName);
      resultString += newPatternString;
    }

    DOMString patternString = "\\setlength{\\**KEY**}{**VALUE**}\n";
    patternString.change("**KEY**", lengthName);
    patternString.change("**VALUE**", p->second);
    resultString += patternString;
    ++p;
  }

  return resultString;
}

DOMString LatexDocumentConfiguration::getSectionsString() {

  DOMString resultString = "\\renewcommand\\thesection{}\n\\renewcommand\\thesubsection{}\n\n\
\\makeatletter\n\
\\renewcommand\\section{\\@startsection{section}{1}{0pt}{-\\baselineskip}{\\baselineskip}{\\sectionStyle}}\n\
\\renewcommand\\subsection{\\@startsection{subsection}{1}{0pt}{-.2\\baselineskip}{.5\\baselineskip}{\\subsectionStyle}}\n\
\\renewcommand\\subsubsection{\\@startsection{subsubsection}{1}{0pt}{-.2\\baselineskip}{.2\\baselineskip}{\\subsubsectionStyle}}\n\
\\makeatother\n";

  // patch with chosen styles: set in setNewDefinitions()
  resultString.change("**SECTION_STYLE**", "\\sectionStyle");
  resultString.change("**SUBSECTION_STYLE**", "\\subsectionStyle");
  resultString.change("**SUBSUBSECTION_STYLE**", "\\subsubsectionStyle");
  resultString += "\n%end sections string\n";

  return resultString;
}

DOMString LatexDocumentConfiguration::getDocumentStartString() {
  DOMString resultString = "\n\\begin{document}\n\\defaultfont\\**ALIGNMENT**\n";

/// set widths for any tables using tabulary:
  resultString += "\\setlength\\tymax{.6\\textwidth}\\setlength\\tymin{.25\\textwidth} % set limits to table columns\n";

  if (documentMap_["title"] == "true") {
    resultString += "\\title{**TITLE**}\n**AUTHOR****DATE**\\maketitle\n\n";
    if (!documentMap_["titleString"].empty())
      resultString.change("**TITLE**", documentMap_["titleString"]);  // we have specified a specific title
  }

  if (!documentMap_["author"].empty()) {
    DOMString insertString = "\\author{";
    insertString += documentMap_["author"]; /// this may be an empty string if there is no author
    insertString += "}\n";

    resultString.change("**AUTHOR**", insertString);
  }

  DOMString dateString = documentMap_["date"];
  if (!dateString.empty()) {
    DOMString insertString = "\\date{";
    if (dateString == "true")
      insertString += "\\today";
    else
      insertString += dateString;
    insertString += "}\n";

    resultString.change("**DATE**", insertString);
  }

  DOMString alignmentString = documentMap_["textAlign"]; // global text alignment
  if (alignmentString == "left")
    alignmentString = "\\raggedright";
  else if (alignmentString == "right")
    alignmentString = "\\raggedleft";
  else if (alignmentString == "center")
    alignmentString = "\\centering";
  else
    alignmentString.clear();  // use the default alignment

  resultString.change("\\**ALIGNMENT**", alignmentString);

  resultString += getHeadersString(false); // not body


  resultString += getTOCString();
  resultString += getHeadersString(true);

  DOMString columnsString = documentMap_["bodyColumns"];
  if (!columnsString.empty() && columnsString != "1") {
    resultString += "\n\\begin{multicols}{**COLUMNS**}\n\n";
    resultString.change("**COLUMNS**", columnsString);
  }

  return resultString;
}

DOMString LatexDocumentConfiguration::getHeadersString(bool isBody) {
    return getFancyMappedString(isBody);
}

DOMString LatexDocumentConfiguration::getFancyMappedString(bool isBody) {
  DOMString resultString = "\n%specify fancy page headings\n";
  DOMString typeString = "BODY";
  if (!isBody)
    typeString = "TOC";

  resultString += "\\fancyhead[L]{**TYPELEFT**}\n";
  resultString += "\\fancyhead[C]{**TYPECENTER**}\n";
  resultString += "\\fancyhead[R]{**TYPERIGHT**}\n";
  resultString += "%end of fancy\n";

  resultString.change("TYPE", typeString);

  return resultString; // this will be further patched in latexconfiguration
}

DOMString LatexDocumentConfiguration::getTOCString() {
  DOMString resultString;

  DOMString tocSpec = documentMap_["tocColumns"];
  if (tocSpec == "0") return ""; // no toc

  if (tocSpec.empty()) return "";

  // add table of contentsL All we need is a TOCHEADERS node; it need not have any children at this point
  if (tocSpec != "1") {
    DOMString patternString = "\\begin{multicols}{**COLS**}\n";
    patternString.change("**COLS**", tocSpec);
    resultString += patternString;
  }

  resultString += "\\tableofcontents\n";

  if (tocSpec != "1") {
    DOMString patternString = "\\end{multicols}\n";
    patternString.change("**COLS**", tocSpec);
    resultString += patternString;
  }

  resultString += "\\vfill\n\\pagebreak\n";

  return resultString;
}

 DOMString LatexDocumentConfiguration::getDefinitionsString() {
  /// this will be upgraded

  DOMString definitionsString = "\\setcounter{secnumdepth}{**DEPTH**} % set depth of numbering of headings\n";
  DOMString numberingString = documentMap_["numberingDepth"];
  if (numberingString.empty())
    numberingString = "2";


  return definitionsString.change("**DEPTH**", numberingString);
}

DOMString LatexDocumentConfiguration::getTailString() {
  DOMString resultString;
  DOMString columnsString = documentMap_["bodyColumns"];

  if (!columnsString.empty() && columnsString != "1")
    resultString = "\\end{multicols}\n";

  resultString += "\\end{document}\n";

  return resultString;
}

void LatexDocumentConfiguration::setMaps(Document& configDocument) {
  NodeStream nodeStream(configDocument, false, false);

  while (true) {
    Node& node = nodeStream.nextNode();
    if (!node.isValid()) break;

    if (node.getNodeName() != "TAG") continue;
    DOMString specType = node.getNodeValue();

    if (specType == "document")
      setSpecs(node, documentMap_);
    else if (specType == "package")
      setSpecs(node, packageMap_);
  }
}

void LatexDocumentConfiguration::setSpecs(Node& refNode, StringMap& map) {
  SiblingStream siblingStream(refNode, true);

  while (true) {
    Node& node = siblingStream.nextNode();
    if (!node.isValid()) break;

    if (node.getNodeName() != "SPEC") continue;
    Element& element = static_cast<Element &>(node);

    map[getCamelBack(element.getAttribute("property"))] = element.getValue();
  }
}

DOMString LatexDocumentConfiguration::getDefaultFont() { return documentMap_["font"]; }
bool LatexDocumentConfiguration::isVerbose() { return documentMap_["verbose"] == "true"; }

bool LatexDocumentConfiguration::isXHTML() {
  DOMString sourceType = documentMap_["source"].toLowerCase();
  return sourceType.empty()  || sourceType == "XHTML";
}

}
