/***************************************************************************
 *   Copyright (C) 2007 by John Redmond   *
 *   john_redmond@optusnet.com.au   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "latexmapper.h"
#include "documentsummary.h"
#include "domstring.h"
#include "nodestream.h"
#include "liststream.h"
#include "siblingstream.h"
#include "stringreader.h"
#include "tokenstream.h"
#include "converter.h"
#include "element.h"
#include "text.h"
#include "comment.h"
#include "pointer.h"
#include "domwriter.h"
#include "matcher.h"
#include "getter.h"
#include "xmlparser.h"
#include "dotparser.h"

namespace limpid {

LatexMapper::LatexMapper(const DOMString& specPath): specPath_(specPath) {}

Document& LatexMapper::processDocument(Document& sourceDocument) {
  resultDocument_ = Document(); // refresh
  setDefaultValues(sourceDocument);
  setDefaultDefinitions();
  updateSpecs();
  setTopStrings(sourceDocument);
  buildDocument(sourceDocument);

  return resultDocument_;
}

void LatexMapper::setDefaultValues(Document& sourceDocument) {
  documentMap_["document"] = "article";
  documentMap_["page"] = "a4paper";
  documentMap_["columns"] = "2";
  documentMap_["numbering"] = "arabic";
  documentMap_["toc"] = "true";
  documentMap_["fancy"] = "true";
  documentMap_["title"] = "'UNTITLED'";

  packageMap_["geometry"] ="a4paper, top=25mm, left=30mm, bottom=45mm, right=0mm, headheight=55pt, includehead, includemp";
  packageMap_["multicol"] = "";

  styleMap_["title"] = "\\bfseries\\large";
  styleMap_["section"] = "\\bfseries\\centering\\Large";
  styleMap_["subsection"] = "\\bfseries\\centering\\large";
  styleMap_["subsubsection"] = "\\bfseries";
  styleMap_["global"] = "\\raggedright\\noindent";

  tocHeadersMap_["lhead"] = "";
  tocHeadersMap_["chead"] = "\\Large\\bfseries <xsl:value-of select='escape-tex(**TITLE_PATH**)'/>";
  tocHeadersMap_["rhead"] = "";

  bodyHeadersMap_["lhead"] = "\\large\\bfseries\\leftmark";
  bodyHeadersMap_["chead"] = "";
  bodyHeadersMap_["rhead"] = "\\large\\bfseries\\rightmark";

  // get a title from the source document
  DOMString titlePath = getTitlePath(sourceDocument.getDocumentElement());
  if (!titlePath.empty())
    documentMap_["title"] = titlePath; // otherwise "UNTITLED" is used

}

void LatexMapper::updateSpecs() {
  if (specPath_.empty()) return; // no further configuration
  if (!specPath_.endsWith(".xml") && !specPath_.endsWith(".txt")) {
    wrt::Errors << "Cannot configure LatexMapper with " << specPath_ << wrt::endl;
    wrt::Errors << "The file must end in '.xml' or '.txt'" << wrt::endl;
    return;
  }

  Reader reader(specPath_);

  if (specPath_.endsWith(".xml")) {
    XMLParser parser(reader);
    Getter getter(parser);
    configure(getter);
  }
  else if (specPath_.endsWith("txt")) {
    DotParser parser(reader);
    Getter getter(parser);
    configure(getter);
  }

  // establish how the sections are to be named
  DOMString sectionName = documentMap_["section-name"];
  if (!sectionName.empty()) {
    DOMString lheadString = "**SECTIONSTYLE** **SECTION** \\Roman{section}";
    lheadString.change("**SECTIONSTYLE**", styleMap_["section"]);
    lheadString.change("**SECTION**", sectionName);
    bodyHeadersMap_["lhead"] = lheadString;
  }

  DOMString subsectionName = documentMap_["subsection-name"];
  if (!subsectionName.empty()) {
    DOMString rheadString = "**SUBSECTIONSTYLE** **SUBSECTION** \\Roman{subsection}"; // avoid problems with long headings
    rheadString.change("**SUBSECTIONSTYLE**", styleMap_["subsection"]);
    rheadString.change("**SUBSECTION**", subsectionName);
    bodyHeadersMap_["rhead"] = rheadString;
  }

}

void LatexMapper::setTopStrings(Document& sourceDocument) {
  bool hasTOC = documentMap_["toc"] == "true";
  bool isFancy = documentMap_["fancy"] == "true";

  // set up the document type, with "article being the default, unless otherwise specified
  documentString_ = "\\documentclass";

  DOMString pageString = documentMap_["page"];
  if (!pageString.empty()) {  // may need a parameter
    documentString_ += "[**PAGESTRING**]";
    documentString_.change("**PAGESTRING**", documentMap_["page"]);
  }

  documentString_ += "{**TYPESTRING**}\n";
  documentString_.change("**TYPESTRING**", documentMap_["document"]);

  documentString_ += getPackageString(packageMap_);

  if (isFancy) {// do we want fancy page headers?
    documentString_ += "\\usepackage{fancyhdr}\n\\pagestyle{fancy}\n";
    if (hasTOC)
      documentString_ += getFancyMappedString(tocHeadersMap_);
  }
  else if (hasTOC)
    documentString_ += getMappedString(tocHeadersMap_, "\\**KEY**{**VAL**}\n");

  // patch with actual title path
  documentString_.change("**TITLE_PATH**", documentMap_["title"]);

  // generally, we want arabic page numbering. roman or Roman can cause overflow problems with the TOC
  DOMString numberingString = "\n\\pagenumbering{**NUMBERING**}\n\n";
  numberingString.change("**NUMBERING**", documentMap_["numbering"]);
  documentString_ += numberingString;

  // add table of contents? All we need is a TOCHEADERS node; it need not have any children at this point
  if (hasTOC)
   tocString_ = "\\tableofcontents\n\\vfill\n\\pagebreak\n";

  if (isFancy) // do we want fancy page headers?
    tocString_ += getFancyMappedString(bodyHeadersMap_);
  else
    tocString_ += getMappedString(bodyHeadersMap_, "\\**KEY**{**VAL**}\n");

  // now define the section styles

  sectionsString_ = "\\renewcommand\\thesection{}\n\\renewcommand\\thesubsection{}\n\n\
\\makeatletter\n\
\\renewcommand\\section{\\@startsection{section}{1}{0pt}{-\\baselineskip}{.6\\baselineskip}{**SECTION_STYLE**}}\n\
\\renewcommand\\subsection{\\@startsection{subsection}{1}{0pt}{-\\baselineskip}{.4\\baselineskip}{**SUBSECTION_STYLE**}}\n\
\\renewcommand\\subsubsection{\\@startsection{subsubsection}{1}{0pt}{-\\baselineskip}{.2\\baselineskip}{**SUBSUBSECTION_STYLE**}}\n\
\\makeatother\n\n";

  // patch with chosen styles: set in setNewDefinitions()
  sectionsString_.change("**SECTION_STYLE**", styleMap_["section"]);
  sectionsString_.change("**SUBSECTION_STYLE**", styleMap_["subsection"]);
  sectionsString_.change("**SUBSUBSECTION_STYLE**", styleMap_["subsubsection"]);

  documentStartString_ = "\\begin{document}\n\n";

// use multiple columns if appropriate for the body of the document, after the TOC (if any)
  DOMString columnsString = documentMap_["columns"];
  DOMString documentStyle = styleMap_["global"];
  if (!columnsString.empty() && columnsString != "1") {
    documentBodyString_ = "\n\\begin{multicols}{**COLUMNS**}\n\n{**DOCUMENTSTYLE**\n";
    documentBodyString_.change("**COLUMNS**", columnsString);
    documentBodyString_.change("**DOCUMENTSTYLE**", documentStyle);

    tailString_ += "}\\end{multicols}\n";
  }
  else {
    documentBodyString_ = "\n\n{**DOCUMENTSTYLE**\n";
    documentBodyString_.change("**DOCUMENTSTYLE**", documentStyle);
    documentBodyString_ += "\n";
    tailString_ += "}\n";
  }

  tailString_ += "\\end{document}\n";
}

void LatexMapper::setDefaultDefinitions() {
  setDefinition("L0", "", "");
  setDefinition("L1", "", "\n");
  setDefinition("L2", "\\medskip\n", "\n\n");
  setDefinition("L3", "\\medskip\n", "\n\n");
  setDefinition("L4", "\\smallskip\n", "\n\n");
  setDefinition("L5", "\\smallskip\n", "\n\n");
  setDefinition("L6", "\\smallskip\n", "\n\n");
  setDefinition("L7", "\\smallskip\n", "\n\n");
  setDefinition("L8", "\\smallskip\n", "\n\n");

  setDefinition("T1", "\\textbf{\\title{", "}}\n");
  setDefinition("T2", "\\section{", "}\n");
  setDefinition("T3", "\\subsection{", "}\n");
  setDefinition("T4", "\\textbf{");
  setDefinition("T5", "\\textbf{");
  setDefinition("T6", "\\textbf{");

  setDefinition("H1", "\\textbf ");
  setDefinition("H2", "\\textbf{");
  setDefinition("H3", "\\textbf{");
  setDefinition("H4", "\\textbf{");
  setDefinition("H5", "\\textbf{");
  setDefinition("H6", "\\textbf{");

  setDefinition("S1", "\\textbf{");
  setDefinition("S2", "\\textbf{");
  setDefinition("S3", "\\textbf{");
  setDefinition("S4", "\\textbf{");
  setDefinition("S5", "\\textbf{");
  setDefinition("S6", "\\textbf{");

  setDefinition("F1", "\\emph{\\Large\\centering ");
  setDefinition("F2", "\\emph{\\large\\centering ");
  setDefinition("F3", "\\textsc{");
  setDefinition("F4", "\\textsc{");
  setDefinition("F5", "\\textbf{");
  setDefinition("F6", "\\textbf{");
  setDefinition("F7", "\\emph{");
  setDefinition("F8", "\\emph{");

  setDefinition("D1", "", ""); // these entries are just fillers
  setDefinition("D2", "", ""); // these entries are just fillers
  setDefinition("D3", "", ""); // these entries are just fillers
  setDefinition("D4", "", ""); // these entries are just fillers
  setDefinition("D5", "", ""); // these entries are just fillers
  setDefinition("D6", "", ""); // these entries are just fillers

  setDefinition("M1", "", ""); // these entries are just fillers
  setDefinition("M2", "", ""); // these entries are just fillers
  setDefinition("M3", "", ""); // these entries are just fillers
  setDefinition("M4", "", ""); // these entries are just fillers
  setDefinition("M5", "", ""); // these entries are just fillers
  setDefinition("M6", "", ""); // these entries are just fillers

  setDefinition("m1", "\\emph{ ", " }");
  setDefinition("m2", "\\emph{ ", " }");
  setDefinition("m3", "\\emph{ ", " }");
  setDefinition("m4", "\\emph{ ", " }");
  setDefinition("m4", "\\emph{ ", " }");
  setDefinition("m6", "\\emph{ ", " }");
}

void LatexMapper::setDefinition(const DOMString& key, const DOMString& preString, const DOMString& postString) {
  preStringMap_[key] = preString;
  postStringMap_[key] = postString;
}

void LatexMapper::buildDocument(Document& sourceDocument) {
  Element documentElement = getDocumentElement();
  resultDocument_.appendChild(documentElement);

  addComment(documentElement, "the following are the templates to handle document details");
  addTopElements(documentElement);
  addRootElement(documentElement);

  addComment(documentElement, "the following are templates for the elements identified in the source document");
  ListStream listStream(sourceDocument.getDocumentElement());

  while (true) {
    Node& node = listStream.nextNode();
    if (!node.isValid()) break;

    if (node.getNodeName() == "ELEMENT")
     addDefinedTemplate(node, documentElement);
  }

  addComment(documentElement, "the following handles text and possible space characters");
  addTextTemplate(documentElement);
}

Element LatexMapper::getDocumentElement() {
  Element documentElement("xsl:stylesheet");
  documentElement.setAttribute("xmlns:xsl", "http://www.w3.org/1999/XSL/Transform");
  documentElement.setAttribute("version", "1.1");

  Element outputElement("xsl:output");
  outputElement.setAttribute("method", "text");
  documentElement.appendChild(outputElement);

  return documentElement;
}

void LatexMapper::addTopElements(Element& parentElement) {
  addTopElement("class", documentString_, parentElement);
  addTopElement("sections", sectionsString_, parentElement);
  addTopElement("document-start", documentStartString_, parentElement);
  addTopElement("toc", tocString_, parentElement);
  addTopElement("document-body", documentBodyString_, parentElement);
  addTopElement("document-end", tailString_, parentElement);
  addNewlineElement(parentElement);
}

void LatexMapper::addRootElement(Element& parentElement) {
  Element rootElement("xsl:template");
  rootElement.setAttribute("match", "/");
  parentElement.appendChild(rootElement);

  addCallElement("class", rootElement);
  addCallElement("sections", rootElement);
  addCallElement("document-start", rootElement);
  addCallElement("toc", rootElement);
  addCallElement("document-body", rootElement);

  Element applyElement("xsl:apply-templates");
  rootElement.appendChild(applyElement);

  addCallElement("document-end", rootElement);
}

void LatexMapper::addTopElement(const DOMString& nameString, const DOMString& contentString, Element& parentElement) {
  Element templateElement("xsl:template");
  parentElement.appendChild(templateElement);
  templateElement.setAttribute("name", nameString);
  addText(templateElement, contentString);
}

void LatexMapper::addTemplateElement(const DOMString& matchString, const DOMString& preString, const DOMString& postString, Element& parentElement, bool addTerminator) {
  Element templateElement("xsl:template");
  parentElement.appendChild(templateElement);
  templateElement.setAttribute("match", matchString);
  addXSLText(templateElement, preString);
  Element applyElement("xsl:apply-templates");
  templateElement.appendChild(applyElement);
  addXSLText(templateElement, postString);

  if (addTerminator)
    addCallElement("newline", templateElement);
}

void LatexMapper::addDefinedTemplate(Node& node, Element& documentElement) {
  Element& element = static_cast<Element &>(node);
  DOMString typeString =  element.getAttribute("type");

  bool needsNewline = true;

  if (typeString.startsWith("L") || typeString.startsWith("m") || typeString == "T1" ||
      typeString == "T2" || typeString == "T3")
    needsNewline = false;

  addTemplateElement(element.getAttribute("path"), preStringMap_[typeString],
    postStringMap_[typeString], documentElement, needsNewline);
}

void LatexMapper::addNewlineElement(Element& parentElement) {
  Element& templateElement = addElement(parentElement, "xsl:template");
  templateElement.setAttribute("name", "newline");

  Element& ifElement = addElement(templateElement, "xsl:if");
  ifElement.setAttribute("test", "position()!=last()");
  addText(ifElement, "\\\\\n");
}

void LatexMapper::addTextTemplate(Element& parentElement) {
  Element& templateElement = addElement(parentElement, "xsl:template");
  templateElement. setAttribute("match", "text()");

  Element& beforeElement = addElement(templateElement, "xsl:if");
  beforeElement.setAttribute("test", "preceding-sibling::*");
  addXSLText(beforeElement, " ");

  Element& valueElement = addElement(templateElement, "xsl:value-of");
  valueElement.setAttribute("select", "escape-tex(.)");

  Element& afterElement = addElement(templateElement, "xsl:if");
  afterElement.setAttribute("test", "following-sibling::*");
  addXSLText(afterElement, " ");
}

void LatexMapper::addCallElement(const DOMString& matchString, Element& parentElement) {
  Element templateElement("xsl:call-template");
  templateElement.setAttribute("name", matchString);
  parentElement.appendChild(templateElement);
}

void LatexMapper::addXSLText(Element& parentElement, const DOMString& value) {
  if (!value.empty()) // ignore an empty string
    addLeaf("xsl:text", value, parentElement);
}

DOMString LatexMapper::getTitlePath(Node& rootNode) {
  ListStream listStream(rootNode);

  while (true) {
    Node& node = listStream.nextNode();
    if (!node.isValid()) break;

    if (node.getNodeType() != Node::ELEMENT_NODE) continue;
    Element& element = static_cast<Element &>(node);
    if (element.getAttribute("type") == "T1") return element.getAttribute("path");
  }

  return "";
}

DOMString LatexMapper::getPackageString(StringMap& sourceMap) {
  DOMString patternString1 = "\\usepackage{**NAME**}\n";
  DOMString patternString2 = "\\usepackage[**PARAMS**]{**NAME**}\n";
  DOMString resultString;

  StringMap::iterator p = sourceMap.begin();
  while (p != sourceMap.end()) {
    DOMString name = p->first;
    DOMString params = p->second;

    if (params.empty()) {
      DOMString line = patternString1;
      line.change("**NAME**", name);
      resultString += line;
    }
    else {
      DOMString line = patternString2;
      line.change("**PARAMS**", params);
      line.change("**NAME**", name);
      resultString += line;
    }
    ++p;
  }

  return resultString;
}

DOMString LatexMapper::getMappedString(StringMap& sourceMap, const DOMString& patternString) {
  DOMString resultString = "\n";

  StringMap::iterator p = sourceMap.begin();
  while (p != sourceMap.end()) {
    DOMString line = patternString;
    line.change("**KEY**", p->first);
    line.change("**VAL**", p->second);
    resultString += line;
    ++p;
  }

  return resultString;
}

DOMString LatexMapper::getFancyMappedString(StringMap& sourceMap) {
  DOMString patternString = "\\fancyhead[**KEY**]{**VAL**}\n";
  DOMString resultString = "\n";

  StringMap::iterator p = sourceMap.begin();
  while (p != sourceMap.end()) {
    DOMString line = patternString;
    DOMString keyString = p->first;

    line.change("**KEY**", keyString.toUpperCase()[0]);
    line.change("**VAL**", p->second);
    resultString += line;
    ++p;
  }

  return resultString;
}

void LatexMapper::setMap(Node& parentNode, StringMap& map) {
  if (!parentNode.isValid()) return;

  ListStream stream(parentNode);

  while (true) {
    Node& node = stream.nextNode();
    if (!node.isValid()) break;

    mapItem(node.getNodeName(), node.getValue().trim().resolveEntities(true), map);
  }
}

void LatexMapper::mapItem(const DOMString& key, const DOMString& value, StringMap& map) {
  map[key] = value;
}

void LatexMapper::configure(Getter& getter) {
  configDocument_ = getter.getDocument();
  Node& documentElement = configDocument_.getDocumentElement();
  if (documentElement.getNodeName() == "root")
    documentElement = documentElement.getFirstChild();

  setMap(getChild("DOCUMENT", documentElement), documentMap_);//, documentList_);
  setMap(getChild("PACKAGE", documentElement), packageMap_);//, packageList_);
  setMap(getChild("TOCHEADERS", documentElement), tocHeadersMap_);
  setMap(getChild("HEADERS", documentElement), bodyHeadersMap_);
  setMap(getChild("STYLE", documentElement), styleMap_);
}

}
