/***************************************************************************
                          lexxiaapp.cpp  -  description
                             -------------------
    begin                : Mon Apr 24 2006
    copyright            : (C) 2006 by John Redmond
    email                : john_redmond@optusnet.com.au
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "lexxiaapp.h"
#include "definitions.h"
#include "transformer.h"
#include "argparser.h"
#include "documentbox.h"
#include "processbox.h"
#include "builder.h"
#include "stepper.h"
#include "cutter.h"
#include "newnamer.h"
#include "saver.h"
#include "getter.h"
#include "dtdgetter.h"
#include "putter.h"
#include "xsloutputter.h"
#include "mutator.h"
#include "joiner.h"
#include "deformer.h"
#include "reformer.h"
#include "reporter.h"
#include "xmlparser.h"
#include "sxfparser.h"
#include "rjparser.h"
#include "csvparser.h"
#include "dotparser.h"
#include "cssparser.h"
#include "xmlwrapper.h"
#include "xmlwriter.h"
#include "csvwriter.h"
#include "stringreader.h"
#include "htmlstyler.h"
#include "schemabuilder.h"
#include "schemainspector.h"
#include "helper.h"
#include "inspector.h"
#include "texconverter.h"
#include "attributeconverter.h"
#include "attributestripper.h"
#include "elementdropper.h"
#include "rootstripper.h"
#include "textstripper.h"
#include "xmlmapper.h"
#include "latexmapper.h"
#include "tokenstream.h"
#include "timer.h"
#include "normalizer.h"

namespace limpid {

LexxiaApp::LexxiaApp():
  documentSource_(0), getterCount_(0), processorCount_(0),
  helpful_(false), verbose_(false), verboseWriter_(wrt::Errors) {}

int LexxiaApp::exec(int argc, char *argv[]) {
  Timer timer;

  verbose_ = isVerbose(argc, argv);

  if (isDocumentProcess(argc, argv)) {

    DocumentBox documentBox;
    if (!configForDocBox(documentBox, argc, argv)) {
      verboseWriter_ << "COMMAND HAS BAD CONFIG [" << argc << "]" << wrt::endl;
      return 1;
    }

    if (helpful_) {
      Helper().help();
      return 0;
    }

    execDocuments(documentBox);
    if (verbose_)
      timer.report("docbox processing");
    return 0;
  }

  // otherwise, we have a stepper and will use it on only one input
    ProcessBox processBox;
    if (!configForStepper(processBox, argc, argv)) {
      verboseWriter_ << "BAD CONFIG FOR STEPPER [" << argc << "]" << wrt::endl;
      return 1;
    }

    processBox.process();
    if (verbose_)
      timer.report("stepper processing");
    return 0;
}

int LexxiaApp::execDocuments(DocumentBox& documentBox) {
  documentBox.start();  // initialise processors

  if (!getterPaths_.size()) {
    setGetter(documentBox, inputType_);
    documentBox.update("DEFAULT");
    documentBox.process();
    return 0;
  }

  for (size_t index = 0; index < getterPaths_.size(); ++index) {
    DOMString sourcePath = getterPaths_[index];
    setGetter(documentBox, sourcePath);

    DOMString sourceName = sourcePath;
    size_t slashIndex = sourcePath.lastIndexOf(SeparatorChar);

    if (slashIndex != sourcePath.npos) {
      // extract a source file name from the full source path
      sourceName = sourcePath.substring(slashIndex + 1,
        sourcePath.getLength() - slashIndex - 1); // substring after '/'
    }

    // the source name is expected to have an extension
    size_t dotIndex = sourceName.indexOf("."); // substring before '.'
    if (dotIndex != sourceName.npos)
      sourceName = sourceName.substring(0, dotIndex);

    // allow output processors to form full source paths
    documentBox.update(sourceName);
    documentBox.process();
  }

  documentBox.finish();

  return 0;
}

int LexxiaApp::execStandardIn(DocumentBox& documentBox) {
  documentBox.update("");
  documentBox.process();
  documentBox.finish();

  return 0;
}


bool LexxiaApp::configForDocBox(DocumentBox& documentBox, int argc, char *argv[]) {
  for (int index = 1; index < argc; ++index) {
    DOMString arg = getArgString(argv[index]);

    if (arg.startsWith("?")) {
      inputType_ = arg;
      continue;
    }

    // if there is no prefix to the arg, it defines a getter, of which there may be many
    if (!arg.startsWith("-")) {
      getterPaths_.push_back(arg);
      continue;
    }

    // drop through to here to specify a processor for a document box
    DOMString prefix = arg.substring(0, 2);
    DOMString param = arg.substring(2, arg.getLength() - 1);

    if (prefix == "-1") {
      SchemaBuilder schemaBuilder(param.trim());
      documentBox.addProcessor(schemaBuilder);
      continue;
    }
    if (prefix == "-2") {
      SchemaInspector schemaInspector(param.trim());
      documentBox.addProcessor(schemaInspector);
      continue;
    }
    if (prefix == "-3") {
      XMLMapper xmlMapper;
      documentBox.addProcessor(xmlMapper);
      continue;
    }
    if (prefix == "-4") {
      LatexMapper latexMapper(param.trim());
      documentBox.addProcessor(latexMapper);
      continue;
    }
/*
    if (prefix == "-D") {
      Deformer deformer;
      documentBox.addProcessor(deformer);
      continue;
    }
*/
    if (prefix == "-E") {
      if (!addTexConverter(documentBox, param))
        return false;
      continue;
    }
    if (prefix == "-e") {
      DOMString newParam = "**VERBOSE**+";
      newParam += param;
      if (!addTexConverter(documentBox, newParam))
        return false;
      continue;
    }
    if (prefix == "-F") {
      Reformer reformer;
      documentBox.addProcessor(reformer);
      continue;
    }
    if (prefix == "-h" || prefix == "-H") {
      helpful_ = true;
      continue;
    }

    if (prefix == "-I") {
      Inspector inspector;
      documentBox.addProcessor(inspector);
      continue;
    }
    if (prefix == "-J") {
      addJoiner(documentBox, param);
      continue;
    }
    if (prefix == "-M") {
      addMutator(documentBox, param);
      continue;
    }
    if (prefix == "-O") {
      addPutter(documentBox, param, false);
      continue;
    }
    if (prefix == "-o") {
      addPutter(documentBox, param, true);
      continue;
    }
    if (prefix == "-q") {
      addReporter(documentBox, param, true); // simple reporting
      continue;
    }
    if (prefix == "-Q") { // full reporting if schema defined
      addReporter(documentBox, param, false);
      continue;
    }
    if (prefix == "-T") {
      addTransformer(documentBox, param);
      continue;
    }
    if (arg.startsWith("-V"))
      continue;

    if (arg.startsWith("-w")) {
      addXMLWrapper(documentBox, param, false);
      continue;
    }
    if (arg.startsWith("-W")) {
      addXMLWrapper(documentBox, param, true);
      continue;
    }

// document editors

    if (prefix == "-A") {
      AttributeConverter converter(param);
      documentBox.addProcessor(converter);
      continue;
    }
    if (prefix == "-a") {
      AttributeStripper stripper(param);
      documentBox.addProcessor(stripper);
      continue;
    }
    if (prefix == "-c") {
      Cutter cutter(param, false); // do not discard children
      documentBox.addProcessor(cutter);
      continue;
    }
    if (prefix == "-C") {
      Cutter cutter(param, true); // discard children
      documentBox.addProcessor(cutter);
      continue;
    }
    if (prefix == "-D") {
      Deformer deformer;
      documentBox.addProcessor(deformer);
      continue;
    }
    if (prefix == "-z") {
      ElementDropper dropper;
      documentBox.addProcessor(dropper);
      continue;
    }
    if (prefix == "-n") {
      addNewNamer(documentBox, param, false); // do not discard children
      continue;
    }
    if (prefix == "-N") {
      addNewNamer(documentBox, param, true); // discard children
      continue;
    }
    if (prefix == "-r") {
      RootStripper stripper;
      documentBox.addProcessor(stripper);
      continue;
    }
    if (prefix == "-t") {
      TextStripper stripper;
      documentBox.addProcessor(stripper);
      continue;
    }
    if (prefix == "-Z") {
      Normalizer normalizer;
      documentBox.addProcessor(normalizer);
      continue;
    }

    // if we have any other processors specified, signal a failure:
    verboseWriter_ << prefix << " with [" << param << "] is unknown in Lexxia" << wrt::endl;
    return false;
  }

  return true;
}

// set, on the fly, a getter for a document box
bool LexxiaApp::setGetter(DocumentBox& documentBox, const DOMString& sourceString) {
  if (sourceString.empty() || sourceString.startsWith("?")) {
    if (verbose_)
      verboseWriter_ << "Processing standard in" << wrt::endl;
    setDefaultInGetter(documentBox, inputType_);
      return true;
  }

  else if (verbose_)
    verboseWriter_ << "Processing source: " << sourceString << wrt::endl;

  DOMString ucString = sourceString;
  ucString.toUpperCase();

  if (ucString.endsWith(".XML") || ucString.endsWith(".XSL") || ucString.endsWith(".XSD")) {
    XMLParser parser(sourceString);
    Getter getter(parser);
    documentBox.setSource(getter);
    return true;
  }
  if (ucString.endsWith(".HTML") || ucString.endsWith(".HTM")) {
    XMLParser parser(sourceString, true); // true == using HTML
    Getter getter(parser);
    getter.setCheckBalance(true);
    documentBox.setSource(getter);
    return true;
  }
  if (ucString.endsWith(".SXF")) {
    SXFParser parser(sourceString);
    Getter getter(parser, true);
    documentBox.setSource(getter);
    return true;
  }
  if (ucString.endsWith(".RJ")) {
    RJParser parser(sourceString);
    Getter getter(parser, true);
    documentBox.setSource(getter);
    return true;
  }

  if (ucString.endsWith(".CSV")) {
    CSVParser parser(sourceString, ',', true); // true == use first row as labels
    Getter getter(parser, true);
    documentBox.setSource(getter);
    return true;
  }

  if (ucString.endsWith(".TXT")) {
    DotParser parser(sourceString); // true == use first row as labels
    Getter getter(parser, true);
    documentBox.setSource(getter);
    return true;
  }

  if (ucString.endsWith(".CSS")) {
    CSSParser parser(sourceString);
    Getter getter(parser, true);
    documentBox.setSource(getter);
    return true;
  }

  if (ucString.endsWith(".DTD")) {
    DTDGetter getter(sourceString);
    documentBox.setSource(getter);
    return true;
  }

  return false;
}

bool LexxiaApp::setDefaultInGetter(DocumentBox& documentBox, const DOMString& sourceString) {
  DOMString ucString = sourceString;
  ucString.toUpperCase();

  if (ucString.empty() || ucString == "?XML" || ucString == "?XSL" || ucString == "?XSD") {
    Reader reader;
    XMLParser parser(reader);
    Getter getter(parser);
    documentBox.setSource(getter);
    return true;
  }
  if (ucString == "?HTML" || ucString == "?HTM") {
    Reader reader;
    XMLParser parser(reader);
    Getter getter(parser);
    getter.setCheckBalance(true);
    documentBox.setSource(getter);
    return true;
  }
  if (ucString == "?SXF") {
    Reader reader;
    SXFParser parser(reader);
    Getter getter(parser, true);
    documentBox.setSource(getter);
    return true;
  }
  if (ucString == "?RJ") {
    Reader reader;
    RJParser parser(reader);
    Getter getter(parser, true);
    documentBox.setSource(getter);
    return true;
  }

  if (ucString == "?CSV") {
    Reader reader;
    CSVParser parser(reader, ',', true); // true == use first row as labels
    Getter getter(parser, true);
    documentBox.setSource(getter);
    return true;
  }

  if (ucString == "?TXT") {
    Reader reader;
    DotParser parser(reader); // true == use first row as labels
    Getter getter(parser, true);
    documentBox.setSource(getter);
    return true;
  }

  if (ucString == "?CSS") {
    Reader reader;
    CSSParser parser(reader);
    Getter getter(parser, true);
    documentBox.setSource(getter);
    return true;
  }
/*
  if (ucString == "?DTD") {
    Reader reader;
    DTDGetter getter(reader);
    documentBox.setSource(getter);
    return true;
  }
*/
  return false;
}

bool LexxiaApp::addPutter(DocumentBox& documentBox, const DOMString& argString, bool toConsole) {
  Putter putter(argString, toConsole);
  documentBox.addProcessor(putter);

  return true;
}

bool LexxiaApp::addTexConverter(DocumentBox& documentBox, const DOMString& argString) {
  ArgParser argParser(argString);
  DOMString configString = argParser.getArg(0);
  DOMString destString = argParser.getArg(1);

  // make sure that the first parameter is for configuration
  if (configString.endsWith(".css") || configString.endsWith(".xml") || configString == "**VERBOSE**") {
    TexConverter texConverter(configString, destString);
    documentBox.addProcessor(texConverter);
    return true;
  }
  // otherwise consider the first parameter to be a destination string with output to standard out
  else if (configString.endsWith(".tex")){
    TexConverter texConverter("", configString);
    documentBox.addProcessor(texConverter);
    return true;
  }

  verboseWriter_ << "PARAM ERROR for TexConverter: [" << configString << "] and [" << destString << "]" <<  wrt::endl;
  return false;
}

bool LexxiaApp::addHTMLStyler(DocumentBox& documentBox, const DOMString& argString, bool expandStyle) {
  ArgParser argParser(argString);
  DOMString screenString = argParser.getArg(0);
  DOMString printString = argParser.getArg(1);
  DOMString tagString = argParser.getArg(2);
  bool useTag = tagString.toUpperCase() == "TRUE";

  HTMLStyler htmlStyler(screenString, printString);
  htmlStyler.setExpanding(expandStyle);
  documentBox.addProcessor(htmlStyler);
return true;
}

bool LexxiaApp::addTransformer(DocumentBox& documentBox, const DOMString& argString) {
  // abort with error if there is no xformString
  if (argString.empty()) {
    verboseWriter_ << "ERROR: " << argString << " should at least have an xml path" << wrt::endl;
    return false;
  }

  ArgParser argParser(argString);
  DOMString xformString = argParser.getArg(0);
  DOMString destString = argParser.getArg(1);
  DOMString buildingString = argParser.getArg(2);
  bool isBuilding = false;

  if (destString == "true" || destString == "false") {
    buildingString = destString;
    destString = ""; // the destString is missing
  }

  isBuilding = buildingString == "true";

  // batch processing from multiple sources ?
  if (getters_.size() > 1 && !destString.empty() && !destString.endsWith("/"))
    destString += "/";

  // otherwise we assume that destString is either empty OR
  // a full and valid path string

  // have we xslstring and dest string?
  if (destString.getLength()) {
    XSLOutputter xslo(xformString, destString, isBuilding);
    documentBox.addProcessor(xslo);
    return true;
  }

  // otherwise output is to standard out
  XSLOutputter xslo(xformString, isBuilding);
  documentBox.addProcessor(xslo);
  return true;
}

bool LexxiaApp::addNewNamer(DocumentBox& documentBox, const DOMString& argString, bool discardChildren) {
  ArgParser argParser(argString);
  DOMString oldPath = argParser.getArg(0);
  DOMString newName = argParser.getArg(1);

  NewNamer newNamer(oldPath, newName, discardChildren);
  documentBox.addProcessor(newNamer);
  return true;
}

bool LexxiaApp::addJoiner(DocumentBox& documentBox, const DOMString& argString) {
  ArgParser argParser(argString);
  DOMString destString = argParser.getArg(0);
  DOMString rootString = argParser.getArg(1);
  DOMString titleString = argParser.getArg(2);

  Joiner mutator(destString, rootString, titleString);
  documentBox.addProcessor(mutator);
  return true;
}

bool LexxiaApp::addMutator(DocumentBox& documentBox, const DOMString& argString) {
  Mutator mutator(argString);
  documentBox.addProcessor(mutator);
  return true;
}

bool LexxiaApp::addXMLWrapper(DocumentBox& documentBox, const DOMString& argString, bool expandStyle) {
  ArgParser argParser(argString);
  DOMString printString = argParser.getArg(0);
  DOMString screenString = argParser.getArg(1);

  XMLWrapper xmlWrapper(printString, screenString);
  xmlWrapper.setExpanding(expandStyle);
  documentBox.addProcessor(xmlWrapper);

  return true;
}

bool LexxiaApp::addReporter(DocumentBox& documentBox, const DOMString& argString, bool simple) {
  Reporter reporter(argString);
  reporter.setSimple(simple);
  documentBox.addProcessor(reporter);
  return true;
}

bool LexxiaApp::isDocumentProcess(int argc, char *argv[]) {
  for (int index = 1; index < argc; ++index) {
    DOMString arg = getArgString(argv[index]);
    if (arg.startsWith("-S")) return false;
  }

  return true;
}

bool LexxiaApp::isVerbose(int argc, char *argv[]) {
  for (int index = 1; index < argc; ++index) {
    DOMString arg = getArgString(argv[index]);
    if (arg.startsWith("-V")) {
      if (arg.endsWith("true"))
	verboseWriter_ = wrt::Console;
      return true;
    }
  }

  return false;
}

bool LexxiaApp::configForStepper(ProcessBox& processBox, int argc, char *argv[]) {
  bool inputSet = false, outputSet = false;

  Builder builder;
  processBox.addProcessor(builder);

  for (int index = 1; index < argc; ++index) {
    DOMString arg = getArgString(argv[index]);

    if (arg.startsWith("-V"))
      continue;

    // is it a source for a parser? use only the first one!!
    if (!arg.startsWith("-") && !inputSet) {
      setParser(processBox, arg);
      if (outputSet) return true;

      inputSet = true;
      continue;
    }

    DOMString prefix = arg.substring(0, 2);
    DOMString param = arg.substring(2, arg.getLength() - 1);

    if (prefix == "-S" && !outputSet) {
      if (!addStepper(processBox, param))
        return false;

      outputSet = true;
      continue;
    }

    // if we have any other processors specified, signal a failure:
    verboseWriter_ << prefix << " and " << param << " are illegal with Stepper" << wrt::endl;

    return false;
  }

  // we lack either an input or stepper
  return inputSet && outputSet;
}

// set a parser for a process box
bool LexxiaApp::setParser(ProcessBox& processBox, const DOMString& sourceString) {
  if (verbose_)
    verboseWriter_ << "Processing source: " << sourceString << wrt::endl;

  if (sourceString.endsWith(".xml")) {
    XMLParser parser(sourceString);
    processBox.setSource(parser);
    return true;
  }
  if (sourceString.endsWith(".html")) {
    XMLParser parser(sourceString, true); // true == using HTML
    processBox.setSource(parser);
    return true;
  }
  if (sourceString.endsWith(".rj")) {
    RJParser parser(sourceString);
    processBox.setSource(parser);
    return true;
  }
  if (sourceString.endsWith(".csv")) {
    CSVParser parser(sourceString, true); // true == use first row as labels
    processBox.setSource(parser);
    return true;
  }

  if (sourceString.endsWith(".txt")) {
    DotParser parser(sourceString); // true == use first row as labels
    processBox.setSource(parser);
    return true;
  }

  return false;
}

bool LexxiaApp::addStepper(ProcessBox& processBox, const DOMString& argString) {
  // abort with error if there is no xformString
  if (argString.empty()) {
    verboseWriter_ << "ERROR: " << argString << " should at least have an xml path" << wrt::endl;
    return false;
  }

  ArgParser argParser(argString);
  DOMString xformString = argParser.getArg(0);
  DOMString destString = argParser.getArg(1);
  DOMString buildingString = argParser.getArg(2).toUpperCase();;

  if (buildingString.empty()) {
    DOMString destCopy = destString;
    destCopy.toUpperCase();
    if (destCopy == "TRUE" || destCopy == "FALSE") {
      buildingString = destCopy;
      destString.clear();
    }
  }

  bool isBuilding = buildingString != "FALSE";

  // have we xslstring and dest string?
  if (destString.getLength()) {
    Stepper stepper(xformString, destString, isBuilding);
    processBox.addProcessor(stepper);
    return true;
  }

  // otherwise output is to standard out
  Stepper stepper(xformString, isBuilding);
  processBox.addProcessor(stepper);
  return true;
}

DOMString LexxiaApp::getArgString(const char* arg) {
  return arg;
}

}
