 /***************************************************************************
                          namednodemap.cpp  -  description
                             -------------------
    begin                : Tue Mar 18 2003
    copyright            : (C) 2003 by John Redmond
    email                : john_redmond@optusnet.com.au
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "list"
#include "attr.h"
#include "element.h"
#include "document.h"
#include "elementcontent.h"
#include "namednodemap.h"
#include "systemmanager.h"
#include "writer.h"

using std::map;
using std::pair;

namespace limpid {

NamedNodeMap::NamedNodeMap():
    refCount_(0), elementContent_(0) {
  try {
    refCount_ = new size_t(1);
    nodes_ = new NodeVector;
  }
  catch (std::bad_alloc) {
    delete refCount_;
    delete nodes_;
    throw;
  }
}

NamedNodeMap::NamedNodeMap(Element &parent):
    refCount_(0), elementContent_(0) {
  try {
    elementContent_ = static_cast<ElementContent *>(parent.nodeContent_);
    refCount_ = new size_t(1);
    nodes_ = new NodeVector;
  }
  catch (std::bad_alloc) {
    delete refCount_;
    delete nodes_;
    throw;
  }
}

NamedNodeMap::NamedNodeMap(const NamedNodeMap& refMap) {
  copyContent(refMap);
}

NamedNodeMap::~NamedNodeMap() {
  if (!--*refCount_) deleteContent();
}

NamedNodeMap& NamedNodeMap::operator=(const NamedNodeMap& refMap) {
  if (this != &refMap) {
    if (!--*refCount_) deleteContent();
    copyContent(refMap);
  }
  return *this;
}

Node& NamedNodeMap::getNamedItem(const DOMString& name) const {
  for (NodeVector::iterator p = nodes_->begin(); p != nodes_->end(); ++p) {
    Node* node = *p;
    if (name == node->getNodeName()) return *node;
  }

  return Attr::failAttr;
}

Node& NamedNodeMap::setNamedItem(Node& node) {
  if (!isValid()) return node;

  if (elementContent_ && elementContent_->ownerNode_ &&  //-//
      !getSetPermission(*elementContent_->ownerNode_, node))
    return Attr::failAttr;

  Node& oldNode = remove(node.getNodeName());
  setItem(node);
  return oldNode;
}

Node& NamedNodeMap::removeNamedItem(const DOMString& name) {
  if (elementContent_ && elementContent_->ownerNode_ &&
      !getRemovePermission(*elementContent_->ownerNode_))
    return Attr::failAttr;
  else return remove(name);
}

Node& NamedNodeMap::removeSpecifiedItem(Node& node) {
  Node* foundNode;

  for (NodeVector::iterator p = nodes_->begin(); p != nodes_->end(); ++p) {
    if (node == **p) {
      foundNode = *p;

    if (elementContent_) {
     foundNode->nodeContent_->parentContent_ = 0;
     foundNode->getOwnerDocument().cacheNode(*foundNode);
     }

     nodes_->erase(p);
     return *foundNode;
    }
  }
  return Attr::failAttr;
}

Node& NamedNodeMap::item(size_t index) const {
  if (index <= nodes_->size()) return *(nodes_->at(index));
  else {
    Node::systemManager.processIndexError(index);
    return Attr::failAttr;
  }
}

size_t NamedNodeMap::getLength() const {
  return nodes_->size();
}

Node& NamedNodeMap::getNamedItemNS(const DOMString& namespaceURI,
                                    const DOMString& name) const {
  for (NodeVector::iterator p = nodes_->begin(); p != nodes_->end(); ++p) {
    Node* node = *p;
    if (name == node->getLocalName() &&
        node->getNamespaceURI() == namespaceURI) return *node;
  }
  return Attr::failAttr;
}

Node& NamedNodeMap::setNamedItemNS(Node& node) {
  if (elementContent_ && elementContent_->ownerNode_ &&
      !getSetPermission(*elementContent_->ownerNode_, node))
    return Attr::failAttr;

  DOMString localName = node.getLocalName();
  if (!localName.size()) return setNamedItem(node);

  DOMString namespaceURI = node.getNamespaceURI();
  if (elementContent_)
    node.nodeContent_->parentContent_ = elementContent_; // give access to ownerElement

  Node& oldNode = remove(namespaceURI, localName);
  setItem(node);
  return oldNode;
}

void NamedNodeMap::setItem(Node& node) {

  if (elementContent_) {
    Node& copy = node.copy();
    nodes_->push_back(&copy);

    copy.nodeContent_->parentContent_ = elementContent_;
    if (elementContent_->documentContent_)
      copy.setDocumentContent(*elementContent_->documentContent_);
  }
  else
    nodes_->push_back(&node);
}

Node& NamedNodeMap::removeNamedItemNS(const DOMString& namespaceURI,
                                      const DOMString& localName) {
  if (elementContent_ && !getRemovePermission(*elementContent_->ownerNode_))
    return Attr::failAttr;
  else
    return remove(namespaceURI, localName);
}

Node& NamedNodeMap::remove(const DOMString& name) {
  for (NodeVector::iterator p = nodes_->begin(); p != nodes_->end(); ++p) {
    Node *node = *p;
    if (name == node->getNodeName()) {
      if (elementContent_) {
        node->nodeContent_->parentContent_ = 0;
        node->getOwnerDocument().cacheNode(*node);
      }
      nodes_->erase(p);
      return *node;
    }
  }
  return Attr::failAttr;
}

Node& NamedNodeMap::remove(const DOMString& namespaceURI,
                                      const DOMString& localName) {
  for (NodeVector::iterator p = nodes_->begin(); p != nodes_->end(); ++p) {
    Node *node = *p;
    if (localName == node->getLocalName() &&
        node->getNamespaceURI() == namespaceURI) {
      if (elementContent_) {
        node->clearLinks();
        node->getOwnerDocument().cacheNode(*node);
      }
      nodes_->erase(p);
      return *node;
    }
  }
  return Attr::failAttr;
}

bool NamedNodeMap::getSetPermission(Node& parent, Node& child) {
  SystemManager& systemManager = Node::systemManager;
  if (!systemManager.checkModificationPermission(parent)) return false;
  if (!systemManager.checkCorrectDocument(parent, child)) return false;

  Attr& attr = static_cast<Attr &>(child);
  return systemManager.checkAttributeNotInUse(attr);
}

bool NamedNodeMap::getRemovePermission(Node& parent) {
  SystemManager& systemManager = Node::systemManager;
  return systemManager.checkModificationPermission(parent);
}

void NamedNodeMap::deleteContent() {
  delete refCount_;
  if (!elementContent_) return;

  for (NodeVector::iterator p = nodes_->begin(); p != nodes_->end(); ++p)
    delete *p;

  delete nodes_;
 }

void NamedNodeMap::copyContent(const NamedNodeMap& refMap) {
  refCount_ = refMap.refCount_;
  elementContent_ = refMap.elementContent_;
  nodes_ = refMap.nodes_;
  refCount_++;
}

size_t NamedNodeMap::getSize() {
  size_t size = sizeof(*this);
  size += sizeof(*nodes_);

  NodeVector::iterator p = nodes_->begin();
  while (p != nodes_->end()) {
    Node *node = *p;
    size += node->getSize();
    ++p;
  }

  return size;
}

NamedNodeMap NamedNodeMap::defaultAttributes;

}
