/***************************************************************************
                          Node.cpp  -  description
                             -------------------
    begin                : Wed Feb 19 2003
    copyright            : (C) 2003 by root
    email                : john_redmond@optusnet.com.au
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <iostream>
#include "node.h"
#include "document.h"
#include "element.h"
#include "text.h"
#include "comment.h"
#include "nodelist.h"
#include "namednodemap.h"
#include "systemmanager.h"
#include "nodestream.h"

namespace limpid {

Node::Node(): nodeContent_(0) {
  try {
    nodeContent_ = new NodeContent(FAIL_NODE);
  }
  catch (std::bad_alloc) {
    delete nodeContent_;
    throw;
  }
}

Node::Node(NodeType): nodeContent_(0) {
}

Node::Node(const Node& refNode): Copied(), nodeContent_(0) {
  try {
    nodeContent_ = refNode.nodeContent_;
    ++nodeContent_->refCount_;
  }
  catch (std::bad_alloc) {
    throw;
  }
}

Node::~Node() {
  if (nodeContent_ && !--nodeContent_->refCount_) {
      delete nodeContent_;
      nodeContent_ = 0;
  }
}

Node& Node::operator=(const Node& refNode) {
  nodeContent_ = 0;
  try {
    replaceContent(refNode);
    return *this;
  }
  catch (...) {
    throw;
  }
}

inline DOMString Node::getNodeName() const {
  if (nodeContent_)
    return nodeNames[nodeContent_->nodeType_];
  else return nodeNames[0];
}

Document& Node::getOwnerDocument() const {
  if (nodeContent_->documentContent_)
    return *nodeContent_->documentContent_->controlDocument_;
  else return Document::failDocument;
}

int Node::getNodeType() const {
  if (!nodeContent_) return BASE_NODE;
  else return nodeContent_->nodeType_;
}

Node& Node::getParentNode() const {
  if (nodeContent_->parentContent_ && nodeContent_->parentContent_->ownerNode_)
    return *nodeContent_->parentContent_->ownerNode_;
  else return failNode;
}

inline NodeList& Node::getChildNodes() const {
  return NodeList::defaultChildNodes;
}

inline NamedNodeMap& Node::getAttributes() const {
   return NamedNodeMap::defaultAttributes;
}

Node& Node::getPreviousSibling() const {
  if (nodeContent_->previousSibling_) return *(nodeContent_->previousSibling_);
  else return failNode;
}

Node& Node::getNextSibling() const {
  if (nodeContent_->nextSibling_) return *(nodeContent_->nextSibling_);
  else return failNode;
}

size_t Node::getSize() {
  size_t size = sizeof(*this);
  size += nodeContent_->getSize();
  return size;
}

DOMString Node::getDisplayString() const {
  DOMString s;

  if (getNodeType() == END_NODE) {
    s += "</";
    s += getNodeName();
    s += ">";
  }
  else {
    s += "<==";
    s += getNodeName();
    s += " NODE==>";
  }
  return s;
}

std::list<DOMString> Node::getPathList() const {
  std::list<DOMString> result;
  const Node *currentNode = this;

  while (true) {
    if (!currentNode || currentNode == &Node::failNode) break;
    int nodeType = currentNode->getNodeType();
    if (nodeType == DOCUMENT_NODE || nodeType == FAIL_NODE) break;
    result.push_front(currentNode->getNodeName());
    if (!currentNode->nodeContent_->parentContent_ ||
        !currentNode->nodeContent_->parentContent_->ownerNode_) break;

    Node *parentNode = currentNode->nodeContent_->parentContent_->ownerNode_;//getParentNode();
    if (!parentNode || parentNode->getNodeType() == DOCUMENT_NODE) break;

    currentNode = parentNode;
  }

  return result;
}

DOMString Node::getPath() const {
  DOMString resultString;
  std::list<DOMString> pathList = getPathList();
  std::list<DOMString>::iterator p = pathList.begin();

  while (p != pathList.end()) {
    resultString += "/";
    resultString += *p;
    p++;
  }

  return resultString;
}

Node& Node::appendElement(Element child) {
  Node& c = child;
  appendChild(c);
  return *this;
}

Node& Node::appendText(Text child) {
  Node& c = child;
  appendChild(c);
  return *this;
}

Node& Node::appendComment(Comment child) {
  Node& c = child;
  appendChild(c);
  return *this;
}

Node& Node::cloneNode(bool) const {
  Node *cpy = new Node;
  getOwnerDocument().cacheNode(*cpy);

  return *cpy;
}

void Node::detach() {
  ParentContent* theContent = nodeContent_->parentContent_;
  if (theContent && theContent->ownerNode_) {
    theContent->ownerNode_->removeChild(*this);
  }
}

void Node::replaceContent(const Node& refNode) {
  if (this != &refNode) {
    if (nodeContent_ && !--nodeContent_->refCount_) delete nodeContent_;
    nodeContent_ = refNode.nodeContent_;
    ++nodeContent_->refCount_;
  }
}

void Node::clearLinks() {
  nodeContent_->parentContent_ = 0;
  nodeContent_->previousSibling_ = 0;
  nodeContent_->nextSibling_ = 0;
}

void Node::reorder(size_t interval) {
  size_t position = interval;
  NodeStream stream(*this);

  while (true) {
    Node& node = stream.nextNode();
    if (!node.isValid()) break;

    node.setPosition(position);
    position += interval;

    if (node.getNodeType() == Node::ELEMENT_NODE) {
      Element element = static_cast<Element &>(node);
      NamedNodeMap attributes = element.getAttributes();
      size_t length = attributes.getLength();

      for (size_t index = 0; index < length; ++index) {
        attributes.item(index).setPosition(position);
        position += interval;
      }
    }
  }
}

SystemManager& Node::systemManager = SystemManager::getSystemManager();

Node Node::failNode;
Node Node::ignoreNode(Node::SYSTEM_NODE);
char Node::nodeNames[][20] = {"#base-node", "#sliced-node", "#sliced-node", "#text",
          "#cdata-section", "#sliced-node", "#sliced-node", "#sliced-node",
          "#comment", "#document", "#document-type", "#document-fragment",
          "#sliced-node", "#end-node", "#system-node", "#fail-node"};
bool Node::verbose = false;

}
