/***************************************************************************
                          nodelist.cpp  -  description
                             -------------------
    begin                : Wed Mar 19 2003
    copyright            : (C) 2003 by John Redmond
    email                : john_redmond@optusnet.com.au
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <iostream>
#include "parentcontent.h"
#include "systemmanager.h"
#include "nodelist.h"
#include "document.h"
#include "writer.h"

namespace limpid {

NodeList::NodeList():
    nodes_(new NodeVector), parentContent_(0), refCount_(new size_t(1)) {}

NodeList::NodeList(ParentNode &parent):
    nodes_(new NodeVector), parentContent_(static_cast<ParentContent *>(parent.nodeContent_)),
    refCount_(new size_t(1)) {}

NodeList::NodeList(const NodeList& refList) {
  try {
    copyContent(refList);
  }
  catch (std::bad_alloc) {
    throw;
  }
}

NodeList::~NodeList() {
  if (!--*refCount_)
    deleteContent();
}

NodeList& NodeList::operator=(const NodeList& refList) {
  if (this != &refList) {
    if (!--*refCount_)
      deleteContent();

    copyContent(refList);
  }
  return *this;
}

NodeList& NodeList::clear() {
  if (parentContent_ && nodes_ && nodes_->size()) {
    NodeVector::iterator p = nodes_->begin();
    while (p != nodes_->end()) {
      Node *node = *p++;
      delete node;
    }
  }
  nodes_->clear();
  return *this;
}

Node& NodeList::item(size_t index) {
  if (index <= nodes_->size()) return *(nodes_->at(index));
  else {
    Node::systemManager.processIndexError(index);
    return Node::failNode;
  }
}

Node& NodeList::addNode(Node& childNode) {
  if (!isValid()) return childNode;

  Node *copy = 0;
  try {
    if (parentContent_ && parentContent_->ownerNode_) {
      if (!getAppendPermission(*parentContent_->ownerNode_, childNode))
        return Node::failNode;

      copy = &childNode.copy();
      copy->detach();

      if (nodes_->size()) {
        nodes_->back()->nodeContent_->nextSibling_ = copy;
        copy->nodeContent_->previousSibling_ = nodes_->back();
      }

      copy->nodeContent_->nextSibling_ = 0;
      nodes_->push_back(copy);
      updateLinks(*copy);
      return *copy;
    }
    else {
      nodes_->push_back(&childNode);
      return childNode;
    }
  }
  catch (std::bad_alloc) {
    delete copy;
    return Node::failNode;
  }
}

Node& NodeList::removeNode(Node& childNode) {
  NodeVector::iterator p = nodes_->begin();
  while (p != nodes_->end()) {
    Node *ptr = *p;
    if (childNode == *ptr) {
      if (parentContent_) {
        if (parentContent_->ownerNode_ &&
            !getRemovePermission(*parentContent_->ownerNode_))
          return Node::failNode;

        Node *previous = childNode.nodeContent_->previousSibling_;
        Node *next = childNode.nodeContent_->nextSibling_;
        if (previous)
          previous->nodeContent_->nextSibling_ = next;
        if (next)
          next->nodeContent_->previousSibling_ = previous;

        childNode.clearLinks();
        childNode.getOwnerDocument().cacheNode(childNode);

      }
      nodes_->erase(p);
      return *ptr;
    }
    ++p;
  }
  return Node::failNode;
}

Node& NodeList::insertNode(Node& newChild, Node& refChild) {
  Node *copy = 0;
  try {
    NodeVector::iterator p = nodes_->begin();
    while (p != nodes_->end()) {
      Node *refPtr = *p;
      if (refChild == *refPtr) {
        if (parentContent_) {
          if (!getAppendPermission(*parentContent_->ownerNode_, newChild))
            return Node::failNode;

          copy = &newChild.copy();
          copy->detach();

          Node *previous = refChild.nodeContent_->previousSibling_;
          if (previous)
            previous->nodeContent_->nextSibling_ = copy;
          copy->nodeContent_->previousSibling_ = previous;

          copy->nodeContent_->nextSibling_ = &refChild;
          refChild.nodeContent_->previousSibling_ = copy;

          nodes_->insert(p, copy);
          updateLinks(*copy);
          return *copy;
        }
        else {
          nodes_->insert(p, &newChild); // no copy is made here
          return newChild;
        }
      }
      ++p;
    }
    return Node::failNode;
  }
  catch (std::bad_alloc) {
    delete copy;
    return Node::failNode;
  }
}

Node& NodeList::getFirstChild() {
  if (nodes_ && nodes_->size()) return *nodes_->front();
  else return Node::failNode;
}

Node& NodeList::getLastChild() {
  if (nodes_ && nodes_->size()) return *nodes_->back();
  else return Node::failNode;
}

inline void NodeList::updateLinks(Node& node) {
  node.nodeContent_->parentContent_ = parentContent_;
  if (parentContent_->documentContent_)
    node.setDocumentContent(*parentContent_->documentContent_);
  node.connect();
}

bool NodeList::getAppendPermission(const Node& parent, const Node& child) {
  SystemManager& systemManager = Node::systemManager;
  if (!systemManager.checkModificationPermission(parent)) return false;
  else if (!systemManager.checkCorrectDocument(parent, child)) return false;
  else return systemManager.checkAppendPermission(*parentContent_->ownerNode_, child);
}

bool NodeList::getRemovePermission(const Node& parent) {
  SystemManager& systemManager = Node::systemManager;
  return systemManager.checkModificationPermission(parent);
}

void NodeList::deleteContent() {
  delete refCount_;
  if (parentContent_) {
    NodeVector::iterator p = nodes_->begin();
    while (p != nodes_->end()) {
      Node *node = *p++;
      delete node;
    }
  }
  delete nodes_;
}

void NodeList::copyContent(const NodeList& refList) {
  refCount_ = refList.refCount_;
  parentContent_ = refList.parentContent_;
  nodes_ = refList.nodes_;
  ++*refCount_;
}

size_t NodeList::getSize() {
  size_t sz = sizeof(*this);
  sz += sizeof(*nodes_);

  return sz;
}

NodeList NodeList::defaultChildNodes;

}
