/***************************************************************************
                          numberformatter.cpp  -  description
                             -------------------
    begin                : Sun Jan 23 2005
    copyright            : (C) 2005 by John Redmond
    email                : john_redmond@optusnet.com.au
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <iostream>
#include <sstream>
#include <string>
#include "writer.h"
#include "numberformatter.h"
#include "romannumberformatter.h"
#include "converter.h"

namespace limpid {

NumberFormatter::NumberFormatter(){
}

NumberFormatter::NumberFormatter(const DOMString& formatString,
    const DOMString& langString, const DOMString& letterString,
    const DOMString& separatorString, const DOMString& sizeString):
        separatorString_(separatorString), letterString_(letterString),
        sizeString_(sizeString),
        /*literals_(new TokenList), formats_(new TokenList), fillers_(new TokenList),
        baseList_(new NumberList), widthList_(new NumberList),*/ outputLocale_(0) {

  parseFormat(formatString);
  setLocale(langString);  // this is not presently used
}

DOMString NumberFormatter::format(size_t number, size_t fieldNumber) {
  // this error condition should never happen
  if (fieldNumber >= fieldSpecifications_.size()) return Converter::toString(number);
  const FieldSpecification& fieldSpecification = fieldSpecifications_[fieldNumber];

  DOMString numberString;

  if (fieldSpecification.isRoman)
    numberString = RomanNumberFormatter::format(number, fieldSpecification.isUpperCase);
  else
    numberString = getNumberString(number, fieldSpecification);

  return numberString += fieldSpecification.postString;
}

DOMString NumberFormatter::getNumberString(size_t number,
      const FieldSpecification& fieldSpecification) {
  DOMString numberString;
  size_t numberBase = fieldSpecification.numberBase;
  Char baseChar = '0'; // for decimal
  if (numberBase == 26) {
    if (fieldSpecification.isUpperCase) baseChar = 'A';
    else baseChar = 'a';
  }

  while (number >= numberBase) {
    int result = number % numberBase;
    number = number / numberBase;
    numberString += Char(result + baseChar);
  }

  if (number) numberString += Char(number + baseChar); // get the final digit

  DOMString resultString;

  size_t extraLength = fieldSpecification.width - numberString.size();
  if (extraLength)
    resultString = fieldSpecification.preString.substr(0, extraLength);

  return resultString += numberString.reverse(); // get the chars in the correct order
}

void NumberFormatter::parseFormat(const DOMString& formatString) {
  size_t formatLength = formatString.size();
  if (!formatLength) return;

  DOMString numberString, postString;
  size_t index = 0;

  // parse out the literal and format substrings from the overall format string.
  // node that the first format string may not have a literal before it.
  while (true) {
    if (index >= formatLength) break;

    numberString.clear();
    postString.clear();

    while (index < formatLength && isalnum(formatString[index])) {
      numberString += formatString[index++];
    }

    if (index < formatLength) {
      while (index < formatLength && !isalnum(formatString[index])) {
        postString += formatString[index++];
      }
    }

    setFieldSpecification(numberString, postString);
  }
}

void NumberFormatter::setFieldSpecification(DOMString& numberString, DOMString& postString) {
    size_t length = numberString.size();
    if (!length) return; // this should never happen

    size_t numberBase = 10, width = 1;  // these are the defaults
    bool isUpperCase = true, isRoman = false;

    Char controlChar = numberString[length - 1];

    // get the number base and case status
    if (controlChar == 'A') numberBase = 26;
    else if (controlChar == 'a') {
      numberBase = 26;
      isUpperCase = false;
    }
    else if (controlChar == 'I')
      isRoman = true;
    else if (controlChar == 'i') {
      isRoman = true;
      isUpperCase = false;
    }
    // otherwise we have arabic chars and base 10

    // get the padding width in '0' chars
    DOMString paddingString;
    if (length > 2 && numberBase == 10 && !isRoman)
      paddingString = DOMString(length - 1, '0');

  FieldSpecification fieldSpec;
  fieldSpec.preString = paddingString;
  fieldSpec.postString = postString;
  fieldSpec.numberBase = numberBase;
  fieldSpec.width = width;
  fieldSpec.isUpperCase = isUpperCase;
  fieldSpec.isRoman = isRoman;
  fieldSpecifications_.push_back(fieldSpec);
}

// This is a potenially very error-prone function. If langString does not
// specify a locale that is available on the system, a runtime_exception will be
// thrown. In this event, the exception is caught and the current global locale
// remains.
// In any case, locale is not presently used.
void NumberFormatter::setLocale(const DOMString& langString) {
  if (langString.empty()) return;

  const char* lang = 0;
  try {
    lang = langString.getCString();
    outputLocale_ = new std::locale(lang);
  }
  catch (...) {
    wrt::Errors << "Unknown locale [" << lang << "]" << wrt::endl;
  }

  delete [] lang;
}

}
