/***************************************************************************
                          numbertemplate.cpp  -  description
                             -------------------
    begin                : Thu Oct 21 2004
    copyright            : (C) 2004 by John Redmond
    email                : john_redmond@optusnet.com.au
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "numbertemplate.h"
#include "nodesource.h"
#include "siblingstream.h"
#include "ancestorstream.h"
#include "followingstream.h"
#include "nodetest.h"
#include "nametest.h"
#include "exprtest.h"
#include "xpathfactory.h"
#include "stringwriter.h"
#include "text.h"

namespace limpid {

NumberTemplate::NumberTemplate(const DOMString& countString,
    const DOMString& valueString, const DOMString& levelString, const DOMString& fromString):
    level_(SINGLE), defined_(true), separator_(". "), numberStack_(new NumberStack) {
  configure(countString, valueString, levelString, fromString);
}

bool NumberTemplate::process(XSLContext& parentContext) {
  parentContext.updateCurrentNode();
  numberStack_->clear();

  if (valueExpr_.isValid()) // this is the simplest and fastest option
    numberStack_->push_back(valueExpr_.getNumber(parentContext).getInt());
  else {
    if (!defined_)  // node test is generated dynamically
      test_ = new NodeTest(parentContext.getNode());

    addCounts(parentContext);
  }

  output(parentContext.getWriter());
  return true;
}

bool NumberTemplate::processNode(XSLContext& parentContext, Node& node) {
  parentContext.updateCurrentNode();
  numberStack_->clear();

  if (valueExpr_.isValid()) // this is the simplest and fastest option
    numberStack_->push_back(valueExpr_.getNumber(parentContext).getInt());
  else {
    if (!defined_)  // node test is generated dynamically
      test_ = new NodeTest(parentContext.getNode());

    addCounts(parentContext);
  }

  StringWriter writer;
  output(writer);
  Text numberText(writer.getString());
  node.appendChild(numberText);

  return true;
}

// collect one or more counts and push them to a stack
void NumberTemplate::addCounts(XSLContext& parentContext) {
  XSLContext testContext;

  numberStack_->push_back(getCount(parentContext.getNode(), stream_));
  if (level_ != MULTIPLE) return;

  // only the multiple option uses more than one stream
  ancestorStream_->setRoot(parentContext.getNode());
  while (true) {
    Node& node = ancestorStream_->nextNode();
    if (!node.isValid()) break;

    numberStack_->push_back(getCount(node, stream_));

    testContext.setNode(node);

    // if there is a valid from test and it matches this node,
    // do not process the stream further
    if (fromExpr_.isValid() &&
      fromExpr_.getNodeSet(testContext, true).getLength()) break;
  }
}

// count the number of steam elements that pass the test
int NumberTemplate::getCount(Node& rootNode, Pointer<NodeSource> stream) {
  stream->setRoot(rootNode);
  size_t hits  = 1;  // the root element is the first to be counted

  while (true) {
    Node& node = stream->nextNode();
    if (!node.isValid()) break;

    if (test_->test(node))
      ++hits;
  }

  return hits;
}

// output the integers on the number on a LIFO basis
void NumberTemplate::output(CharacterWriter& writer) {
  size_t count = 0;

  while (!numberStack_->empty()) {
    writer << numberFormatter_.format(numberStack_->back(), count++);
    numberStack_->pop_back();
  }
}

// configure on the basis of all the parameter strings
void NumberTemplate::configure(const DOMString& countString,
    const DOMString& valueString, const DOMString& levelString,
    const DOMString& fromString) {

  DOMString level = levelString; // make a non-const copy
  level.toUpperCase();
  if (level == "MULTIPLE") level_ = MULTIPLE;
  else if (level == "ANY") level_ = ANY;

  ancestorStream_ = new AncestorStream;

  if (level_ == ANY)
    stream_ = new FollowingStream(false);  // false == preceding
  else
    stream_ = new SiblingStream(false); // false == preceding siblings;

  if (countString.empty()) // the test is set dynamically when defined_ is reset
    defined_ = false;
  else {
    Expr expr = XPathFactory::getReverseExpr(countString);
    test_ = new ExprTest(expr, true); // true for exact matching
  }

  if (!valueString.empty())
    valueExpr_ = XPathFactory::getExpr(valueString, false); // false == silent build

  if (!fromString.empty())
    fromExpr_ = XPathFactory::getReverseExpr(fromString, false);// false == silent build
}

}
