/***************************************************************************
 *   Copyright (C) 2007 by John Redmond   *
 *   john_redmond@optusnet.com.au   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "pathregistry.h"
#include "matchedtemplate.h"
#include "stringreader.h"
#include "tokenstream.h"

namespace limpid {

PathRegistry::PathRegistry(): cssRefChar_('A'), xpathRefChar_('A') {}

void PathRegistry::addPath(const DOMString& pathString) {
  MatchedTemplate matchedTemplate(pathString, "");
  selectionManager_.setMatchedTemplate(pathString, matchedTemplate, false); // not imported
}

DOMString PathRegistry::addPath(Element& element, const DOMString& priorityString, bool isXPathString) {
  setPaths(element, isXPathString);
  DOMString correctedPath = element.getAttribute("path");
  MatchedTemplate matchedTemplate(correctedPath, "");

  if (!priorityString.empty())
    matchedTemplate.setPriority(toDouble(priorityString));
  selectionManager_.setMatchedTemplate(correctedPath, matchedTemplate, false); // not imported

  return correctedPath;
}

DOMString PathRegistry::getPath(Node& node) {
  Template& templ = selectionManager_.getMatchedTemplate(node);
  if (templ.getPriority() == 1234)
    return "**IGNORE**";

  return templ.getMatchValue();
}

void PathRegistry::setPaths(Element& element, bool isXPathString) {
  DOMString valueString = element.getValue();

  element.setAttribute("style", getPathStyle(valueString));

  if (isXPathString)
    element.setAttribute("path", valueString);
  else
    element.setAttribute("path", correctPath(valueString));
}

DOMString PathRegistry::correctPath(DOMString& pathString) {
  if (pathString.startsWith("xpath:"))
    return pathString.change("xpath:", "");

  DOMString valueString = pathString, resultString, signalString, delimiterString = " >+.#:";
  DOMString inputString = valueString;

  valueString.normalizeWhitespace();
  valueString.change(" >", ">");
  valueString.change("> ", ">");
  valueString.change(" +", "+");
  valueString.change("+ ", "+");
  valueString.change(" .", ".");
  valueString.change(". ", ".");
  valueString.change(" #", "#");
  valueString.change("# ", "#");
  valueString.change(" :", ":");
  valueString.change(": ", ":");

  StringReader reader(valueString);
  TokenStream tis(reader);
  tis.setWhitespaceChars("");
  tis.setDelimiters(delimiterString);

  DOMString nextToken, followingToken, extraToken;
  bool colonSet = false;

  while (true) {
    nextToken = tis.nextToken();
    if (nextToken.empty()) break;

    followingToken = tis.peekToken(); 

    if (followingToken == "+") {
      tis.nextToken();  // discard the "+"

      extraToken = tis.nextToken();
      if (extraToken.empty()) break;  // this is an error

      resultString += extraToken;
      resultString += "[name(preceding-sibling::*)=\"";
      resultString += nextToken; // the token before the '+'
      resultString += "\"]";

      continue;
    }

    if (nextToken == ":") {
      nextToken = tis.nextToken();
      if (nextToken == "first-child")
        resultString += "/*[not (preceding-sibling::*)]";
      else if (nextToken == "last-child")
        resultString += "/*[not (following-sibling::*)]";
      else break; // otherwise there is an error

      continue;
    }

    if (nextToken == " ") {
      resultString += "//";

      continue;
    }

    if (nextToken == ">") {
      resultString += "/";

      continue;
    }

    if (!delimiterString.contains(nextToken)) {
      resultString += nextToken;

      continue;
    }

    if (nextToken == "." || nextToken == "#") {
      DOMString attrString = tis.nextToken();
      if (nextToken == ".")
        resultString += "[@class=";
      else
        resultString += "[@id=";

      resultString += "\"";
      resultString += attrString.escapeDelimiters();
      resultString += "\"";

      resultString += "]";
    }
  }

  pathMap_[resultString] = pathString; // map for later back conversion

  return resultString;
}

DOMString PathRegistry::getPathStyle(DOMString& pathString) {
  DOMString valueString = pathString;

  valueString.normalizeWhitespace();
  valueString.change(" >", ">");
  valueString.change("> ", ">");
  valueString.change(">", "C");

  valueString.change(" +", "+");
  valueString.change("+ ", "+");
  valueString.change("+", "P");

  valueString.change(" .", ".");
  valueString.change(". ", ".");
  valueString.change(".", "D");

  valueString.change(" #", "#");
  valueString.change("# ", "#");
  valueString.change("#", "H");

  valueString.change(" ", "");

  return valueString;
}

}
