/***************************************************************************
 *   Copyright (C) 2007 by root   *
 *   john_redmond@optusnet.com.au   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "reformer.h"
#include "element.h"
#include "attr.h"
#include "text.h"
#include "nodestream.h"
#include "liststream.h"
#include "nodelist.h"
#include "stringreader.h"
#include "tokenstream.h"
#include "stringreader.h"
#include "dombuilder.h"
#include "xmlparser.h"
#include "domwriter.h"
#include "stringwriter.h"
#include "getter.h"

namespace limpid {

  Document& Reformer::processDocument(Document& document) {
    upgrade(document);
    return document;
  }

  void Reformer::upgrade(Document& document) {
    static DOMString upgradeNames = "table";
    NodeList matchList;
    NodeStream nodeStream(document.getDocumentElement(), false, false);

    while (true) {
      Node& node = nodeStream.nextNode();
      if (!node.isValid()) break;

      if (node.getNodeType() != Node::ELEMENT_NODE) continue;

      DOMString nodeName = node.getNodeName();

      if (nodeName == "table")
        reformTable(static_cast<Element &>(node));
      if (nodeName == "ul" || nodeName == "ol")
        reformList(static_cast<Element &>(node));
      if (nodeName == "dl")
        reformDefinitionList(static_cast<Element &>(node));
      if (nodeName == "a")
        reformAnchor(static_cast<Element &>(node));
      if (nodeName == "pre")
        reformPreText(static_cast<Element &>(node));
      if (nodeName == "meta")
        reformMetaText(static_cast<Element &>(node));
    }

    replaceMixedContent(document);
  }

  void Reformer::reformTable(Element& tableElement) {
    ListStream childStream(tableElement);
    NodeList tempList;
    int headCount = 0, rowCount = 0, footCount = 0;
    bool headAdded = false, footAdded = false;
    Element headElement("thead");
    Element bodyElement("tbody");
    Element footElement("tfoot");

    while (true) {
      Node& node = childStream.nextNode();
      if (!node.isValid()) break;
      tempList.addNode(node);

      DOMString nodeName = node.getNodeName();
      if (nodeName == "thead" || nodeName == "head") ++headCount;
      if (nodeName == "p") ++rowCount;
      if (nodeName == "tfoot" || nodeName == "foot") ++footCount;
    }

    if (headCount) tableElement.appendChild(headElement);
    if (rowCount) tableElement.appendChild(bodyElement);
    if (footCount) tableElement.appendChild(footElement);

    ListStream tempStream(tempList);
    while (true) {
      Node& node  = tempStream.nextNode();
      if (!node.isValid()) break;

      DOMString nodeName = node.getNodeName();

      if (!headAdded && (nodeName == "thead" || nodeName == "head")) {
        replaceRow(headElement, static_cast<Element &>(node));
        headAdded = true;
        continue;
      }

      else if (!footAdded && (nodeName == "tfoot" || nodeName == "foot")) {
        replaceRow(footElement, static_cast<Element &>(node));
        footAdded = true;
        continue;
      }

      else if (nodeName == "p")
        replaceRow(bodyElement, static_cast<Element &>(node));
    }
  }

  void Reformer::replaceRow(Element& parentNode, Element& sourceElement) {
    Element rowElement("tr");
    parentNode.appendChild(rowElement);
    StringReader contentReader(sourceElement.getValue());
    TokenStream tis(contentReader);
    tis.setWhitespaceChars("|");
    tis.setDelimiters("");

    while (true) {
      DOMString token = tis.nextToken();
      if (token.empty()) break;

      addLeaf("td", token, rowElement);
    }

    sourceElement.getParentNode().removeChild(sourceElement); 
  }

  void Reformer::reformList(Element& listElement) {
    ListStream listStream(listElement);
    NodeList copyList;

    while (true) {
      Node& node = listStream.nextNode();
      if (!node.isValid()) break;

      if (node.getNodeName() == "p")
        copyList.addNode(node);
    }

    ListStream copyStream(copyList);
    while (true) {
      Node& node = copyStream.nextNode();
      if (!node.isValid()) break;

      Element itemElement("li");
      Text itemText = node.getValue();
      itemElement.appendChild(itemText);
      listElement.replaceChild(itemElement, node);
    }
  }

  void Reformer::reformDefinitionList(Element& listElement) {
    ListStream listStream(listElement);
    NodeList copyList;

    while (true) {
      Node& node = listStream.nextNode();
      if (!node.isValid()) break;

      if (node.getNodeName() == "p")
        copyList.addNode(node);
    }

    ListStream copyStream(copyList);
    while (true) {
      Node& node = copyStream.nextNode();
      if (!node.isValid()) break;

      DOMString contentString = node.getValue();
      DOMString tagString = "";
      DOMString dataString = contentString;

      size_t barIndex = contentString.indexOf("|");
      if (barIndex != -1) {
        tagString = contentString.substr(0, barIndex);
        dataString = contentString.substr(barIndex + 1);
      }

      if (!tagString.empty()) {
        Element tagElement("dt");
        Text tagText(tagString);
        tagElement.appendChild(tagText);
        listElement.insertBefore(tagElement, node);
      }

      Element dataElement("dd");
      Text dataText(dataString);
      dataElement.appendChild(dataText);
      listElement.replaceChild(dataElement, node);
    }
  }

  void Reformer::reformAnchor(Element& anchorElement) {
    DOMString contentString = anchorElement.getValue();
    if (anchorElement.getAttributeNode("href").isValid() ||
        anchorElement.getAttributeNode("name").isValid())
      return; // nothing to do

    // set up defaults 
    DOMString refString = contentString;
    DOMString hrefString = contentString;

    // split the string
    int barIndex = contentString.indexOf("|");
    if (barIndex != -1) {
      refString = contentString.substr(0, barIndex);
      hrefString = contentString.substr(barIndex + 1);
    }

    Text refText(refString);
    anchorElement.replaceChild(refText, anchorElement.getFirstChild());
    anchorElement.setAttribute("href", hrefString);
  }

  void Reformer::reformPreText(Element& preElement) {
    DOMString contentString = preElement.getValue();
    contentString.change("<", "&lt;");

    Text newText(contentString);
    preElement.replaceChild(newText, preElement.getFirstChild());
  }

  void Reformer::reformMetaText(Element& metaElement) {
    DOMString metaString = metaElement.getValue();
    DOMString metaType = "name";
    DOMString metaValue;
    DOMString contentString = metaString;

    int barIndex = metaString.indexOf("|");
    if (barIndex != -1) {
      metaValue = metaString.substr(0, barIndex);
      contentString = metaString.substr(barIndex + 1);
    }

    int equalIndex = metaValue.indexOf("=");
    if (equalIndex != -1) {
      metaType = metaValue.substr(0, equalIndex);
      metaValue = metaValue.substr(equalIndex + 1);
    }

    metaElement.setAttribute(metaType, metaValue);
    metaElement.setAttribute("content", contentString);

    metaElement.removeChild(metaElement.getFirstChild());
  }

  void Reformer::replaceMixedContent(Document& document) {
    NodeList matchList;
    NodeStream nodeStream(document.getDocumentElement(), false, false);

    while (true) {
      Node& node = nodeStream.nextNode();
      if (!node.isValid()) break;

      if (node.getNodeType() != Node::TEXT_NODE) continue;
      if (static_cast<Text &>(node).getData().contains("<")) {
        matchList.addNode(node);
      }
    }

    ListStream listStream(matchList);
    listStream.setForward(false); // reverse order

    while (true) {
      Node& node = listStream.nextNode();
      if (!node.isValid()) break;

      replaceText(node.getParentNode());
    }
  }

  void Reformer::replaceText(Node& node) {
    StringWriter stringWriter;
    DOMWriter(stringWriter).write(node);
    DOMString outString = stringWriter.getString();

    StringReader stringReader(outString);
    XMLParser parser(stringReader);
    Getter getter(parser);
    Node& newNode = getter.getDocument().getDocumentElement();

    node.getOwnerDocument().importNode(newNode, true);
    node.getParentNode().replaceChild(newNode, node);
  }

}
