/***************************************************************************
                          rjparser.cpp  -  description
                             -------------------
    begin                : Sun Apr 9 2006
    copyright            : (C) 2006 by John Redmond
    email                : john_redmond@optusnet.com.au
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "rjparser.h"
#include "node.h"
#include "element.h"
#include "attr.h"
#include "text.h"
#include "comment.h"
#include "cdatasection.h"
#include "xmldeclaration.h"
#include "processinginstruction.h"
#include "documenttype.h"
#include "endnode.h"
#include "writer.h"
#include "stringreader.h"

#include "domwriter.h"

namespace limpid {

RJParser::RJParser(const DOMString& fileName):
    endNode_(new EndNode("ENTRY")),
    reader_(new Reader(fileName)), tis_(new TokenStream(*reader_)),
    atStart_(true), inEntry_(false), hasContent_(false), atEnd_(false) {
  tis_->setWhitespaceChars("\n");
  tis_->setDelimiters("");
}

RJParser::RJParser(const Reader& charStream):
    endNode_(new EndNode("ENTRY")),
    reader_(new Reader(charStream)), tis_(new TokenStream(*reader_)),
    atStart_(true), inEntry_(false), hasContent_(false), atEnd_(false) {
  tis_->setWhitespaceChars("\n");
  tis_->setDelimiters("");
}

Node& RJParser::nextNode() {
  if (atStart_) {
    atStart_ = false;
    node_ = new Element("ENTRIES");
    node_->addChildNodes();
    return *node_;
  }

  if (hasContent_) {
    hasContent_ = false;
    atEnd_ = true;
    return *contentNode_;
  }

  if (atEnd_) {
    atEnd_ = false;
    return *endNode_;
  }

  if (!tis_->isValid())
    return Node::failNode;

  if (!inEntry_) {
    inEntry_ = true;
    node_ = new Element("ENTRY");
    node_->addChildNodes();
    return *node_;
  }

  // otherwise we are inside an entry

  if (!setLine()) // the only available lines may have been empty
    return Node::failNode;

  parseLine();
  node_->setPosition(tis_->getLineNo());
  return *node_;
}

void RJParser::parseLine() {
  StringReader stringReader(line_);
  TokenStream tis(stringReader);
  tis.setDelimiters(":");
  tis.setWhitespaceChars(" \r\n\t");

  DOMString nodeName = tis.nextToken().trim();
  if (nodeName != "%%")
    tis.bump();
  
  DOMString contentString = tis.getDelimitedString('\n').trim();
  if (contentString.getLength())
    completeContentString(contentString);

  if (nodeName == "%%") {
    node_ = new Comment(contentString);
    inEntry_ = false;
    atEnd_ = true;
  }
  else {
    node_ = new Element(nodeName);
    node_->addChildNodes();
    contentNode_ = new Text(contentString);
    contentNode_->setPosition(tis_->getLineNo());
    hasContent_ = true;
    atEnd_ = true;
  }
}

bool RJParser::setLine() {
  while (true) {
    if (!tis_->isValid()) return false;

    line_ = tis_->nextToken();
    if (!line_.empty()) {
      line_ += '\n';
      return true;
    }
  }
}

void RJParser::completeContentString(DOMString& contentString) {
  if (!contentString.endsWith('\\')) return;

  contentString[contentString.getLength() - 1] = ' ';

  bool needMore = true;

  while (needMore && tis_->isValid()) {
    DOMString extraString = tis_->nextToken();
    needMore = extraString.endsWith('\\');

    if (needMore)
      extraString[extraString.getLength() - 1] = ' ';

    contentString += extraString;
  }
}

}
