/***************************************************************************
                          rjxparser.cpp  -  description
                             -------------------
    begin                : Sat Mar 11 2006
    copyright            : (C) 2006 by John Redmond
    email                : john_redmond@optusnet.com.au
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "rjxparser.h"
#include "node.h"
#include "element.h"
#include "attr.h"
#include "text.h"
#include "comment.h"
#include "cdatasection.h"
#include "xmldeclaration.h"
#include "processinginstruction.h"
#include "documenttype.h"
#include "endnode.h"
#include "writer.h"
#include "stringreader.h"

#include "domwriter.h"

namespace limpid {

RJXParser::RJXParser(const DOMString& fileName):
    endNode_(new EndNode("END")),
    reader_(new Reader(fileName)), tis_(new TokenStream(*reader_)),
    hasContent_(false), hasEnd_(false), depth_(0), lineNo_(0) {
  tis_->setWhitespaceChars("\n");
  tis_->setDelimiters("");
}

RJXParser::RJXParser(const Reader& charStream):
    endNode_(new EndNode("END")),
    reader_(new Reader(charStream)), tis_(new TokenStream(*reader_)),
    hasContent_(false), hasEnd_(false), depth_(0), lineNo_(0) {
  tis_->setWhitespaceChars("\n");
  tis_->setDelimiters("");
}

Node& RJXParser::nextNode() {
  if (hasContent_) {
    hasContent_ = false;
    hasEnd_ = true;
    return *contentNode_;
  }

  if (hasEnd_) {
    hasContent_ = false;
    hasEnd_ = false;
    --depth_;
    return *endNode_;
  }

  if (!setLine())
    return Node::failNode;

  try {
    parseLine();
    node_->setPosition(lineNo_);
  }
  catch(...) {
    wrt::Errors << "BAD ALLOCATION ERROR IN RJXPARSER" << wrt::endl;
    throw;
  }

  return *node_;
}

void RJXParser::parseLine() {
  StringReader stringReader(line_);
  TokenStream tis(stringReader);
  tis.setDelimiters(";=");
  tis.setWhitespaceChars(" \r\n\t");

  DOMString nodeName = tis.nextToken().trim();
  if (nodeName == "%%") {
    if (depth_) {
      --depth_;
      node_ = endNode_;
    }
    return;
  }

  DOMString delimiter = tis.nextToken(); // this should be ';' or '='

  // process a parent node
  if (nodeName.startsWith("%")) {
    nodeName = nodeName.substr(1, nodeName.getLength());
    node_ = new Element(nodeName);
    node_->addChildNodes(); // signal nesting to the XMLParser
    ++depth_;
    return;
  }

  // process an attribute
  if (nodeName.startsWith("@")) {
    nodeName = nodeName.substr(1, nodeName.getLength());
    DOMString contentString = tis.getDelimitedString('\n');
    completeString(contentString, true);
    contentString.normalizeWhitespace().escapeDelimiters();

    node_ = new Attr(nodeName);
    Text contentText(contentString);
    node_->appendChild(contentText);
    return;
  }

  if (!nodeName.startsWith("#")) { // we have an element with only text content
    DOMString contentString = tis.getDelimitedString('\n');
    node_ = new Element(nodeName); // this node will be returned by nextNode()
    if (!contentString.empty()) {
      node_->addChildNodes(); // signal nesting to the XMLParser

      completeString(contentString, true);
      contentString.escapeBra();
      contentNode_ = new Text(contentString);

      hasContent_ = true; // nextNode() will pick up the text node next time it is called
      ++depth_;
  }
    return;
  }

  if (nodeName.startsWith("#te")) {
    DOMString contentString = tis.getDelimitedString('\n').escapeBra();
    completeString(contentString, true);
    contentString.normalizeWhitespace().escapeBra();
    node_ = new Text(contentString);
    return;
  }

  if (nodeName.startsWith("#co")) {
    DOMString contentString = tis.getDelimitedString('\n');
    completeString(contentString, true);
    node_ = new Comment(contentString);
    return;
  }

  if (nodeName.startsWith("#cd")) {
    DOMString contentString = tis.getDelimitedString('\n');
    completeString(contentString, false);
    node_ = new CDATASection(contentString);
    return;
  }

  if (nodeName.startsWith("#xm")) {
    DOMString version = tis.getDelimitedString(',').trim();
    DOMString encoding  = tis.getDelimitedString(',').trim();
    DOMString standalone = tis.nextToken().trim();
    node_ = new XMLDeclaration(version, encoding, standalone);
    return;
  }

  if (nodeName.startsWith("#pr")) {
    DOMString target = tis.getDelimitedString(',').trim();
    DOMString data  = tis.getDelimitedString('\n').trim();
    completeString(data, false);
    data.normalizeWhitespace().escapeDelimiters();
    node_ = new ProcessingInstruction(target, data);
    return;
  }

  if (nodeName.startsWith("#do")) {
    DOMString name = tis.getDelimitedString(',').trim();
    DOMString publicID = tis.getDelimitedString(',').trim();
    DOMString systemID = tis.getDelimitedString(',').trim();
    DOMString internalSubset = tis.getDelimitedString('\n').trim();
    completeString(internalSubset, true);
    internalSubset.trim();

    DocumentType* doctype = new DocumentType(name, publicID, systemID);
    if (internalSubset.getLength())
      doctype->setInternalSubset(internalSubset);

    node_ = doctype;

    return;
  }
}

bool RJXParser::setLine() {
  while (true) {
    if (!tis_->isValid()) return false;

    line_ = correctEnd(tis_->nextToken());
    ++lineNo_;
    if (!line_.empty()) {
      line_ += '\n';
      return true;
    }
  }
}

// concatenate two or more text lines if they are escaped with a '\\' character
DOMString& RJXParser::completeString(DOMString& dataString, bool insertSpace) {
  if (!dataString.endsWith('\\')) return dataString; // carriage returns have been removed

  while (tis_->isValid()) {
    size_t length = dataString.size();
    
    if (dataString.endsWith("\\\\")) {
      dataString.erase(length - 2, 2);
      dataString += wrt::endl;
    }
    else {
      if (insertSpace)
        dataString[length - 1] = ' ';
      else
        dataString.erase(length - 1);
    }

    if (!setLine()) break;

    dataString += correctEnd(line_);
    if (!dataString.endsWith('\\')) break;
  }

  return dataString;
}

DOMString& RJXParser::correctEnd(DOMString& dataString) {
  if (dataString.empty()) return dataString; // nothing done

  size_t length = dataString.getLength();
  if (dataString[length - 1] == '\r')
    dataString.erase(length - 1);  // drop the carriage return

  return dataString;
}

}
