/***************************************************************************
                          rjxwriter.cpp  -  description
                             -------------------
    begin                : Tue Mar 21 2006
    copyright            : (C) 2006 by John Redmond
    email                : john_redmond@optusnet.com.au
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "rjxwriter.h"
#include "cdatasection.h"
#include "comment.h"
#include "documenttype.h"
#include "element.h"
#include "endnode.h"
#include "processinginstruction.h"
#include "text.h"
#include "xmldeclaration.h"
#include "attributestream.h"

namespace limpid {

RJXWriter::RJXWriter(): writer_(new Writer), depth_(0) {}

RJXWriter::RJXWriter(const DOMString& destString):
  writer_(new Writer(destString)), depth_(0) {}
RJXWriter::RJXWriter(CharacterWriter& characterWriter):
    writer_(&(characterWriter.copy())) {}

void RJXWriter::update(const DOMString& destString) {
  DOMString outputString = destString;
  outputString += ".rjx";
  writer_ = new Writer(outputString);
}

Node& RJXWriter::process(Node& node) {
  int type = node.getNodeType();

  if (type == Node::ELEMENT_NODE) processElement(node);
  if (type == Node::END_NODE) processEndNode(node);
  if (type == Node::TEXT_NODE) processText(node);
  if (type == Node::CDATA_SECTION_NODE) processCDATASection(node);
  if (type == Node::XML_DECLARATION_NODE) processXMLDeclaration(node);
  if (type == Node::PROCESSING_INSTRUCTION_NODE) processProcessingInstruction(node);
  if (type == Node::DOCUMENT_TYPE_NODE) processDocumentType(node);
  if (type == Node::COMMENT_NODE) processComment(node);

  return node;
}

void RJXWriter::processElement(Node& node) {
  indent(node);
  
  DOMString nodeName = node.getNodeName();
  
  bool hasChildren = node.getChildNodes().getLength();
  bool hasAttributes = node.getAttributes().getLength();

  if (!hasChildren && !hasAttributes) {  // an empty element
    *writer_  << nodeName << ";" << wrt::endl;
    return;
  }

  bool hasNonTextChildren = node.hasNonTextChildren();
  if (!hasNonTextChildren && !hasAttributes) { // only text content
    DOMString valueString = node.getRawValue();
    *writer_ << nodeName << "=" << escapeString(valueString) << wrt::endl;
    return;
  }

  // we fall through to here if we have attributes &/or non-text children
  *writer_ << "%" << nodeName << wrt::endl;
  nest();
  processAttributes(node);

  if (!hasNonTextChildren) { // no end node follows, so fake one
    EndNode endNode(nodeName);
    denest();
    indent(endNode);
    *writer_ << "%% " << nodeName << wrt::endl;
  }
}

void RJXWriter::processEndNode(Node& node) {
  if (!static_cast<EndNode &>(node).getStartNode().hasNonTextChildren()) return;
  
  denest();
  indent(node);
  *writer_ << "%% " << node.getNodeName() << wrt::endl;
}

void RJXWriter::processText(Node& node) {
  if (!node.getParentNode().hasNonTextChildren())
    return;

  indent(node);
  DOMString valueString = node.getRawValue();
  *writer_<< "#text=" << escapeString(valueString) << wrt::endl;
}

void RJXWriter::processCDATASection(Node& node) {
  indent(node);
  *writer_ << "#cdata-section=";
  *writer_ << node.getNodeValue() << wrt::endl;
}

void RJXWriter::processXMLDeclaration(Node& node) {
  indent(node);
  *writer_ << "#xml-declaration=";

  XMLDeclaration& xmlDeclaration = static_cast<XMLDeclaration &>(node);
  *writer_<< xmlDeclaration.getVersion() << "," << xmlDeclaration.getEncoding()\
    << "," << xmlDeclaration.getStandAlone() << wrt::endl;
}

void RJXWriter::processProcessingInstruction(Node& node) {
  indent(node);
  *writer_ << "#processing-instruction=";

  ProcessingInstruction& instruction = static_cast<ProcessingInstruction &>(node);
  DOMString dataString = instruction.getData();
  *writer_ << instruction.getTarget() << "," << escapeString(dataString) << wrt::endl;
}

void RJXWriter::processDocumentType(Node& node) {
  indent(node);
  *writer_ << "#document-type=";

  DocumentType& doctype = static_cast<DocumentType &>(node);
  DOMString internalSubset = doctype.getInternalSubset();
  escapeString(internalSubset);
  
  *writer_ << doctype.getName() << "," << doctype.getPublicId() << ","\
  << doctype.getSystemId() << "," << internalSubset << wrt::endl;
}

void RJXWriter::processComment(Node& node) {
  indent(node);
  DOMString valueString = node.getRawValue();
  *writer_ << "#comment=";
  *writer_ << escapeString(valueString) << wrt::endl;
}

void RJXWriter::processAttributes(Node& node) {
  AttributeStream stream(node);

  while (true) {
    Node& attrNode = stream.nextNode();
    if (!attrNode.isValid()) break;

    indent(node);
    *writer_ << "@" << attrNode.getNodeName() \
    << "=" << attrNode.getRawValue() << wrt::endl;
  }
}

DOMString& RJXWriter::escapeString(DOMString& valueString) {
  static DOMString endString("\n");
  static DOMString slashString("\\");

/* The idea here is to escape all INTERNAL newline chars in valueString.
   If it is in the sequence "\r\n", the '\r' char is overwritten, otherwise
   the backslash is inserted before the newline. In either case, if the newline
   is at the end of the string, it is NOT escaped.
*/
  size_t index = 0;
  while (true) {
    index = valueString.indexOf(endString, index);
    if (index == valueString.npos ||
      index == valueString.getLength() - 1) break;

    if (index > 0 && valueString[index - 1] == '\r')
      valueString[index -1] = '\\';
    else
      valueString.insertString(index, slashString);

    index += 2; // step over the "\n"
  }

  return valueString;
}

void RJXWriter::indent(Node& node) {
  /* introduce some whitespace to make the output more intelligible */
  if (depth_ == 1 &&
      node.getNodeType() == Node::ELEMENT_NODE) {
    *writer_ << wrt::endl;
    return;
  }

  if (depth_ == 0 && node.getNodeType() == Node::END_NODE) {
    *writer_ << wrt::endl;
    return;
  }

  for (int count = depth_ - 1; count > 0; --count)
    *writer_ << "  ";
}

}
