/***************************************************************************
                          schemabuilder.cpp  -  description
                             -------------------
    begin                : Mon 14 April 2008
    copyright            : (C) 2008 by John Redmond
    email                : john_redmond@optusnet.com.au
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "schemabuilder.h"
#include "namednodemap.h"
#include "nodestream.h"
#include "liststream.h"
#include "siblingstream.h"
#include "reader.h"
#include "writer.h"
#include "stringwriter.h"
#include "stringreader.h"

namespace limpid {

SchemaBuilder:: SchemaBuilder(): convert_(false) {}

SchemaBuilder:: SchemaBuilder(const DOMString& pathString): outputPath_(pathString), convert_(pathString.empty()) {
  if (!outputPath_.trim().empty() && outputPath_ != "true")
    writer_ = Writer(outputPath_); 
}

Document& SchemaBuilder::processDocument(Document& inputDocument) {
  mapElements(inputDocument.getDocumentElement());
  buildDTD();

  if (!convert_) {
    writer_ << stringWriter_.getString();
    return inputDocument;
  }

  if (false && outputPath_.empty())
    wrt::Console << stringWriter_.getString() << wrt::endl;

  StringReader stringReader(stringWriter_.getString());
  dtdConverter_ = new DTDConverter(stringReader);

  dtdConverter_->process();
  return dtdConverter_->getDocument();
}

void SchemaBuilder::mapElements(Element& rootElement) {
  DOMString rootName = rootElement.getNodeName();
  NodeStream nodeStream(rootElement, false, true);

  while (true) {
    Node& node = nodeStream.nextNode();
    if (!node.isValid()) break;

    if (node.getNodeType() != Node::ELEMENT_NODE) continue;
    Element& element = static_cast<Element &>(node);

    DOMString elementName = element.getNodeName();
    addToMaps(elementName);
    setChildren(elementName, element);
    setAttributes(elementName, element);
  }
}

void SchemaBuilder::setChildren(const DOMString& elementName, Element& element) {
  ListStream listStream(element);
  ChildDetails& childDetails = masterMap_[elementName];

  while (true) {
    Node& node = listStream.nextNode();
    if (!node.isValid()) break;

    DOMString childName = node.getNodeName();
    if (childName == "#text")
      childName = "#PCDATA";

    if (childDetails.childSet.find(childName) == childDetails.childSet.end()) {
      childDetails.childSet.insert(childName);
      childDetails.childList.push_back(childName);
    }
    setChild (elementName, node);
  }
}

void SchemaBuilder::setAttributes(const DOMString& elementName, Element& element) {
  NamedNodeMap& attributes = element.getAttributes();
  if (!attributes.getLength()) return;

  for (int index = 0; index < attributes.getLength(); ++index) {
    Attr& attribute = static_cast<Attr &>(attributes.item(index));
    DOMString attributeName = attribute.getNodeName();

    StringSet& attributeSet = attributeMap_[elementName];
    if (attributeSet.find(elementName) == attributeSet.end())
      attributeMap_[elementName].insert(attributeName);
  }
}

void SchemaBuilder::setChild(const DOMString& elementName, Node& childNode) {
  DOMString childName = childNode.getNodeName();
  DOMString mapName = getMapName(elementName, childNode.getNodeName());

  if (siblingMap_.find(mapName) == siblingMap_.end()) {
    StringSet elementSet;
    siblingMap_[mapName] = elementSet;
    Limits limits;
    limits.first = INT_MAX;
    limits.second = 0;
    countMap_[mapName] = limits;
  }

  StringSet& siblingSet = siblingMap_[mapName];

  Node& previousSibling = childNode.getPreviousSibling();
  if (previousSibling.isValid()) {  // update for this sibling
    DOMString siblingName = previousSibling.getNodeName();
    if (siblingName != childName && siblingSet.find(siblingName) == siblingSet.end())
      siblingSet.insert(siblingName);
  }

  Limits& limits = countMap_[mapName]; // update the multiple status for the child

  if (!previousSibling.isValid() || previousSibling.getNodeName() != childName) {
    int blockLength = countSame(childNode);
    if (blockLength < limits.first) limits.first = blockLength;
    if (blockLength > limits.second) limits.second = blockLength;
  }
}

void SchemaBuilder::buildDTD() {
  for (int index = 0; index < masterList_.size(); ++index) {
    const DOMString& elementName = masterList_[index];

    ChildDetails& childDetails = masterMap_[elementName];
    StringList& childList = childDetails.childList;

    buildElementDTD(elementName, childList);
  }

  for (int index = 0; index < masterList_.size(); ++index) {
    const DOMString& elementName = masterList_[index];

    ChildDetails& childDetails = masterMap_[elementName];
    StringList& childList = childDetails.childList;

    buildAttributeDTD(elementName);
  }
}

void SchemaBuilder::buildAttributeDTD(const DOMString& parentName) {
  StringSet& attributeSet = attributeMap_[parentName];
  if (attributeSet.empty()) return;

  stringWriter_ << "<!ATTLIST " << parentName << " ";

  StringSet::iterator p = attributeSet.begin();
  while (p != attributeSet.end()) {
    const DOMString& attributeName = *p;
    stringWriter_ << attributeName << " CDATA ";
    ++p;
  }

  stringWriter_ << ">\n";
}

void SchemaBuilder::buildElementDTD(const DOMString& parentName, StringList& childList) {
  StringList hitList;
  StringSet usedSet;
  int childCount = childList.size();

// Are there no childNodes?
  if (childList.empty()) {  // we must have at least one first element if there is content
    stringWriter_ << "<!ELEMENT " << parentName << " EMPTY>\n";
    return;
  }
// First dispose of an element with text content. If it has other child nodes, we have
// mixed content, which we express as a choice
  for (int index = 0; index < childList.size(); ++index) {
    DOMString elementName = childList[index];
    if (elementName == "#PCDATA") {
      if (childList.size() == 1)
        stringWriter_ << "<!ELEMENT " << parentName << " (#PCDATA)>\n";
      else {
        stringWriter_ << "<!ELEMENT " << parentName << " ";
        buildDTDChoice(parentName, childList);
        stringWriter_ << ">\n";
      }
      return;
    }
  }

  // First collect the elements that have no previous sibling (and are first in the specification)
  for (int index = 0; index < childList.size(); ++index) {
    DOMString elementName = childList[index];

    if (elementName != "#PCDATA") {
      DOMString mapName = getMapName(parentName, elementName);

      StringSet& siblingSet = siblingMap_[mapName];
      if (!siblingSet.empty()) continue; // we are looking for an element with no previous sibling
    }

    hitList.push_back(elementName);
    usedSet.insert(elementName);  // mark as used
  }

  if (hitList.empty()) { // there is no specified first child, hence we have all children in a choice
    if (childList.empty())
      stringWriter_ << "<!ELEMENT " << parentName << childList.size() << " ***EMPTY>\n";
    else {
      stringWriter_ << "<!ELEMENT " << parentName << " ";
      buildDTDChoice(parentName, childList);
      stringWriter_ << ")>\n";
    }
    return;
  }

  stringWriter_ << "<!ELEMENT " << parentName << " ";
  if (childCount == 1 || hitList.size() < childCount)
    stringWriter_ << "(";

// do we have only one type of first element?
  if (hitList.size() == 1) {
    DOMString mapName = getMapName(parentName, hitList[0]);

    Limits& limits = countMap_[mapName];
    stringWriter_ << hitList[0];
    if (limits.second > 1) stringWriter_ << "+ ";

    if (childList.size() > 1)
      stringWriter_ << ", ";
  }
// if we have more than one first element, we express them as a choice
  else {
    buildDTDChoice(parentName, hitList); // report a choice of elements
    if (hitList.size() == childCount) return; // nothing more to do: all elements dealt with
  }

// If we fall through to here, we look through the remaining elements that follow one of
// the previous siblings. We repeat this until we have an empty hitList.
  while (true) {
    StringList siblingList = hitList;  // copy the previous matches
    hitList.clear();  // ready for a new collection
    collectChoices(parentName, siblingList, childList, hitList, usedSet);
    if (hitList.empty()) break;

    if (hitList.size() == 1) {
      DOMString elementName = hitList[0];

      DOMString mapName = getMapName(parentName, elementName);
      Limits& limits = countMap_[mapName];

      stringWriter_ << elementName;
      if (limits.second > 1) stringWriter_ << "+";
    }
    else
      buildDTDChoice(parentName, hitList);

    if (usedSet.size() < childCount)
      stringWriter_ << ", ";
  }

  stringWriter_ << ")>\n";
}

void SchemaBuilder::collectChoices(const DOMString& parentName, StringList& siblingList, StringList& sourceList, StringList& resultList, StringSet& usedSet) {
  StringList hitList;
  for (int index = 0; index < siblingList.size(); ++index) {
    DOMString& siblingName = siblingList[index];
    for (int sourceIndex = 0; sourceIndex < sourceList.size(); ++sourceIndex) {
      DOMString& sourceName = sourceList[sourceIndex];

      if (usedSet.find(sourceName) != usedSet.end()) continue;
      DOMString mapName = getMapName(parentName, sourceName);

      StringSet& siblingSet = siblingMap_[mapName];
      if (siblingSet.find(siblingName) == siblingSet.end()) continue;

      resultList.push_back(sourceName);
      usedSet.insert(sourceName);
    }
  }
}

void SchemaBuilder::buildDTDChoice(const DOMString& parentName, StringList& hitList) {
  stringWriter_ << "(";
  for (int index = 0; index < hitList.size() - 1; ++index) {
    DOMString elementName = hitList[index];

    DOMString mapName = getMapName(parentName, elementName);
    Limits& limits = countMap_[mapName];

    stringWriter_ << elementName;
    if (limits.second > 1) stringWriter_ << "+";
    stringWriter_  << " | ";
  }
  stringWriter_ << hitList[hitList.size() - 1] << ")+";
}

void SchemaBuilder::addToMaps(const DOMString& elementName) {
  if (masterMap_.find(elementName) == masterMap_.end()) {
    ChildDetails childDetails;
    masterMap_[elementName] = childDetails;
    masterList_.push_back(elementName);

    StringSet attributeSet;
    attributeMap_[elementName] = attributeSet;
  }
}

int SchemaBuilder::countSame(Node& node) {
  DOMString nodeName = node.getNodeName();
  SiblingStream siblingStream(node);
  int count = 1;

  while (true) {
    Node& nextNode = siblingStream.nextNode();
    if (!nextNode.isValid()) break;
    if (nextNode.getNodeName() != nodeName) break;
    ++count;
  }
  return count;
}

DOMString SchemaBuilder::getMapName(const DOMString& parentName, const DOMString& childName) {
  DOMString mapName = parentName;
  return mapName.append(DOMString("/")).append(childName);
}

int SchemaBuilder::getMaxCount(const DOMString& mapString) {
  Limits& limits = countMap_[mapString];
  return limits.second;
}

}
