/***************************************************************************
 *   Copyright (C) 2007 by John Redmond   *
 *   john_redmond@optusnet.com.au   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "schemainspector.h"
#include "nodestream.h"
#include "liststream.h"
#include "ancestorstream.h"
#include "comment.h"
#include "domwriter.h"

namespace limpid {

Document& SchemaInspector::processDocument(Document& sourceDocument) {
  document_ = Document();
  Element resultElement("ELEMENTS");
  document_.appendChild(resultElement);

  Element& sourceElement = sourceDocument.getDocumentElement();

  // set up reference map and list
  registerElements(sourceElement);
  expandElements(resultElement);

  return document_;
}

void SchemaInspector::registerElements(Element& rootElement) {
  NodeStream nodeStream(rootElement, false, true);

  while (true) {
    Node& node = nodeStream.nextNode();
    if (!node.isValid()) break;

    if (node.getNodeType() != Node::ELEMENT_NODE) continue;
    Element& element = static_cast<Element &>(node);
    DOMString localName = element.getLocalName();
    if (localName == "complexType")
      registerComplexType(element);
    else if (localName == "element")
      registerInstance(element);
  }
}

void SchemaInspector::registerComplexType(Element& complexElement) {
  DOMString complexName = complexElement.getAttribute("name");
  if (complexName.empty()) return;

  complexMap_[complexName] = &complexElement;
  complexList_.push_back(complexName);

  if (complexElement.getAttribute("mixed") == true)
    mixedContent_.insert(complexName);
}

void SchemaInspector::registerInstance(Element& instanceElement) {
  if (instanceElement.getParentNode().getNodeName().endsWith(":schema")) {
    topList_.push_back(&instanceElement);
  }
}

void SchemaInspector::expandElements(Element& resultElement) {
  Element holder("HOLDER");
  ElementList::iterator p = topList_.begin();
  while (p != topList_.end()) {
    Element& topElement = **p++;
    updateContent(topElement);
    while (expandElement(topElement, false)) {}
    setPathDetails(topElement, resultElement);
  }
}

bool SchemaInspector::expandElement(Element& parentElement, bool includeRoot) {
  NodeStream listStream(parentElement, false, includeRoot);
  NodeList hitList;

  while (true) {
    Node& node = listStream.nextNode();
    if (!node.isValid()) break;

    if (node.getNodeType() != Node::ELEMENT_NODE) continue;
    Element& element = static_cast<Element &>(node);

    if (node.getLocalName() != "element") continue;

    DOMString typeString = element.getAttribute("type");
    DOMString nameString = element.getAttribute("name");

    if (typeString.endsWith(":string") || typeString == "expanded" || typeString == "ignored") continue; // a simple type

    if (nameString == "cdata-section") {
      element.getParentNode().removeChild(element);
      return true;
    }

    if (isRecursion(element)) {
      element.setAttribute("type", "ignored");
      continue;
    }

    if (nameString != "cdata-section" && complexMap_.find(typeString) != complexMap_.end()) {
      hitList.addNode(element);
    }
    else  {
      element.setAttribute("type", "ignored");
    }
  }

  if (!hitList.getLength())
    return false;

  ListStream hitStream(hitList);

  while (true) {
    Node& node = hitStream.nextNode();
    if (!node.isValid()) break;

    Element& element = static_cast<Element &>(node);
    updateContent(element);
  }

  return true;
}

void SchemaInspector::setPathDetails(Element& rootElement, Element& resultElement) {
  if (topList_.empty()) return;

  NodeStream nodeStream(rootElement, false, true);
  while (true) {
    Node& node = nodeStream.nextNode();
    if (!node.isValid()) break;

    if (node.getNodeType() != Node::ELEMENT_NODE) continue;
    Element& element = static_cast<Element &>(node);

    if (element.getLocalName() != "element") continue;

    StringList pathList;
    getPathList(element, pathList);
    int pathLength = pathList.size();

    DOMString pathString = getPathString(pathList);
    DOMString typeString = getType(element);
    typeString += Converter::toString(pathLength -1);

    appendEntry(pathString, typeString, resultElement);
  }
}

void SchemaInspector::getPathList(Element& leafElement, StringList& resultList) {
  AncestorStream stream(leafElement, true);

  while (true) {
    Node& node = stream.nextNode();
    if (!node.isValid()) break;

    if (node.getNodeType() != Node::ELEMENT_NODE) continue;
    Element& element = static_cast<Element &>(node);
    if (element.getLocalName() != "element") continue;

    DOMString elementName = element.getAttribute("name");
    resultList.push_front(elementName);
  }
}

DOMString SchemaInspector::getPathString(StringList& pathList) {
  int fullPathLength = pathList.size();
  while (pathList.size() > pathLength_)
    pathList.pop_front();

  DOMString pathString;

  StringList::iterator p = pathList.begin();
    if (fullPathLength <= pathLength_)
      pathString = "/";
    pathString += *p++;

  while (p != pathList.end()) {
    pathString += "/";
    pathString += *p++;
  }

  return pathString;
}

void SchemaInspector::appendEntry(const DOMString& pathString, const DOMString& typeString, Element& parentElement) {
  if (outputPaths_.find(pathString) != outputPaths_.end())
    return; // we have already processed this path

  outputPaths_.insert(pathString); // register this 

  DOMString typeCopy = typeString;
  if (typeString.empty()) return;

  Element pathElement("ELEMENT");
  pathElement.setAttribute("path", pathString);
  pathElement.setAttribute("type", typeCopy);

  parentElement.appendChild(pathElement);
}

DOMString SchemaInspector::getType(Element& element) {
  static DOMString reserveTags = "table th tr td ul ol li";

  DOMString elementName = element.getAttribute("name");

  if (reserveTags.contains(elementName)) // this is a reserved tag: don't translate
    return elementName;

  if (element.getAttribute("mixed") == "true")
    return "M"; //mixed

  DOMString typeName = element.getAttribute("type");
  if (typeName == "expanded" || typeName == "ignored")
    return "L";

  if (isEmpty(element))
    typeName =  "E"; // empty

  if (inSequence(element)) {
    if (isFirst(element)&& isSingle(element)) return  "T"; // title
    if (isLast(element)&& isSingle(element)) typeName = "F"; // tail label
    if (isSingle(element)) return "F"; // feature or tail label
    if (isFirst(element) && DOMUtils::getSiblingCount(element) == 1 &&
      !isSingle(static_cast<Element &>(element.getNextSibling()))) return "H"; // multiple feature at top
  }

  if (!inChoice(element) && isSingle(element)) return "F"; // feature

  if (isParentMixed(element))
      return  "m";

  // no need to look after the stragglers; "D" is the default type
  return "D";
}

bool SchemaInspector::isEmpty(Element& element) {
  NodeStream nodeStream(element, false, false);

  while (true) {
    Node& node = nodeStream.nextNode();
    if (!node.isValid()) break;

    if (node.getNodeType() != Node::ELEMENT_NODE) continue;
    DOMString localName = static_cast<Element &>(node).getLocalName();

    if (localName == "complexType" || localName == "element") return false;
  }

  return true;
}

bool SchemaInspector::isRecursion(Element& element) {
  DOMString elementName = element.getAttribute("name");
  AncestorStream stream(element, false);

  while (true) {
    Node& node = stream.nextNode();
    if (!node.isValid()) break;

    if (node.getNodeType() != Node::ELEMENT_NODE) continue;
    if (static_cast<Element &>(node).getAttribute("name") == elementName)
      return true;
  }

  return false;
}

bool SchemaInspector::isParentMixed(Element& element) {
  AncestorStream stream(element, false);

  while (true) {
    Node& node = stream.nextNode();
    if (!node.isValid()) break;

    if (node.getNodeType() != Node::ELEMENT_NODE) continue;
    Element& element = static_cast<Element &>(node);
    if (element.getLocalName() != "element") continue;

    return (element.getAttribute("mixed")) == "true";
  }

  return false;
}


void SchemaInspector::updateContent(Element& element) {
  DOMString typeString = element.getAttribute("type");
  Element complexElement = static_cast<Element &>(complexMap_[typeString]->cloneNode(true));
  moveContent(complexElement, element);

  element.setAttribute("type", "expanded");
  if (complexElement.getAttribute("mixed") == "true")
      element.setAttribute("mixed", "true");
}

void SchemaInspector::clearStructures() {
  complexMap_.clear();
  complexList_.clear();
  mixedContent_.clear();
  outputPaths_.clear();
  topList_.clear();
}

}
