/***************************************************************************
 *   Copyright (C) 2007 by John Redmond   *
 *   john_redmond@optusnet.com.au   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "schemasummary.h"
#include "document.h"
#include "writer.h"
#include "nodestream.h"
#include "liststream.h"
#include "ancestorstream.h"
#include "siblingstream.h"
#include "text.h"
#include "comment.h"
#include "converter.h"

namespace limpid {

const DOMString& SchemaSummary::getPath(size_t index) {
  if (index < stringList_.size()) return stringList_[index];
  else return DOMString::failString;
}

NodeSummary& SchemaSummary::getSummary(const DOMString& path) {
 SummaryMap::iterator p = summaryMap_.find(path);
 return p->second;
}

NodeSummary& SchemaSummary::getSummary(size_t index) {
 SummaryMap::iterator p = summaryMap_.find(getPath(index));
 return p->second;
}

bool SchemaSummary::hasSummary(const DOMString& pathString) {
  SummaryMap::iterator p = summaryMap_.find(pathString);
  return (p != summaryMap_.end());
}

void SchemaSummary::setup(Node& rootNode) {
  NodeStream stream(rootNode);
  while (true) {
    Node& node = stream.nextNode();
    if (!node.isValid()) break;

    if (node.getNodeType() != Node::ELEMENT_NODE)
      continue;

    if (node.getLocalName() == "element")
      setupSummary(static_cast<Element &>(node));
  }
}

void SchemaSummary::setupSummary(Element& element) {
  DOMString matchString;
  Element& parentNode = static_cast<Element &> (getComplexParent(element));

  if (parentNode.isValid())
    matchString += parentNode.getAttribute("name");

  matchString += "/";

  DOMString elementName = element.getAttribute("name");
  if (elementName.empty())
    elementName = element.getAttribute("type");

  matchString += elementName;

//wrt::Console << matchString << " SET UP " << element << wrt::endl;
  if (!haveSummary(matchString))
    enterSummary(element, elementName, matchString);
}

bool SchemaSummary::haveSummary(const DOMString& matchString) {
  SummaryMap::iterator p = summaryMap_.find(matchString);
  return (p != summaryMap_.end());
}

void SchemaSummary::updateSummary(Node& node, const DOMString& matchString) {
  if (true) return;

  SummaryMap::iterator p = summaryMap_.find(matchString);
  NodeSummary& nodeSummary = (p->second);

  nodeSummary.isLeaf = nodeSummary.isLeaf && isLeaf(node);
  nodeSummary.leafFollows = nodeSummary.leafFollows && (isLeaf(node.getNextSibling()) || hasText(node.getNextSibling()));

  nodeSummary.hasText = nodeSummary.hasText || hasText(node); // it may have mixed content
  nodeSummary.textBefore |= node.getPreviousSibling().getNodeType() == Node::TEXT_NODE;
  nodeSummary.textAfter |= node.getNextSibling().getNodeType() == Node::TEXT_NODE;

  nodeSummary.alwaysFirst &= isFirstSibling(node);
  bool isLast = !node.getNextSibling().isValid();
  nodeSummary.alwaysLast &= isLast;
  nodeSummary.alwaysSecond &= getPosition(node) == 1;
  nodeSummary.everLast |= isLast;

  ++nodeSummary.totalCount;
  int selfCount = countCluster(node);
  if (selfCount > nodeSummary.selfMax)
    nodeSummary.selfMax = selfCount;
  if (selfCount < nodeSummary.selfMin)
    nodeSummary.selfMin = selfCount;

  nodeSummary.isFeature &= (selfCount == 1); // this may not be the best assessment

  int descendantCount = countDescendants(node);//.getParentNode());
  if (descendantCount > nodeSummary.descendantMax)
    nodeSummary.descendantMax = descendantCount;

  if (!nodeSummary.isLeaf)
    addChildPaths(node, nodeSummary.childNames, nodeSummary.childPaths, nodeSummary.childSet);

  addSiblingNames(node, node.getPath(), nodeSummary.siblingBefore, nodeSummary.siblingAfter);
}

void SchemaSummary::enterSummary(Element& element, const DOMString& elementName, const DOMString& matchString) {
  stringList_.push_back(matchString); // add to path list in document order
  stringSet_.insert(matchString);

  NodeSummary nodeSummary;
  nodeSummary.nodeName = elementName;
  nodeSummary.matchString = matchString; //getMatchString(node);

  setOccurs(element, nodeSummary);

wrt::Console << matchString << " MIN " << nodeSummary.selfMin << " MAX " << nodeSummary.selfMax << wrt::endl;

  if (false) {

  nodeSummary.path = element.getPath();
  nodeSummary.parentPath = element.getParentNode().getPath();

  Node& previousSibling = element.getPreviousSibling();
  nodeSummary.depth = getDepth(element);
  nodeSummary.isLeaf = isLeaf(element);
  nodeSummary.leafFollows = isLeaf(element.getNextSibling()) || hasText(element.getNextSibling());
  nodeSummary.isFeature = !hasText(element.getPreviousSibling()) ||  !hasText(element.getNextSibling());
  nodeSummary.textBefore = previousSibling.getNodeType() == Node::TEXT_NODE;
  nodeSummary.textAfter = element.getNextSibling().getNodeType() == Node::TEXT_NODE;
  nodeSummary.hasText = hasText(element); // it may have mixed content

  nodeSummary.alwaysFirst = isFirstSibling(element);
  nodeSummary.alwaysSecond = getPosition(element) == 1;
  nodeSummary.alwaysLast = nodeSummary.everLast = !element.getNextSibling().isValid();

  nodeSummary.totalCount = 1;
  int selfCount = countCluster(element);
  nodeSummary.selfMax = selfCount;
  nodeSummary.selfMin = selfCount;
  nodeSummary.isFeature = selfCount == 1;

  int descendantCount = countDescendants(element);
  nodeSummary.descendantMax = descendantCount;

  nodeSummary.depth = getDepth(element);
  if (!nodeSummary.isLeaf)
    addChildPaths(element, nodeSummary.childNames, nodeSummary.childPaths, nodeSummary.childSet);


  addSiblingNames(element, element.getPath(), nodeSummary.siblingBefore, nodeSummary.siblingAfter);
  }
  summaryMap_.insert(std::pair<DOMString, NodeSummary>(matchString, nodeSummary));
}

void SchemaSummary::buildDocument() {
  Element summaryElement("SUMMARY");
  document_.appendChild(summaryElement);

  Element definesElement("DEFINITIONS");
  summaryElement.appendChild(definesElement);
  Comment comment("Insert any new style definitions here");
  definesElement.appendChild(comment);

  Element rootElement("ELEMENTS");
  summaryElement.appendChild(rootElement);

  for (StringList::iterator p = stringList_.begin(); p != stringList_.end(); p++) {
    NodeSummary& nodeSummary = getSummary(*p);

    Element summaryElement("ELEMENT");
    rootElement.appendChild(summaryElement);

    addElementContent(nodeSummary, summaryElement);
  }
}

void SchemaSummary::addElementContent(NodeSummary& nodeSummary, Element& summaryElement) {
  addLeaf("NAME", nodeSummary.nodeName, summaryElement);
  addLeaf("MATCH", nodeSummary.matchString, summaryElement);
  addLeaf("DEPTH", Converter::toString(nodeSummary.depth), summaryElement);
  addLeaf("TAG", nodeSummary.tag, summaryElement);
  addLeaf("PATH", nodeSummary.path, summaryElement);
  addLeaf("TYPE", nodeSummary.type, summaryElement);
  addLeaf("STYLE", nodeSummary.style, summaryElement);

  addBoolean("TEXT", nodeSummary.hasText, summaryElement);
  addBoolean("WITH_TEXT", (nodeSummary.textBefore || nodeSummary.textAfter), summaryElement);

///  addChildren(nodeSummary, summaryElement);

  Element countsElement("COUNTS");
//  summaryElement.appendChild(countsElement);
  addCount("TOTAL_COUNT", nodeSummary.totalCount, summaryElement);//countsElement);

  if (true) return;

  // this extended information is not presently needed
  addCount("SELF_MAX", nodeSummary.selfMax, countsElement);
  addCount("SELF_MIN", nodeSummary.selfMin, countsElement);
  addCount("DEPTH", nodeSummary.depth, countsElement);

  Element flagsElement("FLAGS");
  summaryElement.appendChild(flagsElement);
  addBoolean("SECTION", nodeSummary.isSection, flagsElement);
  addBoolean("IS_LEAF", nodeSummary.isLeaf, flagsElement);
  addBoolean("HAS_TEXT", nodeSummary.hasText, flagsElement);
  addBoolean("TEXT_BEFORE", nodeSummary.textBefore, flagsElement);
  addBoolean("TEXT_AFTER", nodeSummary.textAfter, flagsElement);
  addBoolean("ALWAYS_FIRST", nodeSummary.alwaysFirst, flagsElement);
  addBoolean("ALWAYS_LAST", nodeSummary.alwaysLast, flagsElement);
  addBoolean("EVER_LAST", nodeSummary.everLast, flagsElement);
}

void SchemaSummary::addChildren(NodeSummary& nodeSummary, Element& summaryElement) {
  if (nodeSummary.childNames.empty()) return;

  Element childrenElement("CHILDREN");
  summaryElement.appendChild(childrenElement);

  for (StringList::iterator p = nodeSummary.childNames.begin(); p != nodeSummary.childNames.end(); p++) {
    Element childElement("CHILD");
    Text text(*p);
    childElement.appendChild(text);
    childrenElement.appendChild(childElement);
  }
}

void SchemaSummary::addBoolean(const DOMString& name, bool state, Element& summaryElement) {
 DOMString booleanString(state ? "true" : "false");
 Element stateElement(name);
 summaryElement.appendChild(stateElement);
 Text text(booleanString);
 stateElement.appendChild(text); 
}

void SchemaSummary::addCount(const DOMString& name, int number, Element& summaryElement) {
 Element stateElement(name);
 summaryElement.appendChild(stateElement);
 Text text(Converter::toString(number));
 stateElement.appendChild(text); 
}

bool SchemaSummary::isLeaf(const Node& refNode) {
  return refNode.getChildNodes().getLength() == 1 &&
    refNode.getFirstChild().getNodeType() == Node::TEXT_NODE;
}

bool SchemaSummary::hasText(const Node& refNode) {
  ListStream stream(const_cast<Node&>(refNode));

  while (true) {
    Node& node = stream.nextNode();
    if (!node.isValid()) break;

    if (node.getNodeType() == Node::TEXT_NODE)
      return true;
  }

  return false;
}

bool SchemaSummary::isFirstSibling(Node& node) {
  DOMString refName = node.getNodeName();
  SiblingStream stream(node, false);

  while (true) {
    Node& node = stream.nextNode();
    if (!node.isValid()) break;

    if (node.getNodeName() != refName) return false;
  }

  return true;
}

int SchemaSummary::countCluster(Node& refNode) {
  DOMString refName = refNode.getNodeName();
  SiblingStream stream(refNode, false);
  int count = 1;

  while (true) {
    Node& node = stream.nextNode();
    if (!node.isValid()) break;

    if (node.getNodeType() != Node::ELEMENT_NODE)
      break;

    if (node.getNodeName() != refName)
      break;

    ++count; // bump counter
  }

 return count;
}

int SchemaSummary::countLeaves(Node& refNode) {
  ListStream stream(const_cast<Node&>(refNode));
  int nodeCount = 0;

  while (true) {
    Node& node = stream.nextNode();
    if (!node.isValid()) break;

    if (isLeaf(node))
      ++nodeCount;
  }

  return nodeCount;
}

int SchemaSummary::countDescendants(Node& refNode) {
  NodeStream stream(const_cast<Node&>(refNode));
  int nodeCount = 0;

  while (true) {
    Node& node = stream.nextNode();
    if (!node.isValid()) break;

    if (isLeaf(node))
      ++nodeCount;
  }

  return nodeCount;
}

int SchemaSummary::countSiblings(Node& refNode) {
  return refNode.getParentNode().getChildNodes().getLength();
}

void SchemaSummary::addChildPaths(Node& parentNode, StringList& childNames, StringList& childPaths, StringSet& childSet) {
  ListStream stream(parentNode);
  while (true) {
    Node& node = stream.nextNode();
    if (!node.isValid()) break;

    if (node.getNodeType() != Node::ELEMENT_NODE) continue;

    DOMString nodePath = node.getPath();//NodeName();
    if (childSet.find(nodePath) != childSet.end()) continue;  // already registered
    childNames.push_back(node.getNodeName());
    childPaths.push_back(nodePath);
    childSet.insert(nodePath);
  }
}

void SchemaSummary::addSiblingNames(Node& node, const DOMString& nodePath, StringSet& siblingBefore, StringSet& siblingAfter) {
  if (node.getPreviousSibling().isValid()) {
    DOMString previousName = node.getPreviousSibling().getNodeName();
    siblingBefore.insert(previousName);
  }
  if (node.getNextSibling().isValid()) {
    DOMString nextName = node.getNextSibling().getNodeName();
    siblingAfter.insert(nextName);
  }
}

DOMString SchemaSummary::getMatchString(const Node& node) {
  DOMString matchString;
/*
// include this code if a three-step path match is required, otherwise use two steps
  Node& grandParentNode = node.getParentNode().getParentNode();
  if (grandParentNode.getNodeType() == Node::ELEMENT_NODE) {
    matchString += grandParentNode.getNodeName();
    matchString += "/";
  }
*/
  Node& parentNode = node.getParentNode();
  if (parentNode.getNodeType() == Node::ELEMENT_NODE)
    matchString += parentNode.getNodeName();

  matchString += "/";
  matchString += node.getNodeName();

  return matchString;
}

Node& SchemaSummary::getComplexParent(Node& node) {
  AncestorStream stream(node);
  while (true) {
    Node& node = stream.nextNode();
    if (!node.isValid()) break;

    if (node.getLocalName() == "complexType")
      return node;
  }

  return Node::failNode;
}

void SchemaSummary::setOccurs(Element& element, NodeSummary& nodeSummary) {
  DOMString minString = element.getAttribute("minOccurs");
  if (!minString.empty()) nodeSummary.selfMin = Converter::toInt(minString);
  else {
    Element& parentElement = static_cast<Element &>(element.getParentNode());
    DOMString parentName = parentElement.getLocalName();
    DOMString minString = parentElement.getAttribute("minOccurs");

    if (!minString.empty()) nodeSummary.selfMin = Converter::toInt(minString);
    else nodeSummary.selfMin = 1;
  }

  DOMString maxString = element.getAttribute("maxOccurs");
  if (!maxString.empty()) {
    if (maxString == "unbounded") nodeSummary.selfMax = 2000000000;
    nodeSummary.selfMax = Converter::toInt(maxString);
  }
  else {
    Element& parentElement = static_cast<Element &>(element.getParentNode());
    DOMString parentName = parentElement.getLocalName();
    if (parentName != "choice") {
      nodeSummary.selfMax = 1;
      return;
    }

    DOMString maxString = parentElement.getAttribute("maxOccurs");
    if (maxString.empty()) {
      nodeSummary.selfMax = 1;
      return;
    }

    if (maxString == "unbounded") nodeSummary.selfMax = 2000000000;
    else nodeSummary.selfMax = Converter::toInt(maxString);
  }
}

}

