/***************************************************************************
                          selectionmanager.cpp  -  description
                             -------------------
    begin                : Sat Sep 17 2005
    copyright            : (C) 2005 by John Redmond
    email                : john_redmond@optusnet.com.au
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "selectionmanager.h"
#include "templatestream.h"

namespace limpid {

SelectionManager::SelectionManager(bool isImported): 
  cachedTemplates_(new CacheMap), isImported_(isImported) {}

SelectionManager::SelectionManager(bool isImported, const DOMString& modeString):
  cachedTemplates_(new CacheMap), isImported_(isImported), modeString_(modeString) {}

Template& SelectionManager::getCalledTemplate(const DOMString& calledName) {
  Template& calledTemplate = resources_.getCalledTemplate(calledName, isImported_);
  // if isImported_ we have only one chance
  if (calledTemplate.isValid() || isImported_) return calledTemplate;

  // if not isImported_, we are permitted to try for an imported match
  return resources_.getCalledTemplate(calledName, true);
}

Template& SelectionManager::getMatchedTemplate(Node& node) {
  if (!isCached(node.getNodeName()))
    cachePossibleTemplates(node);

  return getCachedTemplate(node);
}

void SelectionManager::setMatchedTemplate(const DOMString& key, Template& tmp, bool isImported) {
 resources_.setMatchedTemplate(key, tmp, isImported);
}

void SelectionManager::cachePossibleTemplates(Node& node) {
  // a CacheMap is easier to use than a multimap because it is easier to
  // work with a Template list to select templates
  TemplateList matchedList;
  cachedTemplates_->insert(make_pair(node.getNodeName(), matchedList));
  if (!isImported_)  // get templates from the top level
    collectTemplates(node, matchedList, false);

  collectTemplates(node, matchedList, true); // get imported templates

  // we need a stable sort here to ensure that imported templates are always
  // listed after their top-level equivalents, in the event that we have both
  matchedList.doSort(); // highest priority will be found first
}

// here we collect all templates that MAY match the node, regardless of mode
inline void SelectionManager::collectTemplates(Node& node, TemplateList& templateList,
    bool isImported) {
  TemplateStream stream(resources_, isImported);
  while (true) {
    Template& tmpl = stream.nextTemplate();
    if (!tmpl.isValid()) break;
    if (node.getNodeType() == Node::DOCUMENT_NODE && tmpl.matchNode(node, false))
      templateList.addTemplate(tmpl);

    if (tmpl.getMode() != modeString_) continue;  // mode mismatch

    if (tmpl.matchNode(node, false)) // false = loose match for target node alone
      templateList.addTemplate(tmpl);
  }
}

// get the template with the best match and with the correct mode
Template& SelectionManager::getCachedTemplate(Node& node) {
  CacheMap::iterator p = cachedTemplates_->find(node.getNodeName());
  if (p == cachedTemplates_->end()) return Template::failTemplate; // never been cached

  TemplateList& templateList = p->second;
  size_t mapLength = templateList.getLength();

  for (size_t index = 0; index < mapLength; ++index) {
    Template& item = templateList.getItem(index);

    if (node.getNodeType() != Node::DOCUMENT_NODE &&
        item.getMode() != modeString_) continue;

    if (item.matchNode(node, true))
      return item;  // passes a full match, including predicates
  }

  return Template::failTemplate; // no matching template found
}

}
