/***************************************************************************
                          stringtester.cpp  -  description
                             -------------------
    begin                : Sat May 29 2004
    copyright            : (C) 2004 by John Redmond
    email                : john_redmond@optusnet.com.au
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "element.h"
#include "stringtester.h"
#include "boolean.h"
#include "xpathstring.h"
#include "number.h"
#include "ancestorstream.h"
#include "writer.h"

namespace limpid {

StringTester::StringTester(MatchType matchType, bool logic):
    ParentComponent(), logic_(logic), matchType_(matchType) {
}

StringTester::StringTester(const DOMString& typeString): ParentComponent() {
  setType(typeString);
}

Boolean StringTester::getBoolean(Context& context) {
  if (!components->size()) return false;

  if (matchType_ == LANG) return testLang(context);
  else if (components->size() == 1) return components->at(0)->getBoolean(context);
  else {
    XPathExpression& first = *components->at(0);
    XPathExpression& second = *components->at(1);
    return comparePair(first, second, context);
  }
}

bool StringTester::testLang(Context& context) {
  XPathString refString = components->at(0)->getString(context);

  AncestorStream stream(true);
  stream.setRoot(context.getNode());

  while (true) {
    Node& node = stream.nextNode();
    if (!node.isValid()) break;

    if (node.getNodeType() != Node::ELEMENT_NODE) continue;
    Element& element = static_cast<Element &>(node);

    XPathString langString = element.getAttribute("xml:lang");
    if (!langString.size()) continue;
    else return langString.startsWith(refString);
  }

  return false;  // no lang declaration found
}

bool StringTester::comparePair(XPathExpression& first, XPathExpression& second,
      Context& context) {
  if (first.isNodeSet()) {
    if (second.isNodeSet())
      return compareNodeSets(static_cast<CollectorComponent &>(first),
              static_cast<CollectorComponent &>(second), context);
    else
      return compareFirstNodeSet(static_cast<CollectorComponent &>(first), second,
              context);
  }
  else if (second.isNodeSet())
      return compareSecondNodeSet(first, static_cast<CollectorComponent &>(second),
              context);
  else {
    XPathString s0 = first.getString(context);
    XPathString s1 = second.getString(context);
    return compareStrings(s0, s1);
  }
}

bool StringTester::compareNodeSets(CollectorComponent& first,
      CollectorComponent& second, Context& context) {
  size_t length0 = first.getLength(context);
  size_t length1 = second.getLength(context);

  for (size_t index0 = 0; index0 < length0; ++index0) {
    for (size_t index1 = 0; index1 < length1; ++index1) {
      DOMString value0 = first.getItem(index0).getValue();
      DOMString value1 = second.getItem(index1).getValue();
      if (compareStrings(value0, value1)) return true;
    }
  }
  return false;
}

bool StringTester::compareFirstNodeSet(CollectorComponent& first,
      XPathExpression& second, Context& context) {

  NodeSet firstSet = first.getNodeSet(context);
  size_t length0 = firstSet.getLength();
  if (!length0) return false;

  DOMString value1 = second.getString(context);

  for (size_t index0 = 0; index0 < length0; ++index0) {
      DOMString value0 = firstSet.getItem(index0).getValue();
      if (compareStrings(value0, value1)) return true;
  }
  return false;
}

bool StringTester::compareSecondNodeSet(XPathExpression& first,
      CollectorComponent& second, Context& context) {

  NodeSet secondSet = second.getNodeSet(context);
  size_t length1 = secondSet.getLength();
  if (!length1) return false;

  DOMString value0 = first.getString(context);

  for (size_t index1 = 0; index1 < length1; ++index1) {
      DOMString value1 = secondSet.getItem(index1).getValue();
      if (compareStrings(value0, value1)) return true;
  }

  return false;
}

bool StringTester::compareStrings(DOMString& first, DOMString& second) {
  if (matchType_ == EQUALS) return (first == second) == logic_;
  else if (matchType_ == CONTAINS) return first.contains(second) == logic_ ;
  return first.startsWith(second) == logic_;
}

void StringTester::setType(const DOMString& typeString) {
  matchType_ = EQUALS;
  logic_ = true;
  if (typeString == "!=") logic_ = false;
  else if (typeString == "contains") matchType_ = CONTAINS;
  else if (typeString == "starts-with") matchType_ = STARTS_WITH;
  else if (typeString == "lang") matchType_ = LANG;
}
}
