/***************************************************************************
                          stylesheet.cpp  -  description
                             -------------------
    begin                : Mon Sep 27 2004
    copyright            : (C) 2004 by John Redmond
    email                : john_redmond@optusnet.com.au
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "stylesheet.h"
#include "documentmerger.h"
#include "configurationmanager.h"
#include "outputmanager.h"
#include "templatelauncher.h"
#include "liststream.h"
#include "attributestream.h"
#include "processbox.h"
#include "builder.h"
#include "timer.h"

namespace limpid {

StyleSheet::StyleSheet(bool verbose):
    isBuilding_(false), verbose_(verbose),
    configurationSet_(new ConfigurationSet) {} // just a place holder

StyleSheet::StyleSheet(const DOMString& fileName, bool isBuilding, bool verbose):
    isBuilding_(isBuilding), verbose_(verbose),
    configurationSet_(new ConfigurationSet) {
  configure(fileName, isBuilding);
}

void StyleSheet::configure(const DOMString& fileName, bool isBuilding) {
  Timer timer;
  Document templateDocument = DocumentMerger(true).getDocument(fileName);
  templateDocumentElement_ = templateDocument.getDocumentElement();

  whitespaceManager_.configure(templateDocument);
  whitespaceManager_.removeTemplateWhitespace(templateDocument);

  templateDocumentElement_ = templateDocument.getDocumentElement();
  if (!templateDocumentElement_.isValid()) return;

  if (XSLUtility::getNodeName(templateDocumentElement_) != "stylesheet")
    configureSimplifiedStyleSheet(templateDocumentElement_);
  else {
    ConfigurationManager(templateDocumentElement_, true).configure(*configurationSet_, true);

    if (configurationSet_->method == "text")
      isBuilding_ = configurationSet_->building = false;
    else
      isBuilding_ = configurationSet_->building = isBuilding;

    configureTemplates(templateDocumentElement_);
  }

  // set escaping before copying
  templateManager_.getContext().setEscaping(*configurationSet_);
  context_ = templateManager_.getContext();

  if (verbose_) timer.report("stylesheet initialisation");
}

void StyleSheet::setWriter(CharacterWriter& writer) {
  OutputManager().setWriter(context_, writer, *configurationSet_);
}

void StyleSheet::transform(const DOMString& sourcePath) {
  Document sourceDocument = getSourceDocument(sourcePath);
  transformDocument(sourceDocument);
}

void StyleSheet::transform(const DOMString& sourcePath, const DOMString& destPath) {
  Document sourceDocument = getSourceDocument(sourcePath);
  transformDocument(sourceDocument, destPath);
}

void StyleSheet::transformDocument(Document& sourceDocument) {
  OutputManager().setWriter(context_, wrt::Console, *configurationSet_);

  transformNode(sourceDocument, isBuilding_);
  wrt::Console << wrt::endl; // force a flush to the console
}

void StyleSheet::transformDocument(Document& sourceDocument, const DOMString& destPath) {
  Writer writer(destPath);
  OutputManager().setWriter(context_, writer, *configurationSet_);

  transformNode(sourceDocument, isBuilding_);
  writer.flush();
}

void StyleSheet::transformNode(Node& sourceNode, bool isBuilding) {
  sourceNode.reorder();
  if (isBuilding) writeDocument(sourceNode);
  else writeStream(sourceNode);
}

void StyleSheet::transformNode(Node& sourceNode,
    const DOMString& destPath, bool isBuilding) {
  sourceNode.reorder();

  if (isBuilding)
    writeDocument(sourceNode);
  else
    writeStream(sourceNode, destPath);
}

void StyleSheet::writeDocument(Node& sourceNode) {
  buildDocument(sourceNode);  // destinationDocument_ is updated
  context_.getDOMWriter().write(destinationDocument_);
}

void StyleSheet::writeStream(Node& sourceNode) {
  context_.setNode(sourceNode);
  TemplateLauncher(selectionManager_, defaultTemplate_).process(context_);
}

void StyleSheet::writeStream(Node& sourceNode, const DOMString& destPath) {
  context_.setNode(sourceNode);

  OutputManager().setWriter(context_, destPath, *configurationSet_);
  TemplateLauncher(selectionManager_, defaultTemplate_).process(context_);
  context_.getWriter().flush();
}

Document& StyleSheet::buildDocument(Node& sourceNode) {
  sourceNode.getOwnerDocument().reorder();
  Document document;
  destinationDocument_ = document; // clear the stored document

  context_.setNode(sourceNode);
  TemplateLauncher(selectionManager_, defaultTemplate_).processNode(context_, destinationDocument_);

  return destinationDocument_;
}

void StyleSheet::configureTemplates(Element& documentElement) {
  templateManager_.configure(documentElement, *configurationSet_, xslResources_, false);
  selectionManager_.setResources(xslResources_);
  defaultTemplate_.setResources(xslResources_);

  // now configure from any included stylesheets in correct order
  NodeList childSheets = getChildSheets(documentElement, true);
  ListStream stream(childSheets);

  while (true) {
    Node& node = stream.nextNode();
    if (!node.isValid()) break;

    Element& child = static_cast<Element &>(node);
    templateManager_.configure(child, *configurationSet_, xslResources_, true);
  }
}

NodeList StyleSheet::getChildSheets(Node& anchor, bool includeAnchor) {
  NodeList resultList;
  visit(anchor, resultList, includeAnchor);
  return resultList;
}

void StyleSheet::visit(Node& root, NodeList& resultList, bool includeRoot) {
  ListStream stream(root);
  stream.setForward(false);

  while (true) {
    Node& node = stream.nextNode();
    if (!node.isValid()) break;

    if (node.getNodeType() != Node::ELEMENT_NODE) continue;
    DOMString nodeName = XSLUtility::getNodeName(static_cast<Element&>(node));
    if (nodeName == "stylesheet")
      visit(node, resultList, true);
  }

  if (includeRoot) resultList.addNode(root);
}

void StyleSheet::configureSimplifiedStyleSheet(Element& documentElement) {
  Element sheetElement("StyleSheet");
  sheetElement.setAttribute("version", "1.0");
  AttributeStream attributeStream(documentElement);
  while (true) {
    Node& node = attributeStream.nextNode();
    if (!node.isValid()) break;

    Attr attr = static_cast<Attr&>(node.cloneNode(true));
    sheetElement.setAttributeNode(attr);
  }

  Element templateElement("xsl:template");
  templateElement.setAttribute("match", "/");

  sheetElement.appendChild(templateElement);
  templateElement.appendChild(documentElement);

  templateManager_.configure(sheetElement, *configurationSet_, xslResources_, false);
}

Document StyleSheet::getSourceDocument(const DOMString& sourcePath) {
  URL url(sourcePath);
  Reader reader(url);
  XMLParser parser(reader);
  if (whitespaceManager_.isConfigured())
    parser.setProcessingXSLT(true);

  ProcessBox box(parser);
  Builder builder;
  box.addProcessor(builder);
  box.process();

  Document document = builder.getDocument();
  Element& documentElement = document.getDocumentElement();
  if (documentElement.isValid() &&
      !documentElement.getAttributeNode("xml:base").isValid())
    documentElement.setAttribute("xml:base",
      URL(sourcePath).getBasePath());  // may need this path later

  if (whitespaceManager_.isConfigured())
    whitespaceManager_.removeSourceWhitespace(document);

  return document;
}

}
