/***************************************************************************
                          sxfparser.cpp  -  description
                             -------------------
    begin                : Thu May 25 2006
    copyright            : (C) 2006 by John Redmond
    email                : john_redmond@optusnet.com.au
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "sxfparser.h"
#include "node.h"
#include "element.h"
#include "attr.h"
#include "text.h"
#include "comment.h"
#include "cdatasection.h"
#include "xmldeclaration.h"
#include "processinginstruction.h"
#include "documenttype.h"
#include "endnode.h"
#include "writer.h"
#include "stringreader.h"

#include "domwriter.h"

namespace limpid {

SXFParser::SXFParser(const DOMString& fileName):
    endNode_(new EndNode("END")),
    reader_(new Reader(fileName)), tis_(new TokenStream(*reader_)),
    hasContent_(false), hasEnd_(false), depth_(0), lineNo_(0), delimiterString_("##") {
  tis_->setDelimiters("=;");
}

SXFParser::SXFParser(const Reader& charStream):
    endNode_(new EndNode("END")),
    reader_(new Reader(charStream)), tis_(new TokenStream(*reader_)),
    hasContent_(false), hasEnd_(false), depth_(0), lineNo_(0), delimiterString_("##") {
  tis_->setDelimiters("=;");
}

Node& SXFParser::nextNode() {
  // hasContent indicates that there is text content all ready to be returned without further processing
  if (hasContent_) {
    hasContent_ = false;
    return *contentNode_;
  }

  // hasEnd_ indicates that an endnode is to be returned to (eg) signal denesting to a Builder
  if (hasEnd_) {
    hasEnd_ = false;
    return *endNode_;
  }
  
  DOMString nodeName = tis_->nextToken();
  if (nodeName.empty())
    return Node::failNode;

  // we have the end of a compound element
  if (nodeName == ";" || nodeName.startsWith("%%")) {
    if (depth_)
      --depth_;

    return *endNode_;
  }

  // we have a start node for an element
  if (nodeName.startsWith("%")) {
    nodeName = nodeName.substr(1, nodeName.getLength());
    node_ = new Element(nodeName);
    node_->setIsParent(true); // signal nesting to the Builder
    ++depth_;
    return *node_;
  }

  // if we have reached here, we have a node with direct text content

  DOMString delimiterString = tis_->nextToken(); // this should be '='
  DOMString contentString;
  if (delimiterString == "=")
    contentString = tis_->getDelimitedString("##");

  // process an attribute
  if (nodeName.startsWith("@")) {
    nodeName = nodeName.substr(1, nodeName.getLength());
    contentString.normalizeWhitespace().escapeDelimiters();

    node_ = new Attr(nodeName);
    Text contentText(contentString);
    node_->appendChild(contentText);
    return *node_;
  }

  if (!nodeName.startsWith("#")) { // we have an element with only text content
    node_ = new Element(nodeName); // this node will be returned by nextNode()
    if (!contentString.empty()) {
      node_->addChildNodes(); // signal nesting to the XMLParser

///      contentString.escapeBra();
      contentNode_ = new Text(contentString);

      hasContent_ = true; // nextNode() will pick up the text node next time it is called
      static_cast<EndNode *>(endNode_.getPointer())->update(*node_);

      hasEnd_ = true; // the end nod will be picked up on the following call
      ++depth_;
  }
    return *node_;
  }

  if (nodeName.startsWith("#te")) {
///    contentString.escapeBra();
    node_ = new Text(contentString);
    return *node_;
  }

  if (nodeName.startsWith("#co")) {
    node_ = new Comment(contentString);
    return *node_;
  }

  if (nodeName.startsWith("#cd")) {
    node_ = new CDATASection(contentString);
    return *node_;
  }

  // we must have an XML declaration, processing instruction or doctype
  contentString += '\n';
  StringReader stringReader(contentString);
  TokenStream tis(stringReader);
  tis.setDelimiters("");
  tis.setWhitespaceChars("\n");
  
  if (nodeName.startsWith("#xm")) {
    DOMString version = tis.getDelimitedString(',').trim();
    DOMString encoding  = tis.getDelimitedString(',').trim();
    DOMString standalone = tis.nextToken().trim();
    node_ = new XMLDeclaration(version, encoding, standalone);
    return *node_;
  }

  if (nodeName.startsWith("#pr")) {
    DOMString target = tis.getDelimitedString(',').trim();
    DOMString data  = tis.nextToken().trim();
    data.normalizeWhitespace().escapeDelimiters();
    node_ = new ProcessingInstruction(target, data);
    return *node_;
  }

  if (nodeName.startsWith("#do")) {
    DOMString name = tis.getDelimitedString(',').trim();
    DOMString publicID = tis.getDelimitedString(',').trim();
    DOMString systemID = tis.getDelimitedString(',').trim();
    DOMString internalSubset = tis.nextToken().trim();
    internalSubset.trim();

    DocumentType* doctype = new DocumentType(name, publicID, systemID);
    if (internalSubset.getLength())
      doctype->setInternalSubset(internalSubset);

    node_ = doctype;
    return *node_;
  }

  return *node_;
}

DOMString SXFParser::getContentString() {
  return tis_->getDelimitedString("##");
}

}
