/***************************************************************************
                          documentmanager.cpp  -  description
                             -------------------
    begin                : Thu Sep 18 2003
    copyright            : (C) 2003 by John Redmond
    email                : john_redmond@optusnet.com.au
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "systemmanager.h"
#include "node.h"
#include "element.h"
#include "document.h"
#include "attr.h"
#include "writer.h"
#include "domexception.h"
#include "converter.h"

namespace limpid {

SystemManager::SystemManager():
      namespaces(false), active(true),
      trimmingText(false), normalizingText(false),
      permissions(new int[PERMISSIONS_SIZE]), response(WARN) {
  populatePermissions();
}

SystemManager::~SystemManager() {
  delete [] permissions;
}

void SystemManager::processIndexError(size_t index) {
  if (!active || response == IGNORE) return;

  int type = DOMException::INDEX_SIZE_ERR;
  DOMString spec = " ";
  spec += toString(int(index));

  processResponse(type, spec);
}

void SystemManager::processNotFoundError(Node& node) {
  if (!active || response == IGNORE) return;

  int type = DOMException::NOT_FOUND_ERR;
  DOMString spec = "not found: ";
  spec += node.getDisplayString();

  processResponse(type, spec);
}

void SystemManager::processInUseError(Node& node) {
  if (!active || response == IGNORE) return;

  int type = DOMException::INUSE_ATTRIBUTE_ERR;
  DOMString spec = "attribute in use ";
  spec += node.getDisplayString();

  processResponse(type, spec);
}

bool SystemManager::checkAppendPermission(const Node& parent, const Node& child) {
  if (!active) return true;

  if (!parent.isValid() || !child.isValid()) {
    int errorType = DOMException::HIERARCHY_REQUEST_ERR;

    DOMString spec = "[";
    spec += child.getDisplayString();
    spec += "] appended to [";
    spec += parent.getDisplayString();
    spec += "] involves at least one fail node";

    processResponse(errorType, spec);
    return false;
  }

  if (!checkModificationPermission(parent)) return false;
  if (!checkCorrectDocument(parent, child)) return false;

  int parentType = parent.getNodeType();
  int childType = child.getNodeType();
  int mask = *(permissions + parentType);

  if (mask & (1 << childType)) return true;
  else if (response == IGNORE) return false;

  DOMString spec = "type [";
  spec += child.getNodeName();
  spec += "] (";
  spec += child.getDisplayString();
  spec += ") is appended to type [";
  spec += parent.getNodeName();
  spec += "] ";
  spec += parent.getDisplayString();

  processResponse(DOMException::HIERARCHY_REQUEST_ERR, spec);

  return false;
}

bool SystemManager::checkModificationPermission(const Node& parent) {
  if (!parent.isReadOnly()) return true;
  else if (response == IGNORE) return false;

  int type = DOMException::NO_MODIFICATION_ALLOWED_ERR;
  DOMString spec = " for ";
  spec += parent.getDisplayString();

  processResponse(type, spec);

  return false; // keep the compiler happy
}

bool SystemManager::checkCorrectDocument(const Node& parent, const Node& child) {
  if (!active) return true;

  bool result = true;

  if (parent.nodeContent_->documentContent_ && child.nodeContent_->documentContent_)
    result = child.nodeContent_->documentContent_ == parent.nodeContent_->documentContent_;

  if (result || response == IGNORE) return result;

  DOMString spec = " for ";
  spec += child.getDisplayString();
  spec += " appended to ";
  spec += parent.getDisplayString();

  processResponse(DOMException::WRONG_DOCUMENT_ERR, spec);

  return result;
}

bool SystemManager::checkAttributeNotInUse(const Attr& attr) {
  if (!active) return true;
  else if (attr.nodeContent_->parentContent_) {
    int type = DOMException::HIERARCHY_REQUEST_ERR;

    DOMString spec = attr.getDisplayString();
    spec += " already set on element: ";
    spec += attr.getOwnerElement().getDisplayString();

    processResponse(type, spec);
    return false;
  }
  return true;
}

SystemManager& SystemManager::getSystemManager() {
  static SystemManager systemManager;
  return systemManager;
}

void SystemManager::processResponse(int errorType, DOMString& spec) {
  DOMString message = errorNames[errorType - 1];
  if (response == WARN) {
    wrt::Console << "DOM Error(" << (int)errorType << "):\n\t";
    wrt::Console << message << " " << spec << wrt::endl;
  }
  else
    throw DOMException(errorType, spec);
}

void SystemManager::setPermission(int parentType, int childType) {
  *(permissions + parentType) |= (1 << childType);
}

void SystemManager::populatePermissions() {
  int *ptr = permissions;
  for (size_t i = 0; i < PERMISSIONS_SIZE; i++) *ptr++ = 0; // initialize array

  setPermission(Node::DOCUMENT_NODE, Node::ELEMENT_NODE);
  setPermission(Node::DOCUMENT_NODE, Node::COMMENT_NODE);
  setPermission(Node::DOCUMENT_NODE, Node::DOCUMENT_TYPE_NODE);
  setPermission(Node::DOCUMENT_NODE, Node::XML_DECLARATION_NODE);
  setPermission(Node::DOCUMENT_NODE, Node::PROCESSING_INSTRUCTION_NODE);

  setPermission(Node::DOCUMENT_FRAGMENT_NODE, Node::ELEMENT_NODE);
  setPermission(Node::DOCUMENT_FRAGMENT_NODE, Node::PROCESSING_INSTRUCTION_NODE);
  setPermission(Node::DOCUMENT_FRAGMENT_NODE, Node::COMMENT_NODE);
  setPermission(Node::DOCUMENT_FRAGMENT_NODE, Node::TEXT_NODE);
  setPermission(Node::DOCUMENT_FRAGMENT_NODE, Node::CDATA_SECTION_NODE);
  setPermission(Node::DOCUMENT_FRAGMENT_NODE, Node::ENTITY_REFERENCE_NODE);

  setPermission(Node::ENTITY_REFERENCE_NODE, Node::ELEMENT_NODE);
  setPermission(Node::ENTITY_REFERENCE_NODE, Node::PROCESSING_INSTRUCTION_NODE);
  setPermission(Node::ENTITY_REFERENCE_NODE, Node::COMMENT_NODE);
  setPermission(Node::ENTITY_REFERENCE_NODE, Node::TEXT_NODE);
  setPermission(Node::ENTITY_REFERENCE_NODE, Node::CDATA_SECTION_NODE);
  setPermission(Node::ENTITY_REFERENCE_NODE, Node::ENTITY_REFERENCE_NODE);

  setPermission(Node::ELEMENT_NODE, Node::ELEMENT_NODE);
  setPermission(Node::ELEMENT_NODE, Node::COMMENT_NODE);
  setPermission(Node::ELEMENT_NODE, Node::TEXT_NODE);
  setPermission(Node::ELEMENT_NODE, Node::CDATA_SECTION_NODE);
  setPermission(Node::ELEMENT_NODE, Node::PROCESSING_INSTRUCTION_NODE);
  setPermission(Node::ELEMENT_NODE, Node::ENTITY_REFERENCE_NODE);

  setPermission(Node::ATTRIBUTE_NODE, Node::TEXT_NODE);
  setPermission(Node::ATTRIBUTE_NODE, Node::ENTITY_REFERENCE_NODE);

  setPermission(Node::ENTITY_NODE, Node::ELEMENT_NODE);
  setPermission(Node::ENTITY_NODE, Node::PROCESSING_INSTRUCTION_NODE);
  setPermission(Node::ENTITY_NODE, Node::COMMENT_NODE);
  setPermission(Node::ENTITY_NODE, Node::TEXT_NODE);
  setPermission(Node::ENTITY_NODE, Node::CDATA_SECTION_NODE);
  setPermission(Node::ENTITY_NODE, Node::ENTITY_REFERENCE_NODE);
}

char SystemManager::errorNames[][30] = {
          "Index size error", "Domstring size error",
          "Hierarchy request error", "Wrong document error",
          "Invalid character error", "No data allowed error",
          "No modification allowed error", "Not found error",
          "Not supported error", "In-use attribute error",
          "IO error", "Memory error"
};

}
