/***************************************************************************
                          templatemanager.cpp  -  description
                             -------------------
    begin                : Wed Sep 29 2004
    copyright            : (C) 2004 by John Redmond
    email                : john_redmond@optusnet.com.au
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "applytemplate.h"
#include "attributetemplate.h"
#include "documenttemplate.h"
#include "calltemplate.h"
#include "copytemplate.h"
#include "copyoftemplate.h"
#include "elementtemplate.h"
#include "foreachtemplate.h"
#include "formattemplate.h"
#include "iftemplate.h"
#include "matchedtemplate.h"
#include "messagetemplate.h"
#include "numbertemplate.h"
#include "sorttemplate.h"
#include "specifiedattributestemplate.h"
#include "stringtemplate.h"
#include "commenttemplate.h"
#include "templatemanager.h"
#include "texttemplate.h"
#include "valuetemplate.h"
#include "variablesettemplate.h"
#include "documentwritertemplate.h"
#include "keymanager.h"
#include "templateexpr.h"
#include "decimalformatmap.h"

#include "domstring.h"
#include "node.h"
#include "element.h"
#include "document.h"
#include "attr.h"
#include "namednodemap.h"

#include "attributestream.h"
#include "liststream.h"
#include "nodestream.h"

#include "tokenstream.h"
#include "stringreader.h"
#include "stringwriter.h"

#include "converter.h"
#include "domwriter.h"
#include "numberformatter.h"
#include "errorhandler.h"

#include "xpathfactory.h"
#include "xpathstring.h"
#include "parentcomponent.h"
#include "xpathstring.h"
#include "compoundstring.h"
#include "variable.h"
#include "stringgenerator.h"

#include "domwriter.h"

namespace limpid {

TemplateManager::TemplateManager(XSLResources& resources,
      XSLContext& context, bool imported):
      resources_(resources), context_(context), isImported_(imported) {
}

void TemplateManager::configure(Element& documentElement,
      const ConfigurationSet& configurationSet, XSLResources& resources, bool imported) {

  configurationSet_ = configurationSet;
  resources_ = resources;
  isImported_ = imported;

  if (!isImported_)
    resources.setCDATASectionNames(documentElement.getAttribute("cdata-section-elements"));

  processSheet(documentElement);
}

// We process only top-level elements of the sheet element AND
// top-level elements of embedded stylesheet elements.
// We therefore use a ListStream recursively, rather than use
// the automatic recursion of NodeStream
void TemplateManager::processSheet(Element& sheetElement) {
  setAttributeSets(sheetElement);

  ListStream stream(sheetElement);

  while (true) {
    Node& node = stream.nextNode();
    if (!node.isValid()) break;

    if (isXSLElement(node)) {
      Element& element = static_cast<Element &>(node);
      DOMString nodeName = getNodeName(element);

      if (nodeName == "stylesheet") // recurse
        processSheet(element);
      else
        processTopLevelElement(element);
    }
  }
}

// Here we use NodeStream to search through all the document to extract
// references to all the attribute-set elements and add them to the template map
void TemplateManager::setAttributeSets(Element& sheetElement) {
  NodeStream stream(sheetElement, false, false);

  while (true) {
    Node& node = stream.nextNode();
    if (!node.isValid()) break;

    if (!isXSLElement(node)) continue;

    Element& element = static_cast<Element &>(node);
    DOMString nodeName = getNodeName(element);
    if (nodeName != "attribute-set") continue;

    DOMString setName = element.getAttribute("name");
    if (setName.empty()) {
      ErrorHandler::report(element, "Attribute set without a name");
      continue;  // must have a name
    }

    resources_.setAttributeSet(element);
  }
}

void TemplateManager::processTopLevelElement(Element& element) {
  DOMString nodeName = getNodeName(element);
  if (nodeName == "variable")
    addVariable(element, context_);
  else if (nodeName == "template")
    buildTemplate(element);
  else if (nodeName == "decimal-format")
    addDecimalFormat(element);
  // additional top-level elements would go here
}

void TemplateManager::addDecimalFormat(Element& formatElement) {
  DecimalFormat decimalFormat;
  decimalFormat.update(formatElement);
  context_.getDecimalFormatMap().setFormat(formatElement.getAttribute("name"), decimalFormat);
}

void TemplateManager::buildTemplate(Element& templateElement) {
  DOMString nameString = templateElement.getAttribute("name");
  DOMString matchString = templateElement.getAttribute("match");
  DOMString modeString = templateElement.getAttribute("mode");
  DOMString priorityString = templateElement.getAttribute("priority");
  DOMString stepString = templateElement.getAttribute("step");

  MatchedTemplate matchedTemplate = getMatchedTemplate(matchString,
    modeString, priorityString, templateElement);

  if (!nameString.empty())
    resources_.setCalledTemplate(nameString, matchedTemplate, isImported_);
    
  if (stepString == "yes") {
    stepString = matchString;
    size_t slashPos = stepString.rfind('/');
    if (slashPos != stepString.npos)
      stepString = stepString.substring(slashPos + 1,
      stepString.getLength() - slashPos - 1);
    resources_.setStep(stepString);
  }

  if (!matchString.empty()) {
    DOMString keyString = getCompoundTemplateName(matchString, modeString);
    resources_.setMatchedTemplate(keyString, matchedTemplate, isImported_);
  }
}

// we may have a union path
void TemplateManager::setMultiplyMatchedTemplate(DOMString selectString,
      DOMString& modeString, DOMString& priorityString, Element& templateElement) {
  StringReader stringReader(selectString);
  TokenStream tis(stringReader);
  tis.setWhitespaceChars("|");

  while (true) {
    DOMString pathString = tis.nextToken();
    if (pathString.empty()) break;

    MatchedTemplate matchedTemplate =
      getMatchedTemplate(pathString, modeString, priorityString, templateElement);
    DOMString keyString = getCompoundTemplateName(pathString, modeString);
    resources_.setMatchedTemplate(keyString, matchedTemplate, isImported_);
  }
}

MatchedTemplate TemplateManager::getMatchedTemplate(DOMString& matchString,
    DOMString& modeString, DOMString& priorityString, Element& templateElement) {
  MatchedTemplate matchedTemplate(matchString, modeString);
  addParameters(matchedTemplate, templateElement, "param");
  addTemplateContent(matchedTemplate, templateElement);
  if (!priorityString.empty())
    matchedTemplate.setPriority(Converter::toDouble(priorityString));

  return matchedTemplate;
}

void TemplateManager::addTemplateContent(Template& parentTemplate, Element& templateElement) {
  if (!hasChildren(templateElement))
    return;

  ListStream stream(const_cast<Element &>(templateElement));

  while (true) {
    Node& node = stream.nextNode();
    if (!node.isValid()) break;

    if (node.getNodeType() == Node::TEXT_NODE) {
      parentTemplate.setHasText(true);
      DOMString displayValue = node.getValue();
      if (!displayValue.empty())
        if (context_.getEscapingText())
          displayValue.resolveEntities();
        addSimpleString(displayValue, parentTemplate);
      continue;
    }

    if (node.getNodeType() != Node::ELEMENT_NODE) continue;
    Element& element = static_cast<Element &>(node);

    if (isXSLElement(element))
      processXSLElement(parentTemplate, element);

    else // it must be a literal element
      addLiteralElement(parentTemplate, element);
  }
}

void TemplateManager::processXSLElement(Template& parentTemplate,
      Element& xslElement) {
  DOMString name = getNodeName(xslElement);
  DOMString selectString = xslElement.getAttribute("select");

  if (name == "apply-templates") {
    // if selectString contains "key(", it needs to be modified by KeyManager
    if (!selectString.empty())
      selectString = KeyManager().getExprString(configurationSet_, selectString);
    else
      selectString = "* | text()";

    DOMString modeString = xslElement.getAttribute("mode");
    ApplyTemplate applyTemplate(selectString, modeString, isImported_);
    applyTemplate.setResources(resources_);

    addParameters(applyTemplate, xslElement, "with-param");
    parentTemplate.addTemplate(applyTemplate);
    addTemplateContent(applyTemplate, xslElement);
  }
  else if (name == "call-template") {
    DOMString templateName = xslElement.getAttribute("name");
    if (!templateName.size()) return; // cannot work without a name

    CallTemplate callTemplate(templateName, isImported_);
    callTemplate.setResources(resources_);
    addParameters(callTemplate, xslElement, "with-param");
    parentTemplate.addTemplate(callTemplate);
  }

  else if (name == "for-each") {
    ForEachTemplate forEachTemplate(selectString);
    parentTemplate.addTemplate(forEachTemplate);
    addTemplateContent(forEachTemplate, xslElement);
  }

  else if (name == "sort") {
    DOMString typeString = xslElement.getAttribute("data-type");
    DOMString orderString = xslElement.getAttribute("order");
    SortTemplate sortTemplate(selectString, typeString, orderString);
    parentTemplate.addSortTemplate(sortTemplate);
  }

  else if (name == "choose") {
    CompoundTemplate chooseTemplate;
    parentTemplate.addTemplate(chooseTemplate);
    addTemplateContent(chooseTemplate, xslElement);
  }
  else if (name == "if") {
    DOMString testString = xslElement.getAttribute("test");
    IfTemplate ifTemplate(testString, true);
    parentTemplate.addTemplate(ifTemplate);
    addTemplateContent(ifTemplate, xslElement);
  }
  else if (name == "when") {
    DOMString testString = xslElement.getAttribute("test");
    IfTemplate whenTemplate(testString, false);
    parentTemplate.addTemplate(whenTemplate);
    addTemplateContent(whenTemplate, xslElement);
  }
  else if (name == "otherwise")
    addTemplateContent(parentTemplate, xslElement);
  else if (name == "value-of") {
    DOMString disableOutputEscaping = xslElement.getAttribute("disable-output-escaping");
    ValueTemplate valueTemplate(selectString, disableOutputEscaping == "yes");
    valueTemplate.setResources(resources_);
    parentTemplate.addTemplate(valueTemplate);
  }

  else if (name == "cdata-section") {
    ValueTemplate valueTemplate(selectString);
    parentTemplate.addTemplate(valueTemplate);
  }
  else if (name == "element")
    addDefinedElement(parentTemplate, xslElement);
  else if (name == "comment")
    addComment(parentTemplate, xslElement);
  else if (name == "processing-instruction")
    addProcessingInstruction(parentTemplate, xslElement);
  else if (name == "copy")
    addCopyNode(parentTemplate, xslElement);
  else if (name == "copy-of")
    addCopyOfNode(parentTemplate, xslElement);
  else if (name == "text") {
    Expr textExpr = getStringExpr(xslElement.getValue());
    bool resolveText = xslElement.getAttribute("disable-output-escaping") != "yes";
    // resolveText == true means we need to use the default poliicy
    if (resolveText)
      resolveText = !configurationSet_.escape_text;
      
    TextTemplate textTemplate(textExpr, resolveText);
    parentTemplate.addTemplate(textTemplate);
  }
  else if (name == "variable")
    addVariableTemplate(parentTemplate, xslElement);
  else if (name == "message")
    addMessage(parentTemplate, xslElement);
  else if (name == "number")
    addNumber(parentTemplate, xslElement);
  else if (name == "document") {
    DOMString dest = xslElement.getAttribute("href");
    DocumentWriterTemplate documentWriterTemplate(dest, configurationSet_);
    parentTemplate.addTemplate(documentWriterTemplate);
    addTemplateContent(documentWriterTemplate, xslElement);
  }
}

void TemplateManager::addVariableTemplate(Template& parentTemplate,
      Element& xslElement) {
  Expr variableExpr = getVariableExpr(xslElement);
  if (variableExpr.isValid()) {
    DOMString variableName = xslElement.getAttribute("name");
    VariableSetTemplate variableSetTemplate(variableName, variableExpr);
    parentTemplate.addTemplate(variableSetTemplate);
  }
}

void TemplateManager::addDefinedElement(Template& parentTemplate,
      Element& elementElement) {
  DOMString elementName = elementElement.getAttribute("name");
  if (elementName.empty()) {
    ErrorHandler::report(elementElement, "Defined element without a name");
    return;  // must have a name
  }

  Expr nameExpr = XPathString(elementName);

  if (isAttributeValueTemplate(elementName))
    nameExpr = XPathFactory::getExpr(elementName.substring(1, elementName.size() - 2));
  bool isHTML = configurationSet_.method == "html";

  ElementTemplate elementTemplate(elementName, nameExpr, hasChildren(elementElement), isHTML);
  elementTemplate.setNamespaces(configurationSet_);
  parentTemplate.addTemplate(elementTemplate);

  // add attributes here
  addAttributesFromSets(elementTemplate, elementElement);
  addDefinedAttributes(elementTemplate, elementElement);

  DOMString namespaceName = elementElement.getAttribute("namespace");
  if (!namespaceName.empty())
    addLiteralAttribute(elementTemplate, namespaceName, elementName);

  // recurse to add child templates
  addTemplateContent(elementTemplate, elementElement); // recurse
}

void TemplateManager::addLiteralElement(Template& parentTemplate,
      Element& elementElement) {
  bool isHTML = configurationSet_.method == "html";
  DOMString elementName = elementElement.getNodeName();

  Expr nameExpr = XPathString(elementName); // may be just a simple string
  ElementTemplate elementTemplate(elementName, nameExpr, hasChildren(elementElement), isHTML);
  elementTemplate.setNamespaces(configurationSet_);
  parentTemplate.addTemplate(elementTemplate);

  // add attrs present in literal tag
  NamedNodeMap& attributes = elementElement.getAttributes();
  size_t attrCount = attributes.getLength();
  for (size_t index = 0; index < attrCount; ++ index) {
    Attr& attr = static_cast<Attr &>(attributes.item(index));
    XPathString attrValue = attr.getValue();
    if (!context_.getEscapingAttributes())
      attrValue.resolveEntities();

   addLiteralAttribute(elementTemplate, attr.getNodeName(), attrValue);
  }

  // add child templates, including specified attributes,
  // which may overwrite the literal attributes
  addTemplateContent(elementTemplate, elementElement); // recurse
}

void TemplateManager::addAttribute(Template& parentTemplate, Element& attributeElement) {
  DOMString attrName = attributeElement.getAttribute("name");
  Expr nameExpr = getStringExpr(attrName);
  Expr uriExpr = getStringExpr(attributeElement.getAttribute("namespace"));
  Expr valueExpr = getStringExpr(attributeElement.getValue());

  AttributeTemplate attributeTemplate(attrName, nameExpr, uriExpr, valueExpr);
  parentTemplate.addTemplate(attributeTemplate);
}

void TemplateManager::addLiteralAttribute(ElementTemplate& elementTemplate,
      const DOMString& name, const DOMString& value) {
  Expr nameExpr = XPathString(name);
  Expr valExpr = getStringExpr(value);
  AttributeTemplate attributeTemplate(name, nameExpr, valExpr);
  elementTemplate.addTemplate(attributeTemplate);
}

// this function is used both to add attributes that are directly specified as children
// of template element and, recursively, children of specified attribute sets.
void TemplateManager::addDefinedAttributes(Template& parentTemplate,
    Element& templateElement) {
  ListStream stream(templateElement);

  while (true) {
    Node& node = stream.nextNode();
    if (!node.isValid()) break;

    if (!isXSLElement(node)) continue;

    Element element = static_cast<Element &>(node);
    DOMString nodeName = getNodeName(element);
    if (nodeName != "attribute") continue; // ignore

    Element& attrElement = static_cast<Element &>(node);
    DOMString attrName = attrElement.getAttribute("name");
    if (attrName.empty()) {
      ErrorHandler::report(templateElement, "attr without a name");
      continue;  // this is an error
    }
    Expr nameExpr = getStringExpr(attrName);
    Expr uriExpr = XPathString(attrElement.getAttribute("namespace"));
    Expr valExpr = getDefinedAttributeExpr(attrElement);

    AttributeTemplate attributeTemplate(attrName, nameExpr, uriExpr, valExpr);
    parentTemplate.addTemplate(attributeTemplate);
  }
}

// this is a recursive function, because a "use-attribute-sets" attribute may
// be on an attribute-set declaration, at top level, as well as on one or more
// attribute-set elements stored in the template map.
void TemplateManager::addAttributesFromSets(Template& elementTemplate,
      Element& templateElement) {
  if (!templateElement.isValid()) return;

  DOMString setNames = templateElement.getAttribute("use-attribute-sets");
  if (setNames.empty()) return; // there may not be this attribute

  StringReader reader(setNames);
  TokenStream tis(reader);
  tis.setDelimiters(""); // tokenise the string

  while (true) {
    DOMString setName = tis.nextToken();
    if (setName.empty()) break;

    Element& setElement = resources_.getAttributeSet(setName);
    addAttributesFromSets(elementTemplate, setElement); // first recurse if needed

    addDefinedAttributes(elementTemplate, setElement);
  }
}

void TemplateManager::addNamespaceAttr(ElementTemplate& elementTemplate,
      DOMString namespaceName, DOMString elementName) {
  DOMString attrName = " xmlns";
  if (elementName.contains(":")) {
    size_t colonIndex = elementName.indexOf(":");
    attrName += ":";
    attrName += elementName.substring(0, colonIndex);
  }

  Expr nameExpr = XPathString(attrName);
  Expr valExpr = XPathString(namespaceName.escapeDelimiters());

  AttributeTemplate attributeTemplate(attrName, nameExpr, valExpr);
  elementTemplate.addTemplate(attributeTemplate);
}

void TemplateManager::addCopyNode(Template& parentTemplate,
      Element& copyElement) {

  CopyTemplate copyTemplate(true, hasChildren(copyElement)); // can form start tag
  parentTemplate.addTemplate(copyTemplate);

  // add attributes here
  addAttributesFromSets(copyTemplate, copyElement);
  addDefinedAttributes(copyTemplate, copyElement);

  // recurse to add child templates
  addTemplateContent(copyTemplate, copyElement); // recurse
}

void TemplateManager::addCopyOfNode(Template& parentTemplate, Element& copyOfElement) {
  DOMString selectString = copyOfElement.getAttribute("select");
  if (selectString.empty()) {
    ErrorHandler::report(copyOfElement, "attr without a name");
    return;  // this is an error
  }
  CopyOfTemplate copyOfTemplate(selectString);
  parentTemplate.addTemplate(copyOfTemplate);
}

void TemplateManager::addComment(Template& tmp, Element& commentElement) {
  CommentTemplate commentTemplate;
  addTemplateContent(commentTemplate, commentElement);
  tmp.addTemplate(commentTemplate);
}

void TemplateManager::addProcessingInstruction(Template& tmp, Element& instructionElement) {
  DOMString instructionName = instructionElement.getAttribute("name");
  if (instructionName.empty()) {
    ErrorHandler::report(instructionElement, "processing instruction without a name");
    return;  // cannot continue
  }

  CompoundTemplate instructionTemplate;
  tmp.addTemplate(instructionTemplate);

  FormatTemplate nestTemplate(true);
  instructionTemplate.addTemplate(nestTemplate);

  DOMString instructionString = "<?";
  instructionString += instructionName;
  instructionString += " ";
  instructionString += instructionElement.getValue();
  instructionString += "?>";
  addSimpleString(instructionString, instructionTemplate);

  FormatTemplate deNestTemplate = FormatTemplate(false);
  instructionTemplate.addTemplate(deNestTemplate);
}

void TemplateManager::addMessage(Template& tmp, Element& messageElement) {
  DOMString terminateString = messageElement.getAttribute("terminate");

  MessageTemplate messageTemplate(terminateString == "yes");
  tmp.addTemplate(messageTemplate);

  addTemplateContent(messageTemplate, messageElement);
}

// add a number template to a parent template. Because of the large number of
// possible attributes for an xsl:number element, control of number formatting
// is partitioned into a policy class, NumberFormatter.
void TemplateManager::addNumber(Template& parentTemplate, Element& numberElement) {
  DOMString countString = numberElement.getAttribute("count");
  DOMString valueString = numberElement.getAttribute("value");
  DOMString levelString = numberElement.getAttribute("level");
  DOMString fromString = numberElement.getAttribute("from");
  NumberTemplate numberTemplate(countString, valueString, levelString, fromString);

  DOMString formatString = numberElement.getAttribute("format");
  // the following group of attributes is not actually used:

  DOMString langString = numberElement.getAttribute("lang");
  DOMString letterString = numberElement.getAttribute("letter-value");
  DOMString separatorString = numberElement.getAttribute("grouping-separator");
  DOMString sizeString = numberElement.getAttribute("grouping-size");
  NumberFormatter numberFormatter(formatString, langString, letterString,
    separatorString, sizeString);

  numberTemplate.setNumberFormatter(numberFormatter);
  parentTemplate.addTemplate(numberTemplate);
}

// xsl:variable is a top-level element, and simply adds a variable to the variable
// map in the XSLContext. It uses a transient template instance.
void TemplateManager::addVariable(Element& variableElement, XSLContext& context) {
  DOMString variableName = variableElement.getAttribute("name");
  Expr variableExpr = getVariableExpr(variableElement);
  if (variableExpr.isValid())
    VariableSetTemplate(variableName, variableExpr).process(context);
}

Expr TemplateManager::getVariableExpr(Element& variableElement) {
  DOMString variableValue = variableElement.getAttribute("select");
  if (!variableValue.empty())
    return XPathFactory::getExpr(variableValue);

  // drop through to here if there was no variable expression
  // and we cannot set a variable
  return Expr();
}

// this function is called by xsl:apply-templates and xsl:call-template. It
// creates a VariableSetTemplate (with a true parameter to allow it to overwrite
// an existing map entry) and adds it to a parent template.
void TemplateManager::addParameters(Template& parentTemplate,
    Element& templateElement, const DOMString& tagName) {
  ListStream stream(templateElement);
  while (true) {
    Node& node = stream.nextNode();
    if (!node.isValid()) break;

    if (!isXSLElement(node)) continue;

    Element element = static_cast<Element &>(node);
    DOMString nodeName = getNodeName(element);
    if (nodeName != tagName) continue; // error

    Element& paramElement = static_cast<Element &>(node);
    DOMString variableName = paramElement.getAttribute("name");
    if (variableName.empty()) {
      ErrorHandler::report(templateElement, "parameter is incomplete");
      continue;
    }

    DOMString selectString = paramElement.getAttribute("select");
    if (!selectString.empty()) {
      Expr paramExpr =  XPathFactory::getExpr(selectString);
      parentTemplate.setParameter(variableName, paramExpr);
    }
  }
}

void TemplateManager::addSimpleString(DOMString& displayString, Template& parentTemplate) {
  StringTemplate stringTemplate(displayString);
  parentTemplate.addTemplate(stringTemplate);
}

Expr TemplateManager::getStringExpr(DOMString string) {
  if (!hasAttributeValueTemplate(string))
    return XPathString(string);

  CompoundString compoundString;
  StringReader reader(string);
  TokenStream tis(reader);
  tis.setDelimiters("{}");
  tis.setWhitespaceChars("");

  CompoundTemplate holderTemplate;

  while (true) {
    DOMString token = tis.nextToken();
    if (token.empty()) break;

    if (token != "{" && token !="}") {
      XPathString substring(token);
      compoundString.appendChild(substring);
      continue;
    }

    if (token == tis.peekToken()) {
      tis.bump();  // this is a literal curly bracket
      XPathString xpathstring(token);
      compoundString.appendChild(xpathstring);
      continue;
    }

    if (token != "{") {
      wrt::Errors << "ERROR: " << token << " cannot start an attr temp expr" << wrt::endl;
      continue; // this is not good
    }

    // we have just read a "{" token, so now we read an expression token
    DOMString expressionToken = tis.nextToken();
    Expr expr(XPathFactory::getExpr(expressionToken));
    compoundString.appendChild(XPathFactory::getExpr(expressionToken).getContent());

    token = tis.peekToken();
    if (token != "}") {
      wrt::Errors << "ERROR: " << token << " should end an attr temp expr" << wrt::endl;
      continue; // this also is not good
    }

    tis.bump();
  }

  return compoundString;
}

bool TemplateManager::hasAttributeValueTemplate(DOMString testString) {
  testString.change("{{", "}").change("}}", "}");
  size_t startIndex = testString.indexOf("{");

  size_t endIndex = testString.indexOf("}");
  if (endIndex == testString.npos) return false;

  return endIndex > startIndex;
}

bool TemplateManager::isAttributeValueTemplate(DOMString testString) {
  testString.change("{{", "}").change("}}", "}");
  size_t startIndex = testString.indexOf("{");
  if (startIndex != 0) return false;

  size_t endIndex = testString.indexOf("}");
  if (endIndex == testString.npos) return false;

  return endIndex == testString.size() - 1;
}

Expr TemplateManager::getDefinedAttributeExpr(Element& attrElement) {
  if (attrElement.getChildNodes().getLength() == 1 &&
      (attrElement.getFirstChild().getNodeType() == Node::TEXT_NODE ||
       attrElement.getFirstChild().getNodeType() == Node::CDATA_SECTION_NODE))
    return getStringExpr(attrElement.getFirstChild().getNodeValue());

  CompoundTemplate holder;
  addTemplateContent(holder, attrElement);

  return TemplateExpr(holder);
}


bool TemplateManager::hasChildren(Node& parentNode) {
  ListStream stream(parentNode);  // no endnodes, exclude root

  while (true) {
    Node& node = stream.nextNode();
    if (!node.isValid()) break;

    if (node.getNodeType() == Node::ELEMENT_NODE ||
        node.getNodeType() == Node::TEXT_NODE) return true;
  }

  return false;
}


DOMString TemplateManager::getExprString(DOMString& string) {
  if (string.startsWith("{") && string.endsWith("}"))
    return string.substring(1, string.size() - 2);
  else return string;
}

DOMString TemplateManager::addDelimiters(DOMString& string) {
  DOMString delimitedString, delimiterString = "\"";

  delimitedString += delimiterString;
  delimitedString += string.escapeDelimiters();
  delimitedString += delimiterString;

  return delimitedString;
}

DOMString TemplateManager::getCompoundTemplateName(DOMString& pathString,
    DOMString& modeString) {
  DOMString compoundString = modeString;
  if (compoundString.empty())
    compoundString = "null";
  compoundString += "://";
  compoundString += pathString;  // --> "<modeString>://<pathString>"

  return compoundString;
}

}
