/***************************************************************************
 *   Copyright (C) 2007 by John Redmond   *
 *   john_redmond@optusnet.com.au   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "texconverter.h"
#include "documentsummary.h"
#include "domutils.h"
#include "nodestream.h"
#include "liststream.h"
#include "siblingstream.h"
#include "stringreader.h"
#include "tokenstream.h"
#include "converter.h"
#include "element.h"
#include "text.h"
#include "comment.h"
#include "pointer.h"
#include "domwriter.h"
#include "stringwriter.h"
#include "matcher.h"
#include "getter.h"
#include "xmlparser.h"
#include "cssparser.h"
#include "xpathfactory.h"
#include "xslcontext.h"

namespace limpid {

TexConverter::TexConverter(const DOMString& specPath, const DOMString& destPath = ""): verbose_(false), specPath_(specPath), destPath_(destPath) {
  if (specPath == "**VERBOSE**") {
    verbose_ = true;
    specPath_ = ""; // use only the default settings in latexConfiguration_
    latexConfiguration_.setVerbose(true);
  }

  latexConfiguration_.configure(specPath_);
}

Document& TexConverter::processDocument(Document& sourceDocument) {
  writeOutput(sourceDocument);
  return sourceDocument;
}

void TexConverter::writeOutput(Document& sourceDocument) {
  // first try to get a title from an h1 element of the source document
  output_ << latexConfiguration_.getPreambleString(sourceDocument);
  output_ << latexConfiguration_.getDefinitionsString();
  output_ << latexConfiguration_.getSectionsString();
  output_ << latexConfiguration_.getDocumentStartString(sourceDocument);

  Element& documentElement = sourceDocument.getDocumentElement();
  Element& rootElement = documentElement;

  if (documentElement.getNodeName() == "html") {
    Node& bodyElement = getChild("body", rootElement);
    if (bodyElement.isValid())
      rootElement = static_cast<Element &>(bodyElement);
  }

  traverse(rootElement, output_, true);

  output_ << latexConfiguration_.getTailString();
  output_.flush();
}

void TexConverter::traverse(Node& sourceRoot, CharacterWriter& writer, bool includeRoot) {
  NodeStream stream(sourceRoot, true, includeRoot);

  while (true) {
    Node& node = stream.nextNode();
    if (!node.isValid()) break;
    int nodeType = node.getNodeType();

    if (nodeType == Node::TEXT_NODE || nodeType == Node::CDATA_SECTION_NODE) {
      writer << node.getValue().escapeTex();
      continue;
    }

    if (nodeType == Node::ELEMENT_NODE) {
      if (node.getChildNodes().getLength() == 0) {
wrt::Errors << "Empty element found on line " << node.getPosition() << wrt::endl;
        continue; // nasty!
    }

      DOMString nodeName = node.getNodeName();

      if (latexConfiguration_.isHidden(nodeName)) {
        stream.setIncludeChildren(false);
        continue;
      }

      if (nodeName == "head") {  // ignore the head and all its content
        stream.setIncludeChildren(false);
        continue;
      }
      if (nodeName == "linkbox") {
        writeLink(node, nodeName, writer);
        stream.setIncludeChildren(false);
        continue;
      }
      if (nodeName == "pre") {
        writeLiteral(node, nodeName, writer);
        stream.setIncludeChildren(false);
        continue;
      }
      if (nodeName == "ul" || nodeName == "ol") {
        writeList(node, nodeName, writer);
        stream.setIncludeChildren(false);
        continue;
      }
      if (nodeName == "dl") {
        writeDefinitionList(node, nodeName, writer);
        stream.setIncludeChildren(false);
        continue;
      }
      if (nodeName == "table") {
        writeTable(node, nodeName, writer);
        stream.setIncludeChildren(false);
        continue;
      }

      // use default processing for the element
      if (isBlockNode(node))
        openBlockStyle(node, stream, writer);
      else
        openTagStyle(node, stream, writer);
    }

    if (nodeType == Node::TEXT_NODE || nodeType == Node::CDATA_SECTION_NODE) {
      writeText(node, writer);
      continue;
    }

    if (nodeType == Node::END_NODE)
      writeEnd(node, writer);

    writer.flush();
  }
}

void TexConverter::writeText(Node& node, CharacterWriter& writer) {
  // do we need a space before?
  if (hasTextBefore(node))
    writer << " ";

  writer << node.getValue().resolveEntities(true).change("&", "and").escapeTex().trim();

  // we need a space after, provided we are not at the end of a sentence.
  Node& parentNextNode = node.getParentNode().getNextSibling();
  if (parentNextNode.getNodeType() == Node::TEXT_NODE) {
    DOMString nextText = parentNextNode.getValue();
    if (nextText != ".")
      writer << " ";
  }
}

void TexConverter::writeList(Node& node, const DOMString& nodeName, CharacterWriter& writer) {
  DOMString openString = latexConfiguration_.getPrestring(nodeName);
  DOMString closeString = latexConfiguration_.getPoststring(nodeName);

  writer << openString << " % " << node << wrt::endl;

  ListStream listStream(node);
  while (true) {
    Node& childNode = listStream.nextNode();
    if (!childNode.isValid()) break;

    if (childNode.getNodeName() == "li")
      writer << "\\item " << getContentString(childNode) << wrt::endl;
  }

  writer << closeString;
}

void TexConverter::writeDefinitionList(Node& node, const DOMString& nodeName, CharacterWriter& writer) {
  DOMString openString = latexConfiguration_.getPrestring(nodeName);
  DOMString closeString = latexConfiguration_.getPoststring(nodeName);

  writer << openString << wrt::endl;

  ListStream listStream(node);
  while (true) {
    Node& childNode = listStream.nextNode();
    if (!childNode.isValid()) break;

    if (childNode.getNodeName() == "dt")
      output_ << "\\item[" << getContentString(childNode) << "] ";
    else if (childNode.getNodeName() == "dd")
      writer << getContentString(childNode) << wrt::endl;
  }

  writer << closeString;
}

void TexConverter::writeLiteral(Node& node, const DOMString& nodeName, CharacterWriter& writer) {
  DOMString openString = latexConfiguration_.getPrestring("pre");
  DOMString closeString = latexConfiguration_.getPoststring("pre");

  writer << openString << wrt::endl;;
  writer << node.getValue().change("&lt;", "<").trim();
  writer << wrt::endl << closeString;

}

void TexConverter::writeLink(Node& node, const DOMString& nodeName, CharacterWriter& writer) {
  Element& element = static_cast<Element &>(node);

  writer << "{\\footnotesize \\url{";
  writer << node.getValue().resolveEntities(true); // include "&lt;
  writer << "(" << element.getAttribute("href") << ")";
  writer << "}} ";

  if (node.getNextSibling().getNodeName() == "a")
    writer << "\\hspace{5mm} ";
  else
    writer << "\\\\" << wrt::endl;
}

void TexConverter::writeTable(Node& node, const DOMString& nodeName, CharacterWriter& writer) {
  DOMString alignmentOptions = "LLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLL";
  DOMString openFloat = "{\\hfill\\begin{table}[h]\n";
  DOMString openTabular = "\\begin{tabulary}{\\textwidth}{|**ALIGN**|}\n";
  DOMString closeTabular = "\\end{tabulary}\n";
  DOMString closeFloat = "\\end{table}\\hfill}\n\n";

  Node& captionNode = getChild("caption", node);
  NodeList headList = getChildren("thead", node); // there should be only one
  NodeList bodyList = getChildren("tbody", node);
  NodeList footList = getChildren("tfoot", node);
  NodeList rowList = getChildren("tr", node);

  // work out the maximum number of cells per table row
  int width = getMaxLength(headList);

  int currentWidth = getMaxLength(bodyList);
  if (currentWidth > width)
    width = currentWidth;

  currentWidth = getMaxLength(footList);
  if (currentWidth > width)
    width = currentWidth;


  currentWidth = getMaxLength(rowList);
  if (currentWidth > width)
    width = currentWidth;

  DOMString alignmentString = alignmentOptions.substr(0, width);
  openTabular.change("**ALIGN**", alignmentString);

// output the top of the float and tabular, with or without a caption
  writer << openFloat;
  if (captionNode.isValid())
    writer << "{\\bfseries\\caption{" << getContentString(captionNode) << "}}" << wrt::endl;
  writer << openTabular;

  writer << "\\hline\n"; // top of table

  // output the table contents, with a line above and below:
  if (headList.getLength()) { // output header with following hline
    writeRows(headList, true, writer); // write bold
    writer << "\\hline\n";
  }

  if (bodyList.getLength()) writeRows(bodyList, false, writer); // write normal
  if (rowList.getLength()) writeRows(rowList, false, writer); // table rows at the top level are deprecated
  if (footList.getLength()) {
    writer << "\\hline\n";
    writeRows(footList, true, writer); // write bold
  }

  writer << "\\hline" << wrt::endl;

  writer << closeTabular; // output the bottom
  writer << closeFloat;
}

void TexConverter::writeRows(NodeList& rowList, bool highlighting, CharacterWriter& writer) {
  ListStream listStream(rowList);
  while (true) {
    Node& listNode = listStream.nextNode();
    if (!listNode.isValid()) break;

    writeRow(listNode, highlighting, writer);
  }
}

void TexConverter::writeRow(Node& rowNode, bool highlighting, CharacterWriter& writer) {
  if (!getChildCount(rowNode)) return;

  ListStream dataStream(rowNode);
  while (true) {
    Node& dataNode = dataStream.nextNode();
    if (!dataNode.isValid()) break;

    if (dataNode.getNodeName() == "tr") {
      writeRow(dataNode, highlighting, writer);
      continue;
    }

    if (dataNode.getNodeName() == "td") {
      if (highlighting)
        writer << "{\\bfseries " << getContentString(dataNode) << "}";
      else
         writer << getContentString(dataNode);
    }

    if (hasNextSibling(dataNode))
      writer << " & ";
    else
      writer << "\\\\\n";
  }
}

int TexConverter::getMaxLength(NodeList& rowList) {
  int maxCount = 0, currentCount;
  for (int index = 0; index < rowList.getLength(); ++index) {
    Node& currentNode = rowList.item(index);
    DOMString nodeName = currentNode.getNodeName();
    if (nodeName == "tr")
      currentCount = getChildCount(rowList.item(index));
    else
      currentCount = getMaxLength(currentNode.getChildNodes());

    if (currentCount > maxCount)
      maxCount = currentCount;
  }
  return maxCount;
}

void TexConverter::writeEnd(Node& node, CharacterWriter& writer) {
  Node& startNode = static_cast<EndNode &>(node).getStartNode();

  if (isBlockNode(startNode))
    closeBlockStyle(startNode, writer);
  else
    closeTagStyle(startNode, writer);
}

void TexConverter::update(const DOMString& destName) {
  // replace wildcard, if there is one
  DOMString destPath = destPath_;
  destPath.change("*", destName);
  output_ = Writer(destPath);
}

void TexConverter::openBlockStyle(Node& node, NodeStream& stream, CharacterWriter& writer) {
  DOMString pathString = latexConfiguration_.getMatchedPath(node);

  if (pathString == "**IGNORE**") {
    stream.setIncludeChildren(false);
    return;
  }

  if (!isBlockNode(node.getPreviousSibling()) && hasText(node.getPreviousSibling()))
    writer << wrt::endl;

  writer  << "{";

  if (verbose_)
    writer << "BLOCK " << getLabel(node) << "\\\\";

  DOMString preString, postString, nodeName = node.getNodeName();

  if (latexConfiguration_.isUndefined(nodeName)) { /// we have already tried
    endStack_.push_back(postString); /// empty
    return;
  }

  if (pathString.empty()) {
    endStack_.push_back(postString); /// empty
    latexConfiguration_.setUndefined(nodeName);
    return;
  }

  preString = getPrestring(pathString).trim();
  postString = getPoststring(pathString).trim();


  if (!preString.empty())
    writer << preString << " ";

  if (postString.empty())
    postString = "**STYLE**";  /// empty but valid

  endStack_.push_back(postString);
}

void TexConverter::closeBlockStyle(Node& node, CharacterWriter& writer) { // this always closes an environment
  DOMString backString = getBackString();

  if (backString == "**STYLE**")
    backString.clear();

  writer << backString << "}" << " % block end " << node << wrt::endl;

  if (node.getNextSibling().isValid())
    writer << wrt::endl;
}

void TexConverter::openTagStyle(Node& node, NodeStream& stream, CharacterWriter& writer) {
  DOMString pathString = latexConfiguration_.getMatchedPath(node);

  if (pathString == "**IGNORE**") {
    stream.setIncludeChildren(false);
    return;
  }

  DOMString preString = getPrestring(pathString).trim();
  DOMString postString = getPoststring(pathString).trim();
  DOMString sectionString = latexConfiguration_.getSectionString(pathString);

  if (verbose_)
    writer << getLabel(node) << " ";

  bool inMixedContent = hasText(node.getParentNode());

  if (false && inMixedContent && node.getPreviousSibling().isValid())
    writer << " "; //a space between siblings

  if (!sectionString.empty()) { // we have a function for a section
    writer << sectionString << "{";
    if (!preString.empty())
      writer << " ";

    if (postString.empty())
      postString = "**STYLE**";  // signal that there is a closing brace

    sectionStack_.push_back(true);
    endStack_.push_back(postString);
    return;
  }

  // we may or may not have style commands

  if (!preString.empty() || !postString.empty()) {// do we really have a style?
    writer << "{" << preString; // may be empty

    if (!preString.empty())
      writer << " ";

    if (postString.empty())
      postString = "**STYLE**";  // signal that there is a style to close
  }

  // otherwise we have no style and little to do
  sectionStack_.push_back(false);
  endStack_.push_back(postString);

}

void TexConverter::closeTagStyle(Node& node, CharacterWriter& writer) { // this may or may not close and environment
  bool isSectionStyle = getSectionStatus();

  bool inMixedContent = hasText(node.getParentNode());
  DOMString backString = getBackString();

  bool hasStyle = !backString.empty();

  if (backString == "**STYLE**")
    backString.clear();

  if (inMixedContent && hasNextSibling(node)) {
    if (hasStyle) {
      writer << "}";
      writer << backString;
    }

    return; /// mixed content
  }

  else {
    if (hasStyle)
      writer << backString;
    if (hasStyle)
      writer << "}";


    if (hasNextSibling(node) && hasText(node.getNextSibling()) && !backString.contains("\\end{") && !isSectionStyle)
      writer << "\\\\";

    if (isSectionStyle || hasNextSibling(node)) /// not the last leaf in a block
      writer << wrt::endl; /// terminate line
  }
}

bool TexConverter::getSectionStatus() {
  bool isSectionStyle = false;

  if (sectionStack_.size()) {
    isSectionStyle = sectionStack_.back();
    sectionStack_.pop_back();
  }

  return isSectionStyle;
}

DOMString TexConverter::getBackString() {
  DOMString returnString;

  if (endStack_.size()) {
    returnString = endStack_.back();
    endStack_.pop_back();
  }

  return returnString;
}

DOMString TexConverter::getContentString(Node& node) {
  StringWriter writer;
  traverse(node, writer, false);  // ignore the node itself

  return writer.getString();
}

DOMString TexConverter::getLabel(Node& node) {
  DOMString returnString = "{\\bfseries (";

  Node& parentNode = node.getParentNode();
  if (parentNode.getNodeType() == Node::ELEMENT_NODE) {
    Element& parentElement = static_cast<Element &>(parentNode);
    returnString += parentElement.getNodeName();

    DOMString attributeString = parentElement.getAttribute("id");
    if (!attributeString.empty()) {
      returnString += "#";
      returnString += attributeString;
    }
    else {
      attributeString = parentElement.getAttribute("class");
      if (!attributeString.empty()) {
        returnString += ".";
        returnString += attributeString;
      }
    }
  }

  returnString += ">";
  returnString += node.getNodeName();
  returnString += ")} ";

  return returnString;
}

}
