/***************************************************************************
                          tokeninputstream.cpp  -  description
                             -------------------
    begin                : Fri Apr 25 2003
    copyright            : (C) 2003 by John Redmond
    email                : john_redmond@optusnet.com.au
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <list>
#include "tokenstream.h"
#include "reader.h"
#include "writer.h"
#include "iobase.h"
#include "definitions.h"

using namespace std;

/**
 * TokenStream, the stream tokenizer used by text parsers.
 * <P>
 * TokenStream is a general purpose tokenizer, in that it can be configured
 * (on the fly, if necessary) to recognize any SINGLE-CHARACTER token. In
 * summary, when nextToken() is called, its behaviour is:
 * 1. If a defined token is read from the input, it is returned as a string.
 * 2. If a non-token character is read, it is added to a token buffer and
 *    reading continued within a loop. Two situations now apply:
 *  a. if token character is SUBSEQUENTLY read, this is returned to the input
 *      by unget(c) and the already buffered characters are returned as a string;
 *  b. if a  whitespace character is subsequently read, this character is now
 *      cached and the already buffered characters returned as a string;
 *  b. if end of file is SUBSEQUENTLY reached, the buffered characters are
 *      returned as a string.
 * 3. A whitespace character is never returned by nextToken().
 * <P>
 * The above behaviour is sufficient to read tags, but a special method,
 * getDelimitedString(DOMString&), is required for (eg) Comment and CDATASection nodes,
 * which have specific end substrings ("-->" and "]]>"). Note that it
 * continues to read through, and accumulate, whitespace chars.
 * A related version, getDelimitedString(char), looks for a single char
 * and is much more efficient.
 * <P>
 * Methods are provided for altering the token chars on the fly. For example,
 * the tokens are "<>/?" for top-level XML parsing and "<>=\"\'\\" for
 * parsing of attributes within Element start tags. For DTD parsing,
 * different tokens are defined: "<>()=,|*+?!#\"\'%&;"
 * <P>
 * Where the data source is specified by a string, for a file name,
 * TokenStream uses a Reader to attempt to resolve the source and
 * provide a stream of characters, with -1 as the eof mark.
 *
 * @see XMLParser
 * @see PunctuationParser
 * @see DTDParser
 * @see Path
 * @see DefaultStream
 *
 * @author John Redmond
 */

namespace limpid {

TokenStream::TokenStream() {
  initialize();
}

TokenStream::TokenStream(const DOMString& fileName) {
  initialize();
  input_ = new Reader(fileName);
  if (!(input_->isValid()))
    wrt::Console << "Could not set up input from " << fileName << wrt::endl;
}

TokenStream::TokenStream(const URL& url) {
  initialize();
  input_ = new Reader(url);
  if (!(input_->isValid()))
    wrt::Console << "Could not set up input from " << url.getPath() << wrt::endl;
}

TokenStream::TokenStream(CharacterReader& inp) {
  initialize();
  input_ = &inp.copy();
}

DOMString& TokenStream::alignOn(const DOMString& delimSet) {
  while (true) {
    DOMString &nextString = peekToken();
    if (verbose_) wrt::Console << "ALIGN [" << nextString << "]" << wrt::endl;
    if (!nextString.size()) return nextString;
    else if (delimSet.find(nextString) != DOMString::npos) return nextString;
    else bump();
  }
}

DOMString& TokenStream::alignOn(const char* delimSet) {
  DOMString delimString = delimSet;
  return alignOn(delimString);
}

void TokenStream::setDelimiters(const DOMString& delims) {
  delimiters = delims;
}

DOMString& TokenStream::peekToken() {
  if (dirty_) {
    getToken();
    dirty_ = false;
  }
    if (verbose_) wrt::Console << returnString.size() << " PEEK TOKEN ["\
     << returnString << "]" << wrt::endl;
    return returnString;
}

DOMString& TokenStream::nextToken() {
  if (dirty_) getToken();
  dirty_ = true;

  if (verbose_) wrt::Console << returnString.size() << " NEXT TOKEN ["\
   << returnString << "] using delimiters ^" << delimiters << "^" << wrt::endl;

  return returnString;
}

void TokenStream::bump() {
  nextToken();
}

DOMString& TokenStream::getWhitespace() {
  return whitespace;
}

void TokenStream::getToken() {
  whitespace.clear();
  returnString.clear();
  bool buildingStarted = false;

  while (true) {
    Char c = getChar();
    if (c == FAILCHAR) break;
    else if (c == 0) continue;

    if (delimiters.find(c) != DOMString::npos) {
      if (verbose_) wrt::Console << delimiters << " UNGET [" << c << "]" << wrt::endl;
      if (buildingStarted) unget(c);  // it will be the next token
      else {
        returnString += c; // this is the token
        buildingStarted = true;
      }
      break;
    }
    // break on white space
    else if (wsChars.find(c) != DOMString::npos) {
      if (!buildingStarted) whitespace += c;
      else {
        unget(c);
        break;
      }
    }
    // else add the character to the return string
    else {
      returnString += c;
      buildingStarted = true;
    }
  }

  if (verbose_) wrt::Console << returnString.size() << " GET TOKEN ["\
     << returnString << "]" << wrt::endl;
  if (!returnString.size()) returnString = failString;
}

DOMString& TokenStream::getDelimitedString(const DOMString& endString) {
  int startLineNo = lineNo_;
  textString.clear();

  size_t endSize = endString.size();
  textString.clear();
  if (!endSize) return textString;

  Char endChar = endString[endSize - 1];
  int endLength = endString.size();

  while (true) {
    Char c = getChar();
    if (c == FAILCHAR) {
      wrt::Errors << "FATAL INPUT ERROR WITH DELIMITER [" << endString\
       << "] on line " << startLineNo << wrt::endl;
      break;
    }
    if (c != 0) textString += c;
    if (c != endChar) continue;
    if (textString.endsWith(endString)) {
      int totalLength = textString.size();
      textString.resize(totalLength - endLength); ///
      break;
    }
  }
  if (verbose_) wrt::Console << "from[" << endString << "] DELIM read: [" << textString\
   << "]" << wrt::endl;
  return textString;
}

DOMString& TokenStream::getDelimitedString(Char delimiter) {
  int startLineNo = lineNo_;
  textString.clear();

  while (true) {
    Char c = getChar();
    if (c == FAILCHAR) {
      wrt::Console << "SERIOUS INPUT ERROR WITH DELIMITER [" << delimiter\
       << "] on line " << startLineNo << wrt::endl;
      break;
    }
    else if (c == delimiter) break;
    else textString += c;
  }
  if (verbose_) wrt::Console << "*(" << delimiter << ")* TS DELIM read: [" << textString << "]" << wrt::endl;
  return textString;
}

DOMString& TokenStream::getText() {
  textString.clear();
  int count = 0;
  while (true) {
    char c = getChar();
    if (c == FAILCHAR) break;
    else if (c == '<') {
      unget(c);
      break;
    }
    else {
      textString += c;
      ++count;
    }
  }
  if (verbose_)
    wrt::Console << count << " TEXT read: [" << textString << "]"\
   << textString.size() << wrt::endl;
  return textString;
}

void TokenStream::getTokens() {
  while (true) {
    DOMString token = nextToken();
    if (!token.size()) break;
  }
}

void TokenStream::initialize() {
  verbose_ = false;
  dirty_ = true;
  lineNo_ = 1;

  wsChars = DOMString(" \t\n\r");
  setProcessXML();
}

}
