/***************************************************************************
                          tokeninputstream.h  -  description
                             -------------------
    begin                : Fri Apr 25 2003
    copyright            : (C) 2003 by John Redmond
    email                : john_redmond@optusnet.com.au
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef TOKENSTREAM_H
#define TOKENSTREAM_H

#include <iostream>
#include <fstream>
#include <string>
#include <list>
#include "definitions.h"
#include "copied.h"
#include "node.h"
#include "characterreader.h"
#include "iobase.h"
#include "pointer.h"

#include "writer.h"

/**
 * TokenStream, the stream tokenizer used by text parsers.
 * 
 * TokenStream is a general purpose tokenizer, in that it can be configured
 * (on the fly, if necessary) to recognize any SINGLE-CHARACTER token. In
 * summary, when nextToken() is called, its behaviour is:
 * 1. If a defined token is read from the input, it is returned as a string.
 * 2. If a non-token character is read, it is added to a token buffer and
 *    reading continued within a loop. Two situations now apply:
 *  a. if token character is SUBSEQUENTLY read, this is returned to the input
 *      by unget() and the already buffered characters are returned as a string;
 *  b. if a  whitespace character is subsequently read, this character is now
 *      cached and the already buffered characters returned as a string;
 *  b. if end of file is SUBSEQUENTLY reached, the buffered characters are
 *      returned as a string.
 * 3. A whitespace character is never returned by nextToken().
 * 
 * The above behaviour is sufficient to read tags, but a special method,
 * readUntilAndReject(), is required for (eg) Comment and CDATASection nodes,
 * which have specific end substrings ("-->" and "]]>"). Note that it
 * continues to read through, and accumulate, whitespace chars.
 * 
 * Methods are provided for altering the token chars on the fly. For example,
 * the tokens are "<>/?" for top-level XML parsing and "<>=\"\'\\" for
 * parsing of attributes within Element start tags. For DTD parsing,
 * different tokens are defined: "<>()=,|*+?!#\"\'%&;"
 * 
 * Where the data source is specified by a string, for a file name or url,
 * TokenStream uses an EncodedReader to attempt to resolve the source.
 *
 * @see XMLParser
 * @see PunctuationParser
 * @see DTDParser
 * @see Path
 * @see DefaultStream
 *
 * @author John Redmond
 */

namespace limpid {
  class TokenStream: public Copied {
  public:
    TokenStream();
  	TokenStream(const DOMString& fileName);
  	TokenStream(const URL& url);
    TokenStream(CharacterReader& inp);

    void setProcessXML() { delimiters = "<"; }
    void setProcessTag() { delimiters = "<>[]=/\"\'?"; }
    void setProcessAttributes() { delimiters = "<>=\"\'\\"; }
    void setProcessDTD() { delimiters = "<>()=,|*+?!#\"\'%&;"; }
    void setDelimiters(const DOMString& delimiters);
    void setWhitespaceChars(const DOMString& chars) { wsChars = chars; }
    void setVerbose(bool v) { verbose_ = v; }
    void setEncoding(IOBase::Encoding encoding) { input_->setEncoding(encoding); }
    void setFailToken(DOMString& fString) { failString = fString; }
    void setFailToken(const DOMString& fString) { failString = fString; }
    bool isValid() { return input_->isValid(); }

    bool isFailToken(DOMString& fString) { return fString == failString; }

    CharacterReader& getHandler() { return *input_; }
    DOMString& nextToken();
    DOMString& peekToken();
    void unget() { input_->unget(); }
    void bump();
    DOMString& alignOn(const DOMString& delimSet);
    DOMString& alignOn(const char* delimSet);
    DOMString& getWhitespace();
    void clearWhitespace() { whitespace.clear(); }
    DOMString& getDelimitedString(const DOMString& endString);
    DOMString& getDelimitedString(Char delimiter);
    DOMString& getText();
    int getLineNo() { return lineNo_; }
    int getCharCount() { return charCount_; }
    void getTokens();
    TokenStream& copy() { return *(new TokenStream(*this)); }

  private:
    Pointer<CharacterReader>input_;
    bool verbose_, dirty_;
    int lineNo_ , charCount_;
    DOMString returnString, failString, textString, whitespace,
      delimiters, tokenString, wsChars;

    Char getChar() {
      Char c = input_->nextChar();
      ++charCount_;
      if (c == '\n') ++lineNo_;
      return c;
    }

    void unget(Char c) {
      if (c == '\n') --lineNo_;
      input_->unget();
    }

    void getToken();
    void initialize();
  };
}
#endif
