/***************************************************************************
                          url.cpp  -  description
                             -------------------
    begin                : Sat Nov 1 2003
    copyright            : (C) 2003 by John Redmond
    email                : john_redmond@optusnet.com.au
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "url.h"
#include <fstream>
#include <sstream>
#include "writer.h"

namespace limpid {

URL::URL(bool toConsole):
  protocol_(toConsole? IOBase::CONSOLE: IOBase::ERRORS), valid_(true) {}

URL::URL(const DOMString& urlString): valid_(true) {
  setContent(urlString);
}

URL::URL(const DOMString& refPath, const DOMString& fileName): valid_(true) {
  DOMString path = completeBasePath(refPath);
  path += fileName;
  setContent(path);
}

URL::URL(const URL& baseURL, const DOMString& fileName) {
  copyContent(baseURL, fileName);
}

std::istream& URL::getInput() {
  if (protocol_ == CONSOLE || protocol_ == ERRORS) { // no error input!
    return std::cin;
  }
  else {
    const char* sourceString = path_.getCString();
    std::ifstream *fileInput = new std::ifstream(sourceString);
    delete [] sourceString;

    if (fileInput->good()) {
      return *fileInput;
    }
    else {
      fileInput->close();
      wrt::Errors << "BAD INPUT AT: " << path_ << wrt::endl;
      valid_ = false;
      exit(1);
      return std::cin;
    }
  }
}

std::ostream& URL::getOutput() {
  valid_ = true;

  if (protocol_ == CONSOLE)
    return std::cout;
  if (protocol_ == ERRORS)
    return std::cerr;

  // we reach here if there is to be a file stream
  const char* destString = path_.getCString();
  unlink(destString);

  std::ofstream *fileOutput = new std::ofstream(destString);
  delete [] destString;

  if (fileOutput->good()) {
    return *fileOutput;
  }
  else {
    wrt::Errors << "BAD OUTPUT AT: " << path_ << wrt::endl;
    fileOutput->close();
    valid_ = false;
    exit(1);
  }

  return std::cout;
}

void URL::setContent(DOMString urlString) {
  if (urlString.trim().empty()) {
    protocol_ = IOBase::CONSOLE;
    return;
  }

  size_t hashIndex = urlString.indexOf("#");
  if (hashIndex != urlString.npos) {
    fragmentPath_ = "/";
    fragmentPath_ += urlString.substring(hashIndex + 1, urlString.size() - hashIndex - 1);
    urlString = urlString.substring(0, hashIndex);
  }

  DOMString ps = "://";
  size_t pos = urlString.find(ps);

  if (pos == urlString.npos) {
    protocol_ = IOBase::FILE;
    path_ = urlString;
  }
  else {  // these other options are not yet supported as they are system-dependent
    path_ = urlString.substring(pos + 3, urlString.size() - pos - 3);
    DOMString pstring = urlString.substring(0, pos);
    if (pstring == "http")
      protocol_ = IOBase::HTTP;
    else if (pstring == "ftp")
      protocol_ = IOBase::FTP;
    else protocol_ = IOBase::FILE;
  }

  delimiter_ = getDelimiter(path_);
  size_t lastPos = path_.find_last_of(delimiter_);
  if (lastPos != path_.npos)
    basePath_ = path_.substring(0, lastPos + 1); // include the last delimiter
}

Char URL::getDelimiter(DOMString& path) {
  size_t lastPos = path.find_last_of('\\');
  if (lastPos != path.npos)
    return '\\';
  else return '/';
}

DOMString URL::completeBasePath(DOMString refPath) {
  Char delimiter = getDelimiter(refPath);
  if (!refPath.endsWith(delimiter))
    refPath += delimiter;

  return refPath;
}

void URL::copyContent(const URL& baseURL, const DOMString& fileName) {
  if (fileName.startsWith("/") ||
      fileName.startsWith("http://") ||
      fileName.startsWith("ftp://")) {
    setContent(fileName);
    return; // the path is complete, so ignore baseURL
  }

  // otherwise the path is relative to the base path in the baseURL
  protocol_ = baseURL.protocol_;
  delimiter_ = baseURL.delimiter_;
  basePath_ = baseURL.basePath_;
  path_ = basePath_;
  path_ += fileName;
}

}
