/***************************************************************************
                          whitespacemanager.cpp  -  description
                             -------------------
    begin                : Wed Feb 9 2005
    copyright            : (C) 2005 by John Redmond
    email                : john_redmond@optusnet.com.au
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "whitespacemanager.h"
#include "xslutility.h"
#include "writer.h"
#include "liststream.h"
#include "stringreader.h"
#include "tokenstream.h"
#include "nodestream.h"
#include "endnode.h"
#include "domwriter.h"
#include "errorhandler.h"
#include "liststream.h"
#include "nodestream.h"
#include "nametest.h"

namespace limpid {

WhitespaceManager::WhitespaceManager(): preserves_(new NameSet),
      strips_(new NameSet), preserveTests_(new AndTest)  {  // for testing
}

WhitespaceManager::WhitespaceManager(Document& document): preserves_(new NameSet),
      strips_(new NameSet), preserveTests_(new AndTest)   {
  configure(document);
}

void WhitespaceManager::removeTemplateWhitespace(Node& root) {
  NodeList removeList;
  NodeStream stream(root, false, false); // no end nodes, exclude root

  // collect all empty text nodes
  while (true) {
    Node& node = stream.nextNode();
    if (!node.isValid()) break;

    if (node.getNodeType() != Node::TEXT_NODE) continue;

    Node& parentNode = node.getParentNode();
    if (parentNode.getLocalName() == "text" &&
        XSLUtility::isXSLElement(parentNode)) {
      continue; // keep the text
    }
    
    if (node.getValue().trim().empty())  // this resolves entities
      removeList.addNode(node);
  }

  // now remove the empty text nodes from the document tree
  ListStream removeStream(removeList);

  while (true) {
    Node& node = removeStream.nextNode();
    if (!node.isValid()) break;

    Node& parentNode = node.getParentNode();
    if (!parentNode.isValid()) continue; // impossible

    if (!XSLUtility::isXSLElement(parentNode)) {
      parentNode.removeChild(node);
      continue;
    }

    DOMString nodeName = XSLUtility::getNodeName(static_cast<Element &>(parentNode));
    if (nodeName != "text")
      parentNode.removeChild(node);
  }
}

void WhitespaceManager::removeSourceWhitespace(Node& root) {
  NodeList deleteList;
  NodeStream stream(root, false, false); // exclude end nodes and include root

  while (true) {
    Node& node = stream.nextNode();
    if (!node.isValid()) break;

    if (node.getNodeType() != Node::TEXT_NODE) continue;
    if (node.getValue().trim().empty()) deleteList.addNode(node);
  }

  ListStream deleteStream(deleteList);  // remove selected nodes

  while (true) {
    Node& node = deleteStream.nextNode();
    if (!node.isValid()) break;

    Node& parentNode = node.getParentNode();
    if (!parentNode.isValid()) continue;

    if (!preserveTests_->test(parentNode))
      parentNode.removeChild(node);
  }
}

void WhitespaceManager::configure(Document& sheetDocument) {
  Element& documentElement = sheetDocument.getDocumentElement();

  if (!documentElement.isValid()) return;
    configureSheet(documentElement);
}

void WhitespaceManager::configureSheet(Element& sheetElement) {
  ListStream stream(sheetElement);

  while (true) {
    Node& node = stream.nextNode();
    if (!node.isValid()) break;

    if (isXSLElement(node)) {
      Element element = static_cast<Element &>(node);
      DOMString nodeName = getNodeName(element);

      if (nodeName == "stylesheet") // recurse
        configureSheet(element);
      else
        configureWhitespaceElement(element);
    }
  }
}

void WhitespaceManager::configureWhitespaceElement(Element& element) {
  DOMString nodeName = XSLUtility::getNodeName(element);

  if (nodeName == "preserve-space") {
    DOMString elements = element.getAttribute("elements");
    if (!elements.empty())
      addPreserves(elements);
    else
      ErrorHandler::report(element, "MISSING elements attr.");
  }
  else if (nodeName == "strip-space") {
    DOMString elements = element.getAttribute("elements");
    if (!elements.empty())
      addStrips(elements);
    else
      ErrorHandler::report(element, "MISSING elements attr.");
  }

  setPreserveTests();
}

void WhitespaceManager::addPreserves(DOMString& elements) {
  StringReader reader(elements);
  TokenStream tis(reader);

  while (true) {
    DOMString token = tis.nextToken();
    if (token.empty()) break;

    NameSet::iterator p = strips_->find(token);
    if (p != strips_->end()) strips_->erase(p);

    preserves_->insert(token);
  }
}

void WhitespaceManager::addStrips(DOMString& elements) {
  StringReader reader(elements);
  TokenStream tis(reader);

  while (true) {
    DOMString token = tis.nextToken();
    if (token.empty()) break;

    NameSet::iterator p = preserves_->find(token);
    if (p != strips_->end()) preserves_->erase(p);

    strips_->insert(token);
  }
}

void WhitespaceManager::setPreserveTests() {
  NameSet::iterator p = preserves_->begin();
  while (p != preserves_->end()) {
    DOMString nodeName = *p;
    NameTest nameTest(nodeName);
    preserveTests_->addTest(nameTest);
    ++p;
  }
}

}
