/***************************************************************************
 *   Copyright (C) 2007 by John Redmond   *
 *   john_redmond@optusnet.com.au   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "xmlmapper.h"
#include "domwriter.h"
#include "liststream.h"

namespace limpid {

XMLMapper::XMLMapper(): stringMap_(new StringMap), stringSet_(new StringSet) {
  setDefaultMappings();
}

Document& XMLMapper::processDocument(Document& document) {
  document_ = Document(); // refresh in case it is necessary

  Element documentElement = getDocumentElement();
  document_.appendChild(documentElement);
  appendTemplates(document.getDocumentElement(), documentElement);

  return document_;
}

void XMLMapper::setDefaultMappings() {
  setDivs();
  setDefaultGroup("T", "h");
  setDefaultGroup("H", "h");
  setMapping("L0", "div id =\"L0");
  setInvariants("M", "p");
  setInvariants("D", "");
  setInvariants("B", "b");
  setInvariants("S", "em");
  setInvariants("F", "em");
  setInvariants("X", "em");
  setInvariants("m", "em");
  setInvariants("E", "p");
  setInvariants("FAIL", "em");
}

void XMLMapper::setDefaultGroup(const DOMString& keyStart, const DOMString& resultStart) {
  for (int index = 1; index < 9; ++index) {
    DOMString indexString = Converter::toString(index);
    DOMString key = keyStart;
    key += indexString;

    DOMString result = resultStart;
    result += indexString;

    setMapping(key, result);
  }
}

void XMLMapper::setDivs() {
  DOMString keyStart = "L";
  DOMString resultStart = "div class =\"L";
  for (int index = 1; index < 9; ++index) {
    DOMString indexString = Converter::toString(index);
    DOMString key = keyStart;
    key += indexString;

    DOMString result = resultStart;
    result += indexString;
    result += "\"";

    setMapping(key, result);
  }
}

void XMLMapper::setInvariants(const DOMString& keyStart, const DOMString& result) {
  for (int index = 1; index < 9; ++index) {
    DOMString indexString = Converter::toString(index);
    DOMString key = keyStart;
    key += indexString;

    setMapping(key, result);
  }
}

void XMLMapper::setMapping(const DOMString& key, const DOMString& value) {
  (*stringMap_)[key] = value;
}

Element XMLMapper::getDocumentElement() {
  Element documentElement("xsl:stylesheet");
  documentElement.setAttribute("xmlns:xsl", "http://www.w3.org/1999/XSL/Transform");
  documentElement.setAttribute("version", "1.1");

  Element outputElement("xsl:output");
  outputElement.setAttribute("method", "xml");
  outputElement.setAttribute("doctype-public", "-//W3C//DTD XHTML 1.0 Transitional//EN");
  documentElement.appendChild(outputElement);

  return documentElement;
}

void XMLMapper::setTags(Document& sourceDocument) {
  ListStream listStream(sourceDocument.getDocumentElement());

  while (true) {
    Node& node = listStream.nextNode();
    if (!node.isValid()) break;

    if (node.getNodeType() != Node::ELEMENT_NODE) continue;

    Element& element = static_cast<Element &>(node);
    DOMString typeName = element.getAttribute("type");
    DOMString tagString = (*stringMap_)[typeName];

    element.setAttribute("tag", tagString);
  }
}

void XMLMapper::appendTemplates(Element& sourceElement, Element& documentElement) {
  ListStream listStream(sourceElement);
  while (true) {
    Node& node = listStream.nextNode();
    if (!node.isValid()) break;

    if (node.getNodeType() != Node::ELEMENT_NODE) continue;
    Element& element = static_cast<Element &>(node);
    DOMString typeName = element.getAttribute("type");
    DOMString pathName = element.getAttribute("path");

    if (typeName.startsWith("=")) {
      DOMString elementName = typeName.substr(1); // it is a literal tag name
      appendTemplate(pathName, elementName, documentElement);
      continue;
    }

    if (typeName.startsWith("L")) { // we have a div
      Element templateElement("xsl:template");
      documentElement.appendChild(templateElement);
      templateElement.setAttribute("match", pathName);

      Element divElement("div");
      if (typeName == "L0")
        divElement.setAttribute("id", typeName);
      else
        divElement.setAttribute("class", typeName);

      templateElement.appendChild(divElement);
      addElement(divElement, "xsl:apply-templates");
      continue;
    }

    if (typeName == "-1") continue;// it is an empty element

    DOMString tagName = (*stringMap_)[typeName];
    if (tagName.empty()) tagName = "p";
    appendTemplate(pathName, tagName, documentElement);
  }

  appendTextTemplate(documentElement);
}

void XMLMapper::appendTemplate(const DOMString& pathName, const DOMString& tagName, Element& parentElement) {
  Element templateElement("xsl:template");
  parentElement.appendChild(templateElement);

  if (!tagName.empty()) {
    templateElement.setAttribute("match", pathName);
    Element& namedElement = addElement(templateElement, tagName);
    addElement(namedElement, "xsl:apply-templates");
  }
  else
    addElement(templateElement, "xsl:apply-templates");
}

void XMLMapper::appendTextTemplate(Element& parentElement) {
  Element& templateElement = addElement(parentElement, "xsl:template");
  templateElement. setAttribute("match", "text()");

  Element& beforeElement = addElement(templateElement, "xsl:if");
  beforeElement.setAttribute("test", "preceding-sibling::*");
  addXSLText(beforeElement, " ");

  Element& valueElement = addElement(templateElement, "xsl:value-of");
  valueElement.setAttribute("select", "escape-tex(.)");

  Element& afterElement = addElement(templateElement, "xsl:if");
  afterElement.setAttribute("test", "following-sibling::*");
  addXSLText(afterElement, " ");
}

void XMLMapper::addXSLText(Element& parentElement, const DOMString& value) {
  if (!value.empty()) // ignore an empty string
    addLeaf("xsl:text", value, parentElement);
}

}
