/***************************************************************************
                          xmlparser.cpp  -  description
                             -------------------
    begin                : Tue Mar 18 2003
    copyright            : (C) 2003 by John Redmond
    email                : john_redmond@optusnet.com.au
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <iostream>
#include "xmlparser.h"
#include "node.h"
#include "element.h"
#include "documenttype.h"
#include "text.h"
#include "comment.h"
#include "cdatasection.h"
#include "processinginstruction.h"
#include "xmldeclaration.h"
#include "endnode.h"
#include "systemmanager.h"
#include "tokenstream.h"
#include "converter.h"
#include "stringreader.h"
#include "tokenstream.h"

namespace limpid {

XMLParser::XMLParser(const DOMString& fileName, bool isHTML):
      reader_(new Reader(fileName)), tis_(new TokenStream(*reader_)),
      emptySet_(isHTML ? new StringSet: 0), processingXSLT_(false),
      isHTML_(isHTML) {
  setConstants();
}

XMLParser::XMLParser(CharacterReader& charStream, bool isHTML):
      reader_(&(charStream.copy())), tis_(new TokenStream(*reader_)),
      emptySet_(isHTML ? new StringSet: 0), processingXSLT_(false),
      isHTML_(isHTML) {
  setConstants();
}

void XMLParser::setSource(const DOMString& sourceName) {
  reader_ = new Reader(sourceName);
}

void XMLParser::setConstants() {
  delimiterString_ = "'\"";
  trimText_ = SystemManager::getSystemManager().isTrimmingText();
  normalizeText_ = SystemManager::getSystemManager().isNormalizingText();
  if (isHTML_)
    setEmptySet();
  node_ = 0;
}

Node& XMLParser::nextNode() {
  DOMString s = tis_->peekToken();
  if (s.empty()) return Node::failNode;
  if (s == "<") buildTag();
  else buildTextNode(s);

  node_->setPosition(tis_->getLineNo());
  tis_->setProcessXML();

  return *node_;
}

void XMLParser::buildTag() {
  if (processingXSLT_) {
    DOMString ws = tis_->getWhitespace();
    if (!ws.empty()) {
      setNode(new Text(ws));
      tis_->clearWhitespace();
      node_->setPosition(tis_->getLineNo());
      return;
    }
  }

  tis_->bump();
  tis_->setProcessTag();

  DOMString name = tis_->nextToken();

  // consider special cases
  if (name == ("!DOCTYPE"))
    buildDoctype();
  else if (name == "/") completeEndTag();
  else if (name == "?") buildPI();
  else if (name.startsWith("!--")) buildComment(name);
  // otherwise we have a start tag for a comment, CDATA, doctype or element
  else buildNamedTag(name);
}

void XMLParser::buildNamedTag(DOMString& name) {
  if (name == "!") {
    DOMString tag = tis_->nextToken();
    if (tag == "[") {
      buildCDATASection();
    }
  }

  else completeStartTag(name);
}

void XMLParser::completeStartTag(DOMString& name) {
  Element *element = new Element(name);
  addAttributes(*element, name);
  setNode(element);
}

void XMLParser::completeEndTag() {
  DOMString name = tis_->nextToken();
  EndNode* endNode = new EndNode(name);
  setNode(endNode);
  swallow(">");
}
void XMLParser::buildTextNode(DOMString& text) {
  tis_->bump();
  tis_->setProcessXML();

  DOMString result;
  if (!trimText_ && !normalizeText_) result += tis_->getWhitespace();
  result += text;
  result += tis_->getText();

  if (!processingXSLT_) {
    if (normalizeText_) result = result.normalizeWhitespace();
    else if (trimText_) result = result.trim();
  }

  setNode(new Text(result));
}

void XMLParser::buildPI() {
  DOMString target = tis_->nextToken();
  if (target == "xml")
    buildXMLDeclaration();
  else {
    DOMString content = getDelimitedString(">");
    setNode(new ProcessingInstruction(target, content));
  }
}

void XMLParser::buildXMLDeclaration() {
  tis_->setProcessAttributes();

  Element holder("HLDR");
  addAttributes(holder, "HLDR"); // kludge to parse out the declaration details
  DOMString version = holder.getAttribute("version");
  DOMString encodingString = holder.getAttribute("encoding");
  DOMString standalone = holder.getAttribute("standalone");
  setNode(new XMLDeclaration(version, encodingString, standalone));

  if (!encodingString.empty())
    tis_->setEncoding(IOBase::getEncoding(encodingString));

  tis_->setProcessXML();
}

void XMLParser::buildComment(DOMString& commentString) {
  commentString = commentString.substring(3, commentString.length() - 3);
  commentString += getDelimitedString("-->");
  setNode(new Comment(commentString));
}

void XMLParser::buildCDATASection() {
  getDelimitedString("[");  // skip the "CDATA["
  DOMString cdsString = getDelimitedString("]]>");

  setNode(new CDATASection(cdsString));
}

void XMLParser::buildDoctype() {
  tis_->setProcessTag();

  DOMString name = tis_->nextToken();
  DOMString emptyString = "";
  DOMString publicId, systemId, internalSubset;

  DOMString& nextToken = tis_->nextToken();

  if (nextToken == "SYSTEM") {
    systemId = getUnquotedString();
    nextToken = tis_->nextToken();
  }

  else if (nextToken == "PUBLIC") {
    publicId = getUnquotedString();
    systemId = getUnquotedString();
    nextToken = tis_->nextToken();
  }

  // now look for an internal subset
  if (nextToken == "[")
    internalSubset = getDelimitedString("]");

  if (nextToken != ">")
      swallow(">");

  DocumentType* doctype = new DocumentType(name, publicId, systemId);
  setNode(doctype);

  doctype->setInternalSubset(internalSubset);
  URL sourceURL = reader_->getURL();
  doctype->setSourceURL(sourceURL); // needed for dtd access

  tis_->setProcessXML();
}

void XMLParser::addAttributes(Element& element, const DOMString& name) {
  tis_->setProcessAttributes();
  while (true) {
    DOMString nextToken = tis_->peekToken();
    if (nextToken.empty()) break;

    if (nextToken == "/") {
      tis_->bump(); //"/";
      tis_->bump(); //">"
      break;
    }
    else if (nextToken == ">") { // end of element attributes
      tis_->bump();
      if (!isEmpty(name)) element.setIsParent(true);
      break;
    }
    else if (nextToken == "?") {  // end of xml declaration
      tis_->bump();
      tis_->bump();
      break;
    }

    // this is non-DOM for the addition of detail to an XML declaration
    // as well as setting of attributes on an element
    DOMString name = tis_->nextToken();
    DOMString equate = tis_->nextToken();
    if (equate != "=") {
      getDelimitedString(">");
      DOMString next = tis_->nextToken(); // bail out
      break;
    }
    else {
      DOMString delimiter = tis_->nextToken();
      if (delimiter != "'" && delimiter != "\"") {
        alignOn(">");
        DOMString next = tis_->nextToken(); // bail out

        DOMString errorDesc = " bad delimiter for string [";
        errorDesc += (DOMString)delimiter;
        errorDesc += "], on line ";
        errorDesc += Converter::toString(int(tis_->getLineNo()));
        reportError(DOMException::IO_ERR, errorDesc);

        break;
      }
    DOMString value = tis_->getDelimitedString(delimiter[0]);

    Attr attribute(name);
    attribute.setPosition(tis_->getLineNo());
    attribute.setValue(value);
    element.setAttribute(name, value);
    }
  }

  tis_->setProcessXML();
}

void XMLParser::setNode(Node* node) {
// it is safe to delete the pointer if:
// 1. There IS a pointer to delete (!= 0) or
// 2. it has been appended to a parent (and its refcount > 1 or
// 3. it has not been appended and its "parent" is invalid.
if (node_ && (node_->getRefCount() > 1 || !node->getParentNode().isValid()))
  delete(node_);
  node_ = node; // this fits all nodes
}

DOMString XMLParser::getUnquotedString() {
  DOMString quote = tis_->nextToken(); // assume it is an opening quote
  DOMString result = tis_->getDelimitedString(quote);
  return result;
}

void XMLParser::reportError(int type, DOMString& desc) {
  Node::systemManager.processResponse(type, desc);
}

void XMLParser::setEmptySet() {
  DOMString emptyTags =
    "area base basefont br col frame hr img input isindex link meta param ";

  StringReader stringReader(emptyTags);
  TokenStream tis(stringReader);

  while (true) {
    DOMString token = tis.nextToken();
    if (token.empty()) break;

    emptySet_->insert(token);
    emptySet_->insert(token.toUpperCase());
  }
}

}
