/***************************************************************************
 *   Copyright (C) 2007 by root   *
 *   john_redmond@optusnet.com.au   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "xmlwrapper.h"
#include "element.h"
#include "text.h"
#include "xmldeclaration.h"
#include "documenttype.h"
#include "nodestream.h"
#include "reader.h"
#include "stringwriter.h"

namespace limpid {

Document& XMLWrapper::processDocument(Document& document) {
  Element& documentElement = document.getDocumentElement();
  if (!documentElement.isValid()) return document; // surely not

  if (documentElement.getFirstChild().getNodeName() == "head")
    buildHeadedDocument(document);
  else buildSimpleDocument(document);

  return document;
}

void XMLWrapper::buildHeadedDocument(Document& document) {
  Element& documentElement = document.getDocumentElement();
  Element htmlElement("html");

  // add document head
  htmlElement.appendChild(documentElement.getFirstChild());

  // now we need a body
  if (documentElement.getNodeName() == "body") {
    htmlElement.appendChild(documentElement);
    document.appendChild(htmlElement);
  }
  else {
    Element bodyElement("body");
    htmlElement.appendChild(bodyElement);
    copyContent(documentElement, bodyElement);
  }
  // finally, replace the original document element
  document.replaceChild(htmlElement, documentElement);

  setStyles(document);
  setProlog(document);
}

void XMLWrapper::buildSimpleDocument(Document& document) {
  setBody(document);
  setHead(document);
  setStyles(document);
  setProlog(document);
}

void XMLWrapper::setProlog(Document& document) {
  XMLDeclaration xmlDeclaration("1.0", "UTF-8", "");
  document.insertBefore(xmlDeclaration, document.getDocumentElement());
  DocumentType documentType("html", "-//W3C//DTD_XHTML 1.0 Strict//EN","http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd");
  document.insertBefore(documentType, document.getDocumentElement());
}

void XMLWrapper::setBody(Document& document) {
  //make sure that the document element is called "body"
  Element& documentElement = document.getDocumentElement();
  if (documentElement.getNodeName() == "body") return; // all OK

  Element bodyElement("body");
  copyContent(documentElement, bodyElement);

  document.replaceChild(bodyElement, documentElement);
}

void XMLWrapper::setHead(Document& document) {
  // enclose all content in an "html" document element
  Element& documentElement = document.getDocumentElement();
  Element htmlElement("html");

  // there is no defined head, so create one and attempt to give it a title
  Element headElement("head");
  htmlElement.appendChild(headElement); // the head
  addTitle(documentElement, headElement);
  htmlElement.appendChild(documentElement); // this is the body

  document.appendChild(htmlElement);

}

void XMLWrapper::addTitle(Element& sourceElement, Element& headElement) {
  NodeStream childStream(sourceElement, false, false);
  // stream down through the document, looking for some sort of title
  while (true) {
    Node& node = childStream.nextNode();
    if (!node.isValid()) break;

    DOMString nodeName = node.getNodeName();
    if (nodeName == "h1" || nodeName == "h2" || nodeName == "h3") {
      addLeaf("title", node.getValue(), headElement);
      return;
    }
  }

  // there is no title there
  addLeaf("title", "untitled", headElement);
}

void XMLWrapper::setStyles(Document& document) {
  Element& headElement = static_cast<Element &>(document.getDocumentElement().getFirstChild());

  // add style(s)
  if (expandStyle_) {
    if (screenStyle_.empty() && !printStyle_.empty())
      expandStyle(headElement, printStyle_);
    else if (!screenStyle_.empty() && printStyle_.empty())
      expandStyle(headElement, screenStyle_);
    else {
      expandStyle(headElement, printStyle_, "print");
      expandStyle(headElement, screenStyle_, "screen");
    }
  }
  else {
    if (screenStyle_.empty() && !printStyle_.empty())
      addStyle(headElement, printStyle_);
    else if (!screenStyle_.empty() && printStyle_.empty())
      addStyle(headElement, screenStyle_);
    else {
      addStyle(headElement, printStyle_, "print");
      addStyle(headElement, screenStyle_, "screen");
    }
  }
}

void XMLWrapper::addStyle(Element& parentElement, const DOMString& styleName, const DOMString& mediaString) {
  if (styleName.empty()) return;

  Element linkElement("link");
  linkElement.setAttribute("rel", "stylesheet");
  linkElement.setAttribute("href", styleName);
  if (!mediaString.empty())
    linkElement.setAttribute("media", mediaString);
  linkElement.setAttribute("type", "text/css");
  parentElement.appendChild(linkElement);
}

void XMLWrapper::expandStyle(Element& parentElement, const DOMString& styleName, const DOMString& mediaString) {
  if (styleName.empty()) return;

  Element linkElement("style");
  if (!mediaString.empty())
    linkElement.setAttribute("media", mediaString);
  linkElement.setAttribute("type", "text/css");
  parentElement.appendChild(linkElement);

  Reader reader(styleName);
  StringWriter stringWriter;
  stringWriter.putChar('\n');

  while (true) {
    Char c = reader.nextChar();
    if (c == FAILCHAR) break;
    else stringWriter.putChar(c);
  }

  DOMString styleString = stringWriter.getString();
  styleString.trim();
  styleString += "\n";

  Text text(styleString);
  linkElement.appendChild(text);
}

void XMLWrapper::copyContent(Node& sourceNode, Node& destNode) {
  while (true) {
    Node& node = sourceNode.getFirstChild();
    if (!node.isValid()) break;

    destNode.appendChild(node);
  }
}

}
