/***************************************************************************
                          xmlwriter.cpp  -  description
                             -------------------
    begin                : Tue Oct 21 2003
    copyright            : (C) 2003 by John Redmond
    email                : john_redmond@optusnet.com.au
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "xmlwriter.h"
#include "xmldeclaration.h" 
#include "endnode.h"
#include "element.h"
#include "stringreader.h"
#include "tokenstream.h"
#include "liststream.h"
#include "stringreader.h"

namespace limpid {

XMLWriter::XMLWriter(bool isHTML): emptySet_(new StringSet),
    writer_(new Writer(true)), format_(new OutputFormat),  isHTML_(isHTML) {
  setDefaults(false);
}

XMLWriter::XMLWriter(const URL& url, bool isHTML): emptySet_(new StringSet),
    writer_(new Writer(url)), format_(new OutputFormat), isHTML_(isHTML) {
  setDefaults(false);
  setEmptySet();
}

XMLWriter::XMLWriter(CharacterWriter& op, bool isHTML): emptySet_(new StringSet),
    writer_(&op.copy()), format_(new OutputFormat), isHTML_(isHTML) {
  setDefaults(false);
  setEmptySet();
}

void XMLWriter::update(const DOMString& destString) {
  DOMString outputString = destString;
  if (isHTML_)
    outputString += ".html";
  else
    outputString += ".xml";
    
  writer_ = new Writer(outputString);
}

void XMLWriter::setFormat(const DOMString& nl, const DOMString& i, const DOMString& ai) {
  format_->setFormat(nl, i, ai);
}

void XMLWriter::setStartAndFinish(const DOMString& strt, const DOMString& fnish) {
  format_->setStartAndFinish(strt, fnish);
}

Node& XMLWriter::process(Node& node) {
  int type = node.getNodeType();

  if (type == Node::DOCUMENT_NODE) return processDocumentStart(node);
  if (type == Node::XML_DECLARATION_NODE) return processXMLDeclaration(node);
  if (type == Node::DOCUMENT_TYPE_NODE) return processDoctype(node);
  if (type == Node::ELEMENT_NODE) return processStart(node);
  if (type == Node::TEXT_NODE) return processText(node);
  if (type == Node::END_NODE) return processEnd(node);
  return processSimple(node);

  return node;  // keep the compiler happy
}

Node& XMLWriter::processStart(Node& node) {
  if (format_->isStarted &&
      node.getPreviousSibling().getNodeType() != Node::TEXT_NODE) {
    writeFormat();
  }
  format_->isStarted = true; // there just may not be a prolog!

  DOMString displayString = node.getDisplayString();
  if (isHTML_ && isHTMLNode(node))
    displayString.change("/>", ">");

  *writer_ << displayString;
  

  if (node.getChildNodes().getLength()) {
    format_->nest();
  }
  
  return node;
}

Node& XMLWriter::processEnd(Node& node) {
  format_->deNest();
  EndNode& endNode = static_cast<EndNode &>(node);
  if (!containsText(endNode.getStartNode()))
    writeFormat();
  
  *writer_ << node;

  return node;
}

Node& XMLWriter::processText(Node& node) {
  *writer_ << node;
  return node;
}

Node& XMLWriter::processXMLDeclaration(Node& node) {
  // this should be at the start of the output
  DOMString encodingString =
    (static_cast<XMLDeclaration &>(node)).getEncoding();
  if (encodingString.getLength())
    writer_->setEncoding(encodingString);

  *writer_ << node << format_->newline;
  format_->isStarted = true;

  return node;
}

Node& XMLWriter::processDoctype(Node& node) {
  *writer_ << format_->indentation;
  *writer_ << node;
  format_->isStarted = true;

  return node;
}

Node& XMLWriter::processSimple(Node& node) {
  writeFormat();
  *writer_ << node;

  return node;
}

Node& XMLWriter::processDocumentStart(Node& node) {
  return node;  // presently a noop
}

void XMLWriter::start() {  // invoked externally: part of the Processor interface
  *writer_ << format_->startString;
}

void XMLWriter::finish() { // invoked externally: part of the Processor interface
  *writer_ << format_->finishString;
  writer_->close();
}

void XMLWriter::setDefaults(bool isHTML) {
  format_->isHTML = isHTML;
  format_->setFormat("\n", "", " ");
  setStartAndFinish("", "\n");
  format_->isFormatting = true;
}

bool XMLWriter::isHTMLNode(Node& node) {
  const StringSet::iterator p = emptySet_->find(node.getNodeName());
  return p != emptySet_->end();
}

bool XMLWriter::containsText(Node& node) {
  ListStream stream(node);
  while (true) {
    Node& childNode = stream.nextNode();
    if (!childNode.isValid()) break;

    if (childNode.getNodeType() == Node::TEXT_NODE)
      return true;

  }
  return false;
}

void XMLWriter::writeFormat() {
  if (format_->isStarted && format_->isFormatting)
    *writer_ << format_->newline << format_->indentation;
}

void XMLWriter::setEmptySet() {
  DOMString emptyTags =
    "area base basefont br col frame hr img input isindex link meta param";

  StringReader stringReader(emptyTags);
  TokenStream tis(stringReader);

  while (true) {
    DOMString token = tis.nextToken();
    if (token.empty()) break;

    emptySet_->insert(token);
    emptySet_->insert(token.toUpperCase());
  }
}

}
