/***************************************************************************
                          xpathbuilder.cpp  -  description
                             -------------------
    begin                : Sat May 8 2004
    copyright            : (C) 2004 by John Redmond
    email                : john_redmond@optusnet.com.au
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "xpathbuilder.h"
#include "element.h"

#include "nodestream.h"
#include "attributestream.h"
#include "liststream.h"
#include "ancestorstream.h"
#include "singlestream.h"
#include "followingstream.h"
#include "siblingstream.h"

#include "nametest.h"
#include "typetest.h"
#include "andtest.h"
#include "ortest.h"
#include "predicate.h"

#include "arithmeticcomponent.h"
#include "booleancomparison.h"
#include "groupcomponent.h"
#include "stringcomparison.h"
#include "numbercomparison.h"
#include "generalcomparison.h"
#include "booleangenerator.h"
#include "numbergenerator.h"
#include "stringgenerator.h"
#include "nodereporter.h"
#include "stringtranslator.h"
#include "stringtester.h"
#include "variable.h"

#include "collectorcomponent.h"
#include "idcollector.h"
#include "idchecker.h"
#include "nodesetcounter.h"
#include "compoundexpr.h"

#include "currentcollector.h"
#include "idgenerator.h"
#include "numberstyler.h"
#include "availableexpression.h"
#include "unparsedurigenerator.h"
#include "systempropertygenerator.h"
#include "documentcollector.h"

#include "errorhandler.h"

/**
  The mapping between element names in the document and processor/component names
  is generally straight-forward, with the exception:
    UnionExpr ==> OptionExpr

  The spirit of this change is that XPath provides one or more path options,
  as separated by "|" tokens in the source string. Each PathExpr is then an option.

  Note that all objects being added to UnionExpr or PathExpr are NodeSetProcessors.
  Normally OptionExpr has only instances of PathExpr or collection functions as children.
  Less commonly, UnionExpr might have nstances of Literal, Variable or Number; if
  multiples of these are present, the emantics of the overall Expr become dubious.

  At the end of a build, the OptionExpr is wrapped in an Expr instance, which provides
  all the common interface functons.

  The workhorse function is getComponent(...), which adds lower-level components recursively.
*/

namespace limpid {

XPathBuilder::XPathBuilder() {
}

// We are working here at the top level of an expression, and expect to return:
// a literal or
// a variable or
// a number (?) or
// a collector function or
// an option expression that contains one or more path expressions.
// Any other components at this level are errors.

// Note that an id() or key() function equates to a collector, and legitimately can form
// part of a path expression.

Expr XPathBuilder::buildExpr(Document& document, const DOMString& exprString) {
  Element documentElement = document.getDocumentElement();
  if (!documentElement.isValid()) return Expr();

  if (isPathExpr(documentElement))
    return getPathExpr(documentElement).getContent();

  if (isCollectorFunctionCall(documentElement))
    return getFunctionCallExpr(documentElement);

  if (isCollector(documentElement)) {
    return getOptionExpr(documentElement).getContent();
}
  return getComponent(documentElement).getContent();

/**
  These expressions, in turn, can contain one or more steps and/or
  collection functions:
  The functions id(...), key(...), current() and document(...) are collection functions
  and are NodeSetProcessors like Step and Predicate.

  According to the W3C recommendation, document(...) should return a node set
  with all nodes corresponding to documents. The disadvantage of this is the
  potentially heavy use of memory. For this reason, Limpid gives document(...)
  special treatment so that only one document is loaded at any one time.
  As it is a collector, key(...) can coexist with steps in a path expression;
  so, too, can current() coexist with them, though I have real doubts about
  the semantics of such a combination.
*/
}

/**
  getOptionsExpr populates an empty OptionExpr with PathExpr processors. Usually the xpath
  string will start with a location path consisting of one or more steps, with
  embedded predicates. Occasionally, however, the string may start with a collection processor,
  or there may be one or more collection processors embedded among steps. All of these
  possibilities have to be managed. getOptions() must therefore recognize paths that
  include collection functions.
*/

Expr XPathBuilder::getOptionExpr(Element& root) {
  OptionExpr optionExpr;
  ListStream stream(root);

  while (true) {
    Node& node = stream.nextNode();
    if (!node.isValid()) break;

    if (node.getNodeType() != Node::ELEMENT_NODE) continue;

    Element& element = static_cast<Element &>(node);
    if (element.getNodeName() == "PathExpr") {
      Expr pathExpr = getPathExpr(element);
      optionExpr.addProcessor(static_cast<NodeSetProcessor &>(pathExpr.getContent()));
      optionExpr.setPriority(optionExpr.getPriority() + pathExpr.getPriority());
    }
    else if (isCollectorFunctionCall(element)) {
      Expr functionExpr = getFunctionCallExpr(element);
      optionExpr.addProcessor(static_cast<NodeSetProcessor &>(functionExpr.getContent()));
      optionExpr.setPriority(optionExpr.getPriority() + 0.5);
    }
    else {// other function calls are incorrect in this context
      DOMWriter(wrt::Errors).write(root);
      ErrorHandler::report(element, "Incorrect function call in union expression");
    }
  }

  return optionExpr;
}

Expr XPathBuilder::getPathExpr(Element& root) {
  PathExpr pathExpr;
  addPathProcessors(pathExpr, root);

  // after addition of all the steps, the path expression
  // will have the priority of the last step
  return pathExpr;
}

/** Worker function to add Steps and CollectionFunctions directly to PathExpr.
    Other elements are ignored. */
void XPathBuilder::addPathProcessors(PathExpr& parentProcessor, Element& componentElement) {
  ListStream stream(componentElement);
  while (true) {
    Node& node = stream.nextNode();
    if (!node.isValid()) break;

    if (node.getNodeType() != Node::ELEMENT_NODE) continue; // a comment?
    Element& element = static_cast<Element &>(node);
    if (element.getNodeName() == "Step") {
      Step step = getStep(element);
      parentProcessor.addProcessor(step);
      addPredicates(step, element);
      parentProcessor.setPriority(parentProcessor.getPriority() + step.getPriority());
    }
    else if (isCollectorFunctionCall(element)) {
      Expr collectionExpr = getFunctionCallExpr(componentElement);
      parentProcessor.addProcessor(static_cast<NodeSetProcessor &>(collectionExpr.getContent()));
///      parentProcessor.setPriority(0);
    }
    else
      ErrorHandler::report(node, "unknown processor in path expression");
  }
}

/**
  According to XPath Production 4, a predicate is part of a step.
  Here, one or more predicates can be added to a Step.
*/

void XPathBuilder::addPredicates(Step& parentProcessor,
      Element& componentElement) {
  if (!componentElement.hasChildNodes()) return;

  ListStream stream(componentElement);
  while (true) {
    Node& node = stream.nextNode();
    if (!node.isValid()) break;

    if (node.getNodeName() == "PredicateExpr")
      addPredicate(parentProcessor, static_cast<Element &>(node));
  }
}

void XPathBuilder::addPredicate(Step& step, Element& componentElement) {
  size_t childCount = componentElement.getChildNodes().getLength();
  if (childCount != 1)
    ErrorHandler::report(componentElement, "predicate should have only one child expr");

  if (!childCount) return;

  Node& firstChild = componentElement.getFirstChild();
  Expr predicateComponent = getComponent(static_cast<Element &>(firstChild));
  if (!predicateComponent.isValid()) {
    ErrorHandler::report(componentElement, "bad child element in this");
    ErrorHandler::report(firstChild, "bad child element in predicate");
    return;
  }

  Predicate predicate(predicateComponent);
  step.appendPredicate(predicate);
  step.setPriority(step.getPriority() + 0.01);
}

/**
  Examine, in turn, each of the child elements of the parent element, this is mapped to
  an XPath component, and added to the parent component. The hard work
  of the mapping and addition is delegated to addComponent().
*/

void XPathBuilder::addComponents(Element& parentElement, ParentComponent& parent) {
  ListStream stream(parentElement);//, false, false);

  while (true) {
    Node& node = stream.nextNode();
    if (!node.isValid()) break;
    else if (node.getNodeType() == Node::ELEMENT_NODE) {
      if (node.getNodeName() != "Group")
        addComponent(static_cast<Element &>(node), parent);
      else {
        Element& childElement = static_cast<Element &>(node);
        addComponents(childElement, parent);  // recurse
      }
    }
  }
}

/**
  Create an XPath component, based on the specification in componentElement, and
  add the component to the parent component. This is a dispatch function and
  the task of actually creating and adding the component is delegated to lower functions.
  Non-leaf components (ParentComponents) have children added to them recursively, by
  invoking addComponent() again at a lower level.
*/
void XPathBuilder::addComponent(Element& componentElement, ParentComponent& parent) {
  Expr expr = getComponent(componentElement);
  parent.appendChild(expr.getContent());

}

/** Assemble a Step:
  A Step needs \ a stream to fetch the nodes and a test to determine
  whether to keep them.
*/
Step XPathBuilder::getStep(Element& stepElement) {
  Step step;
  setStream(stepElement, step);
  setTest(stepElement, step);

  return step;
}

/**
  Set a stream on a collector.
*/

void XPathBuilder::setStream(Element& componentElement, Step& step) {
  DOMString axis = componentElement.getAttribute("axis");
  if (axis == "child") {
    ListStream stream;
    step.setStream(stream);
  }
  else if (axis == "attribute") {
    AttributeStream stream;
    step.setStream(stream);
  }
  else if (axis == "descendant" ||
      axis == "descendant-or-self" || axis == "relative") {
    bool includeRoot = axis == "descendant-or-self";

    NodeStream stream(false, includeRoot); // false for no end nodes
    step.setStream(stream);
  }
  else if (axis == "ancestor" || axis == "ancestor-or-self") {
    bool includeSelf = axis == "ancestor-or-self";
    AncestorStream stream(includeSelf);
    step.setStream(stream);
  }
  else if (axis == "preceding" || axis == "following") {
    bool logic = axis == "following";
    FollowingStream stream(logic);
    step.setStream(stream);
  }
  else if (axis == "preceding-sibling" || axis == "following-sibling") {
    bool logic = axis == "following-sibling";
    SiblingStream stream(logic);
    step.setStream(stream);
  }
  else if (axis == "self" || axis == "parent" || axis == "root"
          || axis == "test-self" || axis == "test-parent") {
    SingleStream stream(axis);
    step.setStream(stream);
  }
}

/**
  Functions to set a Test on a Step.
*/

void XPathBuilder::setTest(Element& componentElement, Step& step) {
  DOMString axis = componentElement.getAttribute("axis");
  if (axis == "self" || axis == "parent") {
    setTrueTest(step);
    return;
  }

  setNodeTest(componentElement, step);
}

void XPathBuilder::setNodeTest(Element& componentElement, Step& step) {
  DOMString typeName = componentElement.getAttribute("type");
  if (typeName.size()) {
    if (typeName == "node") {
      Test test;
      step.setTest(test); // accept any node
      step.setPriority(-0.5);
    }
    else setNodeTypeTest(componentElement, step);

    return;
  }

  DOMString nodeName = componentElement.getAttribute("name");
  if (nodeName.size()) {
    NameTest test(nodeName);
    step.setTest(test);
    if (nodeName == "*") step.setPriority(0.25);///-0.5);
    else if (nodeName.contains("*")) step.setPriority(0.35);///-0.25);
    else step.setPriority(0.5);
  }
  else setTrueTest(step); // accept all nodes and do not add to priority
}

void XPathBuilder::setNodeTypeTest(Element& componentElement, Step& step) {
  DOMString typeName = componentElement.getAttribute("type");

  if (typeName == "processing-instruction") {
    TypeTest typeTest(Node::PROCESSING_INSTRUCTION_NODE);
    DOMString nodeName = componentElement.getAttribute("name");

    if (!nodeName.size()) {
      step.setTest(typeTest);
      step.setPriority(-0.5);
    }
    else {
      AndTest andTest;
      andTest.addTest(typeTest);
      NameTest nameTest(nodeName);
      andTest.addTest(nameTest);
      step.setTest(andTest);  // test for both type and name
      step.setPriority(0);
    }
  }
  else if (typeName == "text") {
    OrTest orTest;
    TypeTest textTest(Node::TEXT_NODE);
    orTest.addTest(textTest);
    TypeTest cdataTest(Node::CDATA_SECTION_NODE);
    orTest.addTest(cdataTest);
    step.setTest(orTest);
    step.setPriority(0.1);///-0.5);
  }
  else if (typeName == "comment") {
    TypeTest typeTest(Node::COMMENT_NODE);
    step.setTest(typeTest);
    step.setPriority(0.1);///-0.5);
  }
}

void XPathBuilder::setTrueTest(Step& step) {
  Test test;
  step.setTest(test);
  step.setPriority(0.5);
}

/**
  getComponent() operates recursively to build components. It
  delegates downwards to specialized helper functions.
  This helps to maintain a consistent coding level in each function.
*/

Expr XPathBuilder::getComponent(Element& componentElement) {
  Expr failExpr;
  DOMString componentName = componentElement.getNodeName();

  if (componentName == "FunctionCall")
    return getFunctionCallExpr(componentElement);

  if (componentName == "OrExpr" || componentName == "AndExpr") {
    DOMString groupName = componentElement.getNodeName();
    GroupComponent::GroupType groupType = GroupComponent::AND;
    if (groupName == "OrExpr") groupType = GroupComponent::OR;

    GroupComponent groupComponent(groupType);
    addComponents(componentElement, groupComponent);
    return groupComponent;
  }

  if (componentName == "ArithmeticExpr")
    return getArithmeticExpr(componentElement);

  if (componentName == "RelationalExpr" || componentName == "EqualityExpr")
    return getComparison(componentElement);

  if (componentName == "Option")
    return getOptionExpr(componentElement).getContent();

  if (componentName == "Literal")
    return XPathString(componentElement.getValue());

  if (componentName == "Variable")
    return Variable(componentElement.getValue());

  if (componentName == "Number")
    return Number(componentElement.getAttribute("value"));

  if (componentName == "Boolean") {
    DOMString type = componentElement.getAttribute("type");
    bool logic = type == "true";
    return Boolean(logic);
  }

  if (componentName == "PathExpr")
    return getPathExpr(componentElement);

  if (componentName == "UnionExpr")
    return getOptionExpr(componentElement);

  if (componentName == "Group") {
    Node& groupChild = componentElement.getFirstChild();
    if (groupChild.isValid())
      return getComponent(static_cast<Element &>(groupChild));
   }

  ErrorHandler::report(componentElement, "unknown xpath component");
  return failExpr;
}

Expr XPathBuilder::getPredicate(Element& componentElement) {
  size_t childCount = componentElement.getChildNodes().getLength();
  if (childCount != 1)
    ErrorHandler::report(componentElement, "predicate should have exactly one child component");

   if (!childCount) Expr();

  Element& firstChild = static_cast<Element &>(componentElement.getFirstChild());
  Expr predicateExpr = getComponent(firstChild);

  return Predicate(predicateExpr);
}

/**
  getArithmeticExpr() is a helper function to build an arithmentic expression.
*/
Expr XPathBuilder::getArithmeticExpr(Element& componentElement) {
  NodeList& childNodes = componentElement.getChildNodes();
  if (childNodes.getLength() != 2) {
    ErrorHandler::report(componentElement, "an arithmetic component must have two children");
    return Expr();
  }

  DOMString modeString = componentElement.getAttribute("type");

  Element& first = static_cast<Element &>(childNodes.item(0));
  Element& second = static_cast<Element &>(childNodes.item(1));

  ArithmeticComponent arith(modeString);
  addComponent(first, arith);
  addComponent(second, arith);
  return arith;
}

Expr XPathBuilder::getStringTranslator(Element& componentElement) {
  NodeList childNodes = componentElement.getChildNodes();
  if (childNodes.getLength() != 3) {
    ErrorHandler::report(componentElement, "translator requires three child components");
    return Expr();
  }

  DOMString param2 = childNodes.item(1).getNodeName();
  DOMString param3 = childNodes.item(2).getNodeName();
  if (param2 != "literal" || param3 != "literal")
    ErrorHandler::report(componentElement,
      "translator requires literal values for the second and third parameters");
  
  Expr expr = getComponent(static_cast<Element &>(childNodes.item(0)));
  return StringTranslator(expr, childNodes.item(1).getValue(),
    childNodes.item(2).getValue()); // these two params are literals
}

/**
  getComparison() sorts out the combinations of equalities and inequalities between
  node sets, strings, numbers and booleans.
*/
Expr XPathBuilder::getComparison(Element& componentElement) {
  NodeList& childNodes = componentElement.getChildNodes();
  if (childNodes.getLength() != 2) {
    ErrorHandler::report(componentElement, "a comparison requires two child components");
    return Expr();
  }

  DOMString componentName = componentElement.getNodeName();
  DOMString typeString = componentElement.getAttribute("type").resolveEntities();

  Element& first = static_cast<Element &>(childNodes.item(0));
  Element& second = static_cast<Element &>(childNodes.item(1));

  if (componentName == "RelationalExpr") {  // process <, <=, >=, >
    // this MUST be a numeric comparison
    NumberComparison numberComparison(typeString);

    addComponent(first, numberComparison);
    addComponent(second, numberComparison);
    return numberComparison;
  }

  // we reach here if it is an EqualityExpression

  // for this special case, comparison must be by strings
  if ((isCollector(first) && isCollector(second))) {
    StringComparison stringComparison(typeString);
    addComponent(first, stringComparison);
    addComponent(second, stringComparison);
    return stringComparison;
  }

  if (isNumberComparison(componentElement)) {
    NumberComparison exp(typeString);
    addComponent(first, exp);
    addComponent(second, exp);
    return exp;
  }

  if (isBooleanComparison(componentElement)) {
    BooleanComparison exp(typeString);
    addComponent(first, exp);
    addComponent(second, exp);
    return exp;
  }

  // default: do a simple comparison on getString() results
  GeneralComparison exp(typeString);
  addComponent(first, exp);
  addComponent(second, exp);
  return exp;
}

/**
  getFunctionCallExpr() returns function call expressions to getComponent().
*/
Expr XPathBuilder::getFunctionCallExpr(Element& componentElement) {
  Expr failExpr;

  // We may be after an XSLT function expression
  DOMString functionName = componentElement.getAttribute("type");
  if (isXSLTFunctionCall(functionName))
    return getXSLTFunctionCallExpr(componentElement);

  if (functionName == "document") {
    DocumentCollector documentCollector;
    addComponents(componentElement, documentCollector);
    return documentCollector; // collector with one or more children
  }
  // We are actually after a built-in XPath expression
  if (functionName == "contains" || functionName == "starts-with") {
    NodeList& childNodes = componentElement.getChildNodes();
    StringTester tester(functionName);

    if (childNodes.getLength() == 2) {
      DOMString modeString = componentElement.getAttribute("type");
      Element& first = static_cast<Element &>(childNodes.item(0));
      Element& second = static_cast<Element &>(childNodes.item(1));

      addComponent(first, tester);
      addComponent(second, tester);
      return tester;
    }

    ErrorHandler::report(componentElement, "function call requires 2 params");
    return failExpr;
  }
  if (functionName == "lang") {
    NodeList& childNodes = componentElement.getChildNodes();
    StringTester tester(functionName);
    if (childNodes.getLength()) {
      DOMString modeString = componentElement.getAttribute("type");
      Element& first = static_cast<Element &>(childNodes.item(0));

      addComponent(first, tester);
      return tester;
    }

    ErrorHandler::report(componentElement, "function call requires 1 param");
    return tester;
  }
  if (functionName == "boolean" || functionName == "not") {
    BooleanGenerator booleanGenerator(functionName);
    addComponents(componentElement, booleanGenerator);
    return booleanGenerator;
  }
  if (functionName == "string" || functionName == "substring" ||
      functionName == "substring-before" || functionName == "substring-after" ||
      functionName == "concat" || functionName == "normalize-space" || functionName == "escape-xml" || functionName == "escape-tex") {
    StringGenerator stringFunction(functionName);
    addComponents(componentElement, stringFunction);
    return stringFunction;
  }
  if (functionName == "translate")
    return getStringTranslator(componentElement);
  if (functionName == "number" || functionName == "last" ||
      functionName == "position" || functionName == "string-length" ||
      functionName == "floor" || functionName == "ceiling" ||
      functionName == "round" || functionName == "sum") {
    NumberGenerator numberGenerator(functionName);
    addComponents(componentElement, numberGenerator);
    return numberGenerator;
  }
  if (functionName == "name" || functionName == "local-name" ||
      functionName == "namespace-uri" ) {
    Node& firstChild = componentElement.getFirstChild();
    Expr expr = getPathExpr(static_cast<Element &>(firstChild));
    NodeReporter nodeReporter(functionName, expr);
    return nodeReporter;
  }
  if (functionName == "count") {
    size_t parameterCount = componentElement.getChildNodes().getLength();
    if (parameterCount != 1) {
      ErrorHandler::report(componentElement, "missing paramerer for count()");
      return failExpr;
    }
    NodeSetCounter counter;
    addComponents(componentElement, counter);
    return counter;
  }
  if (functionName == "id") {
    size_t parameterCount = componentElement.getChildNodes().getLength();
    if (parameterCount == 1) {
      Element paramElement = static_cast<Element &>(componentElement.getFirstChild());
      Expr paramExpr = getComponent(paramElement);
      IDCollector idCollector(paramExpr);
      return idCollector;
    }

    ErrorHandler::report(componentElement, "missing paramerer for id()");
    return failExpr;
}

  if (functionName == "idchecker") {
    size_t parameterCount = componentElement.getChildNodes().getLength();
    if (parameterCount == 1) {
      Element paramElement = static_cast<Element &>(componentElement.getFirstChild());
      Expr paramExpr = getComponent(paramElement);
      IDChecker idChecker(paramExpr);
      return idChecker;
    }

    ErrorHandler::report(componentElement, "missing parameter for idchecker()");
    return failExpr;
  }

  // If we have dropped through to here, we have a problem
  ErrorHandler::report(componentElement, "unknown function call");
  return failExpr;
}

// These two XSLT-related functions must be maintained for addition of extra features.
// Maintenance must include maintenance of the associated function classes!

bool XPathBuilder::isXSLTFunctionCall(const DOMString& typeString) {
  return (typeString == "element-available" ||
    typeString == "function-available" ||
    typeString == "current" ||
    typeString == "document" ||
    typeString == "key" ||
    typeString == "format-number" ||
    typeString == "generate-id" ||
    typeString == "system-property" ||
    typeString == "unparsed-entity-uri");
}

Expr XPathBuilder::getXSLTFunctionCallExpr(Element& functionElement) {
  Expr failExpr;
  DOMString typeString = functionElement.getAttribute("type");

  if (typeString == "function-available" || typeString == "element-available") {
   AvailableExpression availableExpr(typeString);
   addComponents(functionElement, availableExpr);
   return availableExpr;
  }
  if (typeString == "current") {
   CurrentCollector currentCollector;
   return currentCollector;
  }
  // key() is handled internally by transform::TemplateManager
  // because of its irregular behaviour.

  if (typeString == "format-number") {
   NumberStyler numberStyler;
   addComponents(functionElement, numberStyler);
   return numberStyler;
  }
  if (typeString == "generate-id") {
    IDGenerator idGenerator;
    return idGenerator;
  }
  if (typeString == "system-property") {
    SystemPropertyGenerator systemPropertyGenerator;
    addComponents(functionElement, systemPropertyGenerator);
    return systemPropertyGenerator;
  }
  if (typeString == "unparsed-entity-uri") {
    UnparsedURIGenerator uriGenerator;
    addComponents(functionElement, uriGenerator);
    return uriGenerator;
  }

  return failExpr;
}

/* Utility functions to identify structures in DOM subtrees */

bool XPathBuilder::isNumberComparison(Element& componentElement) {
  ListStream stream(componentElement);

  while (true) {
    Node& node = stream.nextNode();
    if (!node.isValid()) break;

    if (node.getNodeName() == "Number" || isNumberGenerator(node))
      return true;
  }
  return false;
}

bool XPathBuilder::isBooleanComparison(Element& componentElement) {
  ListStream stream(componentElement);

  while (true) {
    Node& node = stream.nextNode();
    if (!node.isValid()) break;

    if (node.getNodeName() == "Boolean" || node.getNodeName() == "RelationalExpr" ||
        node.getNodeName() == "EqualityExpr")
      return true;
  }
  return false;
}

bool XPathBuilder::isEquality(DOMString& typeString) {
  return typeString == "=" || typeString == "!=";
}

bool XPathBuilder::isInequality(DOMString& typeString) {
  return typeString == "<" || typeString == "<=" ||
          typeString == ">" || typeString == ">=";
}

bool XPathBuilder::isFunctionCall(Element& componentElement) {
  return componentElement.getNodeName() == "FunctionCall";
}

bool XPathBuilder::isNumberGenerator(Node& componentNode) {
  if (componentNode.getNodeType() != Node::ELEMENT_NODE) return false;
  else if (componentNode.getNodeName() != "FunctionCall") return false;

  Element& componentElement = static_cast<Element &>(componentNode);
  DOMString functionName = componentElement.getAttribute("type");

  return functionName == "count" || functionName == "position" ||
    functionName == "string-length" || functionName == "floor" ||
    functionName == "ceiling" || functionName == "round" || functionName == "sum";
}

bool XPathBuilder::isCollectorFunctionCall(Element& componentElement) {
  if (!isFunctionCall(componentElement)) return false;

  DOMString type = componentElement.getAttribute("type");
  return type == "id" || type == "idchecker" || type == "key" ||
    type == "current" || type == "document";
}

bool XPathBuilder::isPathExpr(Element& componentElement) {
  DOMString type = componentElement.getNodeName();
  return type == "PathExpr";
}

bool XPathBuilder::isUnionExpr(Element& componentElement) {
  DOMString type = componentElement.getNodeName();
  return type == "UnionExpr";
}

bool XPathBuilder::isCollector(Element& componentElement) {
  return isPathExpr(componentElement) || isUnionExpr(componentElement) ||
  isCollectorFunctionCall(componentElement);
}

bool XPathBuilder::isPredicate(Element& componentElement) {
  return componentElement.getNodeName() == "Predicate";
}

}
