/***************************************************************************
                          xpathbuilder.h  -  description
                             -------------------
    begin                : Sat May 8 2004
    copyright            : (C) 2004 by John Redmond
    email                : john_redmond@optusnet.com.au
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
  The mapping between element names in the document and processor/component names
  is generally straight-forward, with the exceptions:
  UnionExpr ==> OptionExpr
  PathExpr  ==> PathExpr

  The spirit of these changes is that XPath provides one or more path options,
  as separated by "|" tokens in the source string. Each PathExpr is then an option.

  Note that all objects being added to UnionExpr or PathExpr are NodeSetProcessors.
  Normally OptionExpr has only instances of PathExpr or IDCollector as children,
  while PathExpr has only instances of Step. Less commonly, UnionExpr might have
  instances of Literal, Variable or Number; if multiples of these are present, the
  semantics of the overall Expr become dubious.

  At the end of a build, the OptionExpr is wrapped in an Expr instance, which provides
  all the common interface funcions.

  XPath is a simple adapter to provide collection of nodes from the subtree of
  a single root noot node. It is not used in XSL transformation; rather an Expr is used
  to process an XSLContext.

  In XPathBuilder, Complex XPathExpressions are assembled by recursive delegation
  to addXXX() and setXXX() functions. This approach has two advantages:
    1.  If an element cannot be mapped to a processor or component, no action is taken
  and a crash is avoided.
    2.  The polymorphism of XPathExpressions is not complete (and does not need to be).
  It would therefore be difficult to assemble child components and pass them upwards
  through calling functions. (Contrast this to the approach taken in XPathParser,
  where all of the components are elements, and easily assembled by getXXX() functions
  and returned upwards.)
 ***************************************************************************/

#ifndef XPATHBUILDER_H
#define XPATHBUILDER_H


#include "document.h"
#include "pathexpr.h"
#include "optionexpr.h"
#include "parentcomponent.h"
#include "idcollector.h"
#include "step.h"
#include "expr.h"


/**
  *@author John Redmond
  */

namespace limpid {
  class XPathBuilder {
  public:
  	XPathBuilder();

    Expr buildExpr(Document& document, const DOMString& exprString);

  private:
    /* anchor function, which is used recursively within predicates */
    Expr getComponent(Element& componentElement);

    // factoring for the more complex expressions
    Expr getOptionExpr(Element& root);
    Expr getPathExpr(Element& root);
    Expr getPredicate(Element& componentElement);
    Expr getFunctionCallExpr(Element& root);
    Expr getXSLTFunctionCallExpr(Element& root);
    Expr getArithmeticExpr(Element& componentElement);
    Expr getStringTranslator(Element& componentElement);
    Expr getComparison(Element& componentElement);

    /* top-level addXXX() functions */
    void addPathProcessors(PathExpr& parentProcessor, Element& componentElement);
    void addPredicates(Step& parentProcessor, Element& componentElement);
    void addPredicate(Step& step, Element& componentElement);

    /* support addXXX() functions */
    void addComponents(Element& parentElement, ParentComponent& parent);
    void addComponent(Element& componentNode, ParentComponent& parent);

    /* PathExpression-building functions */
    Step getStep(Element& componentElement);
    void setStream(Element& componentElement, Step& step);
    void setTest(Element& componentElement, Step& step);
    void setNameTest(Element& componentElement, Step& step);
    void setNodeTest(Element& componentElement, Step& step);
    void setNodeTypeTest(Element& componentElement, Step& step);
    void setTrueTest(Step& step);

    /* Utility functions to identify structures in DOM subtrees */
    bool isEquality(DOMString& typeString);
    bool isInequality(DOMString& typeString);
    bool isPathExpr(Element& componentElement);
    bool isUnionExpr(Element& componentElement);
    bool isCollector(Element& componentElement);
    bool isPredicate(Element& componentElement);
    bool isNumberComparison(Element& componentElement);
    bool isBooleanComparison(Element& componentElement);
    bool isNumberGenerator(Node& componentNode);
    bool isFunctionCall(Element& componentElement);
    bool isCollectorFunctionCall(Element& componentElement);
    bool isXSLTFunctionCall(const DOMString& typeString);
  };
}


#endif
