/***************************************************************************
                          xpathorganizer.cpp  -  description
                             -------------------
    begin                : Wed May 26 2004
    copyright            : (C) 2004 by John Redmond
    email                : john_redmond@optusnet.com.au
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "xpathorganizer.h"
#include <stack>
#include <vector>
#include "node.h"
#include "nodestream.h"
#include "liststream.h"
#include "writer.h"
#include "domwriter.h"

namespace limpid {

XPathOrganizer::XPathOrganizer() {
}

/*
  organize(document) adds structure to the primitive DOM. It works by pushing
  all elements to a stack, then processes them by working back from the top of the
  stack. This effects a bottom-up treatment of the DOM tree.
*/

void XPathOrganizer::organize(Document& document) {
  std::stack<Element> nodeStack;

  NodeStream stream(document, false);
  while (true) {
    Node& node = stream.nextNode();
    if (!node.isValid()) break;

    if (node.getNodeType() == Node::ELEMENT_NODE)
      nodeStack.push(static_cast<Element &>(node));
  }

  while (nodeStack.size()) {
    Element topElement = nodeStack.top();
    nodeStack.pop();

    Element resultNode = structureNode(topElement);
    if (resultNode == topElement) continue;

    Node& parentNode = topElement.getParentNode();
    if (parentNode.isValid())
      parentNode.replaceChild(resultNode, topElement);
  }

  // finally add top-level structure to the default UnionExpression
  Element& documentElement = document.getDocumentElement();
  Node& firstChild = documentElement.getFirstChild();
  if (!firstChild.isValid()) {
    document.removeChild(documentElement);
    return; // the document is now empty
  }

  Element& grandChild = static_cast<Element &>(firstChild.getFirstChild());

  if (grandChild.isValid() && grandChild.getNodeName() == "FunctionCall" &&
      grandChild.getAttribute("type").startsWith("id"))
    document.replaceChild(grandChild, documentElement); // not a path expression

  else if (documentElement.getChildNodes().getLength() == 1) // a single path expression
    document.replaceChild(documentElement.getFirstChild(), documentElement);
}

/**
  Apply structure to the element children of a parent element. This is introducted
  in order of operator or (for comparison) structural precedence.
  Hence unary minus before *, div, mod before +, -, before XXX cmp YYY,
  before AND, before OR.
*/

Element XPathOrganizer::structureNode(Element& node) {
  Element returnNode = node;
  returnNode = processUnaryMinuses(node);
  returnNode = addStructure(returnNode, "MultiplyDivideMod", "MultiplicativeExpr");
  returnNode = addStructure(returnNode, "PlusMinus", "AdditiveExpr");

  addLRStructure(returnNode, "MultiplicativeExpressionAdditiveExpr", "ArithmeticExpr");
  returnNode = addStructure(returnNode, "Relation", "RelationalExpr");
  addLRStructure(returnNode, "RelationalExpr", "RelationalExpr");
  returnNode = addStructure(returnNode, "Equality", "EqualityExpr");
  addLRStructure(returnNode, "EqualityExpr", "EqualityExpr");
  returnNode = addStructure(returnNode, "and", "AndExpr", false);
  returnNode = addStructure(returnNode, "or", "OrExpr", false);

  return returnNode;
}

/*
  Apply a left-associative structuring of the child elements of a node, provided
  that they match to a groupNames parameter. The new structure is an element with
  name of groupLabel. Where includeOperators is true (the default), the operators
  matching groupNames are included. This is the case for arithmetic and logical
  operators, but false is appropriate for grouping nodes under 'And' or 'Or'.
*/
Element XPathOrganizer::addStructure(Element& parentNode, const char* groupNames,
    const char* groupLabel, bool includeOperators) {
  if (!containsNode(parentNode, groupNames)) return parentNode;
  Element newNode = static_cast<Element &>(parentNode.cloneNode(false));

  DOMString groupString = groupNames;
  bool inGroup = false;
  Element group(groupLabel);

  NodeList& children = parentNode.getChildNodes();
  size_t length = children.getLength();

  for (size_t index = 0; index < length; index++) {
    Node& node = children.item(index);
    Node& newChild = node.cloneNode(true);

    if (inGroup) {
      if (includeOperators || !groupString.contains(node.getNodeName()))
        group.appendChild(newChild);
      if (!groupString.contains(node.getNodeName()) && index < length - 1)
        inGroup = groupString.contains(children.item(index + 1).getNodeName());
    }
    else {
      if (index < length - 1)
        inGroup = groupString.contains(children.item(index + 1).getNodeName());
      if (inGroup) {
        group = Element(groupLabel);
        newNode.appendChild(group);
        group.appendChild(newChild);
      }
      else
        newNode.appendChild(newChild);
    }
  }

  return newNode;
}

void XPathOrganizer::addLRStructure(Element& parentNode,
      const char* groupName, const char* structureName) {
  DOMString groupString = groupName;
  std::stack<Element> nodeStack;

  NodeStream stream(parentNode, false, false);
  while (true) {
    Node& node = stream.nextNode();
    if (!node.isValid()) break;

    if (node.getNodeType() == Node::ELEMENT_NODE)
      nodeStack.push(static_cast<Element &>(node));
  }

  while (nodeStack.size()) {
    Element topElement = nodeStack.top();
    nodeStack.pop();

    DOMString nodeName = topElement.getNodeName();
    if (!groupString.contains(nodeName))
      continue;

    Element resultNode = getLRStructure(topElement, structureName);
    if (resultNode == topElement) continue;

    Node& topParentNode = topElement.getParentNode();
    if (topParentNode.isValid())
      topParentNode.replaceChild(resultNode, topElement);
  }
}

Element XPathOrganizer::getLRStructure(Element& root, const char* structureName) {
  size_t length = root.getChildNodes().getLength();
  if (length < 3) return root; // serious error

  ListStream testStream(root);
  Element structure(structureName);

  ListStream stream(root);
  Node& firstNode = stream.nextNode();
  Node& operatorNode = stream.nextNode();
  Node& secondNode = stream.nextNode();

  Element& operatorElement = static_cast<Element &>(operatorNode);
  DOMString operatorName = operatorElement.getAttribute("type");
  structure.setAttribute("type", operatorName);
  structure.appendChild(firstNode.cloneNode(true));
  structure.appendChild(secondNode.cloneNode(true));


  size_t remainingNodes = length - 3;
  while (remainingNodes >= 2) {
    Element higherStructure("ArithmeticExpr");
    operatorNode = stream.nextNode();
    operatorElement = static_cast<Element &>(operatorNode);

    operatorName = operatorElement.getAttribute("type");
    higherStructure.setAttribute("type", operatorName);

    higherStructure.appendChild(structure);
    secondNode = stream.nextNode();
    higherStructure.appendChild(secondNode);
    remainingNodes -=2;

    structure = higherStructure;
  }

  return structure;
}

Element XPathOrganizer::processUnaryMinuses(Element& parent) {
  if (!containsNode(parent, "Minus")) return parent;

  NodeList& children = parent.getChildNodes();
  size_t length = children.getLength();

  for (size_t index = length - 2; index >= 1; --index) {
    Node& node = children.item(index);
    if (node.getNodeName() != "Minus") continue;

    DOMString previousName = children.item(index - 1).getNodeName();

    if (previousName != "Divide" && previousName != "Multiply" &&
        previousName != "Plus" && previousName != "Minus"
        && previousName != "Mod") continue;

    if (children.item(index + 1).getNodeName() != "Number") continue;

    (static_cast<Element &>(children.item(index + 1))).setAttribute("positive", "false");
    parent.removeChild(node);
  }

  return parent;
}

bool XPathOrganizer::containsNode(Node& root, const char* nodeNames) {
  if (root.getNodeType() != Node::ELEMENT_NODE) return false;
  DOMString nameString = nodeNames;

  ListStream stream(root);
  while (true) {
    Node& node = stream.nextNode();
    if (!node.isValid()) break;
    if (node.getNodeType() != Node::ELEMENT_NODE) continue;

    DOMString nodeName = node.getNodeName();
    if (nameString.contains(nodeName)) return true;
  }

  return false;
}

}
