/***************************************************************************
                          xpathparser.cpp  -  description
                             -------------------
    begin                : Wed Apr 28 2004
    copyright            : (C) 2004 by John Redmond
    email                : john_redmond@optusnet.com.au
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <stack>
#include "xpathparser.h"
#include "tokenstream.h"
#include "nodestream.h"
#include "liststream.h"
#include "text.h"
#include "xmldeclaration.h"
#include "systemmanager.h"
#include "stringreader.h"
#include "xmlwriter.h"
#include "domwriter.h"

namespace limpid {

XPathParser::XPathParser(const DOMString& src) {
  setup(src);
}

void XPathParser::setup(const DOMString& source) {
  XMLDeclaration decl("1.0", "UTF-8", "yes");
  document.appendChild(decl);
  Element unionExpression("UnionExpr");
  document.appendChild(unionExpression);
  setTokenList(source);
  tokenHandler.reset();
  populateUnionExpression(unionExpression);

  // I'm not sure that this simplification is a generally good idea:
/*
  size_t unionLength = unionExpression.getChildNodes().getLength();
  if (unionLength) {
    DOMString firstName = unionExpression.getFirstChild().getNodeName();
    if (firstName != "PathExpr" && firstName != "FunctionCall")
      document.replaceChild(unionExpression.getFirstChild(), unionExpression);
  }
*/
}

void XPathParser::populateUnionExpression(Element& unionExpression) {
  while (true) {
    DOMString token = peekToken();
    if (token.empty()) break;

    if (token == "|") {
      bumpToken();
      continue;
    }

    Element node = getNode();
    if (node.isValid())
      unionExpression.appendChild(node);
  }
}

/**
  Token management functions. Load the tokens from the token stream and add them
  to a preliminary token handler. Then read them back from the handler, combine them
  in light of the parser productions and add them to a global handler for use in
  mapping to XPath components.
*/

void XPathParser::setTokenList(const DOMString& source) {
  StringReader reader(source);
  TokenStream tis(reader);
  tis.setDelimiters("|/^[]()=!@:,\"'$+*><");
  TokenHandler preliminaryHandler;

  while (true) {
    DOMString token = tis.nextToken();
    if (!token.size()) break;

    preliminaryHandler.addToken(token);

    if (isDelimiter(token)) {
      DOMString value = tis.getDelimitedString(token[0]);
      preliminaryHandler.addToken(value); // may be a delimiter to an empty string
      if (!isDelimiter(value))
        preliminaryHandler.addToken(token); // closing delimiter
    }
  }

  preliminaryHandler.reset();

  while (true) {
    DOMString token = preliminaryHandler.nextToken();
    if (!token.size()) break;

    if (processHyphens(token)) continue;

// deal with a numeric predicate
    else if (isNumber(token) && preliminaryHandler.peekToken(-2) == "[" &&
        preliminaryHandler.peekToken() == "]") {
      tokenHandler.addToken("position");
      tokenHandler.addToken("()");
      tokenHandler.addToken("=");
      tokenHandler.addToken(token);
      tokenHandler.addToken("]");

      preliminaryHandler.bumpToken();
    }
    else if (token == "(" && preliminaryHandler.peekToken() == ")") {
      preliminaryHandler.bumpToken();
      tokenHandler.addToken("()");
    }
    else if (token == "/" && preliminaryHandler.peekToken() == "/") {
      preliminaryHandler.bumpToken();
      tokenHandler.addToken("//");
    }
    else if (token == ":" && preliminaryHandler.peekToken() == ":") {
      preliminaryHandler.bumpToken();
      tokenHandler.addToken("::");
    }
    else if (token == "." && preliminaryHandler.peekToken() == ".") {
      preliminaryHandler.bumpToken();
      tokenHandler.addToken("..");
    }
    else if ((token == "!" || token == ">" || token == "<") &&
          preliminaryHandler.peekToken() == "=") {
      preliminaryHandler.bumpToken();
      token += "=";
      tokenHandler.addToken(token);
    }
    else if (isDelimiter(token)) {
      DOMString literalString = preliminaryHandler.nextToken();
      if (!literalString.size()) break; // cannot continue

      tokenHandler.addToken(token);  // the opening delimiter

      if (isDelimiter(literalString))
        literalString = ""; // following delimiter means an empty string
      else {
        DOMString followingToken = preliminaryHandler.peekToken();
        if (isDelimiter(followingToken)) // it should be one!
          token = preliminaryHandler.nextToken();
      }
      tokenHandler.addToken(literalString);
      tokenHandler.addToken(token);  // the closing delimiter
      token = preliminaryHandler.peekToken();
    }
    else {
      if (preliminaryHandler.peekToken() == ":" &&
          preliminaryHandler.peekToken(1) != ":") {
        // assemble a qualified name, but do not touch two consecutive colons
        token += preliminaryHandler.nextToken(); // add the colon
        token += preliminaryHandler.nextToken(); // add the local name
      }
      tokenHandler.addToken(token);
    }
  }

  processOperators();
}

bool XPathParser::processHyphens(DOMString token) {
  static DOMString hyphenString = '-';
  if (!token.contains(hyphenString)) return false;

  StringReader reader(token);
  TokenStream tis(reader);
  tis.setDelimiters("-");
  TokenHandler handler;

  while (true) {
    DOMString tok = tis.nextToken();
    if (tok.empty()) break;

    handler.addToken(tok);
  }

  handler.reset();
  DOMString compoundToken;

  while (true) {
    DOMString token0 = handler.nextToken();
    if (token0.empty()) break;

    // simplest case is where the compound string starts with a hyphen
    if (token0 == hyphenString) {
      tokenHandler.addToken(token0);
      continue;
    }

    // also simple is where the compound string starts with a number:
    // add the number string and following hyphen (minus sign) if there is one.
    if (Converter::isNumber(token0)) {
      tokenHandler.addToken(token0);

      if (handler.peekToken() == hyphenString) {
        tokenHandler.addToken(hyphenString);
        handler.bumpToken();
      }
      continue;
    }

    // the remaining string cannot be fully resolved. It could be of the form:
    // alphastring-numberstring, which is legal in XML. If this is not intended,
    // the source should have had whitespace before or after the hyphen
    DOMString compoundString = token0;
    while (true) {
      DOMString extraToken = handler.nextToken();
      if (extraToken.empty()) break;
      compoundString += extraToken;
     }

    // the final special case is where the compund string ends with a hyphen
    if (compoundString.endsWith("-")) {
      compoundString = compoundString.substr(0, compoundString.size() - 1);
      tokenHandler.addToken(compoundString);
      tokenHandler.addToken(hyphenString);
    }
    else
      tokenHandler.addToken(compoundString);

    break;
  }

  return true;
}

void XPathParser::processOperators() {
  tokenHandler.reset();

  while (true) {
    DOMString& token = tokenHandler.nextToken();
    if (token.empty()) break;
    if (token != "*" && token != "div" && token != "mod") continue;

    DOMString& previousToken = tokenHandler.peekToken(-2);
    if (previousToken.empty()) continue;  // start of top-level expression

    if (previousToken == "@" || previousToken == "::" ||
        previousToken == "(" || previousToken ==  "[" ||
        previousToken == "'" || isOperator(previousToken)) continue;

    if (token == "*") token = "__MULT__";
    else if (token == "div") token = "__DIV__";
    else if (token == "mod") token = "__MOD__";
  }
}

/**
  Token test group: isXXX() of hasXXX()
*/
bool XPathParser::isDelimiter(const DOMString& token) {
  return token == "\"" || token == "'";
}

bool XPathParser::isConjunction(const DOMString& token) {
  return token == "and" || token == "or";
}

bool XPathParser::isNumber(const DOMString& token) {
  if (token.empty()) return false;
  if (token == "." || token == "..") return false; // otherwise, may have decimal pts
  else return Converter::isNumber(token);
}

bool XPathParser::isOperator(const DOMString& token) {
  return token == "and" || token == "or" || token == "__MOD__" || token == "__DIV__" ||
    token == "__MULT__" || token == "/" || token == "//" || token == "|" || token == "+" ||
    token == "-" || token == "=" || token == "!=" || token == "|" ||
    token == "<=" || token == ">=";
}

/**
  Get a qualified name from a handler. It may contain colons and a single hyphen.
*/

DOMString XPathParser::getQName(TokenHandler& handler, DOMString& token) {
  DOMString result = token;//getHyphenatedName(handler, token);
  if (handler.peekToken() == ":" && handler.peekToken() != ":") {
    result += ":";
    handler.bumpToken();
    DOMString nextTok = handler.nextToken();
    result += nextTok;
  }

  return result;
}

DOMString XPathParser::getHyphenatedName(TokenHandler& handler, DOMString& token) {
  DOMString result = token;

  while (true) {
    if (handler.peekToken() != "-") break;

    DOMString twoAhead = handler.peekToken(1);
    if (!twoAhead.size() || isNumber(twoAhead)) break;  // we have a minus sign, not a hyphen

    result += "-";
    result += twoAhead;
    handler.bumpToken();
    handler.bumpToken();
  }
  return result;
}

/**
  Node Get group: getXXX()
*/

Element XPathParser::getPredicate() {
  nextToken();  // skip the "["
  Element predicate("PredicateExpr");
  Node& currentNode = predicate;

  while(true) {
    DOMString token = peekToken();
    if (!token.size() || token == ")") break;
    if (token == "]") {
      nextToken(); // finished with the predicate
      break;
    }
    Element node = getNode();
    if (!node.isValid()) {
      setErrorReport(predicate);
      bumpToken();
    }
    else currentNode.appendChild(node);
  }

  return predicate;
}

/** getNode() is the central function which is used recursively to return
  elements (corresponding to either leaves or intermediate nodes). It delegates
  the task of element creation to specific functions.
*/
Element XPathParser::getNode() {
  DOMString token = peekToken();
  if (token == "(") return getParenthesis();
  else if (delimiterFollows()) return getLiteral();
  else if (variableFollows()) return getVariable();
  else if (numberFollows()) return getNumber();
  else if (conjunctionFollows()) return getConjunction();
  else if (relationFollows()) return getRelation();
  else if (equalityFollows()) return getEquality();
  else if (additiveOperatorFollows()) return getAdditiveNode();
  else if (booleanFollows()) return getBoolean();
  else if (functionCallFollows()) return getFunctionCall();
  else if (stepFollows()) return getSteps();
  else return Element::failElement;
}

Element XPathParser::getLiteral() {
  bumpToken(); // the opening delimiter
  DOMString token = nextToken();

  Element literal("Literal");
  Text textValue = token;
  literal.appendChild(textValue);
  bumpToken(); // the closing delimiter

  return literal;
}

Element XPathParser::getVariable() {
  bumpToken();
  DOMString token = nextToken();
  Element variable("Variable");
  Text textValue = token;
  variable.appendChild(textValue);

  return variable;
}

Element XPathParser::getNumber() {
  DOMString token = nextToken();
  Element number("Number");
  number.setAttribute("value", token);

  return number;
}

Element XPathParser::getConjunction() {
  DOMString token = nextToken();
  Element conjunction(token);
  return conjunction;
}

Element XPathParser::getAdditiveNode() {
  DOMString token = nextToken();
  DOMString label = "Plus";
  DOMString tag = "+";

  if (token == "-") {
    label = "Minus";
    tag = "-";
  }
  else if (token == "__MULT__") {
    label = "Multiply";
    tag = "*";
  }
  else if (token == "__DIV__") {
    label = "Divide";
    tag = "div";
  }
  else if (token == "__MOD__") {
    label = "Mod";
    tag = "mod";
  }
  Element returnElement(label);
  returnElement.setAttribute("type", tag);

  return returnElement;
}

Element XPathParser::getRelation() {
  DOMString token = nextToken();
  DOMString next = peekToken();

  Element operatorNode("Relation");
  operatorNode.setAttribute("type", token);

  return operatorNode;
}

Element XPathParser::getEquality() {
  DOMString token = nextToken();
  DOMString next = peekToken();

  Element operatorNode("Equality");
  operatorNode.setAttribute("type", token);
  return operatorNode;
}

Element XPathParser::getBoolean() {
  DOMString token = nextToken();
  Element boolean("Boolean");
  boolean.setAttribute("type", token);

  bumpToken(); // the "()"
  return boolean;
}

Element XPathParser::getFunctionCall() {
  Element functionCall("FunctionCall");
  DOMString nextTok = nextToken();
  functionCall.setAttribute("type", nextTok);

  nextTok = nextToken();  // the "(" or "()"
  if (nextTok == "()") return functionCall;

  while (true) {          // collect parameters, if any, after the "("
    nextTok = peekToken();
    if (!nextTok.size()) break;

    Element parameterElement = getNode();
    functionCall.appendChild(parameterElement);

    nextTok = peekToken();
    if (nextTok == ")") {
      bumpToken();
      break; // end of parameters
    }
    else if (nextTok == ",") bumpToken(); // another child token follows
  }

  return functionCall;

}

Element XPathParser::getParenthesis() {
  Element group("Group");
  bumpToken();  // skip the ")"

  while(true) {
    DOMString token = peekToken();
    if (!token.size() || token == ")") {
      bumpToken();
      break;
    }

    Element node = getNode();
    group.appendChild(node);
  }

  return group;
}

Element XPathParser::getSteps() {
  Element pathExpression("PathExpr");

  if (peekToken() == "/") {
    Element step("Step");
    step.setAttribute("axis", "root");
    step.setAttribute("name", "#document");
    pathExpression.appendChild(step);
    bumpToken();
  }

  while (true) {
    if (functionCallFollows()) {
      if (peekToken() == "/") bumpToken();

      Element collector = getFunctionCall();
      pathExpression.appendChild(collector);
      addPredicates(pathExpression);
    }
    else if (stepFollows()) {
      Element step = getStep();
      pathExpression.appendChild(step);
      addPredicates(step);///parentElement);
    }
    else break;
  }
  return pathExpression;
}

void XPathParser::addFurtherProcessors(Element& parentElement) {
  while (true) {
    if (functionCallFollows()) {
      Element collector = getFunctionCall();
      parentElement.appendChild(collector);
      addPredicates(parentElement);  // predicates follow function call
    }
    else if (stepFollows()) {
      Element step = getStep();
      parentElement.appendChild(step);
      addPredicates(step); // predicates added to step
    }
    else break;
  }
}

void XPathParser::addPredicates(Element& parentElement) {
  while (predicateFollows()) {
    Element predicate = getPredicate();
    parentElement.appendChild(predicate);
  }
}

  /** Return a step to be added to a location path. This step may contain one
    or more predicates as children.
  */
Element XPathParser::getStep() {
  DOMString axis = "child";

  if (peekToken() == "/" || peekToken() == "//") {
    DOMString slashToken = nextToken();
    if (slashToken == "//") axis = "descendant-or-self";
  }
  /* Process the special case of an abbreviated step */
  if (abbreviatedStepFollows()) {
    DOMString axisString = "self";
    DOMString stepToken = nextToken();
    if (stepToken == "..") axisString = "parent";

    Element step("Step");
    step.setAttribute("axis", axisString);

    return step;
  }

  /* There may be a specified axis, otherwise the default "child" axis applies */
  if (axisFollows())
    axis = getAxisName();

  Element step("Step");
  step.setAttribute("axis", axis);

  /* There may be a node type specification, with "type" attribute specified */
  if (nodeTypeTestFollows())
    setNodeTypeTest(step);
  /* Otherwise it is an element name specification with a "name" attribute */
  else if (nameTestFollows())
    setNameTest(step);

  return step;
}

DOMString XPathParser::getAxisName() {
  DOMString token = nextToken();
  if (token == "@")
    return "attribute";
  else {
    bumpToken(); // skip "::"
    return token;
  }
}

void XPathParser::setNodeTypeTest(Element& parentElement) {
  DOMString token = nextToken();
  if (token == "comment" || token == "text" || token == "node") {
    parentElement.setAttribute("type", token);
    bumpToken(); // the "()"
  }
  else if (token == "processing-instruction") {
    parentElement.setAttribute("type", token);
    DOMString nextTok = nextToken();

    if (nextTok == "(") {
      DOMString name = nextToken();
      parentElement.setAttribute("name", name);
      bumpToken();  // the ")"
    }
  }
}

void XPathParser::setNameTest(Element& parentElement) {
  DOMString name = nextToken();
  parentElement.setAttribute("name", name);
}

/** The xxxFollows() group, which provide the look-ahead to direct the getNode()
    function.
*/
bool XPathParser::numberFollows() {
  DOMString token = peekToken();
  return isNumber(token);
}

bool XPathParser::relationFollows() {
  DOMString token = peekToken();
  return token == "<" || token == ">" || token == "<=" || token == ">=";
}

bool XPathParser::equalityFollows() {
  DOMString token = peekToken();
  return token == "!=" || token == "=" ;
}

bool XPathParser::additiveOperatorFollows() {
  DOMString token = peekToken();
  return token == "+" || token == "-" || token == "__MULT__" ||
    token == "__DIV__" || token == "__MOD__";
}

bool XPathParser::conjunctionFollows() {
  DOMString token = peekToken();
  return token == "and" || token == "or";
}

bool XPathParser::booleanFollows() {
  DOMString token = peekToken();
  return token == "true" || token == "false";
}

bool XPathParser::delimiterFollows() {
  DOMString token = peekToken();
  return token == "'" || token == "\"";
}

bool XPathParser::variableFollows() {
  DOMString token = peekToken();
  return token == "$";
}

bool XPathParser::functionCallFollows() {
  int peekPosition = 0;
  if (peekToken() == "/") ++peekPosition; // may have to look past a slash

  DOMString nextTok = peekToken(peekPosition + 1);
  if (nextTok != "()" && nextTok != "(") return false;

  DOMString token = peekToken(peekPosition);

  if (token == "last") return true;
  else if (token == "position") return true;
  else if (token == "count") return true;
  else if (token == "id") return true;
  else if (token == "name") return true;
  else if (token == "local-name") return true;
  else if (token == "namespace-uri") return true;
  else if (token == "type") return true;

  else if (token == "string") return true;
  else if (token == "concat") return true;
  else if (token == "starts-with") return true;
  else if (token == "contains") return true;
  else if (token == "substring-before") return true;
  else if (token == "substring-after") return true;
  else if (token == "substring") return true;
  else if (token == "string-length") return true;
  else if (token == "normalize-space") return true;
  else if (token == "translate") return true;
  else if (token == "escape-tex") return true;

  else if (token == "boolean") return true;
  else if (token == "not") return true;
  else if (token == "true") return true;
  else if (token == "false") return true;
  else if (token == "lang") return true;

  else if (token == "number") return true;
  else if (token == "sum") return true;
  else if (token == "floor") return true;
  else if (token == "ceiling") return true;
  else if (token == "round") return true;

  // XSLT functions
  else if (token == "element-available") return true;
  else if (token == "function-available") return true;
  else if (token == "document") return true;
  else if (token == "id") return true;
  else if (token == "key") return true;
  else if (token == "current") return true;
  else if (token == "format-number") return true;
  else if (token == "generate-id") return true;
  else if (token == "system-property") return true;
  else if (token == "unparsed-entity-uri") return true;

  else return false;
}

bool XPathParser::abbreviatedStepFollows() {
  return peekToken() == "." || peekToken() == "..";
}

bool XPathParser::axisFollows() {
  DOMString token = peekToken();
  if (token == "@" || token == "." || token == "..") return true;
  else if ((token == "attribute" || token == "ancestor" ||
        token == "child" || token == "descendant" ||
        token == "following" || token == "namespace" || token == "parent" ||
        token == "preceding" || token == "self" ||
        token == "preceding-sibling" || token == "following-sibling" ||
        token == "ancestor" || token == "descendant" ||
        token == "ancestor-or-self" || token == "descendant-or-self") &&
        peekToken(1) == "::")
  return true;
  else return false;
}

bool XPathParser::nodeTypeTestFollows() {
  DOMString token = peekToken();
  if (token == "comment" || token == "text" || token == "node")
    return peekToken(1) == "()";
  else if (token == "processing-instruction")
     return peekToken(1) == "(" || peekToken() == "()";
  else return false;
}

bool XPathParser::nameTestFollows() {
  DOMString token = peekToken();
  return isalpha(token[0]) || token[0] == '_' || token == "*";
}

bool XPathParser::nodeTestFollows() {
  return nameTestFollows() || nodeTypeTestFollows();
}

bool XPathParser::stepFollows() {
  if (peekToken().startsWith("__")) return false; // an additive operator follows

  markPosition();
  if (peekToken() == "/") {
    bumpToken();
    if (peekToken().empty()) {
      resumePosition();
      return true; // this is the special case of an expression = "/"
    }
    resumePosition();
  }

  if (peekToken() == "/" || peekToken() == "//") bumpToken();
  if (peekToken() == "." || peekToken() == ".." || peekToken() == "@") {
    bumpToken();

    bool isStep = peekToken().empty() ||
                  (peekToken() != "(" && peekToken() != "()");

    resumePosition();
    return isStep;
  }

  if (axisFollows()) bumpToken(2);
  bool result = nodeTestFollows();
  resumePosition();

  return result;
}

bool XPathParser::idFollows() {
  if (peekToken() != "id") return false;
  else return peekToken(1) == "(";
}

bool XPathParser::currentFollows() {
  if (peekToken() != "current") return false;
  else return peekToken(1) == "(";
}

bool XPathParser::locationPathFollows() {
  markPosition();
  if (peekToken() == "/" || peekToken() == "//") bumpToken();
  bool result = stepFollows();
  resumePosition();

  return result;
}

void XPathParser::setErrorReport(Element &parentNode) {
  DOMString token = nextToken();
  Element errorReport("ErrorReport");
  errorReport.setAttribute("token", token);
  parentNode.appendChild(errorReport);
}

}
