/***************************************************************************
                          xpathreverser.cpp  -  description
                             -------------------
    begin                : Sat Oct 16 2004
    copyright            : (C) 2004 by John Redmond
    email                : john_redmond@optusnet.com.au
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "xpathreverser.h"
#include "element.h"
#include "comment.h"
#include "nodelist.h"
#include "documentfragment.h"
#include "nodestream.h"
#include "liststream.h"
#include "writer.h"
#include "ancestorstream.h"
#include "domwriter.h"

namespace limpid {

XPathReverser::XPathReverser(){
}

void XPathReverser::reverse(Document& document) {
  NodeList matchList;

  Element documentElement = document.getDocumentElement(); // make a COPY
  NodeStream stream(documentElement);
  while (true) {
    Node& node = stream.nextNode();
    if (!node.isValid()) break;

    if (node.getNodeType() != Node::ELEMENT_NODE) continue;
    if (node.getNodeName() == "PathExpr"  || node.getNodeName() == "FunctionCall")
      matchList.addNode(node);
  }

  ListStream matchStream(matchList);
  while (true) {
    Node& node = matchStream.nextNode();
    if (!node.isValid()) break;
    if (isWithinPredicate(node)) continue; // do not alter path expressions in predicates

    if (node.getNodeType() != Node::ELEMENT_NODE) continue;
    Element& expressionElement = static_cast<Element &>(node);

    if (expressionElement.getNodeName() == "FunctionCall") {
      reverseFunctionCall(expressionElement);
      continue;
    }

    Element reverseElement = getReverse(node);
    node.getParentNode().replaceChild(reverseElement, node);
  }
}

void XPathReverser::reverseFunctionCall(Element& expressionElement) {
  if (expressionElement.getAttribute("type") == "id")
    expressionElement.setAttribute("type", "idchecker");
}

Element XPathReverser::getReverse(Node& node) {
  Element replacementElement("PathExpr");
  NodeList childList = node.getChildNodes();
  size_t nodeCount = childList.getLength();

  if (!nodeCount) return replacementElement; // should not happen
  else if (nodeCount == 1) {
    Node& child = node.getFirstChild();
    reverseSingle(child);
    replacementElement.appendChild(child);

    return replacementElement;
  }

  DOMString axisName;  // start with an empty string
  axisName = getReverseAxis(axisName);

  for (int index = nodeCount - 1; index >= 0; --index) {
    Node& childNode = childList.item(index);
    if (childNode.getNodeType() != Node::ELEMENT_NODE)
      replacementElement.appendChild(childNode);

    Element& step = static_cast<Element &>(childNode);
    DOMString newAxisName = getReverseAxis(step.getAttribute("axis"));

    step.setAttribute("axis", axisName);
    axisName = newAxisName;

    replacementElement.appendChild(step);
  }

  return replacementElement;
}

void XPathReverser::reverseSingle(Node& node) {
  if (node.getNodeType() != Node::ELEMENT_NODE) return;

  Element &step = static_cast<Element &>(node);
  step.setAttribute("axis", "test-self");
}

DOMString XPathReverser::getReverseAxis(DOMString axisName) {
  if (axisName == "child") return "test-parent";
  else if (axisName == "parent") return "child";
  else if (axisName == "descendant-or-self") return "ancestor-or-self";
  else if (axisName == "ancestor-or-self") return "descendant-or-self";
  else if (axisName == "preceding") return "following";
  else if (axisName == "following") return "preceding";
  else return "test-self";
}

bool XPathReverser::isWithinPredicate(Node& node) {
  AncestorStream stream(node);
  while (true) {
    Node& nde = stream.nextNode();
    if (!nde.isValid()) break;

    if (nde.getNodeType() != Node::ELEMENT_NODE) continue;
    if (nde.getNodeName() == "PredicateExpr") return true;
  }

  return false;
}

}
